#ifndef __PKCS12_funcs__
#define __PKCS12_funcs__

#include "x509_structs.h"
#include "..\SMemory\SBlocks.h"

//
//	Enum pkcs #12 Pseudo-Random Generation Method For :
//
enum {
	P12_prgm4key = 1,	// key material
	P12_prgm4IV  = 2,	// Initial Vector
	P12_prgm4MAC = 3	// MAC-ing
};

//
//	Function: P12_producePseudoRandomBytes
//
//	See - pkcs #12, B.2 General method. (p.14)
//

int P12_producePseudoRandomBytes(
			BYTE *outBuff,
			int outBuffSize, 
			BYTE *salt,
			int saltSize,
			int iterations,
			BYTE *pwd,		// Unicode buffer required.
			int pwdSize,
			BYTE ID = P12_prgm4key// See P12prgm enums above
	);

extern const char *oidP9FriendlyNameOID;				// = "1.2.840.113549.1.9.20";		{ pkcs-9 friendlyName }
extern const char *oidP9LocalKeyIdOID;					// = "1.2.840.113549.1.9.21";		{ pkcs-9 localKeyId }
extern const char *oidP12BtPKCS8SroudedKeyBagOID;		// = "1.2.840.113549.1.12.10.1.2";	{ pkcs-12 bagTypes PKCS8SroudedKeyBag }
extern const char *oidP12BtCertBagOID;					// = "1.2.840.113549.1.12.10.1.3";	{ pkcs-12 bagTypes certBag }
extern const char *oidP12PbeWithSHAAnd3KeyTripleDES_CBC;// = "1.2.840.113549.1.12.1.3";		{ pkcs-12 pbe pbeWithSHAAnd3KeyTripleDES_CBC }

struct p12_attribute
{
	const char *oidStr;
	BYTE type;
	BYTE *value;
	int  size;

	struct p12_attribute *next;

	p12_attribute() { memset( this, 0, sizeof( *this ) ); }
	~p12_attribute() { if ( next ) { delete next; next = NULL; } }
	static void freeList( p12_attribute *list ) { if ( list ) delete list; }
};

//
//	P12_createAttributeList
//		Helpful function to create standard pkcs-12 safe bag attribute list.
//
p12_attribute *P12_createAttributeList(
		BYTE *friendlyName, int friendlyNameSize,
		BYTE *localKeyId,	int localKeyIdSize
	);

int P12_makeSafeKeyBagPkt( 
		BYTE *out,	int *pOutSize,
		const char *safeBagOID, // f.e. szPKCS8SroudedKeyBagOID
		BYTE *data,	int dataSize,
		p12_attribute *attrList
	);


//
//	If out - encrypts data and fills buffer including octet string...
//	otherwise fills only pOutSize value.
//
int P12_makePBEncryptedPkt(
		BYTE *out,	int *pOutSize,
		BYTE *data,	int dataSize,

		BYTE *salt, int saltSize,
		BYTE *pwd,
		int iteration,
		const char *pbeAlgOID = oidP12PbeWithSHAAnd3KeyTripleDES_CBC
	);

int P12_makeCertificateKeyBagPkt( BYTE *out, int *pOutSize, BYTE *cert, int certSize );
int P12_makePKCS8ShroudedKeyBagPkt( 
		BYTE *out,	int *pOutSize,	// ASN.1 encoded packet
		BYTE *key,	int keySize,	// Plane text private key packet

		BYTE *salt, int saltSize,
		const char *pwd,
		int iteration,
		const char *pbeAlgOID,
		
		p12_attribute *attrList	// Use P12_createAttributeList function to create default attribute list,
								// Use p12_attribute::freeList function to free the list
		);


#ifdef _DEBUG
int pwdTest();
#endif // _DEBUG


////////////////////////////
//
//	Parser functions
//

int P12_parsePBEAlgorithmID      ( BYTE *pkt, int size, unsigned int *pIterations, x509_struct *salt, BYTE *haID, BYTE *saID, BYTE *saMode );
int P12_PKCS8ShroudedKeyBag_parse( BYTE *pkt, int size, cPlainTextBlock *ptbInfo, cPassword *pwd );
int P12_AuthenticatedSafe_parse  ( BYTE *pkt, int size, cPlainTextBlock *ptbInfo, cPassword *pwd );

//
//	P12_PFX_PDU_makeP7Data
//	 the contentInfo can be prepared f.e. by P12_makeSafeContents.
//
int P12_PFX_PDU_makeP7Data( BYTE *pfx, int *pSize, BYTE *contentInfo, int ciSize );
int P12_PFX_PDU_parse( BYTE *pkt, int pktSize, cPlainTextBlock *ptbInfo, cPassword *pwd = NULL );

struct p12_pfx_pdu_p7data : public x509_struct
{
	x509_integer ver;
	x509_struct  seqOfContInfo;
	x509_OID	  p7dataOID;
	x509_struct   optional;
	x509_struct		octetStr;
//	x509_struct  macData; // OPTIONAL

	int prepareForWritting( int contentInfoSize );
	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL );
	int writeBody( BYTE *contentInfo, int contentInfoSize );
};

//
// P12_makeSafeContents
//	every safeBagsT unit has to be prepared before, f.e by P12_makePKCS8ShroudedKeyBagPkt
//
int P12_makeSafeContents( BYTE *SC, int *pSCSize, cPlainTextBlock *safeBagsT, int safeBagsTSize );

struct p12_safeContent : public x509_struct
{
	//
	//	SafeContents ::= SEQUENCE OF SafeBag
	//
	int prepareForWritting( cPlainTextBlock *safeBagsT, int safeBagsTSize );
	int prepareForWritting( x509_struct *safeBagsT, int safeBagsTSize );
	int writeBody( cPlainTextBlock *safeBagsT, int safeBagsTSize );
};


//
//	Secret key packet creation sample 
//
int P12_MakeSecretKeyPkt(
		cPlainTextBlock *pfxPduPtb,
		HPKA hSka,						// Secret key handle
		wchar_t *friendlyNameBuffer,	// Unicode friendly name like 'John Smit <JohnSmit@sci.fi>'
		int friendlyNameBufferSizeInBytes,
		const char *pwd );


//
//      ConvertToUnicodeString
//              Converts ansi buffer to unicode.
//              buff can the same str buffer
//              Returns really filled bytes in buffer
//
int ConvertToUnicodeString( wchar_t *buff, int buffSize, char *str );

#endif // __PKCS12_funcs__