/////////////////////////////////////////////////////////////////
//
//	Copyright (c) 1994-2008 Jetico, Inc., Finland
//	All rights reserved.
//
//	File: SBlocks.h
//	Revision: 
//	Created:
//	Description: Safe memory blocks...
//
/////////////////////////////////////////////////////////////////

#ifndef __SBlocks__
#define __SBlocks__

#ifdef _WIN32
#include <windows.h>
#else
#include "../publicKeyAlgs/Power/Windefs.h"
#endif

#include <memory.h>
#include <string.h>
#include <wchar.h>

class cUniversalString
{
  public:
	cUniversalString()
	{
		memset( this, 0, sizeof( *this ) );
	}

	cUniversalString(const cUniversalString& rhv);

	const cUniversalString& operator = (const cUniversalString& rhv);

	~cUniversalString();

  	//
	// Set functions
	//	if size == -1 it means that the string is null terminated
	//	All functions return number of allocated and filled bytes,
	//	the allocated buffer will include null terminated symbol,
	//	so if success the functions return non zero valus.
	//	Return zero if error.
	//
	int Set( const char *str, int size = -1 );
	int Set( const wchar_t *wcs, int wcsLen = -1 );
	int SetUnicode( const wchar_t *wStr, int sizeInWideChars = -1, bool checkBigEndian = false );
	int SetUnicode( const BYTE *buff, int sizeInBytes = -1, bool checkBigEndian = false );
	int maxI( int a, int b ) { return ( a > b ? a : b ); };
	int Set( cUniversalString *us )
	{
		int sizeW = ( ( us->wcBuffer	  != NULL ) ? SetUnicode( us->wcBuffer		) : 0 );
		int sizeU = ( ( us->unicodeBuffer != NULL ) ? SetUnicode( us->unicodeBuffer ) : 0 );
		int size  = ( ( us->buffer		  != NULL ) ? Set       ( us->buffer		) : 0 );
		return maxI( size, maxI( sizeU, sizeW ) );
	}

	char *buffer;
	int bufferSize;
	wchar_t *wcBuffer;
	int wcBufferSize;
	BYTE *unicodeBuffer;
	int unicodeBufferSize;

	BYTE *AllocBuffer( int size );
	int AllocateUnicodeBuffer( int sizeInSymbols ); // returns size of current unicode buffer.
	void FreeBuffer( BYTE *buff );
	void ClearBuffers();
	bool IsEmpty() { return ( bufferSize == 0 ) || ( buffer[0] == '\0' ); }
	
	//
	// The conversions assume little endian unicode, as used in asn1tag_UnicodeString ( 0x1E ).
	//
	int GenerateUnicodeString(); // returns unicodeBufferSize in bytes, it is equal (pwd length + 1) * 2

#ifdef _WIN32
	int MakeFromUnicodeString();
#endif
	//
	//	Wide char string / unicode buffer conversions.
	//
	//	wstr2unicodeBuff 
	//		unicode buffer size has to be equal wcslen( ws ) * 2 bytes
	//		returns number of copied wide char symbols
	//
	static int wstr2unicodeBuff( BYTE *unicodeBuff, wchar_t *ws );
	int wstr2unicodeBuff( wchar_t *ws )
	{
		if ( AllocateUnicodeBuffer( (int)wcslen( ws ) + 1 ) == 0 )
			return 0;
		
		return wstr2unicodeBuff( unicodeBuffer, ws );
	}

	//
	//	unicodeBuff2wstr
	//		size of wc buffer has to be one symbol more of unicodeBuffByteSize/2
	//		for null termination symbol
	//
	static int unicodeBuff2wstr( wchar_t *ws, int wsSizeInByte, BYTE *unicodeBuff, int unicodeBuffByteSize );
	int unicodeBuff2wstr( wchar_t *ws, int wsSizeInByte )
	{
		int size = wsSizeInByte;
		if ( size > unicodeBufferSize )
			size = unicodeBufferSize;

		return unicodeBuff2wstr( ws, wsSizeInByte, unicodeBuffer, size );
	}
};

//
//	cPlainTextBlock class is used for allocation, storing and deletion of
//	keys and other sensitive information
//
class cPlainTextBlock
{
	//
	//	The call can be send as void handler ( see static WriteBytes ).
	//	But it has a signature value. So validity of the block can be checked.
	//

public:

	enum { sigVal = 0x00627470 }; // "ptb"
	UINT sig;
	bool checkSig()	{ return sig == sigVal; }

//	HANDLE m_heap;


	cPlainTextBlock()
	{
		memset( this, 0, sizeof( *this ) );
		sig = sigVal;
//		m_heap = GetProcessHeap();
	}
	~cPlainTextBlock();

	//
	//	The block supports sequential filling by means of putByte and writeBytes.
	//	It is useful to allocate required size first.
	//
	int allocatedSize;

	BYTE *buffer;
	int bufferSize;

	void ClearBuffer();
	void SeekToBegin();

	//
	//	WriteBytes
	//		if ok returns number of bytes written, otherwise returns 0.
	//		if data == NULL it adds zero bytes to buffer tail.
	//
	int WriteBytes( const BYTE *data, int dataSize, bool init = false );
	int WriteInt( int data ) { return WriteBytes( (BYTE *)&data, sizeof( data ), false ); }
	static int WriteBytesStatic( const void *ptr, BYTE *data, int dataSize );
	int PutByte( BYTE bt );
	int WriteString( const char *str, bool addZero = false ) { return WriteBytes( (const BYTE*)str, (int)strlen( str ) + ( addZero ? 1 : 0 )  ); }
	int WriteString( const wchar_t *str, bool addZero = true ) { return WriteBytes( (const BYTE*)str, 2 * ( (int)wcslen( str ) + ( addZero ? 1 : 0 ) ) ); }
	void Free();
	
	//
	//	Note, ReAllocate does not change bufferSize - current filled data does not changed...
	//
	BYTE *ReAllocate( int size );
	BYTE *Allocate( int size );
	BYTE *Assign( const BYTE *inBuffer, int size );
};

#endif // __SBlocks__
