/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: container.h
 *	scope: v7 module
 *	author: evgeny yakovlev
 *
 *	description:
 *	Implementation of v7 container-related interfaces.
 *
 *	history:
 *
 */

#include "container.h"
#include "error.h"
#include "kgsha.h"
#include "kg_misc.h"

#include <alg/alg.h>
#include <alg/manager.h>
#include <common/log.h>

#include <backend/fs.h>
#include <backend/crypto.h> // IV_64BIT flag

#include <string.h>
#include <stdio.h>

using namespace bdk;


#define	V7_OEMID_STR	"LOCOS94 "


static void dismiss_native_datablock ( DATA_BLOCK** dst, DATA_BLOCK** src, size_t size )
{
        BC_ASSERT( dst && *dst && src && *src );
        memcpy ( *dst, *src, size );
        FreeDataBlock ( ( BYTE** ) src );
}


static bool is_valid_oemid(const char* oemid)
{
	return (strncmp(oemid, V7_OEMID_STR, strlen(V7_OEMID_STR)) == 0);
}


static bool is_valid_kgid( DWORD kgid )
{
        return (kgid == KGSHA_ID) || (kgid == KGSHA256_ID);
}



BCContainerViewHandle V7Format::create_view()
{
	int 			error = 0;
	DWORD 			data_size = sizeof(DATA_BLOCK);
	DWORD 			offset_sectors = 0;
	DATA_BLOCK* 		pdb = &m_data.db;
	DATA_BLOCK* 		temp_pdb = NULL;
	BCAlgHandle 		alg;

	bool rc = CreateHandleEx(alg, this->algorithm_id(), CFLAG_CREATE_WITH_SINGLE_PASSWORD, (BYTE**)&temp_pdb, &data_size, &error, &offset_sectors);

	if( !rc )
	{
		BC_DEBUG_PRINT(("error: %s\n", sha_error_str(error)));
		return BCContainerView::invalid_view();
	}

	dismiss_native_datablock(&pdb, &temp_pdb, sizeof(DATA_BLOCK));
	return this->init_view(offset_sectors, alg);
}


BCContainerViewHandle V7Format::create_hidden_view(IBCContainerIO* io, bc_ulong_t nbytes)
{
	int 			error = 0;
	DWORD 			data_size = sizeof(DATA_BLOCK);
	DWORD 			offset_sectors = 0;
	DATA_BLOCK* 		pdb = &m_data.db;
	BCAlgHandle 		alg;

	offset_sectors = (this->size_bytes() - nbytes) / 512;

	// historically real hidden part size was computed with extra 2kB
	// we can't fix it, but will block dangerous operations
	if( offset_sectors < 4 /* 2kB */ ) {
		bc_get_log().msg(BC_MSG_CRIT) << "Hidden part size " << nbytes << " bytes is too big for v7 container."
				<< "Maximum availiable size is: container size - 2kB.";

		return BCContainerView::invalid_view();
	}

	BC_DEBUG_PRINT(("v7 create context: got offset in bytes 0x%lx\n", offset_sectors * 512));

	bool rc = CreateHandleEx(alg, this->algorithm_id(), CFLAG_CREATE_HIDDEN_PART, (BYTE**)&pdb, &data_size, &error, &offset_sectors);

	if( !rc )
	{
		bc_get_log().msg(BC_MSG_CRIT) << sha_error_str(error);
		return BCContainerView::invalid_view();
	}

	return this->init_view(offset_sectors, alg);
}


BCContainerViewHandle V7Format::get_view(IBCContainerIO* io)
{
	return this->auth_view(CFLAG_VERIFY_AND_LOAD_KEY);
}


BCContainerViewHandle V7Format::update_view(IBCContainerIO* io, bc_view_change_t op)
{
	switch(op)
	{
	case VIEW_ADD_AUTH:
		return this->auth_view(CFLAG_ADD_NEW_PASSWORD);

	case VIEW_REMOVE_AUTH:
		return this->auth_view(CFLAG_REMOVE_ADDITIONAL_PASSWORD);

	case VIEW_CHANGE_AUTH:
		return this->auth_view(CFLAG_CHANGE_PASSWORD);

	default:
		BC_ASSERT(false);
		return BCContainerView::invalid_view();
	};
}


BCContainerViewHandle V7Format::auth_view(DWORD cflag)
{
	int 			error = 0;
	DWORD 			data_size = sizeof(DATA_BLOCK);
	DWORD 			offset_sectors = 0;
	DATA_BLOCK* 		pdb = &m_data.db;
	BCAlgHandle 		alg;

	bool rc = CreateHandleEx(alg, this->algorithm_id(), cflag, (BYTE**)&pdb, &data_size, &error, &offset_sectors);

	if( !rc ) {
		bc_get_log(BC_MSG_CRIT) << sha_error_str(error);
		return BCContainerView::invalid_view();
	}

	return this->init_view(offset_sectors, alg);
}


BCContainerViewHandle V7Format::init_view(bc_ulong_t offset, BCAlgHandle alg) const
{
	bcu32_t flags = (m_data.hs.version == 0) ? 0 : BC_CRYPTO_IV_64BIT;

	bc_ulong_t offset_bytes;
	bc_ulong_t length_bytes;

	if (offset != 0)
	{
		offset_bytes = offset * 512;
		length_bytes = this->size_bytes() - offset_bytes;
	}
	else
	{
		offset_bytes = m_data.hs.dwDataOffset;
		length_bytes = this->size_bytes();
	}


	bc_space_unit_t space;
	space.offset = offset_bytes;
	space.length = length_bytes;

	return BCContainerView::init(alg, space, flags, (offset != 0));
}


bool V7Factory::parse(IBCContainerIO* src)
{
	v7_data data;
	bc_ulong_t nread;

	nread = src->read_bytes( (bc_byte_t*)&data, sizeof(v7_data), 0 );
	if( nread != sizeof(v7_data) )
	{
		BC_DEBUG_PRINT(("Debug: v7: factory parse failed, read %ull bytes\n", nread));
		return false; // not enough data for complete format
	}

	if( !is_valid_oemid(data.hs.br.OEMid) ) {
	  	BC_DEBUG_PRINT(("Debug: v7: factory parse failed, is_valid_oemid %c\n", *data.hs.br.OEMid));
		return false;
	}


	if( !is_valid_kgid(data.hs.keyGenId) ) {
		BC_DEBUG_PRINT(("Debug: v7: factory parse failed, is_valid_kgid %d\n", data.hs.keyGenId));
		return false;
	}

	return true;
}


IBCContainerFormat* V7Factory::load(IBCContainerIO* src)
{
	v7_data data;                // our data

	bc_ulong_t nread = src->read_bytes( (bc_byte_t*)&data, sizeof(v7_data), 0 );
	if( nread != sizeof(v7_data) )
		return NULL; // not enough data for complete format

	return new V7Format(data);
}


IBCContainerFormat* V7Factory::make( bc_container_info_t const & info )
{
	v7_data data;
	HiddenSector* hs = &data.hs;
	DATA_BLOCK* db = &data.db;

	memset( hs, '\0', sizeof( HiddenSector ) );
	memset( db, '\0', sizeof( DATA_BLOCK ) );

	strncpy(hs->br.OEMid, "LOCOS94 ", 8);
    	strncpy(hs->br.volumeLabel, "CRYPTED_DSK", 11);

	hs->br.jmpCode[0] = 0xEB;
    	hs->br.jmpCode[1] = 0x3C;
    	hs->br.jmpCode[2] = 0x90;

    	hs->dwKeySize = sizeof(DATA_BLOCK);
    	hs->dwDataOffset = ( 512 + sizeof(DATA_BLOCK) + 511 ) & ~511;

    	hs->br.bpb.totalSectorsLong = (DWORD)((info.cont_size) / 512) ;
	memset(hs->description, 0, DESCR_SIZE);

	size_t desc_len = info.cont_description.size();
	if( desc_len != 0 ) {
		memcpy(hs->description, info.cont_description.c_str(), (desc_len > (DESCR_SIZE - 1)) ? (DESCR_SIZE - 1) : desc_len );
	}

    	hs->algorithmId = info.cont_alg_id;
    	hs->keyGenId = KGSHA256_ID;
    	hs->version = 7;

	// create product
	return new V7Format(data);
}

