/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: container.h
 *	scope: v7 module
 *	author: evgeny yakovlev
 *
 *	description:
 *	Definition of container-related interfaces.
 *
 *	history:
 *
 */

#ifndef V7_CONTAINER_H
#define V7_CONTAINER_H

#include <stdlib.h>
#include <string.h>

#include "header.h"
#include "kblock.h"

#include <common/build.h>
#include <common/types.h>

#include <alg/alg.h>

#include <container/format.h>
#include <container/factory.h>



/**
 * @brief 	V7 data structure
 */
BC_PACKED(
struct v7_data
{
 	HiddenSector 	hs;
        DATA_BLOCK 	db;
});


/**
 * @brief 	v7 info adaptor
 * @more	Adapts v7 container meta info structure to common interface.
 */
class V7Format : public bdk::IBCContainerFormat
{
public:


	V7Format( v7_data const & data ) :
		m_data(data)
	{};


	BC_INHERITED	std::string name() const
	{
		return "BCv7";
	}


        BC_INHERITED 	bc_ulong_t size_bytes () const
	{
       		return ((bc_size_t)(m_data.hs.br.bpb.totalSectorsLong)) * 512;
	}


	BC_INHERITED	void set_size_bytes(bc_ulong_t size)
	{
       		m_data.hs.br.bpb.totalSectorsLong = size / 512;
	}


       	BC_INHERITED 	std::string description () const
	{
       		return std::string(m_data.hs.description);
	}


	BC_INHERITED	void set_description(std::string const & desc)
	{
		memset(m_data.hs.description, 0, DESCR_SIZE);
		strncpy(m_data.hs.description, desc.c_str(), DESCR_SIZE);
	}


       	BC_INHERITED 	bc_alg_id_t algorithm_id () const
	{
       		return m_data.hs.algorithmId;
	}


	BC_INHERITED 	void set_algorithm_id(bc_alg_id_t id)
	{
       		m_data.hs.algorithmId = id;
	}


	BC_INHERITED 	bc_alg_mode_id_t algorithm_mode_id() const
	{
		return bdk::AM_CBC; // the only one supported by v7
	}


	BC_INHERITED 	void set_algorithm_mode_id(bc_alg_mode_id_t mod)
	{
		return;
	}


        BC_INHERITED 	bc_fs_id_t fs_id () const
	{
		return m_data.hs.fileSystemId;
	}


	BC_INHERITED 	void set_fs_id(bc_fs_id_t id)
	{
		m_data.hs.fileSystemId = id;
	}


        BC_INHERITED 	bool is_locked () const
	{
        	return (m_data.hs.br.jmpCode[0] == 'B');
	}


	BC_INHERITED 	void lock()
	{
		m_data.hs.br.jmpCode[0] = 'B';
	}


	BC_INHERITED 	void unlock()
	{
		m_data.hs.br.jmpCode[0] = 0xEB;
	}


	BC_INHERITED	void serialize(bdk::IBCContainerIO* dest) const
	{
		bc_ulong_t nwritten;
		nwritten = dest->write_bytes((bc_byte_t*)&m_data, sizeof(v7_data), 0);

		BC_ASSERT(nwritten == sizeof(v7_data));
	}


	BC_INHERITED	bdk::BCContainerViewHandle create_view();


	BC_INHERITED	bdk::BCContainerViewHandle create_hidden_view(bdk::IBCContainerIO* io, bc_ulong_t size);


	BC_INHERITED 	bdk::BCContainerViewHandle get_view(bdk::IBCContainerIO* io);


	BC_INHERITED	bdk::BCContainerViewHandle update_view(bdk::IBCContainerIO* io, bdk::bc_view_change_t op);



protected:


	bdk::BCContainerViewHandle auth_view(DWORD cflag);


	bdk::BCContainerViewHandle init_view(bc_ulong_t offset, bdk::BCAlgHandle alg) const;


private:


	/* This v7 container header */
	v7_data		m_data;

};


/**
 * @brief	V7 container factory
 */
class V7Factory : public bdk::IBCContainerFactory
{
public:

	BC_INHERITED 	bool parse(bdk::IBCContainerIO* src);


	/**
	 * @brief	Initialize and return V7 authenticator.
	 */
	BC_INHERITED	bdk::IBCContainerFormat* load(bdk::IBCContainerIO* src);


	/**
	 * @brief	Create and initialize new V7 data.
	 */
	BC_INHERITED	bdk::IBCContainerFormat* make(bdk::bc_container_info_t const & info);

};


#endif

