/*******************************************************************
 *
 *	Copyright (c) 1994-1999 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: kgsha.h
 *	Revision:  $Id: kgsha.cpp,v 1.2 2006-11-07 06:29:13 nail Exp $
 *	Created:
 *	Description: main file of the libkgsha with functions that 
 *		may be called from external programs
 *
 *******************************************************************/

#include <string.h>
#include "sha_generic.h"
#include "kgsha.h"
#include "kg_defs.h"
#include "kblock.h"
#include "kg_misc.h"
#include "pwldlg.h"
#include "error.h"

#include <common/types.h>
#include <common/log.h>
#include <alg/alg.h>

#define TRUE 1
#define FALSE 0
#define POOL_SIZE_BYTES 512

using namespace bdk;

char kgsha_cpp[]="$Id: kgsha.cpp,v 1.2 2006-11-07 06:29:13 nail Exp $";

/*******************************************************
 *
 *	QueryFunction();
 *
 *  Purpose:: Check if the function is available for the 
 *            data block.
 *
 *******************************************************/

BOOL QueryFunction(
                  const BYTE  **vDataBlock,   /* block of data that is stored inside file-container in encrypted form */
                  const DWORD DataSize,   /* size of the block of data */
                  const DWORD FunctionID  /* identefier of function used as CerateKeyHandle(.., CreateFlag, ... ) */
                  )
{
    switch ( FunctionID )
    {
        
        case CFLAG_CREATE_WITH_SINGLE_PASSWORD  : return TRUE;
        case CFLAG_VERIFY_AND_LOAD_KEY          : return TRUE;
        case CFLAG_CHANGE_PASSWORD              : return TRUE;
        case CFLAG_ADD_NEW_PASSWORD             : return TRUE;
        case CFLAG_REMOVE_ADDITIONAL_PASSWORD : return TRUE;
        default : return FALSE;

    }
}


/*******************************************************
 *
 *	GetNameAndVersion()
 *
 *******************************************************/

/*
BOOL GetNameAndVersion(
                                 char *Name, 
                                 DWORD  nameSize, 
                                 DWORD *Major, 
                                 DWORD *Minor 
                                 )
{
    strncpy(Name, DISPLAY_NAME, nameSize - 1);
    Name[nameSize - 1] = 0;
    *Major = MAJOR_VERSION;
    *Minor = MINOR_VERSION;
    return TRUE; 
}
*/




/*******************************************************
 *
 *	CreateKeyHandle()
 *
 *  IMPORTANT: If CreateKeyHandle returns ERROR, the memory 
 *  reserved for DataBlock may be filled by wrong data
 *
 *******************************************************/

BOOL CreateHandleEx(		BCAlgHandle& handle,
				int algid,
                                 DWORD   CreateFlag,     // whether to create new key or open existing
                                 BYTE    **vDataBlock,   // block of data that is stored inside file-container in encrypted form
                                 DWORD   *DataSize,      // size of the block of data
                                 int     *ErrorCode,     // if the function returns FALSE, look at the Error code
                                 DWORD   *DataOffset
                                 )
{
    BOOL    bResult, bHidPart;
    BYTE    *key, *seed, *pool; //, containerId[ CONTAINER_ID_LENGTH ];
    BYTE    tmpKey[ MAXIMUM_KEY_SIZE_BYTES + 4 ];
    char    *password, *newPassword;
    DWORD   offset;
    DWORD   mask = 0, kbPosition = 0;
    DATA_BLOCK **DataBlock = (DATA_BLOCK **)vDataBlock;
    int iAttempts=0;

    key = seed = NULL;
    password = newPassword = NULL;

    if ( !Allocate( &seed, &password, &newPassword, &key, &pool ) )
    {
        *ErrorCode = ERROR_INTERNAL_PROBLEM;
        return FALSE;
    }

    bResult = FALSE; // Assume error wxill occur

    // Test if given algorithm can be called
/*
    if ( !bc_verify_alg(algid) )
    {
        *ErrorCode = ERROR_INVALID_ALGORITHM;
        Free( seed, password, newPassword, key, pool );
        return FALSE;
    }
*/

    if ( !(CORRECT_CFLAG(CreateFlag)) )
    {
        *ErrorCode = ERROR_INCORRECT_CREATE_FLAG;
        Free( seed, password, newPassword, key, pool );
        return FALSE;
    }


    switch ( CreateFlag )
    {
        // **************  Creating new Data Block and return Key Handle ***************
        
        case CFLAG_CREATE_WITH_SINGLE_PASSWORD:

            if ( !DataBlockAllocate( DataBlock, DataSize, algid, SHA256_ID) )
            {
                *ErrorCode = ERROR_INTERNAL_PROBLEM;
                break;
            }

            if ( ((*ErrorCode) = GetSeedValue(
                                             seed, SEED_LENGTH )) != ERROR_NO )
            {
                DataBlockFree( DataBlock );
                break;
            }

            if ( ((*ErrorCode) = GetPasswordWithVerification( NULL, "new password:", password ))
                 != ERROR_NO )
            {
                DataBlockFree( DataBlock );
                break;
            }


            if ( GenerateKey(key, MAXIMUM_KEY_SIZE_BYTES, 
                             seed, SEED_LENGTH, SHA256_ID) != ERROR_NO )
            {
                *ErrorCode = ERROR_INTERNAL_PROBLEM;
                DataBlockFree( DataBlock );
                break;
            }

            if ( DataBlockWriteKey(*DataBlock, algid, password, key, KATTRIBUTE_KEY_FULL, pool) != ERROR_NO )
            {
                DataBlockFree( DataBlock );
                *ErrorCode = ERROR_INTERNAL_PROBLEM;
                break;
            }

            /* verify all that we did before */
            if ( ((*ErrorCode) = 
                  DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, password, key, pool )
                 ) != ERROR_NO )
            {
                break;
            }

            // Call Algorithm driver for expanding and storing the Key
		handle = BCAlg::init(algid, AM_CBC, key, pool);
		if(!handle)
            	{
                	*ErrorCode = ERROR_INVALID_ALGORITHM;
                	DataBlockFree( DataBlock );
                	break;
            	}

            *ErrorCode = ERROR_NO;
            bResult = TRUE;
            break;


            // **************  Verify password and return Key Handle ***************

        case CFLAG_VERIFY_AND_LOAD_KEY: // 1 - minimum number of keys inside the block

            offset = 0;         // If we find out that user entered password for
                                // hidden container, offset won't be 0


            if ( ((*ErrorCode) = DataBlockCheck(*DataBlock, *DataSize, algid)) != ERROR_NO )
            {
                break;
            }

            if ( ((*ErrorCode) = GetPassword(NULL, "password:", password )) != ERROR_NO )
            {
                break;
            }

            // First Attempt to use password: is the key simply stored inside DataBlock?
            *ErrorCode = DataBlockVerifyPasswordAndGetKey(*DataBlock, algid, password, key, pool );


            // Second Attempt to use password: is the password was entered for Hidden Part?

            if ( (*ErrorCode) == ERROR_INCORRECT_PASSWORD )
            {
                mask = 0, kbPosition = 0;

                if ( ((*ErrorCode) = 
                      DataBlockVerifyPasswordAndGetKey_Hidden( *DataBlock, algid, password, key, pool, &offset, &mask, &kbPosition )
                     ) != ERROR_NO )
                {

                    break;
                }
                else
                {
                    //HDLG_ShowInformationMessage( hWndParent );
                }
            }

            else if ( (*ErrorCode) != ERROR_NO ) { 
		    break; 
	    }

            // Call Algorithm driver for expanding and storing the Key
		handle = BCAlg::init(algid, AM_CBC, key, pool);
		if(!handle)
            	{
                	*ErrorCode = ERROR_INVALID_ALGORITHM;
                	break;
            	}


            *ErrorCode = ERROR_NO;
            *DataOffset = offset;
            bResult = TRUE;
		
            break;


            // **************  Change password for Regular or Hidden container ***************

        case CFLAG_CHANGE_PASSWORD:

            if ( ((*ErrorCode) = DataBlockCheck(*DataBlock, *DataSize, algid)) != ERROR_NO )
            {
                break;
            }

            if ( ((*ErrorCode) = GetPassword(NULL, "Enter old password: "/*Caption*/, password )) != ERROR_NO )
            {
                break;
            }



            bHidPart = FALSE;

            if ( ((*ErrorCode) = 
                  DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, password, key, pool )
                 ) != ERROR_NO )
            {
                // May be, we are going to change password for hidden part?
                if ( ((*ErrorCode) = 
                      DataBlockVerifyPasswordAndGetKey_Hidden( *DataBlock, algid, password, key, pool, &offset, &mask, &kbPosition )
                     ) != ERROR_NO )
                {
                    break; // It is not a password for hidden part :-(
                }
                else
                {
                    bHidPart = TRUE;
                }
            }

            if ( !bHidPart )
            {

                if ( ((*ErrorCode) = GetPasswordWithVerification( NULL, "Enter new password: "/*Caption*/, newPassword ))
                     != ERROR_NO )
                {
                    break;
                }

                if ( ((*ErrorCode) = DataBlockChangePassword( *DataBlock, algid, password, newPassword, pool )) != ERROR_NO )
                {
                    break;
                }

            }
            else
            {
                BYTE tmpKey[ MAXIMUM_KEY_SIZE_BYTES + 4 ];

                if ( ((*ErrorCode) = GetPasswordWithVerification(
                                                                "Changing password for hidden part!",//text
                                                                "Enter new password: "/*Caption*/, 
                                                                newPassword ))
                     != ERROR_NO )
                {
                    break;
                }


                // Check if password for hidden part is the same as for original part
                while ( (DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, newPassword, tmpKey, pool )
                        ) == ERROR_NO )
                {
                    ShredData( tmpKey, MAXIMUM_KEY_SIZE_BYTES + 4);

                    if ( ((*ErrorCode) = GetPasswordWithVerification(
                                                                    "WARNING: Entered password equal main container password!",//text
                                                                    "Enter new password: "/*Caption*/, 
                                                                    newPassword ))
                         != ERROR_NO )
                    {
                        break;
                    }

                }

                ShredData( tmpKey, MAXIMUM_KEY_SIZE_BYTES + 4);

                if ( *ErrorCode != ERROR_NO )
                {
                    break;
                }

                *ErrorCode = DataBlockWriteKey_Hidden(*DataBlock, algid, newPassword, key, KATTRIBUTE_KEY_EMPTY, pool, offset, &mask, kbPosition);

                if ( *ErrorCode != ERROR_NO ) break;
            }

            *ErrorCode = ERROR_NO;
            bResult = TRUE;
            break;


            // **************  Add new password for Regular or Hidden container ***************

        case CFLAG_ADD_NEW_PASSWORD:

            if ( ((*ErrorCode) = DataBlockCheck(*DataBlock, *DataSize, algid)) != ERROR_NO )
            {
                break;
            }

            *ErrorCode = DataBlock_CheckFreeSpaceForNewPassword( *DataBlock );
            if ( (*ErrorCode) == ERROR_NOT_ENOUGH_SPACE_FOR_KEY )
            {
                bResult = FALSE;
                break;
            }

//			LoadString( hModule, IDS_ENTER_EXISTING_PASSWORD, msgText, sizeof(msgText) );

            if ( ((*ErrorCode) = GetPassword(NULL, "Enter existing password: ", password )) != ERROR_NO )
            {
                break;
            }
            bHidPart = FALSE;

            if ( ((*ErrorCode) = 
                  DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, password, key, pool )
                 ) != ERROR_NO )
            {
                // May be, we are going to change password for hidden part?
                if ( ((*ErrorCode) = 
                      DataBlockVerifyPasswordAndGetKey_Hidden( *DataBlock, algid, password, key, pool, &offset, &mask, &kbPosition )
                     ) != ERROR_NO )
                {
                    break; // It is not a password for hidden part :-(
                }
                else
                {
                    bHidPart = TRUE;
                }
            }



            for ( ;; )
            {

                if ( ((*ErrorCode) = GetPasswordWithVerification(NULL, "Enter additional password: ", newPassword)) != ERROR_NO )
                {
                    break;
                }

                // Check if such new password was added earlier - nothing to do in this case
                if ( ((*ErrorCode) = DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, newPassword, tmpKey, pool )) == ERROR_NO )
                {
                    if ( bHidPart )
                    {
                        continue;
                    }
                    else
                    {
                        break; // such password was added earliar - nothing to do
                    }
                }

                if ( ((*ErrorCode) = DataBlockVerifyPasswordAndGetKey_Hidden( *DataBlock, algid, newPassword, key, pool, &offset, &mask, &kbPosition )
                     ) == ERROR_NO )
                {
                    if ( bHidPart )
                    {
                        break; // such password was added earliar - nothing to do
                    }
                    else
                    {
                        continue;
                    }
                }

                break;
            }

            ShredData( tmpKey, MAXIMUM_KEY_SIZE_BYTES + 4);

            if ( ((*ErrorCode) == ERROR_NO) ||              // such password was added earliar - nothing to do
                 ((*ErrorCode) == ERROR_CANCELLED_BY_USER) ) // process is cancelled
            {
                bResult = TRUE;
                break;
            } else if((*ErrorCode) == ERROR_USER_PASSWORDS_MISMATCH) {
		bResult = FALSE;
		break;
	   }

            if ( !bHidPart )
            {
                *ErrorCode = DataBlockAddPassword( DataBlock, DataSize, algid, newPassword, key, KATTRIBUTE_KEY_FULL, pool );
            }
            else
            {
                *ErrorCode = DataBlockAddPassword_Hidden(DataBlock, DataSize, algid, newPassword, key, KATTRIBUTE_KEY_EMPTY, pool, offset, &mask ); // update mask in the call

                if ( (*ErrorCode) == ERROR_NO )
                {
                    *ErrorCode = DataBlockWriteKey_Hidden(*DataBlock, algid, password, key, KATTRIBUTE_KEY_EMPTY,
                                                          pool, offset, &mask, kbPosition); // write updated mask to original Key Block
                }
            }

            if ( (*ErrorCode) == ERROR_NOT_ENOUGH_SPACE_FOR_KEY )
            {
                break;
            }

            if ( *ErrorCode != ERROR_NO ) break;
            bResult = TRUE;
            break;


        case CFLAG_REMOVE_ADDITIONAL_PASSWORD:

            bResult = FALSE;

            if ( ((*ErrorCode) = DataBlockCheck(*DataBlock, *DataSize, algid)) != ERROR_NO )
            {
                break;
            }

            for ( iAttempts=0; iAttempts<3; iAttempts++ )
            {

                if ( ((*ErrorCode) = GetPassword( NULL, "Enter additional password to delete: ", password )) != ERROR_NO )
                {
                    //*ErrorCode = ERROR_NO;
                    break;
                }

                // Check if we have entered password for hidden part. 
                // If so, refuse to delete it
                if ( ((*ErrorCode) = DataBlockVerifyPasswordAndGetKey_Hidden( *DataBlock, algid, password, key, pool, &offset, &mask, &kbPosition )
                     ) == ERROR_NO )
                {
                    *ErrorCode = ERROR_NO;
                    continue;
                }

                *ErrorCode = DataBlockRemoveAdditionalPassword( *DataBlock, algid, password, key, pool );

                if ( (*ErrorCode) == ERROR_INCORRECT_PASSWORD )
                {
                    *ErrorCode = ERROR_NO;
                    continue;
                }
                else if ( (*ErrorCode) == ERROR_MASTER_PASSWORD_ENTERED )
                {
                    *ErrorCode = ERROR_NO;
                    continue;
                }
                else if ( (*ErrorCode) == ERROR_NO )
                {
                    break;
                }
            }

            if ( (*ErrorCode) == ERROR_NO ) bResult = TRUE;
            break;

            // **************  Create Hidden part  ***************

        case CFLAG_CREATE_HIDDEN_PART:

            if ( ((*ErrorCode) = DataBlockCheck(*DataBlock, *DataSize, algid)) != ERROR_NO )
            {
                break;
            }

            *ErrorCode = DataBlock_CheckFreeSpaceForNewPassword( *DataBlock );
            if ( (*ErrorCode) == ERROR_NOT_ENOUGH_SPACE_FOR_KEY )
            {
                bResult = FALSE;
                break;
            }

            if ( ((*ErrorCode) = GetPassword(NULL, "Enter existing password: ", password )) != ERROR_NO )
            {
                break;
            }

            if ( ((*ErrorCode) = 
                  DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, password, tmpKey, pool )
                 ) != ERROR_NO )
            {
                bResult = FALSE;
                break;
            }



            for ( ;; )
            {

                if ( ((*ErrorCode) = GetPasswordWithVerification(NULL, "Enter password for hidden part: ", 
                                                                 newPassword)) != ERROR_NO )
                {
                    break;
                }

                // Check if such new password was added earlier - nothing to do in this case
                if ( ((*ErrorCode) = DataBlockVerifyPasswordAndGetKey( *DataBlock, algid, newPassword, tmpKey, pool )) == ERROR_NO )
                {
                    //Text = "This password already in use!";
                    continue;
                }else *ErrorCode = ERROR_NO;
/*
                if ( ((*ErrorCode) = DataBlockVerifyPasswordAndGetKey_Hidden(
                                                                            *DataBlock, Alg, AlgKeyLength,
                                                                            newPassword, tmpKey, pool, &offset, &mask, &kbPosition )
                     ) == ERROR_NO )
                {
                    *ErrorCode = ERROR_HID_PASSWORD_INVALID;
                }
*/
                break;
            }
            
            if ( *ErrorCode != ERROR_NO ) break;
            
            if ( ((*ErrorCode) = GetSeedValue( seed, SEED_LENGTH )) != ERROR_NO )
                break;

            *ErrorCode = ERROR_INTERNAL_PROBLEM; // assume the error will occur

            if ( GenerateKey(key, MAXIMUM_KEY_SIZE_BYTES, seed, SEED_LENGTH, (*DataBlock)->header.hashAlgorithmId) != ERROR_NO )
                break;

            mask = 0, kbPosition = 0;

            if ( ((*ErrorCode) = DataBlockWriteKey_Hidden(*DataBlock, algid, newPassword, key, KATTRIBUTE_KEY_EMPTY, pool, *DataOffset, &mask, HIDDEN_POSITION_ANY)) != ERROR_NO )
                break;

            /* verify all that we did before */
            if ( ((*ErrorCode) = DataBlockVerifyPasswordAndGetKey_Hidden( *DataBlock, algid, newPassword, key, pool, &offset, &mask, &kbPosition )
                 ) != ERROR_NO )
            {
                break;
            }


	    // Call Algorithm driver for expanding and storing the Key
		handle = BCAlg::init(algid, AM_CBC, key, pool);
		if(!handle)
            	{
                	*ErrorCode = ERROR_INVALID_ALGORITHM;
                	break;
            	}

            if ( *ErrorCode != ERROR_NO ) break;
            bResult = TRUE;
            break;

        default:
            *ErrorCode = ERROR_INCORRECT_CREATE_FLAG;
            bResult = FALSE;
            break;
    }

    ShredData( (BYTE *)password, PASSWORD_MAXIMUM_LENGTH );
    ShredData( (BYTE *)newPassword, PASSWORD_MAXIMUM_LENGTH );
    ShredData( key, MAXIMUM_KEY_SIZE_BYTES );

    Free( seed, password, newPassword, key, pool );

    return bResult;
}

BOOL CreateHandle(
				BCAlgHandle& handle,
				int algid,
                               DWORD   CreateFlag,  /* whether to create new key or open existing */
                               BYTE    **vDataBlock,    /* block of data that is stored inside file-container in encrypted form */
                               DWORD   *DataSize,   /* size of the block of data */
                               int     *ErrorCode   /* if the function returns FALSE, look on the Error code */
                               )
{
    DWORD   DataOffset;
    return CreateHandleEx(handle, algid, CreateFlag, vDataBlock,DataSize,ErrorCode,&DataOffset);
}


/*******************************************************
 *
 *	FreeDataBlock() - frees the memory allocated when 
 *  user calls CreateKeyHandle() procedure with 
 *  CreateFlag equal to CFLAG_CREATE_WITH_SINGLE_PASSWORD
 *
 *******************************************************/

DWORD FreeDataBlock( BYTE **vDataBlock )
{
    if ( !DataBlockFree( (DATA_BLOCK **)vDataBlock ) )
        return ERROR_INTERNAL_PROBLEM;
    return ERROR_NO;
}


/*******************************************************
 *
 *	FreeKeyHandle() - frees the memory allocated for 
 *  encryption key inside algorithm driver when user calls 
 *  CreateKeyHandle() procedure with CreateFlag equal to 
 *  CFLAG_CREATE_WITH_SINGLE_PASSWORD or
 *  CFLAG_VERIFY_AND_LOAD_KEY
 *
 *******************************************************/

/*
extern "C" DWORD FreeKeyHandle(
                              ALG_SERV Alg,
                              DWORD KeyHandle 
                              )
{
    if ( alg_free_key( Alg, KeyHandle ) )
        return ERROR_INTERNAL_PROBLEM;
    return ERROR_NO;
}
*/

