
#include "SSS_block.h"
#include "enc_block.h"
#include "user.h"
#include "datablock.h"

#include "CertificateLib/sss/shamir/shamir.h"
#include "CertificateLib/hashAlgs/hashAlgs.h"
#include "CertificateLib/SMemory/cPassword.h"

using namespace kgghost;
using namespace bdk;


// Implementation-only declarations
namespace kgghost
{

	static const int MaxEncodedSize = sizeof(key_info_t);
	static const int MaxPrimeSize = MaxEncodedSize + 1;
	static const int PrimeAndSaltSize = MaxPrimeSize - 1; 
	static const int SaltSize = 8;



	/**
	 * @brief	Share format version tag
	 */
	enum share_version_t
	{
		SHARE_VERSION_UNKNOWN 	= 0,
		SHARE_VERSION_1		= 1,
		SHARE_VERSION_2		= 2,
	};



	/**
	 * @brief	Version 1 of sss share encoded structure. Deprecated.
	 */
	struct sss_share_V1
	{
		BYTE version;
		BYTE reqNum;
		BYTE totalQ;
		BYTE headOfPrime; // One first byte of prime always contain alder bit, so we hide it in encrypted part
		BYTE valX[ MaxPrimeSize ];
		BYTE valY[ MaxPrimeSize ]; // MAXIMUM_KEY_SIZE_BYTES + 1 
		BYTE hashAlgId;
	};


	/**
	 * @brief	Version 2 of sss share encoded structure
	 */
	struct sss_share_V2
	{
		BYTE version;			// share version tag, 2
		BYTE reqNum;			// number of required total shares to decode the key
		BYTE totalQ;			// number of total shares that were encoded. Can be more than required.
		BYTE headOfPrime;		// first byte of prime.
		BYTE valX[1];			// share order number is used for x value.
		BYTE valY[MaxPrimeSize];	// MAXIMUM_KEY_SIZE_BYTES + 1
		BYTE hashAlgId;			// id of hash algorithm used for share digest operations.
	};


	/**
	 * @brief	Encoded share structure.
	 * @more 	These structures are stored in keyblocks in encoded form.
	 *		They are logically grouped by sss users and decoded\encoded by user's algorithm configuration.
	 */
	struct sss_share_t
	{
		/* share format padded for encryption alg */
		bc_padded_type_t < sss_share_V2, DB_MAX_ENC_BLOCK_SIZE > m_sssblock;

		bc_byte_t m_digest[MAXIMUM_DIGEST_SIZE_BYTES];
		

		/**
		 * @more		SSS Share format version.
		 *
		 * @return		share_version_t tags.
		 */
		const BYTE version() const
		{
			return m_sssblock.type()->version;
		}


		/**
		 * @brief		Get the number of shares required to decode the key.
		 * @more		Each encoded share dublicates this value.
		 */
		unsigned int total_required_shares() const
		{
			return m_sssblock.type()->reqNum;
		}


		/**
		 * @brief		Get the number of total shares encoded.
		 * @more		Each encoded share dublicates this value.
		 *			May be more than the required number.
		 */
		unsigned int total_shares() const
		{
			return m_sssblock.type()->totalQ;
		}


		/**
		 * @brief		Get the first byte of common prime number.
		 * @more		Each encoded share dublicates this value. 
		 *			The rest of the prime can be found at the beginning of the SSS keyblock unencrypted.
		 */
		const BYTE prime_high() const
		{
			return m_sssblock.type()->headOfPrime;
		}


		/**
		 * @brief		Load format-specific X value into power integer.
		 *
		 * @p power_x		Power integer storage.
		 */	
		int commitX(pw_int& power_x) const
		{
			if( power_x.set(sizeof(m_sssblock.type()->valX), m_sssblock.type()->valX) != pw_OK )
				return ERROR_INTERNAL_PROBLEM;

			return ERROR_NO;
		}


		/**
		 * @brief		Store format-specific X value from power integer.
		 *
		 * @p power_x		Power integer storage.
		 */	
		int pullX(const pw_int& power_x)
		{
			if( pw_int2bytes(m_sssblock.type()->valX, sizeof(m_sssblock.type()->valX), &power_x) != pw_OK )
				return ERROR_INTERNAL_PROBLEM;

			return ERROR_NO;
		}


		/**
		 * @brief		Store format-specific Y value.
		 *
		 * @p power_y		Power integer storaget.
		 */	
		int commitY(pw_int& power_y) const
		{
			if( power_y.set(sizeof(m_sssblock.type()->valY), m_sssblock.type()->valY) != pw_OK )
				return ERROR_INTERNAL_PROBLEM;

			return ERROR_NO;
		}


		/**
		 * @brief		Store format-specific Y value from power integer.
		 *
		 * @p power_y		Power integer storage.
		 */	
		int pullY(const pw_int& power_y)
		{
			if( pw_int2bytes(m_sssblock.type()->valY, sizeof(m_sssblock.type()->valY), &power_y) != pw_OK )
				return ERROR_INTERNAL_PROBLEM;

			return ERROR_NO;
		}


		/**
		 * @brief		Decode share.
		 * @more		Decrypt share and verify it's digest.
		 *
		 * @p alg		Algorithm config to perform decryption.
		 * @p shareNum		Number of share in keyblock. Need this to generate iv.
		 * @p lowPrime		Buffer with the low values of prime (we have the upper byte). Need this to init hash alg context.
		 *
		 * @return		ERROR_NO if decoded.
		 *			ERROR_INVALID_ALGORITHM if decoded but digest verification failed..
		 *			Other codes if unexpected errors occur.
		 */
		int decode(BCAlgHandle alg, WORD shareNum, bc_byte_t* lowPrime);


		/**
		 * @brief		Calculate and store share digest.
		 * @more		Will calculate digest of the share without the digest buffer inteself.
		 *			Will also store passed hash algorithm id if it is valid.
		 *
		 * @p lowPrime		Buffer with the low values of prime (we have the upper byte). Need this to init hash alg context.
		 * @p hashAlgId		Id of the hash algorithm to use. Will be stored too if valid.	
		 *
		 * @return		ERROR_NO if digest was generated.
		 *			ERROR_INCORRECT_HASH_ALGORITHM if hash alg id is invalid.
		 *			Other codes if unexpected errors occur.
		 */ 
		int calculate_digest(bc_byte_t* lowPrime, int hashAlgId);


		/**
		 * @brief		Verify share digest.
		 *
		 * @p lowPrime		Buffer with the low values of prime (we have the upper byte). Need this to init hash alg context.
		 *
		 * @return		ERROR_NO if digest was verified.
		 *			ERROR_INCORRECT_HASH_ALGORITHM if digest was not verified.
		 *			Other codes if unexpected errors occur.
		 */
		int verify_digest(bc_byte_t* lowPrime);
	};

} // namespace kgghost


int sss_share_t::decode(BCAlgHandle alg, WORD share_num, bc_byte_t* primeBuffer)
{
	unsigned int size = sizeof(*this);
	bc_byte_t iv[IVECTOR_LENGTH_DB] = {0};

	// Decrypt and verify next share

	// use index of share as an iv
	iv[0] = LOBYTE(share_num);
	iv[1] = HIBYTE(share_num);


	alg->decrypt(iv, (bc_byte_t*)this, size);

	int res = this->verify_digest(primeBuffer);
	if( res == ERROR_INCORRECT_HASH_ALGORITHM ) 
		return ERROR_INVALID_ALGORITHM;

	return res;
}


int sss_share_t::calculate_digest(bc_byte_t* primeBuffer, int hash_alg_id)
{

	HHA ha;
	int res = ha_createAlgorithm( &ha, hash_alg_id );
	if ( res != hae_ok )
		return ERROR_INVALID_ALGORITHM;

	unsigned int digest_size = ha_getDigestSize(ha);

	ShredData((bc_byte_t*)m_digest, sizeof(m_digest));

	if( digest_size > sizeof(m_digest) ) 
		return ERROR_INTERNAL_PROBLEM;

	if( ha_process(ha, primeBuffer, PrimeAndSaltSize) != hae_ok ) 
		return ERROR_INTERNAL_PROBLEM;

	if( ha_process(ha, (BYTE*)this, sizeof(this->m_sssblock)) != hae_ok ) 
		return ERROR_INTERNAL_PROBLEM;

	if( ha_final(ha, m_digest) != hae_ok )
		return ERROR_INTERNAL_PROBLEM;


	ha_free( ha );
	m_sssblock.type()->hashAlgId = hash_alg_id;

	return ERROR_NO;
}

	
int sss_share_t::verify_digest(bc_byte_t* primeBuffer)
{
	HHA ha;
	int res = ha_createAlgorithm( &ha, m_sssblock.type()->hashAlgId );
	if ( res != hae_ok )
		return ERROR_INCORRECT_HASH_ALGORITHM;
	
	BYTE tmpDigest[MAXIMUM_DIGEST_SIZE_BYTES] = {0};
	unsigned int digestSize = ha_getDigestSize(ha);

	if( digestSize > sizeof(m_digest) )
		return ERROR_INTERNAL_PROBLEM;

	if( ha_process(ha, primeBuffer, PrimeAndSaltSize) != hae_ok )
		return ERROR_INTERNAL_PROBLEM;

	if( ha_process(ha, (BYTE *)this, sizeof(m_sssblock)) != hae_ok )
		return ERROR_INTERNAL_PROBLEM;
	
	if( ha_final(ha, tmpDigest) != hae_ok )
		return ERROR_INTERNAL_PROBLEM;

	ha_free( ha );

	return (memcmp(tmpDigest, m_digest, digestSize) == 0 ) ? ERROR_NO : ERROR_INCORRECT_HASH_ALGORITHM;
}


namespace kgghost
{

	/**
	 * @brief 	Shamir Shared Scheme key block
	 */
	struct SSS_KeyBlock
	{
		/* prime number with last 8 bytes interpreted as salt value */
		BYTE primeAndSalt[PrimeAndSaltSize];

		/* encoded shares of chosen format */
		sss_share_t shares[1]; // Real number of shares is unknown at declaration. But every share has this number.
	};


	/**
	 * @brief	SSS User is a logical collection of shares for a single password entry.
	 */
	struct sss_user_t
	{
		typedef std::list<sss_share_t>	 	share_list_t;
		typedef share_list_t::iterator 		share_iter_t;
		typedef share_list_t::const_iterator 	share_citer_t;


		/* decoded shares */
		share_list_t 	m_shares;

		/* configured user alg */
		BCAlgHandle	m_alg;


		/**
		 * @brief		Construct uninitialized sss user data
		 */
		sss_user_t()
		{
			this->clear();
		}


		/**
		 * @brief		Construct user with password
		 */
		sss_user_t(const cPassword& pwd, key_info_t const & enc_ids, bc_byte_t* salt)
		{
			this->reset(pwd, enc_ids, salt);
		}


		/**
		 * @brief		Destroy all allocated state
		 */
		~sss_user_t()
		{
			this->clear();
		}


		/**
		 * @brief		Return total shares accumulated.
		 */
		unsigned int total_assigned_shares() const
		{
			return m_shares.size();
		}


		/**
		 * @brief		Get one of the decoded shares as a sample.
		 * @more		Clients will most likely need any of our shares just to look at global data dublicated in all encoded shared.
		 *			The share-unique data is often the matter of incapsulated routines only.
		 */
		const sss_share_t* get_share_sample() const
		{
			if(m_shares.size() == 0) return NULL;
			else return &m_shares.front();
		}


		/**
		 * @brief		Store share in collection.
		 *
		 * @p share		Share data.
		 *
		 * @return		ERROR_NO if share was stored.
		 *			ERROR_INTERNAL_PROBLEM if no memory.
		 */
		int assign_share(const sss_share_t& share)
		{
			m_shares.push_back(share);
			return ERROR_NO;	
		}


		/**
		 * @brief		Decode my shares from keyblock
		 *
		 * @p kbEncrypted	Encrypted sss keyblock buffer with all encoded shares we need to find ours in.
		 * @p kbSize		Size fo encrypted keyblock. Will be updated with the real size.
		 *
		 * @return 		ERROR_NO if decoding succeded.
		 *			ERROR_INVALID_ALGORITHM if supplied alg handle failed to properly decode any share (most likely due to invalid password).
		 *			Other codes if unexpected error encountered.
		 */
		int decode_shares(const bc_byte_t* kbEncrypted, unsigned int kbSize);


		/**
		 * @brief		Commit my shares into runtime table for key restoration.
		 * 
		 * @p sss_runtime	Runtime table of generated sss shares to restore initial key.
		 *
		 * @return		ERROR_NO, if share commit reached runtime table required size.
		 *			ERROR_NOT_ENOUGH_INIT_KEYS, if share commit succeded but table does not have enough shares to restore key.
		 *			Other codes if unexpected error encountered.
		 */
		int commit_shares(SSShares* sss_runtime) const;


		/**
		 * @brief		Set user password
		 */
		void reset(cPassword const & pwd, key_info_t const & key, bc_byte_t* salt)
		{
			this->clear();
			alg_init_from_password(m_alg, key.alg_id, key.mode_id, NULL /*zero pool*/, &pwd, (char*)salt, SaltSize, key.hash_id, 1);
		}


		void clear()
		{
			m_shares.clear();
		}
	};

} // namespace kgghost


int sss_user_t::decode_shares(const bc_byte_t* kbEncrypted, unsigned int kbSize)
{
	//BC_DEBUG_PRINT(("decoding sss user keyblock sized %d bytes\n", kbSize));

	int res;
	bool found = false;

	unsigned int share_size = sizeof(sss_share_t);

	// at this point we require at least one share and a prime. will correct if will decode anything
	unsigned int required_size = PrimeAndSaltSize + share_size;
	if( required_size > kbSize )
		return ERROR_BUFFER_TOO_SMALL;
	else
		required_size = kbSize;


	SSS_KeyBlock* kb = (SSS_KeyBlock*)kbEncrypted;
	bc_byte_t* lowprime = ((SSS_KeyBlock*)kbEncrypted)->primeAndSalt;


	// current share
	sss_share_t share;
	
	// current share number in keyblock
	WORD share_num = 0;

	// try to decode our shares
	for
	(
		sss_share_t* share_ptr = kb->shares;
		(share_num + 1) * share_size <= (required_size - PrimeAndSaltSize);
		++share_ptr, ++share_num
	)
	{
		//BC_DEBUG_PRINT(("decoding share number %d\n", share_num));

		// copy share buffer and decode it inplace
		memcpy((char*)&share, share_ptr, share_size);
		res = share.decode(m_alg, share_num, lowprime);

		if (res == ERROR_INVALID_ALGORITHM)
			continue; // failed to verify this share, but maybe there are more coming
		else if (res != ERROR_NO)
			return res;

		if( !found ) 
		{
			found = true;

			// correct required size
			required_size = share.total_required_shares() * share_size + PrimeAndSaltSize;
			
			if( required_size > kbSize )
				return ERROR_BUFFER_TOO_SMALL;
			else
				required_size = kbSize;
		}

		// successfuly decoded, assign share to myself 
		res = this->assign_share(share);
		if( ERROR_NO != res ) 
			return res;
	}

	// no exceptional errors if we're here

	if( !found ) 
		return res;

	return ERROR_NO;
}


int sss_user_t::commit_shares(SSShares* sss_runtime) const
{
	pw_int x,y;
	int res;

	share_citer_t i = m_shares.begin();
	share_citer_t end = m_shares.end();
	
	const sss_share_t* cur_share;

	while(i != end)
	{
		cur_share = &(*i);
		++i;

		// add decoded share to runtime table
		if( (cur_share->commitX(x) != ERROR_NO) || (cur_share->commitY(y) != ERROR_NO) )
			return ERROR_INTERNAL_PROBLEM;

		res = SSSAddShare(sss_runtime, &x, &y);
	
		switch (res)
		{
			case pw_OK: return ERROR_NO;
			case pw_InvalidKeyBlock:	/* fall thru */
			case pw_BufferSizeTooSmall:
				continue;
			default:
				return ERROR_INTERNAL_PROBLEM;
		}
	}

	return ERROR_NOT_ENOUGH_INIT_KEYS;
}


//
//
//	CSSSProcessor implementation
//
//


CSSSProcessor::CSSSProcessor() : 
	m_sss_runtime(NULL),
	m_shares_required(0),
	m_shares_total(0)
{

}


CSSSProcessor::~CSSSProcessor()
{
	this->reset();
}


bool CSSSProcessor::decode(const bc_byte_t* kbEncoded, unsigned int & kbSize, key_info_t& key, cPassword const & init_pwd)
{
	BC_DEBUG_PRINT(("Debug: v8: Decoding SSS keyblock sized %d bytes\n", kbSize));
	int res;

	BYTE* primeLow = ((SSS_KeyBlock*)kbEncoded)->primeAndSalt;
	BYTE* salt = primeLow + PrimeAndSaltSize - SaltSize;

	// decode shares for the user of the initial password.
	// we need it to know total shares required to initialize runtime first
	sss_user_t init_user(init_pwd, key, salt);

	res = init_user.decode_shares(kbEncoded, kbSize);
	if( ERROR_NO != res )
	{
		//if( res == ERROR_INVALID_ALGORITHM ) 
			//User_Message("Incorrect initial password\n");
		BC_DEBUG_PRINT(("Debug: v8: SSS failed decoding initial user share\n"));
		return false;
	}

	
	// initialize runtime shares table
	res = SSSAllocateShares(&m_sss_runtime, init_user.get_share_sample()->total_required_shares());
	if( res != ppe_ok )
		return false;


	// initialize sss runtime prime with the first byte from decoded share and the rest from raw keyblock data
	BYTE prime[MaxPrimeSize] = {0};
	prime[0] = init_user.get_share_sample()->prime_high();
	memcpy(prime + 1, primeLow, MaxPrimeSize - 1);

	if (m_sss_runtime->prime.set(MaxPrimeSize, prime) != pw_OK )
		return false;


	// dump initial user's share to initialized runtime table
	init_user.commit_shares(m_sss_runtime);
	this->store_user(init_user);	

	cPassword pwd;

	// decode all other users until required shares number wont be reached
	while(true)
	{
		User_GetPassword("Enter additional shared password:", pwd);
		if(res != ERROR_NO) return false;

		sss_user_t cur_user(pwd, key, salt);

		if(!this->is_unique_user(cur_user))
		{
			User_Message("The same password has already been entered for the current shared scheme.\n");
			continue;
		}

		int retries = 0;
		// decode user shares from keyblock
		do
		{
			res = cur_user.decode_shares(kbEncoded, kbSize);
			if( res == ERROR_INVALID_ALGORITHM ) 
				User_Message("Additional share password is incorrect\n");
			else if( res != ERROR_NO )
				return false;				// unexpected error
		
		}
		while((res == ERROR_INVALID_ALGORITHM) && (++retries < 3));


		// commit decoded shares into sss runtime and parse result
		res = cur_user.commit_shares(m_sss_runtime);
		if( res == ERROR_NO ) 
			break;				// success, we now have enough shares to decode a key
		else if( res == ERROR_NOT_ENOUGH_INIT_KEYS ) 
		{
			this->store_user(cur_user);
			continue;			// success, but not enough shares yet, move on to the next user.
		}
		else 
			return false;			// unexpected error
	}


	// we have decoded all shares, restore encryption key info
	if ( SSSRestoreKey(m_sss_runtime, (BYTE*)&key, MaxEncodedSize) != pw_OK )
		return false;

	return true;
}


unsigned int CSSSProcessor::total_shares_ready() const
{
	unsigned int total = 0;

	user_citer_t i = m_users.begin();
	user_citer_t end = m_users.end();

	while(i != end)
	{
		total += (*i)->total_assigned_shares();
		++i;
	}

	return total;
}


bool CSSSProcessor::is_unique_user(sss_user_t const & user) const
{
	bc_byte_t sig[16] = {0};
	bc_byte_t my_sig[16] = {0};
	bc_byte_t iv[8] = {0};	

	user.m_alg->encrypt(iv, my_sig, sizeof(my_sig));

	BCAlgHandle other_alg;
	user_citer_t i = m_users.begin();
	user_citer_t end = m_users.end();
	
	while(i != end)
	{
		other_alg = (*i)->m_alg;

		other_alg->encrypt(iv, sig, sizeof(sig));
		if(memcmp(my_sig, sig, sizeof(sig)) == 0)
			return false; // we got a duplicate

		memset(sig, 0, sizeof(sig));
		++i;
	}

	return true;
}


void CSSSProcessor::reset()
{
	user_citer_t i = m_users.begin();
	user_citer_t end = m_users.end();
	
	while(i != end)
	{
		delete *i;
		++i;
	}

	m_shares_required = 0;
	m_shares_total = 0;

	if(m_sss_runtime) SSSFree(&m_sss_runtime);
}


void CSSSProcessor::store_user(sss_user_t const & user)
{
	sss_user_t* new_user = new sss_user_t(user);
	m_users.push_back(new_user);
}

//
//
//	REFACTOR
//
//




#if 0
int sss_share_t::encode(bc_byte_t* oBuff, unsigned int* ioBuffSize, CBCAlg_Handle alg, int hash_alg_id, WORD share_num, bc_byte_t* primeBuffer)
{
	unsigned int size = sizeof(*this);
	bc_byte_t iv[IVECTOR_LENGTH_DB] = {0};
	int rc;

	BC_ASSERT(*ioBuffSize >= size);
	if(*ioBuffSize < size)
		return ERROR_INTERNAL_PROBLEM;
			

	// fill share digest
	rc = this->calculate_digest(primeBuffer, hash_alg_id);
	if (rc != ERROR_NO)
		return rc;



	// Encode share by encrypting it into buffer
	// Use share number as initial vector
	iv[0] = LOBYTE(share_num);
	iv[1] = HIBYTE(share_num);

	memcpy(oBuff, (const char*)this, size);
	if ( !alg->encrypt(iv, oBuff, size) )
		return ERROR_INTERNAL_PROBLEM;

	*ioBuffSize = size;

	return ERROR_NO;
}
#endif

//
//	encodeShares
//	 0. Checks for different passwords in itemT table
//	 1. Generates prime p one byte more than the key size
//	 2. Generates (recNum - 1) degree polinom P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= recNum -1,
//	 3. Generates shares array by Xi and Yi where P(Xi) == Yi, the array quantity is Summ of itemT[i].sharesQ, where 0 <= i < itemQ
//	 4. For every i gets N = itemT[i].sharesQ shares from generated shares array and encrypts the shares by passwords from itemT[i].pwd
//
//		Encrypted item format is:
//			encrypted:
//				Ver + RN + PrimeAlderByte + Xi + Yi + hashAlgId + reservedBytes + digest
//		where:
//			Ver ( 1 byte ) is version number
//				Ver 1 has 152 bytes size of encrypted share ( see getShareSize() )
//			RN  ( 1 byte ) is number of required shares.
//			PrimeAlderByte - prime is saved in open state but it's alder bytes has assigned alder bit,
//							it is a reason why the byte was hidden to encrypted part.
//			The values is required to recovering key,
//			but it has to be encrypted, so we save it here.
//
//			reservedBytes is random filled value it used for normal encryption with any block based algorithm 
//
//		digest size is defined as MAXIMUM_DIGEST_SIZE_BYTES
//
//		prime and Xi Yi sizes is defined by key size, see MaxPrimeSize
//
//
//		Maximum estimated shares size for 3K block:
//			V1 is -> 70 + N * ( sizeof( sShareV1 ) == 192 ) -> N = ( 3K - 70 ) ) / 192 > 15 shares
//			V2 is -> 70 + N * ( sizeof( sShareV2 ) == 112  )-> N = ( 3K - 70 ) ) / 112 > 26 shares
//
//
//	encodeSharesEx
//		walks tables with share's passwords ( itemT )
//			1. creates symmetric algorithm from password
//			2. fills shares and calculates it's digest
//				the digest is started from plain text primes low bytes 
//				and finished by prepared shares, so we may check the prime integrity
//				together with the share in decoding porcess.
//			3. encodes corresponded share quantity from generated shares table ( sss )
//
#if 0
int encode_shares(char* oBuff, unsigned int* ioBuffSize, SSShares* sss_runtime)
{
	// get size of the single share
	unsigned int shareSize = sizeof(share_type);
	if ( shareSize <= 0 )
		return ERROR_INCORRECT_FORMAT_VERSION;


	unsigned int totalShares = this->total_shares_ready();
	unsigned int requiredSize = primeAndSaltSize + totalShares * shareSize;
	
	if ( *pBuffSize < requiredSize )
	{
		*pBuffSize = requiredSize;
		return ERROR_BUFFER_TOO_SMALL;
	}


	*pBuffSize = requiredSize;

	unsigned int primeSizeBytes = pw_getSizeInBytes(&sss_runtime->prime);
	if ( buffSize < primeSizeBytes )
		return ERROR_INTERNAL_PROBLEM;


	// Symmetric algorithm for share encryption, key is generated from corresponded shares' password.
/*
	cSymAlgorithm sPwdAlg;

	if ( !sPwdAlg.Init( algId, encModeId ) )
	{
		assert( false );
		return ERROR_INVALID_ALGORITHM;
	}

	int pwdKeySizeBytes = ( sPwdAlg.m_keyLengthInBites + 7 ) / 8;
*/

	//	Writes prime's low bytes to buffer (where the password key will be stored later )
	//	and save alder byte to be hidden to encrypted part.

	int buffSize = max( pwdKeySizeBytes, primeAndSaltSize + 1 );

	// Deploys prime low bytes to buffer start and saves alder byte 
	cPlainTextBlock ptb;
	if ( ptb.Allocate( buffSize ) == NULL )
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}



	pw_int2bytes( ptb.buffer, primeSizeBytes, &sss->prime );

	BYTE *pAlderByte = &ptb.buffer[0];

	int pos = 0;
	memcpy( oBuff, ptb.buffer + 1, primeSizeBytes );
	pos += primeAndSaltSize;



	// Prepare buffer for plain text share
	cPlainTextBlock ptbShare;
	if(ptbShare.Allocate(shareSize) == NULL)
		return ERROR_INTERNAL_PROBLEM;

	share_type* share = (share_type*)ptbShare.buffer;
	share->version		= this->share_version<share_type>();
	share->reqNum		= this->shares_needed();
	share->totalQ		= (BYTE)totalShares;
	share->headOfPrime	= *pAlderByte;

	if ( primeAndSaltSize - saltSize < 0 )
		return ERROR_INTERNAL_PROBLEM;


	BYTE *salt = oBuff + primeAndSaltSize - saltSize;

	unsigned int itemN = 0;	// position in itemT
	unsigned int sssN = 0;	// position in sss

	CBCAlg_Handle alg;

	for(unsigned int i; i < m_shares_total; ++i)
	{

		res = alg_init_from_password(alg, m_algId, m_encModeId, NULL /*zero pool*/, &item[i].pwd, salt, saltSize, m_hashAlgId, iteration);
		if( ERROR_NO != res ) return res;		

		res = this->encode_single_share < share_type > (share, sss_runtime, alg, );
		if( ERROR_NO != res ) return res;		
/*

		for( int i = 0; i < itemT[ itemN ].numberOfShares; i++, sssN++ )
		{
			if ( sssN >= (int)sss->arraySize )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			// Fills x and y values

			if ( !fillValue( sv->valX, sizeof( sv->valX ), &sss->array[ sssN ].x )  ||
					!fillValue( sv->valY, sizeof( sv->valY ), &sss->array[ sssN ].y ) 
					)
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			if ( pos + shareSize > requiredSize )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}
			
			// Calculates digest of
			//	1. plain text part of prime.
			//	2. the share info exluding the digest part.
			int err = sssCalculateDigest( buff, sv, (BYTE)hashAlgId );

			if ( err != ERROR_NO )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			// Encode share directly to represent buffer
			// It uses number of share as initial vector

			iv[0] = LOBYTE( WORD( sssN ) );
			iv[1] = HIBYTE( WORD( sssN ) );

			if ( !sPwdAlg.encrypt( iv, buff + pos, ( BYTE * )sv, shareSize ) )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			pos += shareSize;
		}

*/
	}

	if ( pos != requiredSize )
		return ERROR_INTERNAL_PROBLEM;


	return ERROR_NO;
}
#endif

bool CSSSProcessor::encode(bc_byte_t* kbPlain, unsigned int & kbSize, key_info_t const & key, cPassword const & init_pwd)
{
#if 0
	SSShares *sss_runtime = NULL;

	unsigned int totalSharesQ = this->get_total_shares();

	int res = SSSAllocateShares( &sss_runtime, totalSharesQ );
	if ( res != ppe_ok )
		return ERROR_INTERNAL_PROBLEM;
/**
	if ( version == (BYTE)-1 )
		version = defaultSSSVersion;
*/
/**
	// Version 2 uses one byte x values, so theirs are generated from order number
	if ( ( version == 2 ) && 
			!prepareOrderedX( s_table ) )
	{
		SSSFree( &s_table );
		return false;
	}
*/

	res = this->share_coder()->prepareX(sss_runtime);
	if( ERROR_NO != res ) 
	{
		SSSFree(&sss_runtime);
		return res;
	}

	int seedSize = 0;

	res = SSSGenerateShares(sss_runtime, totalSharesQ, keyBuff, keyBuffSize, NULL, &seedSize);
	if ( res != ppe_randomBufferRequired )
	{
		SSSFree(&sss_runtime);
		return false;
	}


	//
	//	Check if seed value has been generated early.
	//
	if ( seedSize > ptbSeed.bufferSize )
	{
		if ( ptbSeed.Allocate( seedSize ) == NULL )
		{
			SSSFree(&sss_runtime);
			return false;
		}

		res = User_GenerateSeed(ptbSeed.buffer, seedSize);
		if ( res != ERROR_NO )
		{
			SSSFree(&sss_runtime);
			return false;
		}
	}

	BYTE *seed = ptbSeed.buffer;

	//cProgressDlg_ShareGen dlgMT( s, reqQ, keyBuff, keyBuffSize, seed, seedSize );
	//dlgMT.DoModal( hMainInst,IDD_WaitShareGenerationDlg, hWnd );

	//res = dlgMT.m_res;

	if ( res != ppe_ok )
	{
		SSSFree( &s_table );
		return ERROR_INTERNAL_PROBLEM;
	}


	res = this->share_coder()->encode_shares( version, kbEncrypted, &kbEncryptedSize, itemT, itemQ, sss_runtime, m_algId, m_encModeId, m_hashAlgId, reqQ );

	SSSFree(&sss_runtime);

	return res;
#endif
	return ERROR_INCORRECT_FORMAT_VERSION; // stub
}
#if 0
int CSSSProcessor::getSeedSize( int keyBuffSize )
{
	SSShares *s = NULL;
	int totalSharesQ = getTotalSharesQ();

	int res = SSSAllocateShares( &s, itemQ );
	if ( res != ppe_ok )
		return -1;

	if ( version == (BYTE)-1 )
		version = 2;

	// Version 2 uses one byte x values, so theirs are generated from order number
	if ( ( version == 2 ) && 
			!prepareOrderedX( s ) )
	{
		SSSFree( &s );
		return -1;
	}

	int seedSize = 0;

	res = SSSGenerateShares( s, totalSharesQ, NULL, keyBuffSize, NULL, &seedSize );
	SSSFree( &s );

	if ( res != ppe_randomBufferRequired )
	{
		return -1;
	}

	return seedSize;
}
#endif









#if 0

class cProgressDlg_ShareGen : public cProgressBlock, public cDlg
{
	SSShares *m_s;
	int m_reqQ;
	BYTE *m_keyBuff;
	int m_keyBuffSize;
	BYTE *m_seed;
	int m_seedSize;

	bool m_cancelled;
	HWND m_hProgress;
//	bool m_useMarqueeMode; 

public:
	int m_res;

	cProgressDlg_ShareGen( SSShares *s, int reqQ, BYTE *keyBuff, int keyBuffSize, BYTE *seed, int seedSize )
	{
		m_s = s;
		m_reqQ = reqQ;
		m_keyBuff = keyBuff;
		m_keyBuffSize = keyBuffSize;
		m_seed = seed;
		m_seedSize = seedSize;
		m_res = -1;
		m_cancelled = false;
	}

	BOOL InitDialog()
	{
		m_hProgress = GetDlgItem( m_hDlg, IDC_PROGRESS );

		SendMessage( m_hProgress, PBM_SETRANGE, 0, MAKELPARAM( 0, 100 ) );
//		m_useMarqueeMode = SendMessage( m_hProgress, (WM_USER+10), true, 0 ); // PBM_SETMARQUEE
		SendMessage( m_hProgress, PBM_SETSTEP, 1, 0 );

		DWORD threadId;
		HANDLE hThread = CreateThread( NULL, 0, threadFunc, this, 0, &threadId );

		if ( hThread == NULL )
		{
			MessageBoxT( m_hDlg, L"Error in thread creating", LoadStringOnceT( IDS_KeyGen ), MB_ICONSTOP );
			EndDialog( m_hDlg, -1 );
			return false;
		}

		CloseHandle( hThread );

		return TRUE;
	}

	int Generate()
	{
		m_res = SSSGenerateShares( m_s, m_reqQ, m_keyBuff, m_keyBuffSize, m_seed, &m_seedSize, this );
		SendMessage( m_hProgress, PBM_SETPOS, true, 0 );
		return m_res;
	}

	static DWORD WINAPI threadFunc( void *ptr )
	{
		cProgressDlg_ShareGen *pThis = ( cProgressDlg_ShareGen * )ptr;

		pThis->Generate();

		PostMessage( pThis->m_hDlg, WM_COMMAND, IDOK, 0 );

		return 0;
	}
	
	void SetText( const char *text )
	{
		SetDlgItemText( m_hDlg, IDC_Text, text );
	}

	// Progress test result enum
	// enum { stepResult_False_ByTable = -1, stepResult_False_ByPrimeTest = 0, stepResult_True = 1 };
	
	//
	// SetCurrentStepResult
	//	Reports result of every step, see enum above.
	//	Returns false to stop the process, otherwise returns non zero.
	//
	bool SetCurrentStepResult( int res )
	{
		if ( m_cancelled )
			return false;

//		if ( res == stepResult_False_ByPrimeTest )
//			SendMessage( m_hProgress, PBM_SETPOS, 0, 0 );
//		SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

		if ( res == stepResult_False_ByPrimeTest )
			SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

		return true;
	}

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		switch (uMsg) 
		{ 
			case WM_INITDIALOG:
				return InitDialog();

	        case WM_COMMAND: 
			{
				switch(wParam) 
				{ 
					case IDOK: 
						EndDialog( m_hDlg, m_res );
						return TRUE; 
					case IDCANCEL:
					{
						int res = MessageBoxT(
									m_hDlg,
									LoadStringOnceT ( IDS_Q_StopGenerationProcess ),
									LoadStringTwiceT( IDS_KeyGen ),
									MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2 );

						if ( res != IDYES )
								return 0;

						m_cancelled = true;
						EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), false );
						return TRUE;
					}
					default:
						return 0; 
				}
			}
			default:
				return 0; 
		}

		return 0;
	}
};

#endif
