#include "container.h"
#include "kgghost.h"
#include "layout.h"
#include "keyblock.h"
#include "user.h"

#include <alg/alg.h>
#include <alg/manager.h>
#include <common/log.h>
#include <common/memory.h>

#include <backend/fs.h>

#include <string.h>

using namespace bdk;
using namespace kgghost;

class V8Unsupported : public BCException
{
public:
        V8Unsupported(const char* what_is)
        {
                m_msg.assign(what_is);
        }
};

BCContainerViewHandle V8Format::init_view( IBCContainerIO* io, key_info_t const & kinfo, BCAlgHandle alg )
{
        if(kinfo.is_for_hidden_part())
                throw V8Unsupported("Current version does not support kg-ghost hidden part");

	bc_space_unit_t space;
	space.offset = kinfo.cont_info.offset;
	space.length = kinfo.cont_info.length;

	bcu32_t flags = BC_CRYPTO_IV_64BIT;

	return BCContainerView::init(alg, space, flags, false);
}


BCContainerViewHandle V8Format::create_view()
{
        throw V8Unsupported("Current version does not support creating kg-ghost passwords");
	//return BCContainerView::invalid_view(); // stub
}


BCContainerViewHandle V8Format::create_hidden_view(IBCContainerIO* io, bc_ulong_t nbytes)
{
        throw V8Unsupported("Current version does not support creating kg-ghost hidden parts");
	//return BCContainerView::invalid_view(); // stub
}


BCContainerViewHandle V8Format::get_view(IBCContainerIO* io)
{
	// check data block integrity
	DWORD err = DataBlock_Check(&m_db, m_total_size);
	if ( err != ERROR_NO )
	{
		BC_DEBUG_PRINT(("Debug: v8: datablock check failed: %d\n", err));
		return BCContainerView::invalid_view();
	}

	key_info_t kinfo;
	BCAlgHandle handle;

	memset(&kinfo, 0, sizeof(kinfo));
	// ask initial password
	cPassword pwd;
	err = User_GetPassword("Enter initial password:", pwd);
	if (ERROR_NO != err)
	{
		BC_DEBUG_PRINT(("Debug: v8: failed asking password: %d\n", err));
		return BCContainerView::invalid_view();
	}


	// if header is not encrypted, fill out key info directly
	if( !m_db.header()->maybe_encrypted() )
	{
		BC_DEBUG_PRINT(("Debug: v8: header is not encrypted, constructing key info\n"));
		kinfo.alg_id = m_db.header()->contAlgId;
		kinfo.mode_id = m_db.header()->encModeId;
		kinfo.hash_id = m_db.header()->hashAlgId;

		err = DataBlock_GetKeyHandle(&m_db, m_total_size, pwd, handle, kinfo);
		if (ERROR_NO != err)
		{
			bc_get_log(BC_MSG_CRIT) << "v8: failed authorizing: " << err;
			return BCContainerView::invalid_view();
		}

		BC_DEBUG_PRINT(("Debug: v8: got cont_info length %llu, offset %llu\n", kinfo.cont_info.length, kinfo.cont_info.offset));
	}
	// else try every possible config
	else
	{
	        return BCContainerView::invalid_view();
#if 0
	        BCAlgList algs;
	        BCModeList modes;
	        BCHashList hashes;

	        BCAlgService* cur_alg;
	        bc_core()->alg_manager()->list_supported_algorithms(algs);
	        BCAlgList::iterator alg_i = algs.begin();
	        BCAlgList::iterator alg_end = algs.end();

		while(alg_i != alg_end)
		{
		        cur_alg = *alg_i++;
		        kinfo.alg_id = cur_alg->id();

		        BCAlgMode* cur_mode;
		        cur_alg->list_supported_modes(modes);
		        BCModeList::iterator mode_i = modes.begin();
		        BCModeList::iterator mode_end = modes.end();

		        while(mode_i != mode_end)
		        {
                                cur_mode = *mode_i++;
                                kinfo.mode_id = cur_mode->id();

				for(unsigned int cur_hash = 0; cur_hash < hashes.total(); ++cur_hash)
				{
					kinfo.hash_id = hashes[cur_hash];

					BC_DEBUG_PRINT(("Debug: v8 decrypt header: trying next config: alg %s, mode %s, hash 0x%x\n",
						cur_alg->name(),
						cur_mode->name(),
						kinfo.hash_id));

					err = DataBlock_GetKeyHandle(&m_db, m_total_size, pwd, handle, kinfo);
					if (ERROR_NO != err)
						continue;
					else
					{

					        BC_DEBUG_PRINT(("Debug: valid config found: alg 0x%x, mode 0x%x, hash 0x%x\n",
					                kinfo.alg_id,
	                                                kinfo.mode_id,
	                                                kinfo.hash_id));
					        break;
					}

					// try to decrypt header with verified config
					if(m_db.header_eblock.decrypt_to(m_decrypted_header, handle, kinfo.hash_id))
						break;

					break;
				}
			}
		}
#endif
	}


	return this->init_view(io, kinfo, handle);
}


void V8Format::serialize(IBCContainerIO* dest) const
{
        BC_ASSERT(this->m_total_size >= _4KB);
        bc_ulong_t nwritten;

        // write first 4kb
        nwritten = dest->write_bytes((bc_byte_t*)&m_db, _4KB, 0);
        BC_ASSERT(nwritten == _4KB);

        if(m_total_size > _4KB)
        {
                bc_ulong_t ext_size = m_total_size - _4KB;
                bc_ulong_t ext_offset = dest->size() - ROUND(ext_size, _4KB);
                nwritten = dest->write_bytes((bc_byte_t*)&m_db + _4KB, ext_size, ext_offset);
                BC_ASSERT(nwritten == ext_size);
        }
}


BCContainerViewHandle V8Format::update_view(IBCContainerIO* io, bc_view_change_t op)
{
        throw V8Unsupported("Current version does not support changing kg-ghost passwords");
	//return BCContainerView::invalid_view(); // stub
}


static bool is_valid_signature(const char* sig)
{
	return (strncmp(sig, DB_HEADER_SIGNATURE, strlen(DB_HEADER_SIGNATURE)) == 0);
}


static bool is_valid_label(const char* label)
{
	return (strncmp(label, DB_HEADER_VOLUME_LABEL, strlen(DB_HEADER_VOLUME_LABEL)) == 0);
}


static bool is_valid_kgid( DWORD kgid )
{
        return (kgid == KEYGEN_ID);
}


bool V8Factory::parse(bdk::IBCContainerIO* src)
{
        BCStaticBuffer < DB_MIN_SIZE > buf;
	bc_ulong_t nread;

	nread = src->read_bytes(buf, sizeof(buf), 0);
	if( nread != sizeof(buf) )
	{
		BC_DEBUG_PRINT(("v8 factory: parsing failed: incorrect format size\n"));
		return false;
	}

	DATA_BLOCK* db = (DATA_BLOCK*)(buf);
	db_header_t* header = db->header();

	if( !is_valid_signature((const char*)header->signature) )
	{
		BC_DEBUG_PRINT(("v8 factory: parsing failed: invalid signature\n"));
		return false;
	}

	if( !is_valid_label((const char*)header->volumeLabel) )
	{
		BC_DEBUG_PRINT(("v8 factory: parsing failed: invalid volume label\n"));
		return false;
	}

	if( !is_valid_kgid(header->keyGenId) )
	{
		BC_DEBUG_PRINT(("v8 factory: parsing failed: invalid keygen id (0x%x)\n", header->keyGenId));
		return false;
	}

	// recognized
	return true;
}


/*
 *	We serialize DATA_BLOCK inside container file the following way:
 *	The first 4kb are always placed in the file header:
 *		1024 header data + 512 pool data + 10 keyblocks (the "primary keyblock", 10 * 256 bytes) = 4096 bytes.
 *	The rest of the keyblock up to DB_KEYBLOCK_SLOTS_CAPACITY is allocated on demand by 4kb blocks
 *	and attached to the end of file:
 *		4kb + 4kb + 4kb + 4kb ("extended keyblock").
 *	The total possible max size of extended keyblock is more than DATA_BLOCK size,
 *	so the last 4kb extended block is read\written only partly, but serialized as a 4kb block anyway.
 */
IBCContainerFormat* V8Factory::load(bdk::IBCContainerIO* src)
{
	// serialized DATA_BLOCK is rounded to 4kb boundary thus bigger than actual
        BCStaticBuffer < ROUND(sizeof(DATA_BLOCK), _4KB) > db_buf;
	DATA_BLOCK* db = (DATA_BLOCK*) db_buf;
	bc_ulong_t nread;

	BC_ASSERT(_4KB >= DB_MIN_SIZE);

	/*
	 * 	Read primary keyblock
	 */

	nread = src->read_bytes(db_buf, _4KB, 0);
	if(nread != _4KB)
	{
		BC_DEBUG_PRINT(("v8 factory: parsing failed: incorrect format size\n"));
		return NULL;
	}

	/*
	 *	Calculate extended keyblock layout
	 */

	unsigned int ext_blocks_4kb;

	// if header is suspected as encrypted we have to read maximum availiable extended keyblock from the file's tail
	if(db->header()->maybe_encrypted())
	{
	        return NULL; // stub

		//ext_blocks_4kb = src->size() / _4KB - 1;
		//if(ext_blocks_4kb > 4) ext_blocks_4kb = 4;
	}
	// if header is open, deduce the number of extended slots from data size.
	else
	{
		ext_blocks_4kb = (src->size() - db->header()->layout.length) / _4KB - 1;
	}

	/*
	 *	Read extended keyblock and init container
	 */

	bc_ulong_t ext_size = ext_blocks_4kb * _4KB;
	bc_ulong_t ext_offset = src->size() - ext_size;
	bc_ulong_t ext_remainder = (ext_blocks_4kb > 0) ? sizeof(db_buf) - sizeof(DATA_BLOCK) : 0;
	bc_byte_t* ext_db = db_buf.ptr(_4KB);

	nread = src->read_bytes(ext_db, ext_size, ext_offset);
	if(nread != ext_size)
	{
		BC_DEBUG_PRINT(("Debug: v8: failed reading extended keyblock data\n"));
		return NULL;
	}

	BC_DEBUG_PRINT(("Debug: v8: total data block size %llu, extended key blocks %llu\n",
			(ext_size + _4KB - ext_remainder), ext_blocks_4kb));

	IBCContainerFormat* fmt = NOTHROW_NEW V8Format(*db, (ext_size + _4KB) - ext_remainder);
        if( !fmt )
	{
                return NULL;
	}

        return fmt;
}


IBCContainerFormat* V8Factory::make(bdk::bc_container_info_t const & info)
{
        throw V8Unsupported("Current version does not support creating kg-ghost containers");
/*
	DATA_BLOCK db;

	ShredData((bc_byte_t*)&db, sizeof(DATA_BLOCK)); // fills DataBlock by random garbage
	db_header_t* hdr = db.header();

	hdr->normalize(); // fill unstable fields by default values

	hdr->version 	= DB_FORMAT_VERSION;
	hdr->keyGenId 	= KEYGEN_ID;
	hdr->contAlgId 	= info.cont_alg_id;
	hdr->encModeId 	= info.cont_mode_id;
	hdr->hashAlgId 	= info.cont_hash_id;
	hdr->layout.offset = _4KB;
	hdr->layout.length = info.cont_size;

	memset( hdr->description, 0, sizeof(hdr->description) );

	size_t desc_len = info.cont_description.size();
	if( desc_len != 0 )
	{
		memcpy(hdr->description, info.cont_description.c_str(), (desc_len > (CONTAINER_DESC_LENGTH - 1)) ? (CONTAINER_DESC_LENGTH - 1) : desc_len );
	}

	IBCContainerFormat* fmt = NOTHROW_NEW V8Format(db, _4KB);
        if( !fmt )
	{
                return NULL;
	}

        return fmt;
*/
}


