#ifndef V8_DATA_BLOCK_H
#define V8_DATA_BLOCK_H

//#include "enc_block.h"

#include "layout.h"
#include "enc_block.h"
#include "CertificateLib/SMemory/cPassword.h"

#include <common/build.h>
#include <common/types.h>

#define DB_FORMAT_VERSION_2			2
#define DB_FORMAT_VERSION			3

#define CONTAINER_ID_LENGTH			4
#define CONTAINER_DESC_LENGTH			25

#define DB_HEADER_SIGNATURE 			"LOCOS94 "
#define DB_HEADER_VOLUME_LABEL			"BC_KeyGenID"

#define KEYGEN_ID    0x00000004


//
//	Data block layouts
//


/* keyblock slot size in bytes */
#define DB_KEYBLOCK_SLOT_SIZE			256 

/* datablock random data pool size in bytes */
#define DB_POOL_SIZE				512 

/* maximum key map size in bytes */
#define DB_KEY_MAP_SIZE				512

/* maximum keyblock slots capacity */ 
#define DB_KEYBLOCK_SLOTS_CAPACITY		DB_KEY_MAP_SIZE / sizeof(slot_t)

#define DB_HEADER_RESERVED_SIZE			1024

#define DB_MIN_SIZE				(DB_HEADER_RESERVED_SIZE + DB_POOL_SIZE)

#define DB_MAX_SIZE				(DB_MIN_SIZE + DB_KEYBLOCK_SLOTS_CAPACITY * DB_KEYBLOCK_SLOT_SIZE)


namespace kgghost
{

	/*
	 *	Keyblock is a data buffer occupied by encryption key info encoded through different methods (PBE, SSS, PKE).
	 *	Keyblocks in DATA_BLOCK are stored in individual fixed-sized slots in key tab (primary and extended).
	 *	If keyblock spans several slots, they will be allocated adjacently.
	 *
	 *	Keyblock map is an index map of slots that keyblocks are contained in.
	 *	Hidden part keyblocks never show up in keyblock map, which means we need to virtually update it when adding keys.
	 */


	/**
	 * @brief 	Type of a keyblock slot.
	 */
	enum slot_type_t
	{
		SLOT_TAIL 	= -1, 	/** < This slot is a tail of a larger keyblock descriptor's data */
		SLOT_EMPTY 	= 0,  	/** < Vacant slot */
		SLOT_HEAD_PBE 	= 1,  	/** < Slot stores the head of the Password Based Encryption encoded key */
		SLOT_HEAD_PKE 	= 2,  	/** < Slot stores the head of the Public Key Encryption encoded key */
		SLOT_HEAD_SSS 	= 3,  	/** < Slot stores the head of the Shamir Secret Sharing scheme encoded key */
	};


	/**
	 * @brief	Slot map entry.
	 */
	BC_PACKED(
	struct slot_t
	{
		/* total bytes occupied by the keyblock whose part is stored here */
		bcu16_t		kb_size;	

		/* type of the slot */
		bcu16_t		type;

		/* Optional parameter depending on the type */
		bcu32_t		key_param;
	});


	/**
	 * @brief	The type of the logical keyblock.
	 */
	enum keyblock_type_t
	{
		KEYBLOCK_UNDEFINED	= SLOT_EMPTY,	/** < Undefined key type. Used when map head slot is unavailiable */
		KEYBLOCK_PBE		= SLOT_HEAD_PBE,/** < Password Based Encryption encoded key */
		KEYBLOCK_PKE		= SLOT_HEAD_PKE,/** < PKE encoded key */
		KEYBLOCK_SSS		= SLOT_HEAD_SSS,/** < SSS encoded key */
	};



	/**
	 * @brief	Encapsulates keyblock slots' position in key tab.
	 *
	 * @more	Used as an adressing map in key tab storage. 
	 *		Most operation will require descriptor and key tab buffer passed together.
	 *		The latter acts as a slot location template for the former.
	 *
	 * @more	There are two ways to obtain descriptors:
	 *		1) Using key map. In this case the descriptor will be set to the type of key stored in keyblock's head slot.
	 *		2) Manually. This is required when key map is encrypted or we are looking for the hidden keyblocks that are not marked on the map.
	 *
	 * @more	Hidden keyblock descriptors address hidden part encoded keys. 
	 *		They still have their key type set but not recorded to the actual map.
	 */
	struct keyblock_descriptor_t
	{

		/* first element index */
		unsigned int	base_slot;

		/* total elements including first */
		unsigned int	total_slots;

		/* keyblock type. undefined for manually parsed keyblocks*/
		bcu16_t		type;

		/* hidden keyblock flags */
		bool		hidden;


		/** 
		 * @brief	Construct empty key descriptor 
		 */
		keyblock_descriptor_t();


		/**
		 * @brief	Construct key descriptor from range of keyblocks.
		 */
		keyblock_descriptor_t(unsigned int base, unsigned int total, bcu16_t type, bool hidden = false);


		/**
		 * @brief	Is this descriptor addresses empty area
		 */
		bool 		is_empty() const { return total_slots == 0; }


		/**
		 * @brief	If type of the keyblock is not defined
		 */
		bool		is_undefined() const { return type == KEYBLOCK_UNDEFINED; }


		/**
		 * @brief	Is this keyblock hidden?
		 */
		bool		is_hidden() const { return hidden; }

	};


	/**
	 * @brief	Keyblock map.
	 * @more	It is _not_ meant to encapsulate key map format from DATA_BLOCK.
	 *		Instead it provides useful utilities to parse simplest case - defined key map.
	 */
	struct keyblock_map_t
	{
		/* the map itself */
		slot_t	map[DB_KEYBLOCK_SLOTS_CAPACITY]; // keyblock slots map
	

		//
		//	Simplest case utilities.
		//


		unsigned int 		capacity() const
		{
			return sizeof(map);
		}


		/**
		 * @brief		Get a descriptor for a keyblock by slot number
		 *
 		 * @p kb		Any keyblock that belongs to the key chain.
		 *
		 * @return		Required range or an empty range if keyblock does not identify any key chain.
		 */
		keyblock_descriptor_t	parse_keyblock_at(unsigned int kb);


		/**
		 * @brief		Record key descriptor to map.
		 *
		 * @p kd		Descriptor to set.
		 */
		void			set_keyblock(const keyblock_descriptor_t& kd);


		/**
		 * @brief		Find a range of adjacent empty key blocks of required size
		 *
		 * @p kb_num		Number of keyblocks required
		 *
		 * @return 		Range of keyblocks found as close to the beginning as possible. Empty range otherwise.
		 */
		keyblock_descriptor_t 	get_free_keyblock(unsigned int size);
	

		/**
		 * @brief		Find a maximum sized empty region of keyblocks
		 *
		 * @return		Biggest range of empty blocks in this keyblock map or empty range if there is no vacant space at all.
		 */
		keyblock_descriptor_t 	get_max_free_keyblock();


		/**
		 * @brief		Check that map has required vacant space.
		 *
		 * @p kb_num		Number of vacant keyblocks to search for
		 */
		bool 			check_free_keyblock(unsigned int kb_num);
		

		keyblock_descriptor_t 	get_free_keyblock_from(unsigned int pos);
	};


	/**
	 * @brief	Describes container metadata and key map.
	 *
	 * @more	As seen in DATA_BLOCK definition, header can be stored in encrypted form as CEncBlock serialized data.
	 */
	BC_PACKED(
	struct db_header_t
	{
		bcu8_t		busiedFlag[ 3 ];			// BootRecord's jmpCode[3]
		bcu8_t		signature[8];				// BootRecord's OEMid 'LOCOS94 '
		bcu8_t		containerID[CONTAINER_ID_LENGTH]; 	// See DataBlockGetContainerId
		bcu8_t		busiedName[ 28 ];			// 
		bcu8_t		volumeLabel[11];			// "BC_KeyGenID" for this version
		bcu32_t		keyGenId;				// Key generator identifier
		bcu32_t		version;				// Container version
		bcu16_t		description[CONTAINER_DESC_LENGTH];	// Same as old version description
	
		space_unit_t	layout;					// container layout
	
		bcu32_t		contAlgId;				// The algorithm used for container encryption
		bcu32_t		encModeId;				// Encryption mode for container algorithm
		bcu32_t		hashAlgId;				// Hash algorithm used for encryption keys generation

		keyblock_map_t	keyMap; // keyblock slots map

		
		/**
		 * @brief	Normalize unstable fields to default values
		 * @more	For use with CEncBlock digest generation and common initialization.
		 *		Will destroy some fields, so be prepared with a copy.
		 */
		void normalize();


		/**
		 * @brief	Is this header encrypted.
		 * @more	Can (and will) assume an invalid header as encrypted.
		 *		Decrypt and parse the header with the correct secret to know for sure later.
		 */
		bool maybe_encrypted() const
		{
				return (strncmp((const char*)volumeLabel, "BC_KeyGenID", sizeof(volumeLabel)) != 0);
		}
	
	});


	/**
	 * @brief	Data Block encrypted header block type stored on media
	 */
	typedef CEncBlock<db_header_t> db_header_eblock_t;


	/**
	 * @brief	Complete definition of v8 container data block
	 */
	struct DATA_BLOCK
	{
		/* encrypted block */
		db_header_eblock_t			header_eblock;

		/* reserved header space to 1024 boundary */
		bc_byte_t				reserved_1K[1024 - sizeof(db_header_eblock_t)];

		/* random pool */
		bc_byte_t				pool[ DB_POOL_SIZE ];

		/* key tab */
		bc_byte_t				key_tab[DB_KEYBLOCK_SLOTS_CAPACITY][DB_KEYBLOCK_SLOT_SIZE];


		const db_header_t* header() const
		{
			return header_eblock.type();
		}

		
		db_header_t* header()
		{
			return header_eblock.type();
		}
	};


	struct key_info_t;


	/**
	 * @brief	Allocates memory for DATA_BLOCK structure 
	 * @more	The procedure which calls DataBlockAllocate is responsible for releasing the memory with the DataBlockFree function.
	 *		Will fill newly allocated DATA_BLOCK by random data.
	 */
	//BOOL DataBlock_Allocate( DATA_BLOCK** ppDataBlock, DWORD* pDataSize);


	/**
	 * @brief	Simple shreds data block upto dbSize and fills signature if bEncryptIt id false.
	 */
	//BOOL DataBlock_Init(DATA_BLOCK* db, DWORD dbSize, DWORD algId, DWORD encModeId, DWORD hashAlgId,	space_unit_t* pContInfo);


	/**
	 * @brief 	Verifies that DATA_BLOCK structure is not damaged.
	 */
	DWORD DataBlock_Check(DATA_BLOCK* DataBlock, DWORD DataSize);


	/**
	 * @brief	Checks if exists space for new key block in the datablock
	 */
	//DWORD DataBlock_CheckFreeSpace(DATA_BLOCK* dataBlock, DWORD requiredSize, DWORD* pMaxFreeSize);


	/**
	 * @brief	Free and shred memory allocated for data block
	 * @more 	If dataSize is 0 the block is freed without shreding
	 */
	//BOOL DataBlock_Free( DATA_BLOCK **pDataBlock, DWORD *pDataSize = 0 );


	unsigned int DataBlock_TotalKeyBlocks(DWORD dbSize);


	/**
	 * @brief		Checks that keyblock params are appliable to decode any entries in data block, i.e. that some keyblock is encoded with those params.
	 *
	 * @p db		Data block
	 * @p dbSize		Size fo data block
	 * @p kd_in_out		In: expected encryption key ids (algorithm, hash, mode,...).
	 *			Out: real key ids, size and data.
	 *
	 * @return		ERROR_NO and out parameters if verification succeded, 
	 *			ERROR_INVALID_PASSWORD if not,
	 *			other values in case of exceptional errors (data block damage, memory failures, etc).
	 */
	DWORD DataBlock_VerifyKeyblock(DATA_BLOCK* db, DWORD dbSize, const cPassword& pwd, key_info_t& ki_in_out);


	/**
	 * @brief		Get an algorithm handle configured with the key stored in data block
	 *
	 *
	 *
	 *
	 */
	DWORD DataBlock_GetKeyHandle(
			DATA_BLOCK		*dataBlock,
			DWORD			dbSize,
			const cPassword&	pwd,
			bdk::BCAlgHandle&	out_handle,	
			key_info_t&		ki_in_out
		);

	/**
	 * @brief		Writes new key to Data Block.
	 *
	 * @p ppDB		Data Block to write new key into. 
	 *			Can require reallocation so it's memory position and size might change.
	 * @p pdbSize		Data Block size. 
	 *			Can be changed if Data Block is reallocated.
	 * @p key		Key information to wrote
	 * @p key_desc		Map descriptor for new key. 
	 *			Leave NULL and new key will take any vacant space required if there is enough.
	 * @p key_proc		Key block processor. Will process the key data before writing according to key type.
	 * @p key_map		Explicit key map to use. For example, virtual hidden part keymaps. 
	 *			Leave NULL to use Data Block's own map.
	 * @p header_is_enc 	If the Data Block Header is encrypted beside the key tabs. 
	 *			If false, will calculate updated header digest only. 
	 *			If true, will also encrypt all updates in header.
	 * @p dec_header	Return copy of the updated Data Block header, decrypted.
	 *
	DWORD DataBlockWriteKey(DATA_BLOCK** 		ppDB, 
				DWORD* 			pdbSize, 
				key_info_t* 		key, 
				keyblock_descriptor_t* 	key_desc, 
				keyblock_map_t* 	key_map, 
				bool 			header_is_enc, 
				db_header_t* 		dec_header);
*/
} // namespace kgghost


//
//	TODO
//


#if 0

////////////////////////////////////
//
//	extern functions
//

class cKeyBlockParams;



#define _UseSimpleAndExetendedBlockOnly_




//
//	DataBlockVerifyKeyBlockAndGetKey
//
//	1. Tries to decode key info, it walks key blocks and implement kbParams to 
//		PBE - decrypt key by key formed from password.
//		PKE - decrypt by public key
//		SSS - implement every share to the key block
//	 Notes
//		a. If kbParams type is undefined ( kbtEmpty ), so all available 
//			encoding shemes will be implemented to DATA_BLOCK
//		b. If pKeyHandle and pKeyInfoCopy both are NULL, so the function
//			checks if the password in kbParams.kbPBE.pwd is compatible with any key block.
//			It returns ERROR_NO if at least the password is compatible even with one pwd 
//			for block with kbtSSS type
//
//	2. If decoded the key info is for hidden part, it reads next bytes as verification block, 
//			See DataBlockWriteKey_Hidden
//
extern 
DWORD DataBlockVerifyKeyBlockAndGetKey(
		DATA_BLOCK	*dataBlock,					// Data block have read directly from container
		DWORD		dbSize,
		cKeyBlockParams *kbParams,				// Key block paramteres used for key encryption.
		sKeyMapInfo *keyMapInfo = NULL,
		KEY_HANDLE	*pKeyHandle = NULL,			//	creates key handle if the parameter is not NULL 
		sKeyInfo	*pKeyInfoCopy = NULL,
		dbHeader	*pDecryptedHeader = NULL,	//	place for decrypted header copy
		dbContInfo	*pContInfo = NULL,			// place for decrypted cont info for hidden or usual part corresponded to pKeyHandle
		HWND		hProgress = NULL,
		bool		*pCancelled = NULL
	);

//
//	DataBlockWriteKey_Hidden
//
//		1. Writes hidden key info encrypted by kbParams
//		2. Writes simple verification block at the key tail encrypted by hidden key info
//
//	How the key read, see DataBlockVerifyKeyBlockAndGetKey
//
//	Note 1. The function does not extend key block for hidden part key. 
//			It will security nonsens if block is extended but key map reports that no one key presents in extended part.
//
//	Note 2. It shreds old key block information if keyMapInfo specified ( is not NULL or position != -1 )
//
extern DWORD 
	DataBlockWriteKey_Hidden(
		DATA_BLOCK			**ppDB,
		DWORD				*pDBSize,
		sKeyInfo			*hiddenKeyInfo,	//	= NULL	if null, so it does not required to save encoded key info
											//	the existing hidden part from started position is simple shreded
		cKeyBlockParams		*kbParams,		//	= NULL

		sKeyMapInfo			*keyMapInfo,	// = NULL,	See Note 2.

		dbKeyId				*keyMap			// the map is used for detecting busied items.
											// Ifnormation about existing hidden parts positions may be accumulated by and sent to the
											// function create new hidden part. See CreateHiddenPart
	);

// DataBlockGetContainerId()
//		Container ID is initially randomized for main and hidden contInfo.
//		The info is not changed even if main contInfo is encrypted.
//
extern BOOL DataBlockGetContainerId( DATA_BLOCK *DataBlock, BYTE *ContainerId );




extern DWORD 
	DataBlockRemoveKeyBlock( DATA_BLOCK **ppDB, DWORD *pDBSize, cKeyBlockParams *kbParams );



struct sCKHParams;


struct sKGCmdParams
{
	DWORD flags;	// Old version compatible flag
	DWORD size;	// size of the structure

	enum { stInit = 0, stUpdated = 1 };
	DWORD state;	// signaled state, initially is zero,
			// if the command successfully processed state = stUpdated

	dbContInfo mainContInfo;

// Parameters for mounting 
	HWND hWnd;
	dbContInfo contInfo;
	KEY_HANDLE *pKeyHandle;
//	int favDrive;
	BOOL readOnly;
	wchar_t mountPoint[ MAX_PATH ];
	
	sKGCmdParams()
	{
		memset( this, 0, sizeof( *this ) );
		size = sizeof( *this );
	}
};





//
//	GenerateKey
//		Generates keys and initial vectors according to PKCS#12 recommendations.
//		Note! If password is used as seed so the buffer has to be in unicode format.
//
extern DWORD
	GenerateKey(
			BYTE *Key,
			int   KeyLength,
			BYTE *salt,
			int	  saltSize,
			BYTE *Seed,
			int   SeedLength,
			BYTE  hashAlgID,
			bool isKey // determintes is kenereted key or initial vector
		);


extern void 
	ShredData( BYTE *data, int dataLength );


class cPassword;
struct sEncIDs;

//
//	FindSamePassword
//		Tryes to decode every key block in the data block with different algorithms
//		Returns true if one key block is compatible, otherwise return false
//
bool FindSamePassword(
		DATA_BLOCK	*dataBlock,
		DWORD		dataBlockSize,
		cPassword	*pwd,
		ekbType		*pKBType = NULL,	// [Out] key block type were same password has been found
		sEncIDs		*encIds = NULL
);

template < class T > void copyEx( T *out, T *in )
{
//	assert( false ); // check this
	if ( out != NULL )
	{
		if ( in == NULL )
			memset( out, 0, sizeof( T ) );
		else
			memcpy( out, in, sizeof( T ) );
	}	
};

#define memCopyEx( out, in, const_size )\
{\
	if ( out != NULL )\
	{\
		int size = const_size;\
		\
		if ( in == NULL )\
		{\
			if ( size > sizeof( *(out) ) )\
			{\
				assert( false );\
				size = sizeof( *(out) );\
			}\
			\
			memset( out, 0, size );\
		}\
		else\
		{\
			if ( size > sizeof( *(out) ) )\
			{\
				assert( false );\
				size = sizeof( *(out) );\
			}\
			\
			if ( sizeof( *(out) ) != sizeof( *(in) ) )\
			{\
				assert( false );\
			}\
			\
			memcpy( out, in, size );\
		}\
	}\
}

//
//	decryptHeaderEx
//		decrypts header by a key handle, the key handle may be for usual or for hidden part
//	
//		1. If dataBlock is not decrypted ( see DATA_BLOCK::header.isEncrypted ) it verifies header
//			and copies dataBlock to output buffer and returns true.
//		2. Tries to decrypt dataBlock::contInfo with the algoritm id and key handle.
//			If the contInfo is properly decrypted, the function decrypts header with keyMap
//			with same key handler, copies decrypted result to out buff and returns true.
//		3. Tries to decrypt hidden part with the algorithm and key handle.
//			If decrypted hidden part was not properly verified returns false.
//		4. Repeats step 1 and 2 but with key derived from decrypted hidden part.
//		5. Otherwise returns false.
//
bool
	decryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		algId,				// Encryption Algorithm Identificator
		DWORD		*pHashAlgId,		//  = NULL, [Out] pointer to hash algorithm decrypted from encBlock tail
		KEY_HANDLE	keyHandle,			// and Key Handle it may be key handle for hidden part
		encBlock <dbHeader>
					*headerOut   = NULL,// Decrypted header, caller procedure is responsible for secure erasing of the decrypted data
		DWORD		*pErrorCode	 = NULL,// if the function returns FALSE, look on the Error code
		dbContInfo	*contInfoOut = NULL,// Decrypted cont info, the procedure copies usual or hidden cont info here, depends on key handle.
		sKeyInfo	*mainKeyInfo = NULL // [OUT] if hidden part properly decrypted the value is fille by decrypted main key info
	);							

//
//	encryptHeaderEx
//		encrypts entire header or it's tail with digest by a key handle for usual part.
//		Note, if the key handle is for hidden part the usual part key is derived from decrypted hidden part.
//	
//		1. If dataBlock is encrypted ( see DATA_BLOCK::header.isEncrypted )
//			1.1 it checks if key handle is valid for the data block by decryptHeaderEx.
//			1.2 If is valid the function copies dataBlock to output buffer and returns true.
//			If is not valid the function returns false
//		2. Checks if key handle is for hidden part.
//			If so derive new key handle form decrypted hidden part
//		3. Encrypts header by the key handle for usual part
//
bool
	encryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		algId,			// Encryption Algorithm Identificator
		DWORD		encModeId,		// Encryption Algorithm Identificator
		DWORD		hashAlgId,		// Hash algorithm is used for digest
		DWORD		keyHandle,		// and Key Handle it may be key handle for hidden part
		encBlock <dbHeader>
					*dbOut,			// Entire encrypted block with tail
		DWORD		*pErrorCode,	// if the function returns FALSE, look on the Error code
		bool		bEntire			// = true, Defines if entire or tail only encryption is used
	);

//
//	decryptHeaderEx
//		decrypts header by a key handle, the key handle may be for usual part only
//			( hidden part key consists of both hidden and main parts keys ) see DataBlockVerifyKeyBlockAndGetKey
//	
//		1. If dataBlock is not encrypted ( see DATA_BLOCK::header.isEncrypted ) it verifies header
//			and copies dataBlock to output buffer and returns true.
//		2. Tries to decrypt dataBlock::contInfo with the algoritm id and key handle.
//			If the contInfo is properly decrypted, the function decrypts header with keyMap
//			with same key handler, copies decrypted result to out buff and returns true.
//		3. Otherwise returns false.
//
bool
	decryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		contAlgId,	// Encryption Algorithm Identificator
		DWORD		encModeId,
		DWORD		*pHashAlgId,	//  = NULL, [Out] pointer to hash algorithm decrypted from encBlock tail
		DWORD		keyHandle,		// and Key Handle it may be key handle for hidden part
		encBlock <dbHeader>
					*headerOut,		// = NULL, Decrypted header, caller procedure is responsible for secure erasing of the decrypted data
		DWORD		*pErrorCode		// = NULL, if the function returns FALSE, look on the Error code
	);

//
//	EncryptHeader
//		Asks password for encryption/decryption
//		Encrypts decrypted header and decrypts encrypted.
//
extern
	bool EncryptHeader(
		HWND			hWnd,
		const wchar_t	*contPath,
		encBlock< dbHeader >
					*db,
		DWORD		dbSize,
		DWORD		*pErrCode,
		DWORD		*pFlags,
		sCKHParams	*ps
	);

//
//	decryptAndVerifyHeader
//
DWORD
	decryptAndVerifyHeader(
		DATA_BLOCK	*dataBlock,
		sKeyInfo	*keyInfo,
		encBlock< dbHeader > 
					*decryptedHeader,
		KEY_HANDLE	*pDelegateKeyHandle = NULL // key handle to be delegeted
	);


//
//	CreateHandle_MainPart
//		Main part password required.
//
bool CreateHandle_MainPart(

		HWND hWnd,

		const wchar_t *text,
		const wchar_t *caption,

		DATA_BLOCK			**ppDB,
		DWORD				*pDBSize,
		
		sKeyInfo			*keyInfo = NULL,

		KEY_HANDLE			*pKeyHandle = NULL,
		sCKHParams			*ps = NULL,
		dbKeyId				*keyMapT = NULL
	);

extern bool 
	updateKeyParams( sCKHParams *ps, sKeyInfo *keyInfo, dbHeader *decHeader = NULL );



//
//	verifyKeyHandle
//	
//	Every create key handle function may has an ready key handle
//	The funvtion allows to verify the handle before using.
//
//	returns ERROR_NO if success.
//
DWORD verifyKeyHandle( DATA_BLOCK *db, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId, dbHeader *decHeader = NULL );

#endif // 0


#endif

