/*******************************************************************\
*
*	Copyright (c) 2007-2009 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: enc_block.h
*	Revision:
*	Created:
*	Description: utility definition to store a typical
*	encrypted block together with it's supporting data
*	and allow operations on it as a whole
*
\*******************************************************************/

#ifndef V8_ENCRYPTED_BLOCK_H
#define V8_ENCRYPTED_BLOCK_H

#include <string.h>

#include "layout.h"
#include "alg_utils.h"
#include "comm_def/kg_defs.h"

#include <alg/alg.h>
#include <common/smartptr.h>
#include <common/class_defs.h> // bc_padded_type_t
#include <common/entype.h>


//
//	Encryption strategy.
//
//	The trick with decrypting encBlock with zero initial vector first and then decryption with the decrypted initial vector
//	worked properly only for CBC mode.
//	So initial verctor is not encrypted now.
//	It has size as DB_MAX_ENCRYPTED_BLOCK_SIZE and is shredded before encryption.
//


#if ( IVECTOR_LENGTH_DB != DB_MAX_ENC_BLOCK_SIZE )
	#define ENCRYPT_IV
#endif


namespace kgghost
{
	/**
	 * @brief	Atomary block size for encryption operations in bytes
	 */
	enum
	{
		BLOCK_BYTES = 16,
		DIGEST_BYTES = 32,
	};


	/**
	 * @brief	Digest calculation helper
	 */
	void calculate_digest( int hash_id, bc_byte_t* digest, const bc_byte_t* text, unsigned int text_size );


	/**
	 * @brief	Digest verification helper
	 */
	bool verify_digest( int hash_id, const bc_byte_t* digest, bc_byte_t* text, unsigned int text_size );


	/**
 	* @brief	Main block of data that will be encrypted.
 	*/
	template < class T >
	struct eb_data_t
	{
		/* encrypted block data, padded to block cipher algorithm boundary */
		bdk::bc_padded_type_t<T, BLOCK_BYTES> data;

		/* digest of padded type */
		bc_byte_t digest[DIGEST_BYTES];

		/* initial vector, unecrypted */
		unsigned char iv[IVECTOR_LENGTH_DB];

		eb_data_t()
		{
		}

		eb_data_t(eb_data_t<T> const & rhv)
		{
			memcpy(this, &rhv, sizeof(rhv));
		}

		/**
		 * @brief	Get stored type data
		 */
		T* type() { return data.type(); }
		const T* type() const { return data.type(); }

		/**
		 * @brief	Verify encrypted data digest
		 */
		bool verify(int hash_id)
		{
			return verify_digest(hash_id, digest, (bc_byte_t*)this->type(), sizeof(T));
		}


		/**
		 * @brief	Update encrypted data digest
		 */
		void update(int hash_id)
		{
			calculate_digest(hash_id, digest, (bc_byte_t*)this->type(), sizeof(T));
		}


		/**
		 * @brief	Store info from external type buffer
		 */
		void store(const T* from, int hash_id)
		{
			this->shred();
			memmove(this->type(), from, sizeof(T));
			this->update(hash_id);
		}


		/**
		 * @brief	Move data into external type buffer
		 */
		void move(T* to)
		{
			memmove(to, this->type(), sizeof(T));
		}


		/**
		 * @brief	Shred all data
		 */
		void shred()
		{
			ShredData((bc_byte_t*)this, sizeof(*this));
		}


		void debug_trace()
		{
#ifdef BC_DEBUG
		        BC_DEBUG_PRINT(("Debug: encrypted block trace begin:\n"));
		        BC_DEBUG_PRINT((" data:\n"));
		        for(unsigned int i = 0; i < sizeof(T); ++i)
		        {
		                BC_DEBUG_PRINT((" 0x%x", *(((bc_byte_t*)&data) + i) ));
		        }
		        BC_DEBUG_PRINT(("\n"));

		        if(sizeof(T) < sizeof(data))
		        {
		                BC_DEBUG_PRINT((" padding:\n"));
		                for(unsigned int i = sizeof(T); i < sizeof(data); ++i)
		                {
		                        BC_DEBUG_PRINT((" 0x%x", *(((bc_byte_t*)&data) + i) ));
		                }
		                BC_DEBUG_PRINT(("\n"));
		        }

                        BC_DEBUG_PRINT((" digest:\n"));
                        for(unsigned int i = 0; i < sizeof(digest); ++i)
                        {
                                BC_DEBUG_PRINT((" 0x%x", digest[i]));
                        }
                        BC_DEBUG_PRINT(("\n"));

                        BC_DEBUG_PRINT((" iv:\n"));
                        for(unsigned int i = 0; i < sizeof(iv); ++i)
                        {
                                BC_DEBUG_PRINT((" 0x%x", iv[i]));
                        }
                        BC_DEBUG_PRINT(("\n"));
                        BC_DEBUG_PRINT(("Debug: encrypted block trace end\n"));
#endif
		}
	};


	template < class T >
	class CEncBlock
	{
	public:

		typedef T block_type;

		/* padded encrypted data + digest + iv*/
		typedef	eb_data_t < T > enc_block_type;


		/**
	 	* @brief		Create empty encrypted block
	 	*/
		CEncBlock()
		{
		}


		block_type* type()
		{
			return m_block.type();
		}


		const block_type* type() const
		{
			return m_block.type();
		}


		/**
	 	* @brief		Decrypt the body of this encrypted block to external buffer.
	 	* @more		Will not change the internal invariant.
	 	*
	 	* @p to_body		External buffer to store decrypted data.
	 	* @p alg_handle	Algorithm configuration for decryption.
	 	*
	 	* @return		True if block was decrypted and verified.
	 	*			False if verification failed.
	 	*/
		bool decrypt_to(block_type *to_body, bdk::BCAlgHandle alg_handle, int hash_id)
		{
			bdk::BCScopedPtr < enc_block_type > copy(new enc_block_type(m_block));
			bc_byte_t* iv = m_block.iv;
			alg_handle->decrypt(iv, (bc_byte_t*)copy.get(), sizeof(enc_block_type) - IVECTOR_LENGTH_DB);

			copy->debug_trace();

			if( !copy->verify(hash_id) )
				return false;

			copy->move(to_body);
			return true;
		}
		/**
	 	* @brief		Initialize encrypted block from external body buffer.
	 	*
	 	* @p from_body		Data buffer to encrypt and store in encrypted form with verification info.
	 	* @p hash_id		Id of the hash algorithm to use for plaintext body digest generation
	 	* @p alg_handle	Configured algorithm handle to perform encryption with
	 	* @p iv		Initial vector to use for encryption. Will be stored as a part of verification.
	 	*
	 	* @return		ERROR_NO if encryption and digest generation went ok.
	 	*			ERROR_INTERNAL_PROBLEM otherwise.
		 */
		void encrypt_from(const block_type *from, bdk::BCAlgHandle alg_handle, int hash_id, unsigned char* iv)
		{
			// encrypt and replace iv with plain copy
			m_block.store(from, hash_id);
			alg_handle->encrypt(iv, (bc_byte_t*)&m_block, sizeof(enc_block_type));
			memmove(m_block.iv, iv, IVECTOR_LENGTH_DB);
		}





	private:


		/* stored data */
		enc_block_type 	m_block;
	};
}


#if 0

/**
 * @brief	Encryption block tail
 *
 * @more	Supporting structure for storing main encrypted block body digest and IV
 */
struct eb_tail_t
{
	unsigned char	m_digest[ MAXIMUM_DIGEST_SIZE_BYTES ];

#ifdef ENCRYPT_IV

	// record here the
	int m_hash_alg_id;


	// insert padding to make iv end at block size boundary, so that eb_tail_t is padded to block size and iv is at it's end
	// tail_padding_t < ROUND_TAIL( (sizeof(int) + IVECTOR_LENGTH_DB + MAXIMUM_DIGEST_SIZE_BYTES), DB_MAX_ENC_BLOCK_SIZE) > m_padding;

	// iv last and aligned
	unsigned char m_iv[IVECTOR_LENGTH_DB];

#else

	unsigned char m_iv[ IVECTOR_LENGTH_DB ];

#endif

	unsigned char* iv()
	{
		return m_iv;
	}

	bool calculate_digest(int hash_alg_id, unsigned char* buf, unsigned int bufsiz);

	bool verify_digest(int* p_hash_alg_id, unsigned char* buf, unsigned int bufsiz);
};



/**
 * @brief	A block of data stored in encrypted form with all nessesary service information for verification.
 * @more	Will not store encryption keys, client code will have to initialize algorithm configurations manually.
 *		Will store initial vector that block was encrypted with.
 *		Whether iv will be stored encrypted or decrypted is specific to implementation version (ENCRYPT_IV macro).
 */
template < class body_type >
class CEncBlock
{
public:

	typedef CEncBlock<body_type> eblock_type;


	/**
	 * @brief		Create empty encrypted block
	 */
	CEncBlock()
	{
		BC_DEBUG_PRINT(("enc_block: body size 0x%x, body padded size 0x%x (padding should be 0x%x), tail size 0x%x, total size 0x%x\n",
					sizeof(body_type),
					sizeof(m_body),
					ROUND_TAIL(sizeof(body_type), DB_MAX_ENC_BLOCK_SIZE),
					sizeof(m_tail),
					sizeof(*this)));
	}


	/**
	 * @brief		Get encrypted body pointer
	 */
	body_type* body()
	{
		return m_body.data();
	}


	/**
	 * @brief		Get read-only encrypted body pointer
	 */
	const body_type* body() const
	{
		return m_body.data();
	}


	eb_tail_t* tail()
	{
		return &m_tail;
	}

	/**
	 * @brief		Initialize encrypted block from external body buffer.
	 *
	 * @p from_body		Data buffer to encrypt and store in encrypted form with verification info.
	 * @p hash_id		Id of the hash algorithm to use for plaintext body digest generation
	 * @p alg_handle	Configured algorithm handle to perform encryption with
	 * @p iv		Initial vector to use for encryption. Will be stored as a part of verification.
	 *
	 * @return		ERROR_NO if encryption and digest generation went ok.
	 *			ERROR_INTERNAL_PROBLEM otherwise.
	 */
	int encrypt_from(body_type *from_body, bccore::CBCAlg_Handle alg_handle, int hash_id, unsigned char* iv)
	{
		unsigned int nbytes = sizeof(*this);

		// shred current data
		ShredData((bc_byte_t*)this, nbytes);

		// copy input body
		memcpy( this->body(), from_body, sizeof(body_type) );

		// calculate plaintext body digest
		if ( !this->calculate_digest(hash_id) )
			return ERROR_INTERNAL_PROBLEM;

#ifdef ENCRYPT_IV	// store iv encrypted
		memcpy(m_tail->iv(), iv, IVECTOR_LENGTH_DB);
		alg_handle->encrypt(iv, (char*)this, nbytes);
#else 			// store iv not encrypted
		alg_handle->encrypt(iv, (unsigned char*)this, nbytes);
		memcpy(m_tail.iv(), iv, IVECTOR_LENGTH_DB);
#endif // ENCRYPT_IV

		return ERROR_NO;
	}


	/**
	 * @brief		Decrypt the body of this encrypted block to external buffer.
	 * @more		Will not change the internal invariant.
	 *
	 * @p to_body		External buffer to store decrypted data.
	 * @p alg_handle	Algorithm configuration for decryption.
	 *
	 * @return		ERROR_NO if decryption and verified were successful.
	 *			ERROR_INVALID_ALGORITHM if verification failed for that domain reason.
	 *			ERROR_INTERNAL_PROBLEM otherwise.
	 */
	int decrypt_to(body_type *to_body, bccore::CBCAlg_Handle alg_handle)
	{
		unsigned char iv[IVECTOR_LENGTH_DB];
		unsigned int nbytes = sizeof( *this );

		bccore::TBCScopedPtr < eblock_type > copy(NOTHROW_NEW eblock_type);
		if( !copy.get() ) return false;

		memcpy((char*)copy.get(), this, nbytes);

#ifdef ENCRYPT_IV

		if( !alg_handle->decrypt(this->m_tail.iv(), copy.get(), nbytes) )
		{
			BC_DEBUG_PRINT(("eblock: iv decryption failed\n"));
			return false;
		}

		memcpy(iv, copy->m_tail.iv(), IVECTOR_LENGTH_DB);
		memcpy((char*)copy.get(), this, nbytes);

#else

		memcpy(iv, this->m_tail.iv(), IVECTOR_LENGTH_DB);

#endif // ENCRYPT_IV

/*
#ifdef BC_DEBUG

		BC_DEBUG_PRINT(("eblock: iv:\n"));
		for(unsigned int i = 0; i < IVECTOR_LENGTH_DB; ++i)
		{
			BC_DEBUG_PRINT(("0x%x ", iv[i]));
		}
		BC_DEBUG_PRINT(("\n"));
#endif
*/
		// we have a valid iv, decrypt
		alg_handle->decrypt(iv, (unsigned char*)copy.get(), nbytes);

/*
#ifdef BC_DEBUG

		BC_DEBUG_PRINT(("eblock: decrypted block:\n"));
		for(unsigned int i = 0; i < sizeof(*this); ++i)
		{
			if( i == sizeof(body_type) )
				BC_DEBUG_PRINT(("\nbody padding:"));

			if( i == sizeof(m_body) )
				BC_DEBUG_PRINT(("\ntail:"));

			BC_DEBUG_PRINT(("0x%x ", ((unsigned char*)copy->body())[i]));
		}
		BC_DEBUG_PRINT(("\n"));

#endif
*/
		// verify digest of decrypted body
		if( !copy->verify_digest(NULL) )
		{
			BC_DEBUG_PRINT(("Digest verification failed - assuming incorrect algorithm\n"));
			return ERROR_INVALID_ALGORITHM;
		}

		memcpy(to_body, copy->body(), sizeof(body_type));

		return ERROR_NO;
	}


	/**
	 * @brief		Calculate current body digest and store in supporting tail structure
	 *
	 * @p hash_alg_id	Id of hash algorithm to use for body digest generation
	 */
	bool calculate_digest(int hash_alg_id)
	{
		return m_tail.calculate_digest(hash_alg_id, (unsigned char*)this->body(), sizeof(body_type));
	}


	/**
	 * @brief		Verify current body digest and store in supporting tail structure
	 *
	 * @p hash_alg_id	Id of hash algorithm to use for body digest generation
	 */
	bool verify_digest(int* hash_alg_id)
	{
		return m_tail.verify_digest(hash_alg_id, (unsigned char*)this->body(), sizeof(body_type));
	}


private:


	// Encrypted body, padded to block boundary.
	// Will always be considered as an encrypted invariant.
	bccore::bc_padded_type_t<body_type, DB_MAX_ENC_BLOCK_SIZE> m_body;


	// Verification data tail
	eb_tail_t	m_tail;

};


#endif // #if 0

#endif
