/*******************************************************************\
*
*	Copyright (c) 2006 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: cKeyBlockParams.h
*	Revision: 
*	Created:		6.10.2004
*	Description:	The key block is used for getting new and existing key parameters
*					It works with three different types of key parameters.
*						1. Password based encryption
*						2. Public key encryption
*						3. Shamir secret shared scheme
*
\*******************************************************************/

#ifndef V8_KEYBLOCK_PROCESSOR_H
#define V8_KEYBLOCK_PROCESSOR_H

#include "layout.h"


#include <common/class_defs.h> // BC_INTERFACE
#include <common/build.h>

class cPassword;


namespace kgghost 
{


	/**
	 * @brief	Defines information about symmetric algorithm encryption key.
	 * @more	This structure is the basic data encoded and decoded by different key block processors into keyblocks.
	 */
	BC_PACKED(
	struct key_info_t
	{
		enum { sizeWithoutKey = sizeof(char) * 2 + sizeof(int) * 2 + sizeof(space_unit_t)};

		bc8_t		size;		// real size of the key info in bytes
		bc8_t		hash_id;	// Digest algorithm used for key block verification ( same as pgp id ( f.e. pgphaSHA1 or bchaSHA256 see pktdefs.h )
		bc32_t		alg_id;		// Algorithm is used for encryption of header, contInfo, hiddenContInfo, keyMap and container body
		bc32_t		mode_id;	// The mode is used for encryption for encryption algorithm above
		space_unit_t 	cont_info;	// Information about container data space this key addresses

		bc_byte_t	key[MAXIMUM_KEY_SIZE_BYTES];	// actual encryption key data


		/**
		 * @brief	Calculate and set the size field by size of the encryption key that is stored here.
		 */
		void calc_size( unsigned int key_size )
		{
			size = sizeWithoutKey + key_size;
		}


		/**
		 * @brief	See if this info describes the key for hidden part
		 */
		bool is_for_hidden_part() const
		{
			return this->cont_info.is_hidden();
		}
	});


	/**
	 * @brief	Keyblock coding interface.
	 */
	class IKeyBlockProcessor
	{
	BC_INTERFACE(IKeyBlockProcessor) // virtual d-tor here
	public:


		/**
		 * @brief	Decode keyblock buffer.
		 * 
		 * @p io_kbEncoded	
		 *		Encoded keyblock data. 
		 *		In: encoded data 
		 *		Out: decoded data if successful.
		 *
		 * @p io_kbSize	
		 *		Encoded keyblock data size.
		 *		In: full buffer size
		 *		Out: decoded buffer size if successful.
		 *
		 * @p key	Decoded key info.
		 * @p init_pwd	Initial user password.
		 *
		 * @return	True on successful decoding, false otherwise.
		 */
		virtual		bool decode(const bc_byte_t* io_kbEncoded, unsigned int & io_kbSize, key_info_t& key, cPassword const & init_pwd) = 0;


		/**
		 * @brief	Encode keyblock buffer.
		 * 
		 * @p io_kbPlain	
		 *		Plain keyblock data. 
		 *		In: data buffer to encode into 
		 *		Out: encoded data if successful.
		 *
		 * @p io_kbSize	
		 *		Encoded keyblock data size.
		 *		In: full buffer size
		 *		Out: encoded buffer size if successful.
		 *
		 * @p key	Key info to encode.
		 * @p init_pwd	Initial user password.
		 *
		 * @return	True on successful encoding, false otherwise.
		 */
		virtual 	bool encode(bc_byte_t* io_kbPlain, unsigned int & io_kbSize, key_info_t const & key, cPassword const & init_pwd) = 0;
	};


	class CKeyBlock
	{
	public:


		CKeyBlock(bc_byte_t* data, unsigned int size, bcu16_t type);


		~CKeyBlock();


		/**
		 * @brief	Verify password against this keyblock
		 *
		 * @p pwd	Password to verify.
		 *
		 * @return	Passed or failed verification.
		 */
		bool 		verify(const cPassword* pwd);
	

		/**
		 * @brief	
		 *		Decode key block data into encryption key
		 *
		 * @p in_out_key
		 *		In: known or suspected key info (alg, mode, hash ids).
		 *		Out: decoded key infoon success.
		 *
		 * @p initial_pwd	
		 *		First password entered by user.
		 *
		 * @return	
		 *		True if encoded key block was decrypted and stored descriptor will be updated.
		 *		False if decoding failed.
		 */
		bool 		decode(key_info_t & in_out_key, cPassword const & initial_pwd);


		/**
		 * @brief	Encode encryption key into key block data
		 *
		 * @p key	Encryption key info to encode.
		 *
		 * @return	Pass\fail.
		 */
		bool 		encode(key_info_t const & key, cPassword const & init_pwd);


		/**
		 * @brief	Keyblock size
		 */
		unsigned int 	size()
		{
			return m_size;
		}


		/**
		 * @brief	Keyblock data buffer
		 */
		bc_byte_t*	data() const
		{
			return m_data;
		}

	
		/**
		 * @brief	Keyblock encoding type
		 */
		bcu16_t type() const
		{
			return m_type;
		}


	protected:


		IKeyBlockProcessor* construct_processor();


	private:


		/* keyblock data buffer */
		bc_byte_t*	m_data;
		
		/* keyblock data size */
		unsigned int	m_size;

		/* keyblock encoding type */
		bcu16_t 	m_type;
		
		IKeyBlockProcessor* m_proc;
	};


#if 0
	struct keyblock_descriptor_t;

	/**
	 * @brief	Abstract keyblock.
	 * @more	Encoded key information data type.
	 */
	class IKeyBlock
	{
	BC_INTERFACE(IKeyBlock)
	public:


		/**
		 * @brief	Verify password against this keyblock
		 *
		 * @p pwd	Password to verify.
		 *
		 * @return	Passed or failed verification.
		 */
		virtual		bool verify(const cPassword* pwd) = 0;
	

		/**
		 * @brief	
		 *		Decode key block data into encryption key
		 *
		 * @p in_out_key
		 *		In: known or suspected key info (alg, mode, hash ids).
		 *		Out: decoded key infoon success.
		 *
		 * @p initial_pwd	
		 *		First password entered by user.
		 *
		 * @return	
		 *		True if encoded key block was decrypted and stored descriptor will be updated.
		 *		False if decoding failed.
		 */
		virtual 	bool decode(key_info_t & in_out_key, const cPassword * initial_pwd, keyblock_descriptor_t& out_kbdesc) = 0;


		/**
		 * @brief	Encode encryption key into key block data
		 *
		 * @p key	Encryption key info to encode.
		 *
		 * @return	Pass\fail.
		 */
		virtual 	bool encode(key_info_t const & key, keyblock_descriptor_t& out_kbdesc, const cPassword* init_pwd) = 0;
	};

	/**
	 * @brief	Processors collection.
	 * @more	Experimental.
	 */
	class CKeyBlockProcessor : public IKeyBlockProcessor
	{
	BC_FORBID_COPY(CKeyBlockProcessor)
	public:


		/**
		 * @brief	Initialize processor state 
		 */
		CKeyBlockProcessor();


		/**
		 * @brief	Release processor state
		 */
		~CKeyBlockProcessor();


		/**
		 * @brief	Decode keyblock.
		 */
		BC_INHERITED	bool decode(const bc_byte_t* kbEncoded, keyblock_descriptor_t& kb_desc, key_info_t& key, const cPassword* init_pwd);


		/**
		 * @brief	Encode keyblock.
		 */
		BC_INHERITED 	bool encode(bc_byte_t* kbPlain, keyblock_descriptor_t& kb_desc, key_info_t const & key, const cPassword* init_pwd);


		/**
		 * @brief	Get processor assosiated with keyblock type
		 */
		IKeyBlockProcessor* get_processor_for_keyblock(bcu16_t kb_type);


	private:


		struct proc_entry_t
		{
			IKeyBlockProcessor* proc;
			bcu16_t type;

			proc_entry_t(IKeyBlockProcessor* p, bcu16_t t) : proc(p), type(t) {};
		};

		typedef std::list<proc_entry_t> 	proc_list_t;
		typedef proc_list_t::iterator		proc_iter_t;

		proc_list_t 				m_procs;
		
	};

#endif





} // namespace kgghost

#endif

