#ifndef V8_RANGE_H
#define V8_RANGE_H

namespace kgghost
{

	/**
	 * @brief	Inclusive range. Used as a location descriptor for keyblocks.
	 * @more	Functions as a contiguous range now, but don't rely on it.
	 */
	struct range_t
	{
		/* first element index */
		unsigned int	m_base;


		/* total elements including first */
		unsigned int	m_total;


		/**
		 * @brief	Constructs empty range
		 */
		range_t();


		/**
		 * @brief	Constructs contigious range
		 *
		 * @p base 	First index
		 * @p total	Total elements, including base
		 */
		range_t(unsigned int base, unsigned int total);


		/**
		 * @brief	Get range size
		 */
		unsigned int 	size() const;


		/**
		 * @brief	Get range base
		 */
		unsigned int 	base() const;


		/**
		 * @brief	Is this an empty range.
		 */
		bool 		is_empty() const;


		/**
		 * @brief	Extract subrange from this range
		 * 
		 * @p l_first	First element of new subrange relative to our coordinates
		 * @p length	Number of elements in new range.
		 *
		 * @return	Required subrange or empty range if required span is out of range's bounds
		 */
		range_t		subrange(unsigned int l_first, unsigned int length);

		
		/**
		 * @brief	Get a subrange starting from next vacant block and spanning to the end of current range
		 *
		 * @return	Required subrange or empty if no vacant blocks present
		 */
		//range_t 	next_vacant();


		/**
		 * @brief	Get a subrange starting from next occupied block and spanning to the end of current range
		 *
		 * @return	Required subrange or empty if no occupied blocks present
		 */
		//range_t 	next_occupied();


		/**
		 * @brief	Find vacant subrange inside this range
		 *
 		 * @p req_size	Require subrange of this size. Should not be larger that our size
		 *
		 * @return	Required subrange or empty range if required space is unavailiable
		 */
		//range_t	find_vacant_subrange( unsigned int req_size );


		/**
		 * @brief	Find a vacant subrange of maximum size
		 *
		 * @return	Required subrange or empty range if there are no vacant elements
		 */
		//range_t	find_biggest_vacant_subrange();
	};


	/**
	 * @brief	Construct a contigious range
	 *
	 * @p base 	First keyblock index
	 * @p length	Total number of keyblocks including base
	 *
	 * @return	Required range
	 */
	range_t		contigious_range(unsigned int base, unsigned int length);

	
	/**
	 * @brief	Construct an empty range.
	 */
	range_t		empty_range();


	/**
	 * @brief	See if a range fits the limits
	 */
	bool		range_fits(const range_t& range, unsigned int limit);

} // namespace kgghost

#endif

