/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: alg.h
 *	scope: core\alg
 *	author: evgeny yakovlev
 *
 *	description: 
 *	Algorithm module facade.
 *
 *	history:
 *
 */

#ifndef BDK_ALG_ALG_H
#define BDK_ALG_ALG_H

#include <common/class_defs.h>
#include <common/smartptr.h>

#include <alg/service.h>


namespace bdk
{

	class BCAlg;


	/**
	 * @brief	Deleter functor for shared pointer.
	 */
	struct bc_alg_deleter
	{
		/**
		 * @brief	Unregister instance from service and delete it.
		 */
		void operator() (BCAlg* alg);
	};


	/**
	 * @brief	Shared algorithm handle
	 */
	typedef BCSharedPtr<BCAlg> BCAlgHandle;
}


namespace bdk
{

	/**
	 * @brief	Configured algorithm ready to use.
	 * @more	Owns all resources allocated by driver and binds them to it's lifetime.
	 */
	class BCAlg 
	{
	BC_FORBID_COPY(BCAlg)

	public:
		

		/**
		 * @brief	Configure algorithm, allocate driver resources
		 *
		 * @p service	Callback service object that spawned us
		 * @p mode	Our mode id
		 * @p key	Encryption key
		 * @p pool	Random data pool
		 *
		 * @throws	CBCException if driver initialization fails
		 */
		BCAlg( BCAlgService* service, int mode, bc_byte_t* key, bc_byte_t* pool );


		/**
		 * @brief	Init algorithm handle
		 *
		 * @p id	Algorithm id
		 * @p mode	Algorithm mode
		 * @p key	Encryption key
		 * @p pool	Random data pool
		 *
		 * @return	Configured shared handle.
		 */
		static 		BCAlgHandle init(int id, int mode, bc_byte_t* key, bc_byte_t* pool);


		/**
		 * @brief	Init algorithm handle
		 *
		 * @p name	Algorithm name
		 * @p mode	Algorithm mode
		 * @p key	Encryption key
		 * @p pool	Random data pool
		 *
		 * @return	Configured shared handle.
		 */
		static 		BCAlgHandle init(std::string const & name, int mode, bc_byte_t* key, bc_byte_t* pool);


		/**
		 * @brief	Deallocate driver resources
		 */
		~BCAlg() throw();


		/**
		 * @brief	Is this algorithm valid?
		 * 
		 * @sa 		release
		 */
		bool is_valid() const;


		/**
		 * @brief	Release resources and invalidate
		 */
		void release();


		/**
		 * @brief	Get algorithm id number
		 */
		int id() const
		{
			return m_service->id();
		}


		/**
		 * @brief	Return algorithm printable name
		 */
		const char* name() const
		{
			return m_service->name();
		}


		/**
		 * @brief	Return mode number the algorithm was configured with
		 */
		int mode() const 
		{
			return m_mode_id;
		}


		/**
		 * @brief	Return mode name
		 */
		const char* mode_name() const
		{
			return m_service->mode_name(m_mode_id);
		}


		/**
		 * @brief	Encrypt block of data
		 *
		 * @p iv	Initial vector
		 * @p buffer	In\out data to encrypt
		 * @p bufsiz	Size of buffer
		 */
		void 		encrypt( const bc_byte_t* iv, bc_byte_t* buffer, size_t bufsiz ) const;


		/**
		 * @brief	Decrypt block of data
		 *
		 * @p iv	Initial vector
		 * @p buffer	In\out data to decrypt
		 * @p bufsiz	Size of buffer
		 *
		 * @return	pass\fail
		 */
		void 		decrypt( const bc_byte_t* iv, bc_byte_t* buffer, size_t bufsiz ) const;


		/**
		 * @brief	Get system key handle.
		 */
		bc_key_handle_t	key_handle()
		{
			return m_key_handle;
		}


	private:


		/* mode id */
		int m_mode_id;
		
		/* service we've been spawned by */
		BCAlgService* m_service;

		/* driver key handle */
		bc_key_handle_t m_key_handle;

		/* voluntary release support */
		bool m_valid;
		
	};




}

#endif 

