/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: bind.h
 *	scope: bdk\alg
 *	author: evgeny yakovlev
 *
 *	description: 
 *	BC crypto loop binding
 *
 *	history:
 *	- experimental state.
 *
 */

#ifndef BDK_ALG_BIND_H
#define BDK_ALG_BIND_H

#include <common/types.h>

#include <alg/device.h>
#include <alg/alg.h>
#include <alg/loop.h>

namespace bdk
{

	/*
	 *	Partition table info
	 */
	struct part_info_t
	{
		unsigned int part_count;
		unsigned int ext_part_count;
		bc_space_unit_t	table[4];
	};


	/**
	 * @brief	Trusted entity to collect all required binding context from involved parties.
	 * @more	Binding a loop device requires a number of system-level private handles from all involved parties.
	 *		This entity provides a single trusted focal point for other entities to pass their low-level private details.
	 */
	class BCLoopBinder
	{
	public:


		/**
		 * @brief	Algorithm context callback. Called from configured alg handles.
		 */
		void collect_alg_context(bc_key_handle_t key_handle)
		{
			m_key_handle = key_handle;
		}


		/**
		 * @brief	Device context callback. Called from configured system device handles.
		 */
		void collect_device_context(backend::bc_crypto_dev_handle_t device)
		{
			m_device = device;
		}


		/**
		 * @brief	Top level binding routine. Walks all parties and asks them to issue their callbacks to collect context.
		 */
		template < class binder_source >
		void bind(BCCryptoLoop < binder_source > & loop, BCDeviceHandle device)
		{
			// collect trusted data
			loop.algorithm()->accept_binder(this);
			device->accept_binder(this);

			// read first sector to look for partiotion table
			bc_sector_t first;
			struct part_info_t info;

			loop.read_sector(first, 0);
			unsigned int start_sector = this->partition_info(first, info);

			BC_DEBUG_PRINT(("Loop: new start sector number %d\n", start_sector));

			// we got our start sector, start binding source
			loop.source().bind(m_device, m_key_handle, loop.algorithm(), loop.flags(), loop.layout(), start_sector);
		}


		template < class binder_source >
		void bind(binder_source const & source, BCDeviceHandle device, bc_data_context_t const & ctx)
		{
			BCCryptoLoop < binder_source > loop(source, ctx);
			this->bind < binder_source > (loop, device);
		}


	protected:


		/**
		 * @brief	Helper to parse first sector for partiotion table info.
		 * @return	Start sector of the last partiotion entry.
		 */
		unsigned int partition_info(bc_sector_t const & first, part_info_t& info) const;


	private:


		/* collected system-level context */
		backend::bc_crypto_dev_handle_t	m_device;
		bc_key_handle_t			m_key_handle;
	};

}

#endif
