/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: loop.h
 *	scope: bdk\alg
 *	author: evgeny yakovlev
 *
 *	description: 
 *	virtual BC crypto loop device interface.
 *
 *	history:
 *
 */

#ifndef BDK_ALG_CRYPTOLOOP_H
#define BDK_ALG_CRYPTOLOOP_H

#include <common/types.h>
#include <common/smartptr.h>

#include <alg/alg.h>


namespace bdk
{

	/**
	 * @brief	Loop context is a collection of data that defines how crypto loop will behave on any given data source.
	 */
	struct bc_data_context_t
	{

		/* configured block cipher */
		BCAlgHandle 		alg;

		/* data sector map */
		bc_space_unit_t 	layout;

		/* loop device flags */
		bcu32_t 		flags;

		/* is the context hidden: 
			TODO:
			refactor
		 */
		bool			is_hidden;
	};		


	/**
	 * @brief	BestCrypt virtual cryptoloop device.
	 * @more	Used as a mimic for what BestCrypt is actually about - seamless access to transparently encrypted data.
	 *		Presents a plain sector-addressed data source.
	 *		Can be set as an actual system virtual device through binding method.
	 *		
	 * @more	Templated by source driver type and requires it to implement compile-time protocol:
	 *		void read_sector(bc_sector_t & sec, bc_ulong_t secnum);
	 *		void write_sector(bc_sector_t & sec, bc_ulong_t secnum);
	 *
	 * @more	If you also plan to create system-wide loop devices through binding,
	 *		your source driver should also provide bind method:
	 *		void bind( 	backend::bc_crypto_dev_handle_t dev, 
	 *				bc_key_handle_t key_handle, CBCAlg_Handle alg, bcu32_t flags, 
	 *				const bc_space_unit_t& layout, unsigned int start_sector )
	 *
	 */
	template < class source_driver >
	class BCCryptoLoop
	{
	public:


		/**
		 * @brief	Initialize loop device
		 *
		 * @p data	Data source access.
		 * @p alg	Algorithm configuration to work with loop device.
		 * @p layout	Space layout of encrypted data.
		 * @p flags	Loop device flags.
		 *
		 * @throws	CBCException if initialization fails
		 */
		BCCryptoLoop( source_driver const & data, BCAlgHandle const & alg, bc_space_unit_t const & layout, bcu32_t flags ) :
			m_data(data),
			m_alg(alg),
			m_layout(layout),
			m_flags(flags)
		{
		}


		/**
		 * @brief	Initialize loop device by preloaded context.
		 * 
		 * @p data	Data source access.
		 * @p ctx	Loop context.
		 *
		 * @throws	CBCException if initialization fails
		 */
		BCCryptoLoop( source_driver const & data, bc_data_context_t const & ctx ) :
			m_data(data),
			m_alg(ctx.alg),
			m_layout(ctx.layout),
			m_flags(ctx.flags)
		{
		}


		/**
		 * @brief	Release loop device
		 */
		~BCCryptoLoop()
		{
		}


		/**
		 * @brief	Read a sector directly bypassing decryption.
		 *
		 * @p sector	Sector data to read into
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t	direct_read_sector(bc_sector_t& sector, bc_ulong_t secnum)
		{
			return m_data.read_sector( sector, this->map_sector(secnum) );
		}


		/**
		 * @brief	Read and decrypt sector
		 *
		 * @p sector	Sector data to read into
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t 	read_sector(bc_sector_t& sector, bc_ulong_t secnum)
		{
			bc_ulong_t rc = this->direct_read_sector(sector, secnum);
			bc_sector_iv_t iv;

			this->prepare_iv(iv, secnum);
			m_alg->decrypt( (bc_byte_t*)&iv.iv, (bc_byte_t*)&sector.sec_data, this->sector_size() );
		
			return rc;
		}


		/**
		 * @brief	Write a sector directly bypassing encryption.
		 *
		 * @p sector	Sector data to write
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t 	direct_write_sector(const bc_sector_t& sector, bc_ulong_t secnum)
		{
			return m_data.write_sector( sector, this->map_sector(secnum) );
		}


		/**
		 * @brief	Encrypt and write sector
		 *
		 * @p sector	Sector data to write
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t 	write_sector(const bc_sector_t& sector, bc_ulong_t secnum)
		{
			bc_sector_t data = sector;
			bc_sector_iv_t iv;

			this->prepare_iv(iv, secnum);
			m_alg->encrypt( (bc_byte_t*)&iv.iv, (bc_byte_t*)&data.sec_data, this->sector_size() );

			return this->direct_write_sector(data, secnum);
		}


		/**
		 * @brief	Get total sectors in data source.
		 */
		bc_ulong_t 	total_sectors() const
		{
			return (m_layout.length - m_layout.offset) / this->sector_size();
		}


		/**
		 * @brief	Get single sector size
		 */
		unsigned int	sector_size() const
		{
			return 512; // sector size resetting not supported for now
		}


		/**
		 * @brief	Get data offset in bytes
		 */
		bc_ulong_t	start_offset() const
		{
			return m_layout.offset;
		}	


		/**
		 * @brief	Get algorithm used for this device
		 */
		BCAlgHandle	algorithm() const
		{
			return m_alg;
		}


		/**
		 * @brief	Get crypto flags.
		 */
		bcu32_t	 	flags() const
		{
			return m_flags;
		}


		source_driver	source() const
		{
			return m_data;
		}

		
		bc_space_unit_t	layout() const
		{
			return m_layout;
		}


	protected:

			
		/**
		 * @brief	Prepare iv buffer for sector
		 * @more 	Depends on BC_FLAGS_IV_64BIT flag
		 */
		void prepare_iv(bc_sector_iv_t& iv, bc_ulong_t secnum)
		{
			if( m_flags & BC_CRYPTO_IV_64BIT ) 
				iv.iv = secnum + 1;
			else 
				iv.word_iv[0] = iv.word_iv[1] = iv.word_iv[2] = iv.word_iv[3] = secnum & 0xFFFF; // first 16 bits copy to all parts.
		}


		/**
		 * @brief	Map virtual sector to real data source sector based on supplied sector map.
		 *
	 	 * @p sec	Virtual sector
		 *
		 * @return	Real sector.
		 */
		bc_ulong_t	map_sector(bc_ulong_t sec)
		{
			return m_layout.offset / this->sector_size() + sec;
		}


	private:


		/* data source handle */
		source_driver			m_data;

		/* initialized algorithm handle */
		BCAlgHandle			m_alg;

		/* data layout */
		bc_space_unit_t			m_layout;

		/* loop device flags */
		bcu32_t				m_flags;
	};


	
}


#endif
