/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: device.h
 *	scope: bdk\alg
 *	author: evgeny yakovlev
 *
 *	description: 
 *	Backend crypto device wrapper.
 *
 */

#ifndef BDK_ALG_DEVICE_H
#define BDK_ALG_DEVICE_H

#include <backend/crypto.h>
#include <common/smartptr.h>


namespace bdk
{
	class BCLoopBinder;
	class BCDevice;

	/**
	 * @brief	Shared device instance handle.
	 */
	typedef BCSharedPtr<BCDevice> BCDeviceHandle;
}



namespace bdk
{

	/**
	 * @brief 	Platform crypto device.
	 */
	class BCDevice
	{	
	BC_FORBID_COPY(BCDevice)
	public:


		/**
		 * @brief 	Get bc device by name
		 */
		explicit BCDevice( const char* devname );


		/**
		 * @brief	Get bc device by system handle.
		 */
		explicit BCDevice( backend::bc_crypto_dev_handle_t sysdev );


		/**
		 * @brief	Get device handle by system name.
		 *
		 * @p name	Crypto device system name.
		 */
		static 		BCDeviceHandle get(std::string const & name);

		
		/**
		 * @brief	Get any free device handle
		 *
		 * @p name	Crypto device system name.
		 */
		static		BCDeviceHandle get_free();
		

		/**
		 * @brief	Return invalid device.
		 */
		static		BCDeviceHandle invalid_device();


		/**
		 * @brief 	Put bc device
		 */
		~BCDevice() throw();


		/**
		 * @brief	Get crypto device system name.
		 */
		const char* system_name()
		{
			return backend::bc_crypto_dev_name(m_dev_handle);
		}
	

		/**
		 * @brief	Lock crypto device.
		 */
		void lock()
		{
			return backend::bc_crypto_dev_lock(m_dev_handle);
		}


		/**
		 * @brief	Unlock crypto device.
		 */
		void unlock()
		{
			return backend::bc_crypto_dev_unlock(m_dev_handle, false);
		}


		/**
		 * @brief	Unlock crypto device.
		 */
		void force_unlock()
		{
			return backend::bc_crypto_dev_unlock(m_dev_handle, true);
		}


		/**
		 * @brief	Is anyone references this device's system handle?
		 */
		bool is_free()
		{
			return backend::bc_crypto_dev_is_free(m_dev_handle);
		}


		/**
		 * @brief	Get device backing file path.
		 * @path	File path will be stored here on success.
		 * @return	false if device is not configured (i.e. is free).
		 */
		bool 		backing_file_path(std::string& path);


		/**
		 * @brief	Remove any permanent data sources binded with this device.
		 */
		bool clear()
		{
			try {
				backend::bc_crypto_dev_clear(m_dev_handle, false);
				return true;
			} catch (BCSystemException const & e) {
				return false;
			}
		}


		/**
		 * @brief	Remove any permanent data sources binded with this device even if it is in use.
		 */
		void detach()
		{
			return backend::bc_crypto_dev_clear(m_dev_handle, true);
		}


		backend::bc_crypto_dev_handle_t system_handle()
		{
			return m_dev_handle;
		}


	private:


		/* platform backend handle */
		backend::bc_crypto_dev_handle_t m_dev_handle;
	};

}



#endif
