/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: service.h
 *	scope: bdk\alg
 *	author: evgeny yakovlev
 *
 *	description:
 *	Algorithm service provider wrapper interface.
 *
 *	history:
 *
 */

#ifndef BDK_ALG_SERVICE_H
#define BDK_ALG_SERVICE_H

#include <list>
#include <map>

#include <common/class_defs.h>
#include <alg/device.h>

#include <backend/crypto.h>

namespace bdk
{
	/**
	 * @brief	Algorithm mode definition
	 */
	struct bc_alg_mode_t
	{
		/* mode tag */
		bc_alg_mode_id_t	id;

		/* is this mode supported ? */
		bool			is_supported;

		/* total key size for encryption mode */
		size_t			key_size;
	};


	/**
	 * @brief 	Encryption mode tag.
	 */
	enum bc_alg_mode_tag_t
	{
	       	AM_LRW = 0xbc000001,     ///< LRW tweakable narrow-block encryption
       		AM_CBC = 0xbc000002,     ///< Cypher Block Chaining mode
       		AM_XTS = 0xbc000004,     ///< XTS tweakable narrow-block encryption
	};


	/** @brief 	Collection of supported modes */
	typedef std::map<unsigned int, bc_alg_mode_t> bc_alg_mode_map_t;


	/**
	 * @brief	Algorithm service info
	 *
	 * Client should init these from config providers
	 */
	struct bc_alg_info_t
	{
		bc_alg_info_t()
		{
			name.clear();
			prvname.clear();
			id = 0;
			key_size = 0;
			block_size = 0;
			ext_key_size = 0;
			mode_map.clear();
		}


		/* alg name (des, aes, etc.) */
		std::string name;

		/* provider name (usually driver name) */
		std::string prvname;

		/* unique alg id number */
		int id;

		/* key size */
		size_t			key_size;

		/* cypher block size */
		size_t			block_size;

		/* extention key size 	*/
		size_t			ext_key_size;

		/* supported modes */
		bc_alg_mode_map_t mode_map;
	};


	/**
	 * @brief	Configured algorithm handle.
	 *
	 * This is what we give to outside world.
	 * When the last copy is destroyed the driver resources are released
	 *
	 * It is exception safe to create non-initialized handles.
	 * All exception from initializing the handle are caught in service objects
	 */
	class BCAlg;


	/**
	 * @brief	Algorithm service
	 *
	 * Stores info about single algorithm service registered in system
	 * For using the algorithm to encrypt\decrypt see $CBCAlg$
	 */
	class BCAlgService
	{
	BC_FORBID_COPY(BCAlgService)

	public:


		/**
		 * @brief	Verify and store information about the service
		 *
		 * @throws	CBCException if there was no memory to initialize service
		 * 		CBCResourceException if service provider (driver) is unavailiable
		 */
		BCAlgService( const bc_alg_info_t& info );


		/**
		 * @brief	Kill the service and release all hanging resources
		 *
		 * @tip		Will force invalidation of all current handles
		 */
		~BCAlgService() throw();


		/**
		 * @brief 	Get alg service name
		 */
		const char* name() const
		{
			return m_name.c_str();
		}


		/**
		 * @brief	Get alg service id
		 */
		int id() const
		{
			return m_id;
		}


		size_t key_size() const
		{
			return m_key_size;
		}


		size_t mode_key_size(unsigned int mode) const
		{
			if( !is_mode_supported(mode) ) return 0;

			return m_modes.find(mode)->second.key_size;
		}


		size_t block_size() const
		{
			return m_block_size;
		}


		/**
		 * @brief	Check encryption mode support
		 *
		 * @p mode	Mode id to check
		 *
		 * @return	pass\fail
		 */
		bool is_mode_supported( unsigned int mode ) const
		{
			if( (mode == AM_CBC) || (mode == AM_LRW) || (mode == AM_XTS) ) return true;
			else return false;
		}


		/**
		 * @brief	Get a human readable mode decription
		 *
		 * @p mode	Mode id
		 */
		const char* mode_name( int mode )
		{

			switch (mode)
			{
			case AM_CBC: 	return "Cypher Block Chaining (CBC)";
			case AM_LRW: 	return "LRW tweakable narrow-block";
			case AM_XTS: 	return "XTS tweakable narrow-block";
			default:	BC_ASSERT(false); return "Unknown mode"; // asserts in debug mode, gracefully errors in release mode
			};
		}


		/**
		 * @brief	Get mode info
		 *
		 * @p mode	Put mode here
		 * @p id	Requested mode id
		 *
		 * @return 	true if mode is supported, false otherwise
		 */
		bool mode_info( bc_alg_mode_t& mode, int id )
		{
			if( !this->is_mode_supported(id) ) return false;

			mode = m_modes[id];
			return true;
		}


		/**
		 * @brief	Configure service for usage
		 *
		 * @p handle	Shareable handle to store configured algorithm
		 * @p mode	Block cypher mode
		 * @p key	Block cypher key
		 * @p pool	Random data pool
		 *
		 * @return	true if configuration succeeded, false otherwise.
		 */
		BCAlg* allocate_alg(int mode, bc_byte_t* key, bc_byte_t* pool );


		/**
		 * @brief	Release configures service
		 *
		 * @p alg	Alg service handle to release
		 *
		 * @tip		All real release is done in CBCAlg destructor
		 * 		We need this to remove handle from our storage
		 */
		void release_alg( BCAlg* alg );


		/**
		 * @brief	Verify service integrity
		 *
		 * @return	true if service is intact, false otherwise
		 */
		bool verify();

	private:

		/* name */
		std::string		m_name;

		/* provider (driver) name */
		std::string		m_provider_name;

		/* id */
		int 			m_id;

		/* key size */
		unsigned int		m_key_size;

		/* block size */
		unsigned int		m_block_size;

		/* supported modes and their key sizes */
		bc_alg_mode_map_t	m_modes;

		typedef std::list<BCAlg*> 	owners_list;
		typedef owners_list::iterator	owners_iter;

		/* algorithm handles ownership */
		owners_list		m_owners;
	};

};


#endif
