/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: crypto.h
 *	scope: bdk\backend interface
 *	author: evgeny yakovlev
 *
 *	description:
 *	Implementation-specific crypto functionality provider to BestCrypt core.
 *	Conforming implementation should be capable of:
 *	- safe encryption key storage
 *	- one-time encryption requests
 *	- persistant encryption requests using system container file through crypto device entry point and allocation of said devices
 *	- auxillary requests (version, verification)
 *
 *	history:
 *	- 22.06.09.
 *	Created unified device + driver as crypto backend.
 */


#ifndef BDK_BACKEND_CRYPTO_H
#define BDK_BACKEND_CRYPTO_H


#include <common/types.h>
#include <backend/fs.h>		// bc_file_handle_t


/**
 * @brief	Crypto provider flags
 */
enum bc_crypto_flags_t
{
	BC_CRYPTO_READONLY	= 0x00000001, /**< Access to cryptoloop in readonly mode */
	BC_CRYPTO_MULTIPART	= 0x00000002, /**< Binded data source has a partition table */
	BC_CRYPTO_IV_64BIT	= 0x00000004, /**< one 64 bit iv starting from 1 */
};


namespace bdk
{
	namespace backend
	{

		/**
		 * @brief 	Crypto device handle
		 */
		typedef struct bc_os_dev_t*	bc_crypto_dev_handle_t;


		/**
		 * @brief 	Get control of a crypto device
		 *
		 * @p devname	Platform-dependant device name
		 *
		 * @throws	System exceptions.
		 * @return 	Initialized device handle.
		 */
		bc_crypto_dev_handle_t 	bc_crypto_dev_get ( const char* devname );


		/**
		 * @brief	Get control of any free crypto device
		 *
		 * @throws	System exceptions.
		 * @return 	Initialized device handle.
		 */
		bc_crypto_dev_handle_t	bc_crypto_dev_get_free();


		/**
		 * @brief	Release control of the crypto device
		 *
		 * @p handle	Device handle
		 *
		 * @throws	System exceptions.
		 */
		void 		bc_crypto_dev_put ( bc_crypto_dev_handle_t& handle );


		/**
		 * @brief	Lock crypto device to forbid any other references.
		 *
		 * @p handle	Device handle
		 *
		 * @throws	System exceptions.
		 */
		void 		bc_crypto_dev_lock ( bc_crypto_dev_handle_t handle );


		/**
		 * @brief	Unlock crypto device to allow any other references
		 *
		 * @p handle	Device handle
		 * @p force	Ask implementation to force unlock if possible.
		 *
		 * @throws	System exceptions.
		 */
		void 		bc_crypto_dev_unlock ( bc_crypto_dev_handle_t handle, bool force );


		/**
		 * @brief	Check if this crypto device is free
		 *
		 * @p handle	Device handle
		 *
		 * @throws	System exceptions.
		 * @return 	false device is busy, true if free.
		 */
		bool 		bc_crypto_dev_is_free ( bc_crypto_dev_handle_t handle );


		/**
		 * @brief	Get crypto device name
		 *
		 * @p handle	Device handle
		 *
		 * @throws	System exceptions.
		 * @return 	Device name string. Will be valid as long as device handle is valid.
		 */
		const char* 	bc_crypto_dev_name ( bc_crypto_dev_handle_t handle );


		/**
		 * @brief 	Bind data file source through crypto device
		 *
		 * @p dev	Crypto device handle to use for persistent binding.
		 * @p file	File handle to bind.
		 * @p key	Key handle that addresses required encryption key in provider memory.
		 * @p algid	Encryption algorithm id to use.
		 * @p mode	Mode of the encryption algorithm.
		 * @p flags	Encryption flags.
		 * @p layout	Container space layout.
		 * @p startsec	First sector in data space (for patiotioning)
		 *
		 * @throws	System exceptions.
		 */
		void 		bc_crypto_set_file (
		    bc_crypto_dev_handle_t 	dev,
		    bc_file_handle_t 	file,
		    bc_key_handle_t 	key,
		    int			algid,
		    int 			mode,
		    bcu32_t 		flags,
		    const bc_space_unit_t* 	layout,
		    unsigned int		startsec
		);


		/**
		 * @brief	Clear crypto device to remove any persistent data sources binded to it
		 *
		 * @p handle	Device handle
		 *
		 * @throws	System exceptions.
		 */
		void		bc_crypto_dev_clear ( bc_crypto_dev_handle_t dev, bool force );


		/**
		 * @brief	See if device is linked to backing file and get this file name.
		 *
		 * @dev		Device.
		 * @file_path	Caller buffer to store backing file path. 
		 * @pathlen	Size of the path buffer in bytes.
		 *
		 * @return	true if device is configured, false otherwise.
		 * 		if function returns true but file path has not changed this means the buffer is too short.
		 */
		bool 		bc_crypto_stat_dev(bc_crypto_dev_handle_t dev, char* file_path, unsigned pathlen);


		/**
		 * @brief 	Store encryption key inside provider memory and return a handle to it
		 *
		 * @p algid	Encryption algorithm id to store key for
		 * @p key	Encryption key to store.
		 * @p key_len	Encryption key length.
		 * @p pool	Random data pool 512 bytes length.
		 *
		 * @throws	System exceptions.
		 * @return 	Key handle to address stored key later.
		 */
		bc_key_handle_t bc_crypto_store_key ( int algid, bc_byte_t* key, unsigned int key_len, bc_byte_t* pool );


		/**
		 * @brief	Relese encryption key inside provider memory.
		 *
		 * @p algid	Encryption algorithm id that stores the key.
		 * @p handle	Key handle.
		 *
		 * @throws	System exceptions.
		 */
		void 		bc_crypto_release_key ( int algid, bc_key_handle_t handle );


		/**
		 * @brief	Perform a one-time encryption of data using encryption key stored in provider's memory.
		 *
		 * @p algid	Encryption algorithm id to use.
		 * @p handle	Key handle for algorithm encryption key stored in provider memory.
		 * @p mode	Encryption mode
		 * @p iv	64 bit buffer of initial vector.
		 * @p buffer	Data to encrypt.
		 * @p bufsiz	Size of data to encrypt in bytes.
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_crypto_encrypt ( int algid, bc_key_handle_t handle, int mode, const bc_byte_t* iv, bc_byte_t* buffer, unsigned int bufsiz );


		/**
		 * @brief	Perform a one-time decryption of data using encryption key stored in provider's memory.
		 *
		 * @p algid	Encryption algorithm id to use.
		 * @p handle	Key handle for encryption key stored in provider memory.
		 * @p mode	Encryption mode
		 * @p iv	64 bit buffer of initial vector.
		 * @p buffer	Data to decrypt.
		 * @p bufsiz	Size of data buffer.
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_crypto_decrypt ( int algid, bc_key_handle_t handle, int mode, const bc_byte_t* iv, bc_byte_t* buffer, unsigned int bufsiz );


		/**
		 * @brief 	Get provider version
		 *
		 * @p major	Put major version number here
		 * @p minor	Put minor version number here
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_crypto_version ( char* major, char* minor );


		/**
		 * @brief 	Verify that provider's encryption algorithm is present.
		 *
		 * @p algid	Encryption algorithm id to verify.
		 *
		 * @throws	System exceptions.
		 * @return 	false if verification failed, true if passed.
		 */
		bool	 	bc_crypto_verify ( int algid );

	}
}

#endif

