/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: dso.h
 *	scope: bestcrypt core\backend interface
 *	author: evgeny yakovlev
 *
 *	description:
 *	dynamic symbol object backend interface
 *
 *	history:
 *
 */

#ifndef BDK_BACKEND_DSO_H
#define BDK_BACKEND_DSO_H

#include <common/types.h>


/**
 * @brief	DSO symbol resolution method
 *
 * When dso loads it may have dependences on other object code loaded.
 * This values define when to perform evaluation.
 * Pick one.
 */
enum bc_dso_resolve_t
{
	BC_DSO_RES_NOW		= 0x0001, /**< Resolve all symbols that dso depends on at load time */
	BC_DSO_RES_LAZY		= 0x0002, /**< Postpone symbol resolution until it is needed */
};


/**
 * @brief	DSO symbols visibility.
 *
 * how loaded symbols will affect all futher resolution
 */
enum bc_dso_visibility_t
{
	BC_DSO_VIS_GLOBAL	= 0x0001, /**< Use dso symbols in all resolution attempts, including future libraries */
	BC_DSO_VIS_LOCAL	= 0x0002, /**< Use dso symbols only to resolve currently loaded object code */
};

namespace bdk
{
	namespace backend
	{


		/**
		 * @brief 	A handle to dso
		 */
		typedef struct bc_os_dso_t* 	bc_dso_handle_t;


		/**
		 * @brief 	Load a dso
		 *
		 * @p libname 	Name of the dso file.
		 * @p rflags	Resolve mode.
		 * @p vflag	Visibility mode.
		 *
		 * @throws	System exceptions.
		 * @return 	Initialized dso handle.
		 */
		bc_dso_handle_t bc_dso_load ( const char* libname, bc_dso_resolve_t rflag, bc_dso_visibility_t vflag );


		/**
		 * @brief 	Unload a dso
		 *
		 * @p handle 	DSO loaded handle to close
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_dso_release ( bc_dso_handle_t& handle );


		/**
		 * @brief 	Load a symbol from dso
		 *
		 * @p handle 	DSO handle to load symbol from
		 * @p symname 	Symbol name
		 *
		 * @throws	System exceptions.
		 * @return 	Loaded symbol address. Can be NULL if symbol address is actually NULL.
		 */
		void*	 	bc_dso_find ( bc_dso_handle_t handle, const char* symname );


		/**
		 * @brief	Check if DSO is already resident
		 *
		 * @p libname	Library path
		 *
		 * @throws	System exceptions.
		 * @return	true if library is loaded, false if not.
		 */
		bool	 	bc_dso_is_resident ( const char* libname );



		/**
		 * @brief 	Get dso handle library file path
		 *
		 * @p handle	Loaded handle
		 *
		 * @return	Path string
		 */
		const char* 	bc_dso_path ( bc_dso_handle_t handle );

	}
}

#endif

