/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: io.h
 *	scope: bestcrypt core\backend interface
 *	author: evgeny yakovlev
 *
 *	description:
 *	file io backend interface
 *
 *	history:
 *
 */

#ifndef BDK_BACKEND_IO_H
#define BDK_BACKEND_IO_H

#include <vector>
#include <string>

#include <common/types.h>


/**
 * @brief File mode flags
 */
enum bc_file_flags_t
{
	BC_FILE_READ		= 0x0001, /**< Open to read */
	BC_FILE_WRITE		= 0x0002, /**< Open to write */
	BC_FILE_CREATE 		= 0x0004, /**< Create if file doesn't exist*/
	BC_FILE_APPEND		= 0x0008, /**< Append writes to the file's end */
	BC_FILE_TRUNC 		= 0x0010, /**< Truncate file contents when opening */
	BC_FILE_EXCL		= 0x0020, /**< If file exists and is being opened with BC_FILE_CREATE, then fail */
	BC_FILE_LARGE		= 0x0080, /**< Try to open file in largefile mode */
	BC_FILE_NONBLOCK	= 0x0100, /**< Non-blocking mode */
};


/**
 * @brief File types
 */
enum bc_file_type_t
{
	BC_TYPE_UNKNOWN		= 0x0000, /**< Could not determine */
	BC_TYPE_FILE		= 0x0001, /**< Regular file */
	BC_TYPE_DIR		= 0x0002, /**< Directory */
	BC_TYPE_LINK		= 0x0004, /**< Symlink */
	BC_TYPE_CDEV		= 0x0010, /**< Character device */
	BC_TYPE_BDEV		= 0x0020, /**< Block device */
};


/**
 * @brief File permissions flags
 */
enum bc_file_perm_t
{
	BC_PERM_UREAD 		= 0x0001, /**< Read by user */
	BC_PERM_UWRITE 		= 0x0002, /**< Write by user */
	BC_PERM_UEXEC		= 0x0004, /**< Exec by user */
	BC_PERM_SETUID		= 0x0008, /**< Set user id */

	BC_PERM_GREAD		= 0x0010, /**< Read by user's group */
	BC_PERM_GWRITE		= 0x0020, /**< Write by user's group */
	BC_PERM_GEXEC		= 0x0040, /**< Execute by user's group */
	BC_PERM_SETGID		= 0x0080, /**< Set user's group id */

	BC_PERM_OREAD		= 0x0100, /**< Read by others */
	BC_PERM_OWRITE		= 0x0200, /**< Write by others */
	BC_PERM_OEXEC		= 0x0400, /**< Execute by others */

	BC_PERM_COPY		= 0x1000, /**< When copying the file, set source file's permissions */
	BC_PERM_DEFAULT		= 0x2000, /**< Use OS's default permission values */
};


/**
 * @brief File seek flags
 */
enum bc_file_seek_t
{
	BC_SEEK_CUR		= 0x0001, /**< Seek relative to current position*/
	BC_SEEK_SET		= 0x0002, /**< Seek absolute position */
	BC_SEEK_END		= 0x0004, /**< Fast forward to the end */
};


/**
 * @brief File access mode
 */
enum bc_file_access_t
{
	BC_ACCESS_EXISTS	= 0x0001, /**< File exists */
	BC_ACCESS_READ		= 0x0002, /**< File can be read */
	BC_ACCESS_WRITE		= 0x0004, /**< File can be written */
	BC_ACCESS_EXEC		= 0x0008, /**< File can be executed */
};

namespace bdk
{
	namespace backend
	{

		/**
		 * @brief 	System file handle.
		 */
		typedef struct bc_os_file_t* 	bc_file_handle_t;


		/**
		 * @brief 	Open\create file.
		 *
		 * @p fname 	System file path.
		 * @p mode 	Access flags as defined in bc_file_mode_t.
		 * @p perm 	Permissions when creating new file as defined in bc_file_perm_t.
		 *
		 * @throws	System exceptions.
		 * @return 	Initialized file handle.
		 */
		bc_file_handle_t bc_file_open ( const char* fname, bcu32_t mode, bcu32_t perm = BC_PERM_DEFAULT );


		/**
		 * @brief	Create and open new file.
		 *
		 * @p path	File path
		 * @p mode	File open mode
		 * @p perm	New file permissions.
		 *
		 * @throws 	System exceptions.
		 * @return	Initialized system file handle.
		 */
//		bc_file_handle_t bc_file_create( const char* path, bcu32_t mode, bcu32_t perm = BC_PERM_DEFAULT );


		/**
		 * @brief 	Close file
		 *
		 * @p handle 	File handle to close
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_file_close ( bc_file_handle_t& handle );


		/**
		 * @brief 	Read from file.
		 *
		 * @p handle 	File handle to read from.
		 * @p buf 	Buffer to read into.
		 * @p nbytes 	Number of bytes to read.
		 *
		 * @throws	System exceptions.
		 * @return 	Number of bytes really read.
		 *		If number of bytes read is less than asked this is an EOF, check bc_file_eof to verify.
		 */
		bc_ulong_t 	bc_file_read ( bc_file_handle_t handle, bc_byte_t* buf, size_t nbytes );


		/**
		 * @brief 	Write to file
		 *
		 * @p handle 	File handle to write to
		 * @p buf 	Buffer to write from
		 * @p nbytes 	Number of bytes to write
		 *
		 * @throws	System exceptions.
		 * @return 	Number of bytes written.
		 */
		bc_ulong_t 	bc_file_write ( bc_file_handle_t fd, const bc_byte_t* buf, size_t nbytes );


		/**
		 * @brief 	Read file at specified offset
		 *
		 * @p handle 	File handle to read from
		 * @p buf 	Buffer to read into
		 * @p nbytes 	Number of bytes to read
		 * @p offset 	File position to read from
		 *
		 * @throws	System exceptions.
		 * @return 	Number of bytes really read.
		 *		If number of bytes read is less than asked this is an EOF, check bc_file_eof to verify.
		 */
		bc_ulong_t 	bc_file_pread ( bc_file_handle_t fd, bc_byte_t* buf, size_t nbytes, bc_ulong_t offset );


		/**
		 * @brief 	Write to file at specified offset
		 *
		 * @p handle 	File handle to write to
		 * @p buf 	Buffer to write from
		 * @p nbytes 	Number of bytes to write
		 * @p offset 	File position to write from
		 *
		 * @throws	System exceptions.
		 * @return 	Number of bytes written.
		 */
		bc_ulong_t 	bc_file_pwrite ( bc_file_handle_t handle, const bc_byte_t* buf, size_t nbytes, bc_ulong_t offset );


		/**
		 * @brief 	Delete file
		 *
		 * @p path 	System file path
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_file_remove ( const char* path );


		/**
		 * @brief 	Rename file
		 *
		 * @p old_path 	Old (existing) file path
		 * @p new_path 	New file path
		 *
		 * @throws	System exceptions.
		 */
		void	 	bc_file_rename ( const char* old_path, const char* new_path );


		/**
		 * @brief 	Check file access mode
		 *
		 * @p fname 	File path
		 * @p flags 	Access flags as in bc_file_access_t
		 *
		 * @throws	System exceptions.
		 * @return	true if specified access is allowed, false if not.
		 */
		bool	 	bc_file_touch ( const char* fname, bc_uword_t flags );


		/**
		 * @brief 	Get file type
		 *
		 * @p fname 	File name
		 *
		 * @throws	System exceptions.
		 * @return 	bc_file_type_t value.
		 */
		bc_file_type_t 	bc_file_type ( const char* fname );


		/**
		 * @brief 	Seek file position
		 *
		 * @p handle 	File handle
		 * @p where 	Seek flag
		 * @p offset 	Seek position
		 *
		 * @throws	System exceptions.
		 * @return	New offset.
		 */
		bc_ulong_t 	bc_file_seek ( bc_file_handle_t handle, bc_file_seek_t where, bc_ulong_t offset );


		/**
		 * @brief 	Tell file's current position
		 *
		 * @p handle	File handle
		 *
		 * @throws	System exceptions.
		 * @return	Current offset.
		 */
		bc_ulong_t 	bc_file_tell ( bc_file_handle_t handle );


		/**
		 * @brief 	Tell if EOF was hit
		 *
		 * @p handle	File handle
		 *
		 * @return	true if hit, false if not.
		 */
		bool		bc_file_is_eof ( bc_file_handle_t handle );


		/**
		 * @brief	Get assosiated file name
		 *
		 * @p handle	File handle to get file name for.
		 *
		 * @return	File name string pointer.
		 */
		const char* 	bc_file_name ( bc_file_handle_t handle );


		/**
		 * @brief	List directory
		 *
		 * @p path	Directory path
		 * @p list	Entry names storage. Will be cleared.
		 *
		 * @throws	System exceptions.
		 * @return	Number of elements.
		 */
		size_t		bc_file_dirlist ( const char* path, std::vector<std::string>& list );


		/**
		 * @brief	Canonize fs path.
		 *
		 * @p path	Path to canonize.
		 */
		void		bc_file_canonize_path ( std::string & path );

		bool 		bc_file_is_owner ( const char* path );

	} // namespace backend
}

#endif


