#ifndef BDK_ENCRYPTED_TYPE_H
#define BDK_ENCRYPTED_TYPE_H

#include <common/build.h>
#include <alg/alg.h>

namespace bdk
{

	/**
	 * @brief	Atomary block size for encryption operations in bytes
	 */
	enum 
	{ 
		BLOCK_BYTES = 16,
		DIGEST_BYTES = 32,
	};


	/**
	 * @brief	Type storage in encrypted form.
	 * @more	Will not store encryption key, iv or any other algorithm information, must be user supplied.
	 * @more	Usage in tandem with TBCVerifiedType is recommended to add digest verification after decryption.
	 */
	template < class T >
	class BCEncryptedType
	{
	public:


		typedef bc_padded_type_t<T, BLOCK_BYTES> padded_type;


		BCEncryptedType()
		{
			BC_DEBUG_PRINT(("Debug: encrypted type: body size 0x%x, padding size 0x%x\n", sizeof(T), sizeof(m_data) - sizeof(T)));
			BC_ASSERT((sizeof(padded_type) % BLOCK_BYTES) == 0);
			BC_ASSERT(sizeof(*this) == sizeof(padded_type));
		}


		T* type()
		{
			return m_data.type();
		}

		const T* type() const
		{
			return m_data.type();
		}

		/**
		 * @brief	Initialize this encrypted block by encrypting data buffer
		 *
		 * @p from	Block data buffer
		 * @p alg	Configured algorithm
		 * @p iv	Initial vector
		 */
		void encrypt(BCAlgHandle alg, bc_byte_t const * iv)
		{
			alg->encrypt(iv, (bc_byte_t*)&m_data, sizeof(m_data));
		}


		/**
		 * @brief	Decrypt current block data to external buffer
		 *
		 * @p to	Block buffer to decrypt to
		 * @p alg	Algorithm configuration
		 * @p iv	Initial vector
		 */
		void decrypt(BCAlgHandle alg, const bc_byte_t* iv)
		{
			alg->decrypt(iv, (bc_byte_t*)&m_data, sizeof(m_data));
		}


	private:

		/* 	we need to give algorithms a contigious space of everything we need to encrypt\decrypt 
			or else we will have to deal with manual iv synchronization problems.
		*/
		padded_type 	m_data;
	};


	void calculate_digest( int hash_id, bc_byte_t* digest, const bc_byte_t* text, unsigned int text_size );


	bool verify_digest( int hash_id, const bc_byte_t* digest, bc_byte_t* text, unsigned int text_size );


	/**
	 * @brief	Type data stored with it's digest.
	 */
	template < class T >
	class BCVerifiedType
	{
	public:

		typedef bc_padded_type_t<T, BLOCK_BYTES> padded_type;


		BCVerifiedType()
		{
			//BC_DEBUG_PRINT(("Debug: verified type: body size 0x%x, padding size 0x%x, digest size 0x%x\n", sizeof(T), sizeof(m_data) - sizeof(T), sizeof(m_digest)));
			BC_ASSERT((sizeof(padded_type) % BLOCK_BYTES) == 0);
			BC_ASSERT(sizeof(*this) == (sizeof(padded_type) + sizeof(m_digest)));
		}


		T* type()
		{
			return m_data.type();
		}

		const T* type() const
		{
			return m_data.type();
		}

		void store(T const * from, int hash_id )
		{
			//shred((unsigned char*)&m_data, sizeof(padded_type));
			memcpy(this->type(), from, sizeof(padded_type));
			calculate_digest(hash_id, m_digest, (bc_byte_t*)this->type(), sizeof(T));
		}


		bool verify(int hash_id)
		{
			//BC_DEBUG_PRINT(("Debug: sizeof verified type 0x%x\n", sizeof(T)));
			return verify_digest(hash_id, m_digest, (bc_byte_t*)this->type(), sizeof(T));
		}
		
	private:

	
		/* type is padded due to historical reasons - no real need */
		bc_padded_type_t<T, BLOCK_BYTES> m_data;

		/* digest of padded type */
		bc_byte_t m_digest[DIGEST_BYTES];
	};

}


#endif

