/*
 * file.h
 *
 *  Created on: Sep 9, 2009
 *      Author: inso
 */

#ifndef BDK_COMMON_FILE_H
#define BDK_COMMON_FILE_H

#include <common/smartptr.h>
#include <common/exception.h>
#include <common/fs.h>
#include <backend/fs.h>

namespace bdk
{

        class BCFile;

        /**
         * @brief	File shared handle.
         */
        typedef BCSharedPtr <BCFile> BCFileHandle;

        /**
         * @brief 	File abstraction.
         * @more	Provides lifetime control for system file object.
         */
        class BCFile
        {
                BC_FORBID_COPY( BCFile);
        public:

                /**
                 * @brief 	Open\create file.
                 * @more	Sets object in "opened" invariant.
                 *		To represent closed files and other static file system objects use BCFSObject.
                 *
                 * @p fname	File path
                 * @p flags	File open flags
                 * @p mode	Mode flags, if file is opened to create
                 *
                 * @throws	BCSystemException if system object initialization fails.
                 */
                BCFile(std::string const & fname, bc_word_t flags, bc_word_t mode = BC_PERM_DEFAULT) throw(BCSystemException);

                /**
                 * @brief 	Close file
                 */
                ~BCFile() throw();

                /**
                 * @brief	Static constructor to open existing file
                 *
                 * @p fname	File name.
                 * @p flags	File open flags.
                 *
                 * @return 	Handle to file object.
                 */
                static BCFileHandle open(std::string const & fname, bc_word_t flags)
                                throw(BCSystemException);

                /**
                 * @brief	Static constructor to open existing file
                 *
                 * @p fsobj	File fs object.
                 * @p flags	File open flags.
                 *
                 * @return 	Handle to file object.
                 */
                static BCFileHandle open(BCFSObject const & fsobj, bc_word_t flags)
                {
                        return BCFile::open(fsobj.path(), flags);
                }

                /**
                 * @brief	Static constructor to create and open new file
                 *
                 * @p fname	File name.
                 * @p flags	File open flags.
                 * @p mode	Permission mode for new file.
                 *
                 * @return 	Handle to file object.
                 */
                static BCFileHandle create(std::string const & fname, bc_word_t flags,
                                           bc_word_t mode = BC_PERM_DEFAULT)
                                throw(BCSystemException);

                /**
                 * @brief	Static constructor to create and open new file
                 *
                 * @p fsobj	File fs object.
                 * @p flags	File open flags.
                 * @p mode	Permission mode for new file.
                 *
                 * @return 	Handle to file object.
                 */
                static BCFileHandle create(BCFSObject const & fsobj, bc_word_t flags,
                                           bc_word_t mode = BC_PERM_DEFAULT)
                {
                        return BCFile::create(fsobj.path(), flags, mode);
                }

                /**
                 * @brief	Get invalid file handle instance.
                 */
                static BCFileHandle invalid_file()
                {
                        return BCFileHandle(NULL);
                }

                /**
                 * @brief 	Read from file's current position
                 *
                 * @p dst	Destination buffer
                 * @p size	Size of single unit to read
                 * @p len	Number of units to read
                 */
                bc_ulong_t read(void *dst, size_t size, size_t len) throw(BCSystemException);

                /**
                 * @brief 	Write to file at current position
                 *
                 * @p src	Source buffer
                 * @p size 	Size of single unit to write
                 * @p len	Number of units to write
                 */
                bc_ulong_t write(const void* src, size_t size, size_t len) throw(BCSystemException);

                /**
                 * @brief 	Read from file at position
                 *
                 * @p dst	Destination buffer
                 * @p size	Size of single unit to read
                 * @p len	Number of units to read
                 * @p offset	File absolute position offset
                 */
                bc_ulong_t pread(void* dst, size_t size, size_t len, bc_ulong_t offset)
                                throw(BCSystemException);

                /**
                 * @brief 	Write to file at position
                 *
                 * @p src	Source buffer
                 * @p size	Size of single unit to read
                 * @p len	Number of units to read
                 * @p offset	File absolute position offset
                 */
                bc_ulong_t pwrite(const void* src, size_t size, size_t len, bc_ulong_t offset)
                                throw(BCSystemException);

                /**
                 * @brief 	Tell file's current position
                 */
                bc_ulong_t tell() const throw(BCSystemException);

                /**
                 * @brief 	Seek file position
                 *
                 * @p offset	Offset to seek
                 * @p where	Where to seek offset from
                 */
                bc_ulong_t seek(bc_ulong_t offset, bc_file_seek_t where) throw(BCSystemException);

                /**
                 * @brief 	Was end of file hit?
                 */
                bool is_eof() const;

                /**
                 * @brief	Get file platform handle.
                 * @more	Use at your own risk.
                 */
                backend::bc_file_handle_t system_handle() const
                {
                        return m_fd;
                }

                /**
                 * @brief	Return file system path.
                 */
                std::string system_path() const;

                /**
                 * @brief	Return static fs object of this file.
                 */
                BCFSObject fs_object() const
                {
                        return BCFSObject(this->system_path());
                }

        private:

                /* backend file handle */
                backend::bc_file_handle_t m_fd;
        };

}

#endif
