/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: io.h
 *	scope: bdk\common
 *	author: evgeny yakovlev
 *
 *	description:
 *	Filesystem objects.
 *
 *	history:
 *
 */

#ifndef BDK_COMMON_FS_H
#define BDK_COMMON_FS_H

#include <backend/fs.h>

#include <common/types.h>
#include <common/class_defs.h>
#include <common/smartptr.h>
#include <common/memory.h>

#include <alg/alg.h>

namespace bdk
{

/**
 * @brief	Filesystem object.
 * @more	Represents static file system objects (i.e. a file, directory, link, etc.).
 */
class BCFSObject
{
public:

	/**
	 * @brief	Empty fs object
	 */
	BCFSObject()
	{
	}
	;

	/**
	 * @brief	Construct fs object from path
	 */
	explicit BCFSObject(std::string const &path) :
		m_path(path)
	{
	}
	;

	/**
	 * @brief	Set fs object path
	 */
	void set_path(std::string const & path)
	{
		m_path = path;
	}

	/**
	 * @brief	Canonize fs object path
	 */
	void canonize();

	/**
	 * @brief	Get type of filesystem object
	 */
	bc_file_type_t type() const;

	/**
	 * @brief	Check fs object access mode
	 */
	bool access(bcu32_t mode) const;

	/**
	 * @brief	Does this fs object exist?
	 */
	bool is_valid() const
	{
		return this->access(BC_ACCESS_EXISTS);
	}

	/**
	 * @brief	Is this a special file (i.e., device file)?
	 */
	bool is_special() const
	{
		return (this->type() == BC_TYPE_BDEV) || (this->type() == BC_TYPE_CDEV);
	}

	/**
	 * @brief	Is this a regular file?
	 */
	bool is_file() const
	{
		return (this->type() == BC_TYPE_FILE);
	}

	/**
	 * @brief	Is this a block device?
	 */
	bool is_directory() const
	{
		return (this->type() == BC_TYPE_DIR);
	}

	/**
	 * @brief	Get path string
	 */
	std::string path() const
	{
		return m_path;
	}

	/**
	 * @brief	Get fs object access rights.
	 */
	bcu32_t rights() const;

	bool is_owner() const;

	/**
	 * @brief	Get fs object i\o blocks map.
	 */
	//bc_space_unit_t	block_map() const;


	/**
	 * @brief	Remove fs object.
	 */
	void remove() const;

	/**
	 * @brief	Check two fs objects for path equality
	 */
	bool operator ==(BCFSObject const & other)
	{
		return path() == other.path();
	}

	/**
	 * @brief	Check two fs objects for path inequality
	 */
	bool operator !=(BCFSObject const & other)
	{
		return !(*this == other);
	}

private:

	/* object path */
	std::string m_path;
};

/**
 * @brief	Serialize fs object info.
 */
std::ostream& operator <<(std::ostream & stream, BCFSObject const & fsobj);



}

#endif

