#ifndef BDK_COMMON_LOG_H
#define BDK_COMMON_LOG_H

#include <list>
#include <sstream>

#include <common/class_defs.h>

namespace bdk 
{


	/**
	 * @brief Log message level
	 */
	enum bc_msg_level
	{
		BC_MSG_INFO	= 0,		/**< Info message */
		BC_MSG_WARN	= 1,		/**< Warning message */
		BC_MSG_CRIT	= 2,		/**< Critical message */
	};
	

	/**
	 * @brief	Log subscriber interface to implent actual logging output.
	 */
	class BCLogListener
	{
	BC_INTERFACE(BCLogListener)
	public:


		/**
		 * @brief	Callback notification about new message to log
		 */
		virtual 	void log_message(std::string const & msg, bc_msg_level lvl) = 0;

	};

	

	/**
	 * @brief 	Log object as a message publisher.
	 */
	class BCLog
	{
	public:


		/**
		 * @brief	Log stream.
		 * @more	Lives during single message logging session.
		 *		Caches the message during lifetime, flushes to log on destruction.
		 */
		class BCMessageStream
		{
		public:


			/**
			 * @brief	Create new log message stream.
			 */
			BCMessageStream(BCLog* parent, bc_msg_level lvl) : m_parent(parent), m_lvl(lvl)
			{
				m_stream.clear();
			}


			/**
			 * @brief	Clone log message stream.
			 */
			BCMessageStream(BCMessageStream const & rhv) : m_parent(rhv.m_parent), m_lvl(rhv.m_lvl)
			{
				m_stream.str(rhv.m_stream.str());
			}


			/**
			 * @brief	Insert in log message stream
			 */
			template < class T >
			BCMessageStream& operator << (T const & obj)
			{
				m_stream << obj;
				return *this;
			}

		
			/**
			 * @brief	Flush and close log message stream.
			 */
			~BCMessageStream() throw()
			{
				m_parent->put(m_stream.str(), m_lvl);
			}


		private:

			/* cached stream */
			std::ostringstream m_stream;
			
			/* parent log */
			BCLog* m_parent;

			/* message level */
			bc_msg_level m_lvl;

		};


		/**
		 * @brief	Register log subscriber
		 */
		void register_listener(BCLogListener* listener);


		/**
		 * @brief	Unregister log subscriber
		 */
		void unregister_listener(BCLogListener* listener);


		/**
		 * @brief 	Close log.
		 */
		~BCLog() throw();


		/**
		 * @brief 	Put single message string
		 */
		void put(std::string const & str, bc_msg_level lvl);
			

		/**
		 * @brief	Start logging stream.
		 */
		BCMessageStream msg(bc_msg_level lvl)
		{
			return BCMessageStream(this, lvl);
		}


	private:


		typedef std::list<BCLogListener*> 	log_listeners;
		typedef log_listeners::iterator		log_listeners_iter;
		
		/* subscribers list */
		log_listeners				m_listeners;

	};


	/**
	 * @brief Get default bdk log
	 */
	BCLog& bc_get_log();


	BCLog::BCMessageStream bc_get_log(bc_msg_level lvl);
}



#endif 
