/**

	Copyright (c) 2009 Jetico, Inc., Finland
	All rights reserved.

	file: smartptr.h
	scope: bdk\common
	author: evgeny yakovlev
	purpose: definitions of resource controlled objects with pointer semantics.
		stolen shamelessly from boost.

	history:
	- 23.06.09
	Merged from shared.h, scoped.h.
	Added weak pointer semantics.

*/

#ifndef BDK_COMMON_SMARTPTR_H
#define BDK_COMMON_SMARTPTR_H

#include <stdlib.h>
#include <common/exception.h>
#include <common/build.h>
#include <common/class_defs.h>

namespace bdk
{


	/**
	 * @brief	Exceptional error case of addressing uninitialized\invalidated pointers
	 */
	class BCBadPtr : public BCException
	{
	public:
		BCBadPtr(const std::string& str) : BCException(str) {}
	};


	// ------------------------------------------------------------
	//
	//	Scoped
	//
	// ------------------------------------------------------------


	/**
	 * @brief	Single object ownership, no-strong reference semantics.
	 * @more	Ensures that at any time there exists only one single strong owner of the object.
	 *		Object's lifetime is bound to owner's lifetime.
	 */
	template < class T >
	class BCScopedPtr
	{
	BC_FORBID_COPY(BCScopedPtr<T>);
	public:


		/**
		 * @brief	Construct empty scoped ptr.
		 */
		BCScopedPtr() : m_pointee(NULL)
		{

		}


		/**
		 * @brief	Aquire strong non-shareable ownership of an object
		 */
		explicit BCScopedPtr( T* pointee ) : m_pointee(pointee)
		{

		}


		/**
		 * @brief	Destroy referenced object.
		 */
		~BCScopedPtr() throw()
		{
			if( m_pointee ) {
				BC_DELETE( T, m_pointee );
			}
		}


		T* operator -> () const
		{
			BC_ASSERT(m_pointee != NULL);
			return m_pointee;
		}


		T& operator * () const
		{
			BC_ASSERT(m_pointee == NULL);
			return *m_pointee;
		}


		T* get() const
		{
			return m_pointee;
		}


		void reset( T* value )
		{
			BC_ASSERT( (m_pointee != value ) ); // self-reset is an error
			BCScopedPtr<T>(value).swap(*this); // swap and delete temporary with our pointer
		}


		void swap( BCScopedPtr<T>& rhv )
		{
			T* tmp = m_pointee;
			m_pointee = rhv.m_pointee;
			rhv.m_pointee = tmp;
		}

	private:

		T* m_pointee;

	};


	template < class T >
	bool operator == (const BCScopedPtr<T>& lhv, const BCScopedPtr<T>& rhv)
	{
		return lhv.get() == rhv.get();
	}


	// ------------------------------------------------------------
	//
	//	Shared
	//
	// ------------------------------------------------------------


	/**
	 * @brief	Interface for reference counted objects.
	 */
	class IBCCounted
	{
	public:


		IBCCounted() : m_count(0)
		{
		}


		/**
		 * @brief	Release reference.
		 */
		void release()
		{
			ATOMIC_DEC(m_count);
			BC_ASSERT(m_count >= 0);
			if(m_count == 0)
			{
				this->on_delete();
				delete this;
			}
		}


		/**
		 * @brief	Retain reference.
		 */
		void retain()
		{
			ATOMIC_INC(m_count);
		}


		/**
		 * @brief	Get reference count.
		 */
		unsigned int count() const
		{
			return m_count;
		}


		virtual ~IBCCounted() {};


	protected:


		/**
		 * @brief	Final release hook
		 */
		virtual void on_delete() = 0;


	private:

		/* refcount */
		int volatile m_count;
	};


	/**
	 * @brief	Reference counted pointer.
	 */
	template < class T >
	class BCCountedP : public IBCCounted
	{
	BC_FORBID_COPY(BCCountedP);
	public:

		BCCountedP(T* p) : m_p(p) {};

	protected:

		void on_delete()
		{
			BC_DELETE(T,m_p);
		}

	private:

		T*	m_p;
	};


	/**
	 * @brief	Reference counted pointer with custom deleter functor
	 */
	template < class T, class D >
	class BCCountedPD : public IBCCounted
	{
	BC_FORBID_COPY(BCCountedPD);
	public:

		BCCountedPD(T* p, D const & d) : m_p(p), m_deleter(d) {}

	protected:

		void on_delete()
		{
			m_deleter(m_p);
		}

	private:

		T* m_p;
		D m_deleter;
	};


	class BCCountedNULL : public IBCCounted
	{
	BC_FORBID_COPY(BCCountedNULL);
	public:

		BCCountedNULL() {};

	protected:

		void on_delete() {};
	};


	/**
	 * @brief 	Smart pointer with refcount.
	 */
	template < class T >
	class BCSharedPtr
	{
	public:


		BCSharedPtr() : m_p(NULL)
		{
			//BC_DEBUG_PRINT(("Debug: creating NULL shared ptr\n"));
			m_shared_count = new BCCountedNULL;
			m_shared_count->retain();
		}


		explicit BCSharedPtr( T* p ) : m_p(p)
		{
			//BC_DEBUG_PRINT(("Debug: creating P shared ptr\n"));
			m_shared_count = new BCCountedP <T> (p);
			m_shared_count->retain();
		}


		template < class Y, class D >
		BCSharedPtr(Y* p, D const & d) : m_p(p)
		{
			//BC_DEBUG_PRINT(("Debug: creating PD shared ptr\n"));
			m_shared_count = new BCCountedPD <Y,D> (p,d);
			m_shared_count->retain();
		}


		BCSharedPtr( BCSharedPtr const & rhv ) : m_p(rhv.m_p)
		{
			//BC_DEBUG_PRINT(("Debug: copying shared ptr\n"));
			m_shared_count = rhv.m_shared_count;
			m_shared_count->retain();
		}


		BCSharedPtr& operator = ( BCSharedPtr const & rhv )
		{
			//BC_DEBUG_PRINT(("Debug: assigning shared ptr\n"));
        		IBCCounted* tmp = rhv.m_shared_count;

        		if( tmp != m_shared_count )
        		{
            			tmp->retain();
            			m_shared_count->release();
            			m_shared_count = tmp;
				m_p = rhv.m_p;
        		}

        		return *this;
		}


		~BCSharedPtr()
		{
			//BC_DEBUG_PRINT(("Debug: releasing shared ptr\n"));
			m_shared_count->release();
		}


		T* get() const
		{
			return m_p;
		}


		T* operator -> () const
		{
			BC_ASSERT(m_p);
			return m_p;
		}


		T& operator * () const
		{
			BC_ASSERT(m_p);
			return *m_p;
		}


		operator bool () const
		{
			return m_p != NULL;
		}


		unsigned int refcount() const
		{
			return m_shared_count->count();
		}


	private:

		T* m_p;

		IBCCounted*	m_shared_count;

#if 0
		/* object we're pointing at */
		reference_type*	m_pointee;

		/* optional deleter */
		deleter*	m_deleter;

		/* shared refcounter for this object, source of exceptions since needs heep allocation */
		BCRefcount* 	m_refcounter;
#endif
	};

	/**
	 * @brief	Create shared handle instance
	 */
	template < class T > BCSharedPtr<T> bc_make_handle(T* obj)
	{
		return BCSharedPtr<T>(obj);
	}

	template < class T>
	bool operator == (const BCSharedPtr<T>& lhv, const BCSharedPtr<T>& rhv)
	{
		return lhv.get() == rhv.get();
	}


	template < class T >
	bool operator != (const BCSharedPtr<T>& lhv, const BCSharedPtr<T>& rhv)
	{
		return !(lhv == rhv);
	}

}


#endif
