/**
*
*	Copyright (c) 2009 Jetico, Inc., Finland
*	All rights reserved.
*
*	file: format.h
*	scope: core\container
*	author: evgeny yakovlev
*
*	description:
*	Container format adaptor interface.
*
*/


#ifndef BDK_CONTAINER_FORMAT_H
#define BDK_CONTAINER_FORMAT_H

#include <string>

#include <common/types.h>
#include <container/io.h>
#include <container/view.h>



namespace bdk
{

	enum bc_view_change_t
	{
		VIEW_ADD_AUTH		= 0,
		VIEW_REMOVE_AUTH	= 1,
		VIEW_CHANGE_AUTH	= 2,
	};


	/**
	 * @brief 	Loaded container interface
	 * @more 	This interface is availiable after authentication by IBCContainer_Authenticator.
	 *		It represents loaded and authenticated container with specific data context initialized by authentication.
	 *		It allows modifications to container's internal storage and meta info.
	 */
	class IBCContainerFormat
	{
	BC_FORBID_COPY(IBCContainerFormat)
	BC_INTERFACE(IBCContainerFormat)


	//-----------------------------------------------------------------
	//	Requests
	public:


		/**
		 * @brief	Format name.
		 */
		virtual		std::string name() const = 0;


                /**
		 * @brief 	Get container description string
		 */
                virtual 	std::string description() const = 0;


		/**
		 * @brief	Get container whole useful size in bytes, including all contexts
		 */
                virtual 	bc_ulong_t size_bytes () const = 0;


                /**
		 * @brief 	Container algorithm id number
		 */
                virtual 	bc_alg_id_t algorithm_id() const = 0;


		/**
		 * @brief 	Container algorithm encryption mode id
		 */
		virtual 	bc_alg_mode_id_t algorithm_mode_id() const = 0;


                /**
		 * @brief 	Ask if loaded container is locked
		 */
                virtual 	bool is_locked() const = 0;


 		/**
                 * @brief 	Serialize container service info.
                 *
                 * @p source	Data buffer for storage
                 */
                virtual 	void serialize( IBCContainerIO* dest ) const = 0;




	//-----------------------------------------------------------------
	//	Operations
	public:



		/**
                 * @brief 	Lock loaded container
                 */
                virtual 	void lock() = 0;


		/**
                 * @brief 	Unlock loaded container
                 */
                virtual 	void unlock() = 0;


		/**
		 * @brief 	Set container description string
		 */
		virtual		void set_description(std::string const & desc) = 0;


		/**
		 * @brief	Set container whole useful size in bytes, including all contexts
		 */
		virtual		void set_size_bytes(bc_ulong_t size) = 0;


		/**
		 * @brief 	Set container algorithm id number
		 */
		virtual 	void set_algorithm_id(bc_alg_id_t id) = 0;


		/**
		* @brief 	Set container algorithm encryption mode id
		*/
		virtual 	void set_algorithm_mode_id( bc_alg_mode_id_t ) = 0;


		/**
		 * @brief	Authentificate container.
		 *
		 * @p fmt	Container format implementation for info request.
		 * @p ctx	Loaded data context.
		 *
		 * @return	Passed\failed authentication.
		 */
	//	virtual		bool auth(bc_data_context_t& ctx) = 0;


		/**
		 * @brief	Create authentication for container and load context.
		 *
		 * @p fmt	Container format implementation for info request.
		 * @p ctx	Loaded data context.
		 *
		 * @return	Passed\failed authentication.
		 */
	//	virtual		bool create_auth(bc_data_context_t& ctx) = 0;


		/**
		 * @brief	Add hidden auth context.
		 *
		 * @p ctx	Context to initialize.
		 * @p size	Size of data that hidden context will address.
		 */
	//	virtual		bool add_hidden_auth(bc_data_context_t& ctx, bc_ulong_t size) = 0;


		/**
		 * @brief	Authentificate container and add another auth method to it.
		 *
		 * @p fmt	Container format implementation for info request.
		 * @p ctx	Loaded data context.
		 *
		 * @return	Passed\failed authentication.
		 */
	//	virtual		bool auth_and_add(bc_data_context_t& ctx) = 0;


		/**
		 * @brief	Authentificate container and change auth method.
		 *
		 * @p fmt	Container format implementation for info request.
		 * @p ctx	Loaded data context.
		 *
		 * @return	Passed\failed authentication.
		 */
	//	virtual 	bool auth_and_change(bc_data_context_t& ctx) = 0;


		/**
		 * @brief	Authentificate container and remove auth method.
		 *
		 * @p fmt	Container format implementation for info request.
		 * @p ctx	Loaded data context.
		 *
		 * @return	Passed\failed authentication.
		 */
	//	virtual 	bool auth_and_remove(bc_data_context_t& ctx) = 0;


		/**
		 * @brief	Create main container view.
		 */
		virtual		BCContainerViewHandle create_view() = 0;


		/**
		 * @brief	Create hidden container view.
		 */
		virtual		BCContainerViewHandle create_hidden_view(IBCContainerIO* io, bc_ulong_t size) = 0;


		/**
		 * @brief	Get a stored view handle.
		 */
		virtual 	BCContainerViewHandle get_view(IBCContainerIO* io) = 0;


		/**
		 * @brief	Get a stored view handle and update auth binding for it (change\add\remove auth secrets).
		 */
		virtual 	BCContainerViewHandle update_view(IBCContainerIO* io, bc_view_change_t op) = 0;

	};

}


#endif

