/**
*
*	Copyright (c) 2009 Jetico, Inc., Finland
*	All rights reserved.
*
*	file: manager.h
*	scope: bdk\container
*	author: evgeny yakovlev
*
*	description: 
*	Manager of registered container types and their factories
*
*/

#ifndef BDK_CONTAINER_MANAGER_H
#define BDK_CONTAINER_MANAGER_H

#include <list>
#include <string>

#include <container/io.h>

namespace bdk
{
	class IBCContainerFactory;
}
	
namespace bdk
{

	/**
	 * @brief	Single registered product type entry
	 */
	struct bc_type_entry_t
	{
		/* name of the container type */
		std::string		type_name;
		
		/* implementing factory */
		IBCContainerFactory*	type_factory;
		
		/* does this source support unrecognized data ? */
		bool			type_unrecognized;


		/**
		* @brief	Check container types equality 
		*/
		bool operator == (bc_type_entry_t const & rhv)
		{
			return type_name == rhv.type_name;
		}
		
		
		/**
		* @brief	Check container types inequality 
		*/
		bool operator != (bc_type_entry_t const & rhv)
		{
			return !(*this == rhv);
		}
	};
	
		
	/**
	 * @brief	Manages registered container types and their factories.
	 */
	class BCContainerManager
	{
	BC_FORBID_COPY(BCContainerManager)
	public:
			

		BCContainerManager()
		{
		}


		/**
		 * @brief	Register new container product type
		 *
		 * @p name	Product name to search for when creating
		 * @p impl	Implementing factory
		 * @p support_unrecognized
		 *		This type can support unrecognized data sources.
		 */
		void 		register_type(std::string const & name, IBCContainerFactory* impl, bool support_unrecognized);
			
			
		/**
		 * @brief	Unregister container product
		 */
		void 		unregister_type(std::string const & name);
			
			
		/**
		 * @brief	Query container type by it's name.
		 * @more	Typically used to make new containers knowing desired type.
		 *
		 * @p name	Container type name.
		 * @p out_type	If requested container type is registered, this structure will be filled with needed details.
		 *
		 * @return	True if requested container type registered, false if not.
		 */
		bool		query_type_by_name(std::string const & name, bc_type_entry_t& out_type);
			
			
		/**
		 * @brief	Query container type by it's data.
		 * @more	This answers to the typical problem of loading container files of module-defined format.
		 *
		 * @p data	Source data to parse as a container format.
		 * @p out_type	If requested container type is registered, this structure will be filled with needed details.
		 *
		 * @return	true if container source was parsed by any factory, false otherwise.
		 */
		bool		query_type_by_data(IBCContainerIO* data, bc_type_entry_t& out_type);
			
			
	private:
			
			
		typedef std::list<bc_type_entry_t>	products_t;
		typedef products_t::iterator		products_iter_t;
			
			
	protected:
			
			
		/**
		 * @brief	Utility to get product iterator
		 *
		 * @return	Valid iterator or end() if not found
		 */
		products_iter_t _product_iter(std::string const & name)
		{
			products_iter_t i = m_products.begin();
			products_iter_t end = m_products.end();
				
			while(i != end)
			{
				if(i->type_name == name)
					break;
				++i;
			}
				
			return i;
		}
			
			
	private:

			
		/* factory module extentions list */
		products_t	m_products;
		
	};
	
}

#endif

