/**
*
*	Copyright (c) 2009 Jetico, Inc., Finland
*	All rights reserved.
*
*	file: type.h
*	scope: bdk\container
*	author: evgeny yakovlev
*
*	description:
*	Container type as a format and authenticator implementors combination.
*
*/

#ifndef BDK_CONTAINER_TYPE_H
#define BDK_CONTAINER_TYPE_H

#include <string>

#include <container/format.h>
#include <common/smartptr.h>


namespace bdk
{
	class BCContainer;

	/**
	 * @brief	Shared instance handle.
	 */
	typedef BCSharedPtr < BCContainer > BCContainerHandle;
}


namespace bdk
{
	/**
	* @brief	Serializable container info
	*/
	struct bc_container_info_t
	{
		std::string	cont_type;
		int 		cont_alg_id;
		int 		cont_mode_id;
		int		cont_hash_id;
		int		cont_fs_id;
		int		cont_auth_id;

		bc_ulong_t	cont_size;
		bool		cont_locked;

		std::string	cont_description;
	};



	/**
	 * @brief	Virtual encrypted data storage.
	 */
	class BCContainer
	{
	BC_FORBID_COPY(BCContainer)
	public:


		/**
		 * @brief	Create container of specified format authenticated by specified authenticator.
		 */
		BCContainer(IBCContainerFormat* fmt, std::string const & type, bool recognized) :
			m_format(fmt),
			m_type(type),
			m_recognized(recognized)
		{
		}


		/**
		 * @brief	Release container.
		 */
		~BCContainer()
		{
		}


		/**
		 * @brief	Initialize shared container handle.
		 */
		static		BCContainerHandle init(IBCContainerFormat* fmt, std::string const & type, bool recognized);


		/**
		 * @brief	Static construction of container from type name and data
		 *
		 * @p data	Virtual data buffer.
		 * @p type	Container type name.
		 *
		 * @return	Initialized handle or NULL if type name is unknown.
		 */
		static 		BCContainerHandle load_by_type(IBCContainerIO* data, std::string const & type);


		/**
		 * @brief	Static construction of container from data only
		 *
		 * @p data	Virtual data buffer.
		 *
		 * @return	Initialized handle or NULL if container type could not be parsed from data.
		 */
		static 		BCContainerHandle load_by_data(IBCContainerIO* data);


		/**
		 * @brief	Static construction of new container type.
		 *
		 * @p type	Container type name.
		 * @p algid	Block cypher id to use with container
		 * @p algmode	Block cypher mode
		 * @p size	Container useful size in bytes
		 * @p desc	Optional description
		 * @p modname	Name of the module that will create the container
		 */
		static		BCContainerHandle make( std::string const & type, int algid, int algmode, bc_size_t size, std::string const & desc );


		/**
		 * @brief	Get container format implementor.
		 */
		IBCContainerFormat* format() const
		{
			return m_format.get();
		}


		/**
		 * @brief	Get container type name.
		 */
		std::string	type() const
		{
			return m_type;
		}


		/**
		 * @brief	Was container format recognized?
		 */
		bool 	is_recognized() const
		{
			return m_recognized;
		}


	private:

		/* format implementation */
		BCScopedPtr < IBCContainerFormat > 	m_format;

		/* type name */
		std::string				m_type;

		/* recognized flag */
		bool					m_recognized;
	};
}


#endif

