#ifndef BDK_CONTAINER_VIEW_H
#define BDK_CONTAINER_VIEW_H

#include <common/smartptr.h>
#include <container/io.h>

struct part_info_t;

namespace bdk
{
	class BCContainerView;

	/**
	 * @brief	Container view shared handle.
	 */
	typedef BCSharedPtr<BCContainerView> BCContainerViewHandle;
}

namespace bdk
{

	/**
	 * @brief	Container persistent binding.
	 * @more	Initializes system crypto loop device on construction, releases on destruction.
	 *		Use locking to make the binding persistent.
	 */
	class BCContainerAutoBinding
	{
	public:


		/**
		 * @brief	Init binding from already linked device
		 */
		BCContainerAutoBinding(BCDeviceHandle device);


		/**
		 * @brief	Init binding.
		 *
		 * @p io	Container data.
		 * @p view	Container view.
		 * @p device	Crypto device to bind to.
		 */
		BCContainerAutoBinding(IBCContainerIO* io, BCContainerView* view, BCDeviceHandle device);


		/**
		 * @brief	Unless it is locked, release binding and free device resources.
		 */
		~BCContainerAutoBinding();


		/**
		 * @brief	Lock binding to make it outlast it's livetime.
		 */
		void lock()
		{
			m_locked = true;
		}


		/**
		 * @brief	Unlock binding to make it auto-unbind on destruction. Default.
		 */
		void unlock()
		{
			m_locked = false;
		}


		/**
		 * @brief	Is the binding locked?
		 */
		bool is_locked() const
		{
			return m_locked;
		}


	protected:


		/**
		 * @brief	Helper to parse first sector for partiotion table info.
		 * @return	Start sector of the last partiotion entry.
		 */
		unsigned int partition_info(bc_sector_t const & first, part_info_t& info) const;


	private:


		/* binded device */
		BCDeviceHandle	m_device;

		/* locked binding flag */
		bool		m_locked;
	};



	/**
	 * @brief	Window into container data.
	 * @more	Combines encryption config, sector map and data source
	 *		to provide a unified transparent access into container data.
	 */
	class BCContainerView
	{
	public:


		/**
		 * @brief	Initialize loop device
		 *
		 * @p data	Data source access.
		 * @p alg	Algorithm configuration to work with loop device.
		 * @p layout	Space layout of encrypted data.
		 * @p flags	Loop device flags.
		 *
		 * @throws	CBCException if initialization fails
		 */
		BCContainerView(BCAlgHandle const & alg, bc_space_unit_t const & layout, bcu32_t flags, bool is_hidden) :
			m_alg(alg),
			m_layout(layout),
			m_flags(flags),
			m_hidden(is_hidden)
		{
			BC_DEBUG_PRINT(("Debug: created view: offset %llu, length %llu, flags 0x%x (%s,%s)\n",
				layout.offset,
				layout.length,
				flags,
				(flags & BC_CRYPTO_READONLY) ? "ro" : "rw",
				(flags & BC_CRYPTO_IV_64BIT) ? "iv64" : "iv16x4"
				));
		}


		/**
		 * @brief	Release loop device
		 */
		~BCContainerView()
		{
		}


		/**
		 * @brief	Init container view
		 */
		static BCContainerViewHandle init(BCAlgHandle const & alg, bc_space_unit_t const & layout, bcu32_t flags, bool is_hidden);


		/**
		 * @brief	Get invalid view
		 */
		static BCContainerViewHandle invalid_view();


		/**
		 * @brief	Read a sector directly bypassing decryption.
		 *
		 * @p sector	Sector data to read into
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t	direct_read_sector(IBCContainerIO* io, bc_sector_t& sector, bc_ulong_t secnum)
		{
			return io->read_bytes( (bc_byte_t*)&sector, this->sector_size(), this->map_sector(secnum) * this->sector_size() );
		}


		/**
		 * @brief	Read and decrypt sector
		 *
		 * @p sector	Sector data to read into
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t 	read_sector(IBCContainerIO* io, bc_sector_t& sector, bc_ulong_t secnum);


		/**
		 * @brief	Write a sector directly bypassing encryption.
		 *
		 * @p sector	Sector data to write
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t 	direct_write_sector(IBCContainerIO* io, const bc_sector_t& sector, bc_ulong_t secnum)
		{
			return io->write_bytes((bc_byte_t*)&sector, this->sector_size(), this->map_sector(secnum) * this->sector_size() );
		}

		/**
		 * @brief	Encrypt and write sector
		 *
		 * @p sector	Sector data to write
		 * @p secnum	Sector number. Will be considered to begin with 0.
		 *
		 * @return	lower level system result
		 */
		bc_ulong_t 	write_sector(IBCContainerIO* io, const bc_sector_t& sector, bc_ulong_t secnum);


		/**
		 * @brief	Bind view to system device
		 */
		BCContainerAutoBinding create_binding(IBCContainerIO* io, BCDeviceHandle dev, bool ro);


		/**
		 * @brief	Get total sectors in data source.
		 */
		bc_ulong_t 	total_sectors() const
		{
			//return (m_layout.length - m_layout.offset) / this->sector_size();
			return m_layout.length / this->sector_size();
		}


		/**
		 * @brief	Get single sector size
		 */
		unsigned int	sector_size() const
		{
			return 512; // sector size resetting not supported for now
		}


		/**
		 * @brief	Get data offset in bytes
		 */
		bc_ulong_t	start_offset() const
		{
			return m_layout.offset;
		}


		/**
		 * @brief	Get algorithm used for this device
		 */
		BCAlgHandle	algorithm() const
		{
			return m_alg;
		}


		/**
		 * @brief	Get crypto flags.
		 */
		bcu32_t	 	flags() const
		{
			return m_flags;
		}


		bc_space_unit_t	layout() const
		{
			return m_layout;
		}


		bool is_hidden() const
		{
			return m_hidden;
		}


	protected:


		/**
		 * @brief	Prepare iv buffer for sector
		 * @more 	Depends on BC_FLAGS_IV_64BIT flag
		 */
		void prepare_iv(bc_sector_iv_t& iv, bc_ulong_t secnum)
		{
			if( m_flags & BC_CRYPTO_IV_64BIT )
				iv.iv = secnum + 1;
			else
				iv.word_iv[0] = iv.word_iv[1] = iv.word_iv[2] = iv.word_iv[3] = secnum & 0xFFFF; // first 16 bits copy to all parts.
		}


		/**
		 * @brief	Map virtual sector to real data source sector based on supplied sector map.
		 *
	 	 * @p sec	Virtual sector
		 *
		 * @return	Real sector.
		 */
		bc_ulong_t	map_sector(bc_ulong_t sec)
		{
			return m_layout.offset / this->sector_size() + sec;
		}


	private:


		/* initialized algorithm handle */
		BCAlgHandle			m_alg;

		/* data layout */
		bc_space_unit_t		m_layout;

		/* loop device flags */
		bcu32_t				m_flags;

		/* hidden view */
		bool				m_hidden;

	};



}



#endif

