#ifndef BDK_MOD_MANAGER_H
#define BDK_MOD_MANAGER_H

#include <string>

#include <common/types.h>
#include <common/class_defs.h>
#include <common/smartptr.h>
#include <mod/module.h>


namespace bdk
{


	/**
	 * @brief Module manager
	 *
	 * Keeps track of all the registered modules.
	 * Supports dynamic management to add\remove modules in runtime
	 * All real work of loading the module is done in CBCModule objects' initialization
	 *
	 */
	class BCModuleManager
	{
	BC_FORBID_COPY(BCModuleManager);

	public:	


		/**
		 * @brief 	Construct manager
		 */
		BCModuleManager();


		/**
		 * @brief 	Destruct manager by releasing all registered modules
		 */
		~BCModuleManager();


		/**
		 * @brief 	Load and register new module library
		 *
		 * @p modpath	File system path to module library
		 *
		 * @return 	Module raw pointer
		 */
		BCModule* load_module( const char* modpath );


		/**
		 * @brief 	Unload module
		 *
		 * @p handle	Module handle, previously saved from load_module
		 *
		 * @tip 	Module will not be released if it has hanging references.
		 * 		Module will be locked for the release operation.
		 */
		void release_module( BCModule* mod );


		/**
		 * @brief	Get module by id
		 *
		 * @p id	Module id
		 *
		 * @return 	Pointer to module
		 */
		BCModule* find_module( bcu32_t id );


		/**
		 * @brief	Get module by name
		 *
		 * @p name	Module name
		 *
		 * @return 	Pointer to module
		 */
		BCModule* find_module( std::string const & name );


		/**
		 * @brief 	Ask if module is loaded
		 */
		bool module_loaded( const BCModule* mod );


		/**
		 * @brief	
		 */
		

	private:

		typedef std::list<BCModule*>		modules_t;

		modules_t m_entries;
	};

}


#endif
