#ifndef BDK_MODULE_H
#define BDK_MODULE_H

#include <common/types.h>
#include <common/class_defs.h>
#include <common/smartptr.h>
#include <common/dso.h>

#include <mod/delegate.h>


namespace bdk 
{

	/**
	 * @brief	Status of the module.
	 */
	enum bc_module_status_t
	{
		BC_MOD_UNKNOWN	= 0, ///< Unknown status 
		BC_MOD_LOAD	= 1, ///< Module is loading it's dso
		BC_MOD_INST	= 2, ///< Module has loaded it's dso and is inside on_load overloaded call. All extention installations are presumed to happen here.
		BC_MOD_READY	= 3, ///< Module has finished all preparations and is ready for use.
		BC_MOD_UNINST	= 4, ///< Module is performing on_release overloaded call.
		BC_MOD_UNLOAD	= 5, ///< Module is unloading it's dso.
	};
	

	/**
	 * @brief 	BC External Module
	 */
	class BCModule
	{
	BC_FORBID_COPY(BCModule)

        public:


		/**
		 * @brief 	Load and init module from dso library
		 *
		 * @throws 	CBCException if loading failed on some stage
		 */
		explicit BCModule( const char* libname );
		
	
		/**
		 * @brief 	Release all module resources
		 */
                ~BCModule() throw();
		

		/**
		 * @brief 	Get module name
		 */
		const char* name() const;


		/**
		 * @brief 	Get module id number
		 */
		bc_mod_id_t id() const;


		/**
		 * @brief 	Get module description
		 */
		const char* description() const;


		/**
		 * @brief 	Get module reference count
		 */
		bcu32_t refcount() const;
		

		/**
		 * @brief	Increase module reference.
		 */
		void get_reference();


		/**
		 * @brief	Release reference
		 */
		void put_reference();

		
		/**
		 * @brief	Get current status
		 */
		bc_module_status_t status() const
		{
			return m_status;
		}


		/**
		 * @brief 	Get module library path
		 */
		const char* dso_path() const;


        private:


		/* module dso. must be declared first, to destroy last */
                BCScopedPtr<BCDSO>		m_dynamic;

		/* module delegate object from dso */		
		BCScopedPtr<IBCModuleDelegate> 	m_export;

		/* module reference count */
		unsigned int 			m_refcount;

		/* module current status */
		bc_module_status_t		m_status;

	};

}

#endif
