/**

	Copyright (c) 2009 Jetico, Inc., Finland
	All rights reserved.

*/

#include <string.h>
#include <algorithm>
#include <functional>

#include <alg/manager.h>
#include <alg/alg.h>
#include <common/log.h>
#include <common/exception.h>
#include <common/core.h>


using namespace bdk;


/*
 *	Predicate to compare algorithm by it's id
 */
struct pred_is_alg_id : public std::binary_function < BCSharedPtr<BCAlgService>, int, bool >
{
	bool operator() (BCSharedPtr<BCAlgService>& alg, int id) const
	{
		return alg->id() == id;
	}
};


/*
 *	Predicate to compare algorithm by it's id
 */
struct pred_is_alg_name : public std::binary_function < BCSharedPtr<BCAlgService>, std::string, bool >
{
	bool operator() (BCSharedPtr<BCAlgService>& alg, std::string const & name) const
	{
		return alg->name() == name;
	}
};



BCAlgManager::BCAlgManager()
{
}


BCAlgManager::~BCAlgManager() throw()
{
}


BCAlgService* BCAlgManager::load_algorithm( const bc_alg_info_t& info )
{
	BC_ASSERT( !this->find_algorithm( info.id ) );

	try
	{
		BCAlgService* service = new BCAlgService(info);

		BC_DEBUG_PRINT(("alg manager: loaded alg id 0x%2x, key size %3d, block size %2d, ext key size %3d, name %12s, CBC %d, LRW %d, XTS %d\n",
				info.id, info.key_size, info.block_size, info.ext_key_size, info.name.c_str(),
				info.mode_map.find(AM_CBC)->second.is_supported,
				info.mode_map.find(AM_LRW)->second.is_supported,
				info.mode_map.find(AM_XTS)->second.is_supported
				));

		m_entries.push_back( serptr_t(service) );
    		return service;
	}
	catch( const BCException& e )
	{
		return NULL;
	}
}


void BCAlgManager::release_algorithm( BCAlgService* alg )
{
	BC_ASSERT(this->find_algorithm(alg->id()));
	m_entries.remove( serptr_t(alg) );
}


BCAlgService* BCAlgManager::find_algorithm( int id )
{
	entries_t::iterator res;
	return (res = std::find_if( m_entries.begin(), m_entries.end(), std::bind2nd(pred_is_alg_id(), id))) != m_entries.end() ? res->get() : NULL;
}


BCAlgService* BCAlgManager::find_algorithm( std::string const & name )
{
	entries_t::iterator res = std::find_if( m_entries.begin(),
	                                        m_entries.end(),
	                                        std::bind2nd(pred_is_alg_name(), name));

	return (res != m_entries.end()) ? res->get() : NULL;
}


void BCAlgManager::list_supported_algorithms(BCAlgList & e)
{
	entries_t::iterator i = m_entries.begin();
	entries_t::iterator end = m_entries.end();

	e.clear();

	while(i != end)
	{
		e.push_back(i->get());
		++i;
	}
}


