#include <string.h>
#include <limits.h>
#include <container/view.h>
#include <common/log.h>

using namespace bdk;

/*
 *	Partition table info
 */
struct part_info_t
{
	unsigned int part_count;
	unsigned int ext_part_count;
	bc_space_unit_t table[4];
};

// check supplied sector buffer for partition table, fill info and return last non-extended partiotion sector offset
static bool parse_partition_table(const bc_byte_t* buf, part_info_t* out_info)
{

	struct partition
	{
		unsigned char active;
		unsigned char start_head;
		unsigned char start_sec;
		unsigned char start_cyl;
		unsigned char os_type;
		unsigned char end_head;
		unsigned char end_sec;
		unsigned char end_cyl;
		unsigned int lba_start_sec;
		unsigned int lba_nr_sec;
	}__attribute__((packed));

	memset(out_info, '\0', sizeof(part_info_t));

	// check msdos partition signature
	if (0x55 != buf[0x1FE] || 0xAA != buf[0x1FF])
		return 0;

	// read partiotion table
	struct partition* prt = (struct partition *) (buf + 0x1BE);
	bool found = false;

	for (unsigned int i = 0; i < 4; ++i, ++prt)
	{
		//x[i] = 0;
		if (0 == prt->lba_nr_sec)
			continue; // no partition in this slot

		// found one
		switch (prt->os_type)
		{
		case 0x05: // DOS extended
		case 0x0F: // Windows 95 extended
		case 0x85: // Linux extended
		{
			// found extended partition
			found = true;
			out_info->ext_part_count++;
			break;
		}
		case 0x07:
		{
			found = true;
			out_info->part_count++;
			out_info->table[i].offset = prt->lba_start_sec;
			out_info->table[i].length = prt->lba_nr_sec;
		}
		};
	}

	return found;
}

BCContainerViewHandle BCContainerView::init(BCAlgHandle const & alg,
		bc_space_unit_t const & layout, bcu32_t flags, bool is_hidden)
{
	return BCContainerViewHandle(new BCContainerView(alg, layout, flags, is_hidden));
}

BCContainerViewHandle BCContainerView::invalid_view()
{
	return BCContainerViewHandle(NULL);
}

bc_ulong_t BCContainerView::read_sector(IBCContainerIO* io,
		bc_sector_t& sector, bc_ulong_t secnum)
{
	bc_ulong_t rc = this->direct_read_sector(io, sector, secnum);
	bc_sector_iv_t iv;

	this->prepare_iv(iv, secnum);
	m_alg->decrypt((bc_byte_t*) &iv.iv, (bc_byte_t*) &sector.sec_data,
			this->sector_size());

	return rc;
}

bc_ulong_t BCContainerView::write_sector(IBCContainerIO* io,
		const bc_sector_t& sector, bc_ulong_t secnum)
{
	bc_sector_t data = sector;
	bc_sector_iv_t iv;

	this->prepare_iv(iv, secnum);
	m_alg->encrypt((bc_byte_t*) &iv.iv, (bc_byte_t*) &data.sec_data,
			this->sector_size());

	return this->direct_write_sector(io, data, secnum);
}

BCContainerAutoBinding BCContainerView::create_binding(IBCContainerIO* io,
		BCDeviceHandle dev, bool ro)
{
	if (ro)
	{
		m_flags |= BC_CRYPTO_READONLY;
		BC_DEBUG_PRINT(("Debug: view read only\n"));
	}

	return BCContainerAutoBinding(io, this, dev);
}

BCContainerAutoBinding::BCContainerAutoBinding(BCDeviceHandle device) :
	m_device(device), m_locked(false)
{
}

BCContainerAutoBinding::BCContainerAutoBinding(IBCContainerIO* io,
		BCContainerView* view, BCDeviceHandle device) :
	m_device(device), m_locked(false)
{
	BC_DEBUG_PRINT(("Debug: initializing view binding\n"));

	// read first sector to look for partiotion table
	bc_sector_t first;
	struct part_info_t info;

	view->read_sector(io, first, 0);

	BC_DEBUG_PRINT(("Debug: looking for partition table...\n"));
	unsigned int start_sector = this->partition_info(first, info);
	BC_DEBUG_PRINT(
			("Debug: view binder: new start sector number %d\n", start_sector));

	// we got our start sector, start binding source
	io->bind(m_device->system_handle(), view->algorithm()->key_handle(),
			view->algorithm(), view->flags(), view->layout(), start_sector );
}

BCContainerAutoBinding::~BCContainerAutoBinding()
{
	if (!m_locked)
	{
		m_device->clear();
	}
}

unsigned int BCContainerAutoBinding::partition_info(bc_sector_t const & first,
		part_info_t& info) const
{
	unsigned int start_sector = INT_MAX;

	memset(&info, 0, sizeof(part_info_t));
	if (parse_partition_table(first.sec_data, &info))
	{
#ifdef BC_DEBUG
		BC_DEBUG_PRINT(("Loop Binder: partition table info:\nPrimary count: %d\nExtended count: %d\n", info.part_count, info.ext_part_count));
		for(unsigned int i = 0; i < 4; ++i)
		{
			BC_DEBUG_PRINT(("Entry %d: start sector %d, total sectors %d\n", i, info.table[i].offset, info.table[i].length));
		}
#endif
		bc_get_log().msg(BC_MSG_WARN) << "Warning: " << info.part_count
				<< " partitions detected inside container";

		// find earliest primary partiotion
		for (unsigned int i = 0; i < info.part_count; ++i)
		{
			if (start_sector > info.table[i].offset)
				start_sector = info.table[i].offset;
		}

	}
	else
	{
		start_sector = 0;
	}

	return start_sector;
}

