#include "io_backend.h"
#include <assert.h>
#include <string.h>
#include <stdio.h>

#include <common/types.h>
#include <common/build.h>
#include <backend/io.h>


using namespace backend;

#define TEST_FILE_NAME "/tmp/test_file_bc"


#define BC_ASSERT_MSG(cond, msg) BC_DEBUG_PRINT(("test message: %s\n", msg)); BC_ASSERT(cond);

static void io_create_test()
{
	bc_file_handle_t file;
	bc_sysres_t rc;

	// try to create a valid file
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_CREATE, BC_FILE_PERM_DEFAULT);
	BC_ASSERT(rc);
	rc = bc_file_touch(TEST_FILE_NAME, BC_FILE_ACCESS_EXISTS);
	BC_ASSERT(rc && (bc_syscode(rc) == 1));
	bc_file_close(file);
	
	// try to create a file that is already there
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_CREATE | BC_FILE_MODE_EXCL, BC_FILE_PERM_DEFAULT);
	BC_ASSERT_MSG(!rc, bc_sysmsg(rc));
	bc_file_remove(TEST_FILE_NAME);

	// create a ro file and test write access
	bcu32_t mode = BC_FILE_PERM_UREAD;
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_CREATE, mode);
	BC_ASSERT(rc);
	rc = bc_file_touch(TEST_FILE_NAME, BC_FILE_ACCESS_READ);
	BC_ASSERT(rc && (bc_syscode(rc) == 1));
	rc = bc_file_touch(TEST_FILE_NAME, mode | BC_FILE_ACCESS_WRITE);
	BC_ASSERT(rc && (bc_syscode(rc) != 1));
}


static void io_open_test()
{
	bc_file_handle_t file;
	bc_sysres_t rc;

	// create our file
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_CREATE, BC_FILE_PERM_UREAD);
	BC_ASSERT(rc);
	bc_file_close(file);

	// open file
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_READ, BC_FILE_PERM_DEFAULT);
	BC_ASSERT(rc);
	bc_file_close(file);

	// open nonexistant file
	rc = bc_file_open(file, "/tmp/idonotexist", BC_FILE_MODE_READ, BC_FILE_PERM_DEFAULT);
	BC_ASSERT_MSG(!rc, bc_sysmsg(rc));

	// open ro file for write
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_WRITE, BC_FILE_PERM_DEFAULT);
	BC_ASSERT_MSG(!rc, bc_sysmsg(rc));

	bc_file_remove(TEST_FILE_NAME);
}


static void io_readwrite_test()
{
	bc_file_handle_t file;
	bc_sysres_t rc;
	
	bc_byte_t buf[0xff] = {0};
	bc_byte_t buf2[0xff] = {0};

	// fill buffer with tags
	memset(buf, 0xBC, sizeof(buf));

	// create our file for reading
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_CREATE | BC_FILE_MODE_READ, BC_FILE_PERM_DEFAULT);
	BC_ASSERT(rc);

	// write buffer to ro file
	rc = bc_file_write(file, buf, sizeof(buf));
	BC_ASSERT_MSG(!rc, bc_sysmsg(rc));

	// reopen file for writing
	bc_file_close(file);
	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_READ | BC_FILE_MODE_WRITE, BC_FILE_PERM_DEFAULT);
	BC_ASSERT(rc);

	// write buffer
	rc = bc_file_write(file, buf, sizeof(buf));
	BC_ASSERT(rc && (bc_syscode(rc) == sizeof(buf)));

	// check current offset
	rc = bc_file_tell(file);
	BC_ASSERT(rc && (bc_syscode(rc) == sizeof(buf)));

	// pwrite from half of the file
	rc = bc_file_pwrite(file, buf, sizeof(buf), sizeof(buf) / 2);
	BC_ASSERT(rc && (bc_syscode(rc) == sizeof(buf)));

	// check current offset, should not be changed since pwrite
	rc = bc_file_tell(file);
	BC_ASSERT(rc && (bc_syscode(rc) == sizeof(buf)));

	// try to read full buffer
	rc = bc_file_read(file, buf2, sizeof(buf2));
	BC_ASSERT(rc && (bc_syscode(rc) == (sizeof(buf2) / 2)));
	
	// check what's read
	BC_ASSERT(memcmp(buf, buf2, sizeof(buf2) / 2) == 0);
	BC_ASSERT(memcmp(buf2 + sizeof(buf2) / 2, buf + sizeof(buf2) / 2, sizeof(buf2) / 2) != 0);

	// check for eof
	rc = bc_file_is_eof(file);
	BC_ASSERT(rc && (bc_syscode(rc)));

	// try to pread full buffer from correct offset
	memset(buf2, 0, sizeof(buf2));
	rc = bc_file_pread(file, buf2, sizeof(buf2), sizeof(buf2) / 2);
	BC_ASSERT(rc && (bc_syscode(rc) == sizeof(buf)));

	// check what's read
	BC_ASSERT(memcmp(buf, buf2, sizeof(buf2)) == 0);

	// try to pread full buffer from incorrect offset
	memset(buf2, 0, sizeof(buf2));
	rc = bc_file_pread(file, buf2, sizeof(buf2), sizeof(buf2));
	BC_ASSERT(rc && (bc_syscode(rc) == sizeof(buf2) / 2));
	
	// check what's read
	BC_ASSERT(memcmp(buf, buf2, sizeof(buf2) / 2) == 0);
	BC_ASSERT(memcmp(buf2 + sizeof(buf2) / 2, buf + sizeof(buf2) / 2, sizeof(buf2) / 2) != 0);
	
	// rewind the file and read it all back by parts
	rc = bc_file_seek(file, BC_FILE_SEEK_SET, 0);
	for(int i = 1; i <= 3; ++i)
	{
		rc = bc_file_read(file, buf2, sizeof(buf2) / 2);
		BC_ASSERT(rc);
		rc = bc_file_tell(file);
		BC_ASSERT(rc && (bc_syscode(rc) == (int)(sizeof(buf2) / 2 * i)));
	}

	bc_file_close(file);
	bc_file_remove(TEST_FILE_NAME);
}


static void io_seek_test()
{
	bc_file_handle_t file;
	bc_sysres_t rc;
	bc_byte_t buf[0xff];

	rc = bc_file_open(file, TEST_FILE_NAME, BC_FILE_MODE_CREATE | BC_FILE_MODE_READ | BC_FILE_MODE_WRITE, BC_FILE_PERM_DEFAULT);
	BC_ASSERT(rc);

	// write data to file and rewind it
	rc = bc_file_write(file, buf, sizeof(buf));
	BC_ASSERT(rc);
	rc = bc_file_seek(file, BC_FILE_SEEK_SET, 0);
	BC_ASSERT(rc && (bc_syscode(rc) == 0));

	// set file to some positions
	rc = bc_file_seek(file, BC_FILE_SEEK_SET, 0xf);
	BC_ASSERT(rc && (bc_syscode(rc) == 0xf));
	rc = bc_file_tell(file);
	BC_ASSERT(rc && (bc_syscode(rc) == 0xf));

	rc = bc_file_seek(file, BC_FILE_SEEK_CUR, 0xf);
	BC_ASSERT(rc && (bc_syscode(rc) == 0xf * 2));
	rc = bc_file_tell(file);
	BC_ASSERT(rc && (bc_syscode(rc) == 0xf * 2));
	

	rc = bc_file_seek(file, BC_FILE_SEEK_END, 0xf);
	BC_ASSERT(rc && (bc_syscode(rc) == 0xff + 0xf));
	rc = bc_file_tell(file);
	BC_ASSERT(rc && (bc_syscode(rc) == 0xff + 0xf));

	bc_file_close(file);
	bc_file_remove(TEST_FILE_NAME);
}


namespace bdk
{

	void bc_test_io_backend()
	{
		BC_DEBUG_PRINT(("Debug: Running io backend tests\n"));
		io_create_test();
		io_open_test();
		io_seek_test();
		io_readwrite_test();
		BC_DEBUG_PRINT(("Debug: io backend tests passed\n"));
	}
}

