/**
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Description:	BestCrypt driver as a crypto provider implementation.
 *	Scope:		Backend implementation.
 *	Author:		evgeny yakovlev.
 *
 *	History:
 *	- 22.06.09, created.
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>

#include "../mod/bc_ioctl.h"	// driver ioctl interface
#include "os_crypto.h"
#include "../src/config.h"

#include <backend/crypto.h>	// interface to implement
#include <common/types.h>
#include <common/build.h>
#include <common/log.h>
#include <common/exception.h>


using namespace bdk;
using namespace bdk::backend;


class IoctlException : public BCSystemException
{
public:
	IoctlException(std::string const & devname, int reqnum, std::string const & drverr)
	{
		m_msg = devname + ":failed ioctl request \"" + reqnum2str(reqnum) + "\":" + drverr;
	}
protected:
	const char* reqnum2str(int reqnum)
	{
		switch (reqnum)
		{
		case 0xBC00: return "get info";
		case 0xBC01: return "set fd";
		case 0xBC02: return "clear fd";
		case 0xBC03: return "lock dev";
		case 0xBC04: return "unlock dev";
		case 0xBC05: return "check fd";
		case 0xBC06: return "notify pid";
//		case 0xBC07: return "get priv";
		case 0xBC09: return "force unlock";
		case 0xBC0A: return "check activity";
		case 0xBC10: return "verify alg";
		case 0xBC20: return "make key";
		case 0xBC21: return "test key";
		case 0xBC22: return "free key";
		case 0xBC30: return "encrypt block";
		case 0xBC31: return "decrypt block";
		case 0xBC40: return "is free";
		case 0xBC41: return "get version";
		case 0xBC42: return "stat device";

		default: return "unknown request";
		};
	}
};

// open device as a linux device file
static bc_crypto_dev_handle_t dev_get ( const char* devpath )
{
	bc_crypto_dev_handle_t handle = new bc_os_dev_t;
	handle->file = bc_file_open ( devpath, BC_FILE_READ /* | BC_FILE_MODE_NONBLOCK */ );
	return handle;
}


// open bc device by minor number
static bc_crypto_dev_handle_t dev_get_number ( unsigned int devnum )
{
	char devname[MAX_DEV_PATH] = {0};

	snprintf ( devname, MAX_DEV_PATH, "/dev/bcrypt%d", devnum );
	return dev_get ( devname );
}


// free device and it's resources
static void dev_put ( bc_crypto_dev_handle_t& handle )
{
	bc_file_handle_t file = handle->file;
	delete handle;
	handle = NULL;
	bc_file_close ( file );
}


// make ioctl request with device file fd
static void make_request ( bc_crypto_dev_handle_t handle, int request, void* data )
{
	int rc = ioctl ( handle->file->fd, request, data );
	if ( rc != 0 )
		throw IoctlException(bc_crypto_dev_name(handle), request, strerror(-rc));
}


// make request using any temporary free device and then release it
static void make_request_free ( int request, void* data )
{
	bc_crypto_dev_handle_t dev = bc_crypto_dev_get ( "/dev/bcrypt0" );
	make_request ( dev, request, data );
	bc_crypto_dev_put ( dev );
}


namespace bdk
{
	namespace backend
	{

		bc_crypto_dev_handle_t bc_crypto_dev_get ( const char* devpath )
		{
			return dev_get ( devpath );
		}


		bc_crypto_dev_handle_t bc_crypto_dev_get_free()
		{
			bc_crypto_dev_handle_t dev;
			bc_dev_conf_t cont = get_dev_config();

			for ( int i = 0; i < cont.total_devs; ++i )
			{
				dev = dev_get_number ( i );

				if ( bc_crypto_dev_is_free ( dev ) )
					return dev;

				bc_crypto_dev_put ( dev );
			}

			return NULL;
		}


		void bc_crypto_dev_put ( bc_crypto_dev_handle_t& handle )
		{
			dev_put ( handle );
		}


		void bc_crypto_dev_lock ( bc_crypto_dev_handle_t handle )
		{
			make_request ( handle, BC_LOCK_DEV, 0 );
		}


		void bc_crypto_dev_unlock ( bc_crypto_dev_handle_t handle, bool force )
		{
			int request = ( force ) ? BC_FORCE_UNLOCK : BC_UNLOCK_DEV;
			make_request ( handle, request, 0 );
		}


		void bc_crypto_dev_clear ( bc_crypto_dev_handle_t dev, bool force )
		{
			make_request ( dev, BC_CLR_FD, ( void* ) ( ( force ) ? 1 : 0 ) );
		}


		bool bc_crypto_dev_is_free ( bc_crypto_dev_handle_t handle )
		{
			int is_free = 0;
			make_request ( handle, BC_IS_FREE, &is_free );
			return ( is_free == 1 );
		}


		const char* bc_crypto_dev_name ( bc_crypto_dev_handle_t handle )
		{
			return bc_file_name ( handle->file );
		}


		bc_key_handle_t bc_crypto_store_key ( int algid, bc_byte_t* key, unsigned int key_len, bc_byte_t* pool )
		{
			struct bc_key arg;
			memset ( &arg, 0, sizeof ( struct bc_key ) );
			arg.magic 	= BC_MAGIC;
			arg.error 	= 0;
			arg.alg_id 	= algid;
			arg.key 	= key;
			arg.key_len 	= key_len;
			arg.pool 	= pool;
			arg.key_handle 	= 0xFFFFFFFF;


			make_request_free ( BC_MAKE_KEY, &arg );
			return arg.key_handle;
		}


		void bc_crypto_release_key ( int algid, bc_key_handle_t handle )
		{
			struct bc_key arg;
			memset ( &arg, 0, sizeof ( struct bc_key ) );
			arg.magic 	= BC_MAGIC;
			arg.error 	= 0;
			arg.alg_id 	= algid;
			arg.key_handle 	= handle;

			make_request_free ( BC_FREE_KEY, &arg );
		}


		void bc_crypto_encrypt ( int algid, bc_key_handle_t handle, int mode, const bc_byte_t* iv, bc_byte_t* buffer, unsigned int bufsiz )
		{
			struct bc_block arg;
			memset ( &arg, 0, sizeof ( struct bc_block ) );
			arg.magic 	= BC_MAGIC;
			arg.error 	= 0;
			arg.alg_id 	= algid;
			arg.alg_mode	= mode;
			arg.key_handle 	= handle;
			arg.buffer 	= buffer;
			arg.buffer_len 	= bufsiz;

			for ( int i = 0; i < 8; i++ ) arg.iv[i] = iv[i];

			make_request_free ( BC_ENCRYPT_BLOCK, &arg );
		}


		void bc_crypto_decrypt ( int algid, bc_key_handle_t handle, int mode, const bc_byte_t* iv, bc_byte_t* buffer, unsigned int bufsiz )
		{
			struct bc_block arg;
			memset ( &arg, 0, sizeof ( struct bc_block ) );
			arg.magic 	= BC_MAGIC;
			arg.error 	= 0;
			arg.alg_id 	= algid;
			arg.alg_mode	= mode;
			arg.key_handle 	= handle;
			arg.buffer 	= buffer;
			arg.buffer_len 	= bufsiz;

			for ( int i = 0; i < 8; i++ ) arg.iv[i] = iv[i];

			make_request_free ( BC_DECRYPT_BLOCK, &arg );
		}


		void bc_crypto_set_file (
		    bc_crypto_dev_handle_t 	dev,
		    bc_file_handle_t 	file,
		    bc_key_handle_t 	key,
		    int			algid,
		    int 			mode,
		    bcu32_t 		flags,
		    const bc_space_unit_t* 	layout,
		    unsigned int 		start_sector
		)
		{
			struct bc_file64  arg2;
			memset ( &arg2, 0, sizeof ( arg2 ) );
			arg2.magic  		= BC_MAGIC;
			arg2.fd     		= file->fd;
			arg2.offset 		= layout->offset;
			arg2.start_sector 	= start_sector;
			arg2.num_sectors 	= 0;
			arg2.key_handle 	= key;
			arg2.alg_id 		= algid;
			arg2.alg_mode		= mode;
			arg2.flags  		= flags;

			make_request ( dev, BC_SET_FD, &arg2 );
		}


		bool bc_crypto_verify ( int algid )
		{
			struct bc_alg arg;
			memset ( &arg, 0, sizeof ( struct bc_alg ) );
			arg.magic 	= BC_MAGIC;
			arg.error 	= 0;
			arg.alg_id 	= algid;
			arg.verified	= 0;

			make_request_free ( BC_VERIFY_ALG, &arg );
			return ( arg.verified == 1 );
		}


		bool bc_crypto_stat_dev(bc_crypto_dev_handle_t dev, char* backing_file_path, unsigned pathlen)
		{
			struct bc_dev_info arg;
			arg.pathbuf = backing_file_path;
			arg.bufsiz = pathlen;
			arg.is_configured = 0;
			
			make_request(dev, BC_STAT_DEV, &arg);
			return (arg.is_configured == 1);
		}

		void bc_crypto_version ( char* major, char* minor )
		{
			struct bc_info  arg;
			memset ( &arg, 0, sizeof ( struct bc_info ) );
			arg.magic = BC_MAGIC;

			make_request_free ( BC_GET_INFO, &arg );
			*major = arg.ver_major;
			*minor = arg.ver_minor;
		}
	}
}

