/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: dso.cpp
 *	scope: linux backend
 *	author: evgeny yakovlev
 *
 *	description:
 *	dynamic symbol object linux backend implementation
 *
 *	history:
 *
 */

#include <dlfcn.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "os_dso.h"

#include <backend/dso.h>
#include <common/build.h>
#include <common/exception.h>

using namespace bdk;
using namespace bdk::backend;

class DSOException: public BCSystemException
{
public:
	DSOException(std::string const & dso, std::string const & action,
			std::string const & syserr)
	{
		m_msg = dso + ":" + action + ":" + syserr;
	}
};

static unsigned int convert_res(bc_dso_resolve_t rflag)
{
	unsigned int unix_rflag = 0;

	switch (rflag)
	{
	case BC_DSO_RES_LAZY:
		unix_rflag = RTLD_LAZY;
		break;
	case BC_DSO_RES_NOW:
		unix_rflag = RTLD_NOW;
		break;
	default:
		BC_ASSERT(false);
	};

	return unix_rflag;
}

static unsigned int convert_vis(bc_dso_visibility_t vflag)
{
	unsigned int unix_vflag = 0;

	switch (vflag)
	{
	case BC_DSO_VIS_GLOBAL:
		unix_vflag = RTLD_GLOBAL;
		break;
	case BC_DSO_VIS_LOCAL:
		unix_vflag = RTLD_LOCAL;
		break;
	default:
		BC_ASSERT(false);
	};

	return unix_vflag;
}

static bc_dso_handle_t dso_get(const char* libname, void* ptr)
{
	bc_dso_handle_t handle = new bc_os_dso_t;

	handle->ptr = ptr;
	handle->path.assign(libname);

	return handle;
}

static void dso_put(bc_dso_handle_t& handle)
{
	delete handle;
	handle = NULL;
}

namespace bdk
{
namespace backend
{

static bc_dso_handle_t dso_invalid_handle = NULL;

bc_dso_handle_t bc_dso_load(const char* libname, bc_dso_resolve_t rflag, bc_dso_visibility_t vflag)
{
	void* ptr = dlopen(libname, convert_res(rflag) | convert_vis(vflag));

	if (ptr == NULL)
		throw DSOException(libname, "dlopen failed", dlerror());

	return dso_get(libname, ptr);
}

void bc_dso_release(bc_dso_handle_t& handle)
{
	BC_ASSERT( handle != dso_invalid_handle );

	int rc = dlclose(handle->ptr);
	std::string name = handle->path;
	dso_put(handle);

	if (rc != 0)
		throw DSOException(name, "dlclose failed", dlerror());
}

void* bc_dso_find(bc_dso_handle_t handle, const char* symname)
{
	BC_ASSERT( handle != dso_invalid_handle );

	// clear any past errors, get symbol address and check for error again. symbol can actually be NULL.
	dlerror();
	void* sym = dlsym(handle->ptr, symname);
	char* err = dlerror();

	// success
	if (err == NULL)
		return sym;
	else
		throw DSOException(symname, "dlsym failed", dlerror());
}

bool bc_dso_is_resident(const char* libname)
{
	// see if library is resident
	dlerror();
	void* handle = dlopen(libname, RTLD_NOLOAD);
	const char* error = dlerror();

	if (error != NULL)
		throw DSOException(libname, "dlopen failed", dlerror());

	return (handle != NULL);
}

const char* bc_dso_path(bc_dso_handle_t handle)
{
	BC_ASSERT(handle != dso_invalid_handle);

	return handle->path.c_str();
}


}
}

