/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: io.cpp
 *	scope: linux backend
 *	author: evgeny yakovlev
 *
 *	description:
 *	file io linux backend implementation
 *
 *	history:
 *
 */


#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include "os_fs.h"
#include <backend/fs.h>
#include <common/build.h>
#include <common/exception.h>


using namespace bdk;
using namespace bdk::backend;

char fsexpbuf[0xff];

class FSException : public BCSystemException
{
public:
	FSException(const char* target, const char* action, const char* syserr)
	{
		int rc = snprintf(fsexpbuf, sizeof(fsexpbuf), "%s: %s: %s", target, action, syserr);
		m_msg.assign(fsexpbuf, rc);
	}

};

static bcu32_t convert_mode( bcu32_t mode )
{
	// will store the mode here
	bcu32_t unix_mode = 0;

	// read\write
	if( (mode & BC_FILE_READ) && (mode & BC_FILE_WRITE) )
	{
		unix_mode |= O_RDWR;
	}
	else if (mode & BC_FILE_READ)
	{
		unix_mode |= O_RDONLY;
	}
	else if (mode & BC_FILE_WRITE)
	{
		unix_mode |= O_WRONLY;
	}

	// create
	if( mode & BC_FILE_CREATE )
	{
		unix_mode |= O_CREAT;
		if( mode & BC_FILE_EXCL )
		{
			unix_mode |= O_EXCL;
		}
	}

	// mods
	if( mode & BC_FILE_APPEND )
	{
		unix_mode |= O_APPEND;
	}

	if( mode & BC_FILE_TRUNC )
	{
		unix_mode |= O_TRUNC;
	}

#if defined(_LARGEFILE64_SOURCE)	// we've been _told_ to make it largefile

	unix_mode |= O_LARGEFILE;

#elif defined(O_LARGEFILE)		// we were asked to, but let's see if it's possible

	if( mode & BC_FILE_LARGE )
	{
		unix_mode |= O_LARGEFILE
	}

#endif

	if( mode & BC_FILE_NONBLOCK ) unix_mode |= O_NONBLOCK;

	return unix_mode;
}


static bcu32_t convert_perm( bcu32_t perm )
{
	// first, check for defaults
	if( perm & BC_PERM_DEFAULT )
	{
		return 0666;
	}

	// will store perms here
	bcu32_t unix_perm = 0;

	if( perm & BC_PERM_UREAD ) 	unix_perm |= S_IRUSR;
	if( perm & BC_PERM_UWRITE ) 	unix_perm |= S_IWUSR;
	if( perm & BC_PERM_UEXEC ) 	unix_perm |= S_IXUSR;
	if( perm & BC_PERM_SETUID ) 	unix_perm |= S_ISUID;

	if( perm & BC_PERM_GREAD ) 	unix_perm |= S_IRGRP;
	if( perm & BC_PERM_GWRITE ) 	unix_perm |= S_IWGRP;
	if( perm & BC_PERM_GEXEC ) 	unix_perm |= S_IXGRP;
	if( perm & BC_PERM_SETGID ) 	unix_perm |= S_ISGID;

	if( perm & BC_PERM_OREAD ) 	unix_perm |= S_IROTH;
	if( perm & BC_PERM_OWRITE ) 	unix_perm |= S_IWOTH;
	if( perm & BC_PERM_OEXEC ) 	unix_perm |= S_IXOTH;

	return unix_perm;
}


static bcu32_t convert_access( bcu32_t access )
{
	bcu32_t unix_access = 0;

	if( access & BC_ACCESS_EXISTS ) unix_access |= F_OK;
	if( access & BC_ACCESS_READ ) 	unix_access |= R_OK;
	if( access & BC_ACCESS_WRITE ) 	unix_access |= W_OK;
	if( access & BC_ACCESS_EXEC ) 	unix_access |= X_OK;

	return unix_access;
}


static bcu32_t convert_system_type( mode_t mode )
{
	if(S_ISREG(mode)) return BC_TYPE_FILE;
	if(S_ISDIR(mode)) return BC_TYPE_DIR;
	if(S_ISLNK(mode)) return BC_TYPE_LINK;
	if(S_ISBLK(mode)) return BC_TYPE_BDEV;
	if(S_ISCHR(mode)) return BC_TYPE_CDEV;

	return BC_TYPE_UNKNOWN;
}


static int convert_seek( bc_file_seek_t seek )
{
	switch (seek) {
	case BC_SEEK_CUR: return SEEK_CUR;
	case BC_SEEK_SET: return SEEK_SET;
	case BC_SEEK_END: return SEEK_END;
	default: 		BC_ASSERT(false);
	};

	return -1;
}


static bc_file_handle_t get_file( const char* fname, int mode, int perm )
{
	bc_file_handle_t handle;

	int fd = open(fname, mode, perm);

	if( fd < 0 )
		throw FSException(fname, "failed open", strerror(errno));

	handle = new bc_os_file_t;
	handle->fd = fd;
	handle->eof_hit = false;
	handle->fname.assign(fname);

	return handle;
}


static void put_file( bc_file_handle_t& handle )
{
	int fd = handle->fd;
	std::string fname = handle->fname;
	delete handle;
	handle = NULL;

	if(0 != close(fd))
		throw FSException(fname.c_str(), "failed close", strerror(errno));
}


namespace bdk
{
namespace backend
{


	static bc_file_handle_t file_invalid_handle = NULL;


	bc_file_handle_t bc_file_open( const char* fname, bcu32_t mode, bcu32_t perm )
	{
		bcu32_t unix_mode = convert_mode( mode );
		bcu32_t unix_perm = convert_perm( perm );

		return get_file(fname, unix_mode, unix_perm);
	}


	void bc_file_close( bc_file_handle_t& handle )
	{
		BC_ASSERT(handle != file_invalid_handle);

		return put_file( handle );
	}


	bc_ulong_t bc_file_read( bc_file_handle_t handle, bc_byte_t* buf, size_t nbytes )
	{
		BC_ASSERT( handle != file_invalid_handle );
		BC_ASSERT( buf );
		BC_ASSERT( nbytes >= 0 );

		// nothing to read
		if( nbytes == 0 ) return 0;

		// get current file position
		bc_ulong_t offset = bc_file_tell(handle);

		size_t nleft = nbytes;
		ssize_t rc;
		int err;

		do
		{
			rc = read(handle->fd, buf, nleft);

			if( rc == -1 )
			{
				err = errno;
				// offset is undefined in some error cases, correct it manually
				bc_file_seek(handle, BC_SEEK_SET, offset);

				if( (err == EINTR) || (err == EAGAIN) || (err == EWOULDBLOCK) ) continue; // we can try again
				else throw FSException(bc_file_name(handle), "read failed", strerror(err));
			}

			nleft -= rc;
			buf += rc;
		} while ( nleft && rc );

		// EOF is not an exception
		if( rc == 0 ) handle->eof_hit = true;

		return nbytes - nleft;
	}


	bc_ulong_t bc_file_write( bc_file_handle_t handle, const bc_byte_t* buf, size_t nbytes )
	{
		BC_ASSERT( handle != file_invalid_handle );
		BC_ASSERT( buf );
		BC_ASSERT( nbytes >= 0 );

		// nothing to write
		if( nbytes == 0 ) return 0;

		// get current file position
		bc_ulong_t offset = bc_file_tell(handle);

		size_t nleft = nbytes;
		ssize_t rc;
		int err;

		// do writes
		do
		{
			rc = write(handle->fd, buf, nleft);

			if ( rc == -1 )
			{
				err = errno;
				// offset is undefined in some error cases, correct it manually
				bc_file_seek(handle, BC_SEEK_SET, offset);

				if( (err == EINTR) || (err == EAGAIN) || (err == EWOULDBLOCK) ) continue; // we can try again
				else throw FSException(bc_file_name(handle), "write failed", strerror(err));
			}

			nleft -= rc;
			buf += rc;

		} while( nleft && rc );

		return nbytes - nleft;
	}


	bc_ulong_t bc_file_pread( bc_file_handle_t handle, bc_byte_t* buf, size_t nbytes, bc_ulong_t offset )
	{
		BC_ASSERT( handle != file_invalid_handle );
		BC_ASSERT( buf );
		BC_ASSERT( nbytes >= 0 );
		BC_ASSERT( offset >= 0 );

		// nothing to read
		if( nbytes == 0 ) return 0;

		size_t nleft = nbytes;
		ssize_t rc;
		int err;

		do
		{
			rc = pread(handle->fd, buf, nleft, offset);

			if( rc == -1 )
			{
				err = errno;
				if( (err == EINTR) || (err == EAGAIN) || (err == EWOULDBLOCK) ) continue; // we can try again
				else throw FSException(bc_file_name(handle), "pread failed", strerror(err));
			}

			nleft -= rc;
			offset += rc;
			buf += rc;
		} while ( nleft && rc );

		// EOF is a special error case
		if( rc == 0 ) handle->eof_hit = true;

		return nbytes - nleft;
	}


	bc_ulong_t bc_file_pwrite( bc_file_handle_t handle, const bc_byte_t* buf, size_t nbytes, bc_ulong_t offset )
	{
		BC_ASSERT( handle != file_invalid_handle );
		BC_ASSERT( buf );
		BC_ASSERT( nbytes >= 0 );
		BC_ASSERT( offset >= 0 );

		if( nbytes == 0 ) return 0;

		size_t nleft = nbytes;
		ssize_t rc;
		int err;


		// do writes
		do
		{
			rc = pwrite(handle->fd, buf, nleft, offset);

			if ( rc == -1 )
			{
				err = errno;
				if( (err == EINTR) || (err == EAGAIN) || (err == EWOULDBLOCK) ) continue;
				else throw FSException(bc_file_name(handle), "pwrite failed", strerror(err));
			}

			nleft -= rc;
			offset += rc;
			buf += rc;

		} while( nleft && rc );

		return nbytes - nleft;
	}


	void bc_file_remove( const char* path )
	{
		int rc = unlink(path);

		if( rc != 0 )
			throw FSException(path, "unlink failed", strerror(errno));
	}


	void bc_file_rename( const char* old_path, const char* new_path )
	{
		if( rename( old_path, new_path ) != 0 )
			throw FSException(new_path, "rename failed", strerror(errno));
	}


	bool bc_file_touch( const char* fname, bc_uword_t flags )
	{
		int unix_flags = convert_access( flags );

		int rc = access( fname, unix_flags );
		int err = errno;

		// access allowed, success
		if (0 == rc)  {
			return true;
		// access denied, fail
		} else if ( err == EACCES ) {
			return false;
		// system error, exceptional case
		} else if ( (err == ENOENT) && (flags == BC_ACCESS_EXISTS) ){
			return false;
		} else if ((err == EROFS) && (flags == BC_ACCESS_WRITE)) {
			return false;
		} else {
			throw FSException(fname, "access failed", strerror(err));
		}
	}


	bc_file_type_t bc_file_type( const char* fname )
	{
		struct stat st;
    		int res = stat( fname, &st );

		if( res != -1 )
			return (bc_file_type_t)convert_system_type(st.st_mode);
		else
			throw FSException(fname, "stat failed", strerror(errno));
	}


	bc_ulong_t bc_file_seek( bc_file_handle_t handle, bc_file_seek_t where, bc_ulong_t offset )
	{
		BC_ASSERT(handle != file_invalid_handle );

		int seek_flag = convert_seek(where);

		bc_long_t off = lseek( handle->fd, offset, seek_flag);

		if( off != -1 )
			return off;
		else
			throw FSException(bc_file_name(handle), "lseek failed", strerror(errno));
	}


	bc_ulong_t bc_file_tell( bc_file_handle_t handle )
	{
		BC_ASSERT(handle != file_invalid_handle);

		return bc_file_seek( handle, BC_SEEK_CUR, 0 );
	}


	bool bc_file_is_eof( bc_file_handle_t handle )
	{
		BC_ASSERT(handle != file_invalid_handle);

		return handle->eof_hit;
	}


	const char* bc_file_name( bc_file_handle_t handle )
	{
		BC_ASSERT( handle != file_invalid_handle );

		return handle->fname.c_str();
	}


	size_t bc_file_dirlist(const char* path, std::vector<std::string>& listing)
	{
		listing.clear();

		DIR* dir;
		dir = opendir(path);
		if(!dir)
			throw FSException(path, "opendir failed", strerror(errno));

		struct dirent* entry;

		while( (entry = readdir(dir)) != NULL )
				listing.push_back(entry->d_name);

		return listing.size();
	}


	void bc_file_canonize_path(std::string & path)
	{
		char* cpath = realpath(path.c_str(), NULL);
		if(!cpath)
			throw FSException(path.c_str(), "realpath failed", strerror(errno));

		path.assign(cpath);
		free(cpath);
	}


	bool bc_file_is_owner(const char* path)
	{
		struct stat st;
		if(lstat(path, &st) != 0 )
			throw FSException(path, "failed lstat", strerror(errno));

		return (st.st_uid == getuid());
	}

}
}


