/*******************************************************************
 *	Copyright (c) 1994-1998 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_alg.c
 *
 *	Description:	algorithm wrapper for module building
 *
 *	Scope:		BestCrypt algorithms drivers
 *
 *	Platforms:	Linux
 *
 *	Author:		Nail R. Kaipov
 *
 *	Created:	10-Nov-1998
 *
 *	Revision:	??-??-2009 Evgeny Yakovlev.
 *			lrw, xts added.
 *			25-jun-2009 evgeny yakovlev.
 *			refactored into single algorithm core
 *
 *******************************************************************/

#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/uaccess.h>

#include "gf_mul.h"
#include "bc_cfg.h"
#include "bc_alg.h"


#define POOL_SIZE_BYTES		512

#define XTS_KEY2_OFFSET 	32


typedef enum
{
	IV_4X16_0 = 0,	// 4 sections by 16 bits starting from 0
	IV_1X64_1 = 1,	// 1 section by 64 bits starting from 1
} bc_iv_format_t;


static void shred (u_char **ptr, u_int len)
{
	if (!ptr || !*ptr) return;
	memset(*ptr, 0, len);
	kfree(*ptr);
	*ptr = NULL;
}


int bc_alg_test_key (struct bc_algorithm* alg, bc_key_handle kh)
{
	if (kh >= KEY_TABLE_LENGTH)
		return -EINVAL;

        if (!alg->key_table[kh].key_extended)
		return -EINVAL;

	return 0;
}


int bc_alg_lock_key (struct bc_algorithm* alg, bc_key_handle kh, int lock)
{
	int error;

	error = bc_alg_test_key(alg, kh);

	if (error)
		return error;

	if (lock)
		alg->key_table[kh].lock++;
	else
		alg->key_table[kh].lock--;

	return 0;
}


int bc_alg_free_key (struct bc_algorithm* alg, bc_key_handle kh)
{
	int err;
	struct bc_key_element* key_table = alg->key_table;

	err = bc_alg_test_key(alg, kh);
	if (err)
		return err;

	key_table[kh].lock--;
	if(key_table[kh].lock > 0)
		return 0;

	shred(&(key_table[kh].key_extended), alg->ops->ext_key_size_bytes);
	shred(&(key_table[kh].pool), POOL_SIZE_BYTES);
	shred(&(key_table[kh].lrw_context), sizeof(GfmulCtx));
	shred(&(key_table[kh].xts_context), alg->ops->ext_key_size_bytes);

	memset(&(key_table[kh]), 0, sizeof(struct bc_key_element));

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	module_put(alg->ops->owner);
#else
	__MOD_DEC_USE_COUNT(alg->ops->owner);
#endif
	return 0;
}


int bc_alg_make_key(struct bc_algorithm* alg, u_char* __user key, u_int keylen, u_char* __user pool, bc_key_handle* out_handle)
{
	bc_key_handle		kh;
	int			error = 0;
	u_char*			key_tmp = NULL;
	u_char* 		key_2 = NULL;
	struct bc_key_element* 	table;
	struct bc_key_element* 	elem;


	// check length
	if (keylen < alg->ops->key_size_bits)
		return -EINVAL;

	keylen = (keylen+7) >> 3; /* convert to bytes */
	table = alg->key_table;


	// find a place for a key
	for (kh = 0; kh < KEY_TABLE_LENGTH; ++kh) {
		if ( table[kh].key_extended == NULL ) break;
	}

	if (kh >= KEY_TABLE_LENGTH || table[kh].key_extended || table[kh].pool)
		return -EBUSY;


	// allocate table entries
	elem = &table[kh];
	memset(elem, 0, sizeof(struct bc_key_element));
	elem->key_extended 	= kmalloc(alg->ops->ext_key_size_bytes, GFP_KERNEL);
	elem->pool 		= kmalloc(POOL_SIZE_BYTES, GFP_KERNEL);
	elem->lrw_context	= kmalloc(sizeof(GfmulCtx), GFP_KERNEL);
	elem->xts_context	= kmalloc(alg->ops->ext_key_size_bytes, GFP_KERNEL);

	key_tmp = kmalloc(keylen, GFP_KERNEL);

	if (!elem->key_extended || !elem->pool || !elem->lrw_context || !elem->xts_context || !key_tmp) {
		error = -ENOMEM;
		goto error_out;
	}

	if( (error = copy_from_user(elem->pool, pool, POOL_SIZE_BYTES)) ) {
		goto error_out;
	}

	// copy exactly as user told us the buffer size is
	if( (error = copy_from_user(key_tmp, key, keylen)) ) {
		goto error_out;
	}

	memset(elem->key_extended, 0, alg->ops->ext_key_size_bytes);
	memset(elem->lrw_context, 0, sizeof(GfmulCtx));

	// get Key2 after Key1 in raw buffer
	key_2 = &key_tmp[keylen - alg->ops->block_size_bytes];

	if( alg->ops->block_size_bytes == 8 ) {
		GFMUL_InitCtxBlock8( key_2, (GfmulCtx*) elem->lrw_context );
	} else if (alg->ops->block_size_bytes == 16 ) {
		GFMUL_InitCtxBlock16( key_2, (GfmulCtx*) elem->lrw_context );
	} else {
		error = -EINVAL;
		goto error_out;
	}


	alg->ops->key_extend(key_tmp, (DWORD *)elem->key_extended);
	alg->ops->key_extend(&(key_tmp[XTS_KEY2_OFFSET]), (DWORD*)elem->xts_context );

	shred(&key_tmp, keylen);
	elem->lock = 1;

	*out_handle = kh;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	try_module_get(alg->ops->owner);
#else
	__MOD_INC_USE_COUNT(alg->ops->owner);
#endif
	return 0;

error_out:
	shred(&(elem->key_extended), alg->ops->ext_key_size_bytes);
	shred(&(elem->pool), POOL_SIZE_BYTES);
	shred(&(elem->lrw_context), sizeof(GfmulCtx));
	shred(&(elem->xts_context), alg->ops->ext_key_size_bytes);
	shred(&key_tmp, keylen);
	return error;
}

//---------------------------------------------
//	Encryption\decryption
//---------------------------------------------


inline void transform_iv (DWORD *iv, DWORD *pool, bc_iv_format_t ver)
{
	unsigned int i;
	if (IV_4X16_0 == ver) {
		i = (iv[0] & 0x3F) << 1;
		iv[0] ^= pool[i];
		iv[1] ^= pool[i+1];
	} else {
		i = __le32_to_cpu(iv[0]);
		i = (i & 0x3F) << 1;
		iv[0] ^= pool[i];
		iv[1] ^= pool[i+1];
	}

}

inline void increment_iv(WORD *iv, bc_iv_format_t ver)
{
	u64	iv64, *piv64;

	if (IV_4X16_0 == ver) {
		iv[0]++;
		iv[3] = iv[2] = iv[1] = iv[0];
	} else {
		piv64 = (u64 *)iv;
		iv64 = __le64_to_cpu(*piv64);
		iv64++;
		*piv64 = __cpu_to_le64(iv64);
	}
}


inline int guess_iv_format(u_char* iv)
{
	union
	{
		WORD words[4];
		uint64_t integer;
	} u_iv;

	u_iv.integer = *((uint64_t*)iv);

	if( (u_iv.words[0] == u_iv.words[3]) && (u_iv.words[0] == u_iv.words[1]) && (u_iv.words[0] == u_iv.words[2]) ) return IV_4X16_0;
	else return IV_1X64_1;
}


inline void transform_buf (DWORD *pool, DWORD *src, DWORD *dst, DWORD len)
{
	while (len--) *dst++ = *src++ ^ *pool++;
}


static int encrypt_cbc(bc_vencrypt_cbc alg_encrypt, u_char* key_extended, u_char *iv, int iv_ver, u_char *src, u_char *dst, u_int len, u_char* pool)
{
	WORD	*s, ivector[4], ivectorS[4];
	DWORD	i, offset, size;

	s = (WORD *)iv;
	for (i = 0; i < 4; i++) ivector[i] = ivectorS[i] = s[i];

	for (offset = 0; ; offset += 512) {
		transform_iv((DWORD *)ivector, (DWORD*)pool, iv_ver);
		size = len > 512 ? 512 : len;
		len -= size;
		transform_buf(
			(DWORD *)pool,
			(DWORD *)(&src[offset]),
			(DWORD *)(&dst[offset]),
			size >> 2);
		alg_encrypt((DWORD *)ivector,
			(DWORD *)key_extended,
			(DWORD *)(&dst[offset]),
			(DWORD *)(&dst[offset]),
			size);
		if (!len) break;

		/* We need to update IVector every 512 bytes */
		increment_iv(ivectorS, iv_ver);
		for(i = 0; i < 4; i++) ivector[i] = ivectorS[i];
	}
	return 0;
}


static int decrypt_cbc(bc_vencrypt_cbc alg_decrypt, u_char* key_extended, u_char *iv, int iv_ver, u_char *src, u_char *dst, u_int len, u_char* pool)
{
	WORD	*s, ivector[4], ivectorS[4];
	DWORD	i, offset, size;

	s = (WORD *)iv;
	for (i = 0; i < 4; i++) ivector[i] = ivectorS[i] = s[i];

	for (offset = 0; ; offset += 512) {
		transform_iv((DWORD *)ivector, (DWORD *)pool, iv_ver);
		size = len > 512 ? 512 : len;
		len -= size;
		alg_decrypt((DWORD *)ivector,
			(DWORD *)key_extended,
			(DWORD *)(&src[offset]),
			(DWORD *)(&dst[offset]),
			size);
		transform_buf(
			(DWORD *)pool,
			(DWORD *)(&dst[offset]),
			(DWORD *)(&dst[offset]),
			size >> 2);
		if (!len) break;

		//  We need to update IVector every 512 bytes
		increment_iv(ivectorS, iv_ver);
		for(i = 0; i < 4; i++) ivector[i] = ivectorS[i];
	}
	return 0;
}


static int encrypt_lrw(	bc_vencrypt_ecb alg_encrypt_ecb,// algorithm specific encrypt int ecb mode fptr
			u_char* key_extended, 		// extended algorithm key
			u_int block_size, 		// size of sequential block to process at a time
			u_char *iv, 			// initial vector
			u_char *src, 			// data source to encrypt from
			u_char *dst, 			// data dest to encrypt into
			u_int len, 			// size of data
			u_char* pool, 			// random pool
			u_char* lrw_ctx			// lrw mode key context
)
{
	int            blockSize_words = 0, length_words, j;
	unsigned short *in_w, *out_w;
	unsigned short ivLocal[8] = {0}, t_lrw[8];
	uint64_t*	liByteAddress;

	liByteAddress = (uint64_t *)ivLocal;

	*liByteAddress = *((uint64_t*)iv);

	// make the address from sector number to byte offset
	(*liByteAddress) <<= 9; // *= 512

	blockSize_words = block_size / sizeof(unsigned short);
	length_words    = (int)len / sizeof(unsigned short);

	in_w  = (unsigned short *)src;
	out_w = (unsigned short *)dst;

	for( j = 0; j < length_words; j += blockSize_words)
	{
		// calculate t and then IV for next block
		if ( block_size == 8)
		{
			GFMUL_MulBlock8( (unsigned char *)ivLocal, (unsigned char *)t_lrw, (GfmulCtx *)lrw_ctx);

			// add 8 (size of block) to ivLocal
			(*liByteAddress) += 8;

			out_w[0] = in_w[0] ^ t_lrw[0];
			out_w[1] = in_w[1] ^ t_lrw[1];
			out_w[2] = in_w[2] ^ t_lrw[2];
			out_w[3] = in_w[3] ^ t_lrw[3];

			alg_encrypt_ecb( (unsigned char *)out_w, (unsigned char *)out_w, key_extended );

			out_w[0] ^= t_lrw[0];
			out_w[1] ^= t_lrw[1];
			out_w[2] ^= t_lrw[2];
			out_w[3] ^= t_lrw[3];

			in_w  = &(in_w[4]);
			out_w = &(out_w[4]);
		}
		else
		{
			GFMUL_MulBlock16( (unsigned char *)ivLocal, (unsigned char *)t_lrw, (GfmulCtx *)lrw_ctx);

			// add 16 (size of block) to ivLocal
			(*liByteAddress) += 16;

			out_w[0] = in_w[0] ^ t_lrw[0];
			out_w[1] = in_w[1] ^ t_lrw[1];
			out_w[2] = in_w[2] ^ t_lrw[2];
			out_w[3] = in_w[3] ^ t_lrw[3];
			out_w[4] = in_w[4] ^ t_lrw[4];
			out_w[5] = in_w[5] ^ t_lrw[5];
			out_w[6] = in_w[6] ^ t_lrw[6];
			out_w[7] = in_w[7] ^ t_lrw[7];

			alg_encrypt_ecb( (unsigned char *)out_w, (unsigned char *)out_w, key_extended );

			out_w[0] ^= t_lrw[0];
			out_w[1] ^= t_lrw[1];
			out_w[2] ^= t_lrw[2];
			out_w[3] ^= t_lrw[3];
			out_w[4] ^= t_lrw[4];
			out_w[5] ^= t_lrw[5];
			out_w[6] ^= t_lrw[6];
			out_w[7] ^= t_lrw[7];

			in_w  = &(in_w[8]);
			out_w = &(out_w[8]);
		}

        }
	return 0;
}


static int decrypt_lrw(	bc_vencrypt_ecb alg_decrypt_ecb,// algorithm specific encrypt int ecb mode fptr
			u_char* key_extended, 		// extended algorithm key
			u_int block_size, 		// size of sequential block to process at a time
			u_char *iv, 			// initial vector
			u_char *src, 			// data source to encrypt from
			u_char *dst, 			// data dest to encrypt into
			u_int len, 			// size of data
			u_char* pool, 			// random pool
			u_char* lrw_ctx			// lrw mode key context
)
{
	int            blockSize_words = 0, length_words, j;
	unsigned short *in_w, *out_w;
	unsigned short ivLocal[8] = {0}, t_lrw[8];
	uint64_t*	liByteAddress;

	liByteAddress = (uint64_t *)ivLocal;

	*liByteAddress = *((uint64_t*)iv);

	// make the address from sector number to byte offset
	(*liByteAddress) <<= 9; // *= 512

	blockSize_words = block_size / sizeof(unsigned short);
	length_words    = (int)len / sizeof(unsigned short);

	in_w  = (unsigned short *)src;
	out_w = (unsigned short *)dst;

	for( j = 0; j < length_words; j += blockSize_words)
	{
		// calculate t and then IV for next block
		if ( block_size == 8)
		{
			GFMUL_MulBlock8( (unsigned char *)ivLocal, (unsigned char *)t_lrw, (GfmulCtx *)lrw_ctx);

			// add 8 (size of block) to ivLocal
			(*liByteAddress) += 8;

			out_w[0] = in_w[0] ^ t_lrw[0];
			out_w[1] = in_w[1] ^ t_lrw[1];
			out_w[2] = in_w[2] ^ t_lrw[2];
			out_w[3] = in_w[3] ^ t_lrw[3];

			alg_decrypt_ecb( (unsigned char *)out_w, (unsigned char *)out_w, key_extended );

			out_w[0] ^= t_lrw[0];
			out_w[1] ^= t_lrw[1];
			out_w[2] ^= t_lrw[2];
			out_w[3] ^= t_lrw[3];

			in_w  = &(in_w[4]);
			out_w = &(out_w[4]);
		}
		else
		{
			GFMUL_MulBlock16( (unsigned char *)ivLocal, (unsigned char *)t_lrw, (GfmulCtx *)lrw_ctx);

			// add 16 (size of block) to ivLocal
			(*liByteAddress) += 16;

			out_w[0] = in_w[0] ^ t_lrw[0];
			out_w[1] = in_w[1] ^ t_lrw[1];
			out_w[2] = in_w[2] ^ t_lrw[2];
			out_w[3] = in_w[3] ^ t_lrw[3];
			out_w[4] = in_w[4] ^ t_lrw[4];
			out_w[5] = in_w[5] ^ t_lrw[5];
			out_w[6] = in_w[6] ^ t_lrw[6];
			out_w[7] = in_w[7] ^ t_lrw[7];

			alg_decrypt_ecb( (unsigned char *)out_w, (unsigned char *)out_w, key_extended );

			out_w[0] ^= t_lrw[0];
			out_w[1] ^= t_lrw[1];
			out_w[2] ^= t_lrw[2];
			out_w[3] ^= t_lrw[3];
			out_w[4] ^= t_lrw[4];
			out_w[5] ^= t_lrw[5];
			out_w[6] ^= t_lrw[6];
			out_w[7] ^= t_lrw[7];

			in_w  = &(in_w[8]);
			out_w = &(out_w[8]);
		}

        }
	return 0;
}

#define SECTOR_SIZE		512
#define XTS_BLOCK_SIZE		16
#define SECTOR_SIZE_POWER 	9
#define BLOCK_LENGTH_POWER	4
#define ULONGLONG 		unsigned long long

static int encrypt_decrypt_xts(	bc_vencrypt_ecb	EncryptProc,
                 		bc_vencrypt_ecb	EncryptOrDecryptProc,
                 		ULONGLONG       SectorNumber, // or IV - Initial Vector
	                 	unsigned char                          *InBuffer,
        	         	unsigned char                          *OutBuffer,
			 	unsigned long                          BufferLength,
                 		unsigned char                          *Context_1,
	                 	unsigned char                          *Context_2 )
{
	ULONGLONG     alpha_J_array[ SECTOR_SIZE / sizeof(ULONGLONG) ];
	ULONGLONG     ulSectorCounter, ulBlockCounter, ulBlockMax;
	unsigned long long    i, ul0, ul1, ulSpecial;
	ULONGLONG     ulPP[ XTS_BLOCK_SIZE / sizeof(ULONGLONG) ];
	ULONGLONG     *ulInBuffer, *ulOutBuffer;

	ulInBuffer  = (ULONGLONG *)InBuffer;
	ulOutBuffer = (ULONGLONG *)OutBuffer;

	ulBlockCounter = BufferLength >> BLOCK_LENGTH_POWER;

	for (ulSectorCounter = 0; ulSectorCounter < ((BufferLength >> SECTOR_SIZE_POWER) + 1); ulSectorCounter++)
	{
		alpha_J_array[0] = SectorNumber + ulSectorCounter;
		alpha_J_array[1] = 0;

		EncryptProc( (unsigned char *)alpha_J_array, (unsigned char *)alpha_J_array, Context_2 );

		for (i = 2; i < (SECTOR_SIZE / XTS_BLOCK_SIZE) * 2; i = i + 2 )
		{
			ul0 = alpha_J_array[ i - 2 ];
			ul1 = alpha_J_array[ i - 1 ];

			ulSpecial = 0;

			if (ul1 & (unsigned long long)0x8000000000000000ULL)
			{
				ulSpecial = 135;
			}

			ul1 <<= 1;

			if (ul0 & (unsigned long long)0x8000000000000000ULL)
			{
				ul1 |= 1;
			}

			ul0 <<= 1;

			ul0 ^= ulSpecial;

			alpha_J_array[i]   = ul0;
			alpha_J_array[i+1] = ul1;
		}

		if (ulBlockCounter < (SECTOR_SIZE / XTS_BLOCK_SIZE))
		{
			ulBlockMax = ulBlockCounter;
		}
		else
		{
			ulBlockMax = SECTOR_SIZE / XTS_BLOCK_SIZE;
		}

		for ( i = 0; i < (ulBlockMax << 1); i += 2 )
		{
			ulPP[0] = alpha_J_array[i]   ^ (*ulInBuffer++);
			ulPP[1] = alpha_J_array[i+1] ^ (*ulInBuffer++);

			EncryptOrDecryptProc( (unsigned char *)ulPP, (unsigned char *)ulPP, Context_1 );

			*ulOutBuffer++ = alpha_J_array[i]   ^ ulPP[0];
			*ulOutBuffer++ = alpha_J_array[i+1] ^ ulPP[1];

			ulBlockCounter--;
		}
	}

	return 0;
}




int bc_alg_encrypt (struct bc_algorithm* alg, bc_key_handle kh, u_char *iv, u_char *src, u_char *dst, u_int buflen, u_int mode)
{
	int err;
	int iv_ver;
	struct bc_key_element* elem;

	if (!buflen)
		return 0;

	if (buflen & (alg->ops->block_size_bytes - 1))
		return -EINVAL;

	if( (err = bc_alg_test_key(alg, kh)) )
		return err;

	iv_ver = guess_iv_format(iv);

	elem = &alg->key_table[kh];

	switch(mode)
	{
	case EX_ENCRYPT_CBC_MODE_FLAG:
		return encrypt_cbc(alg->ops->encrypt_cbc, elem->key_extended, iv, iv_ver, src, dst, buflen, elem->pool);

	case EX_ENCRYPT_LRW_MODE_FLAG:
		return encrypt_lrw(alg->ops->encrypt_ecb, elem->key_extended, alg->ops->block_size_bytes, iv, src, dst, buflen, elem->pool, elem->lrw_context);

	case EX_ENCRYPT_XTS_MODE_FLAG:
		return encrypt_decrypt_xts(alg->ops->encrypt_ecb, alg->ops->encrypt_ecb, *((ULONGLONG*)iv), src, dst, buflen, elem->key_extended, elem->xts_context);

	default:
		return -EINVAL;
	}
}


int bc_alg_decrypt (struct bc_algorithm* alg, bc_key_handle kh, u_char *iv, u_char *src, u_char *dst, u_int len, u_int mode)
{
	int err;
	int iv_ver;
	struct bc_key_element* elem;

	if (!len)
		return 0;

	if (len & (alg->ops->block_size_bytes - 1))
		return -EINVAL;

	if( (err = bc_alg_test_key(alg, kh)) )
		return err;

	iv_ver = guess_iv_format(iv);

	elem = &alg->key_table[kh];

	switch(mode)
	{
	case EX_ENCRYPT_CBC_MODE_FLAG:
		return decrypt_cbc(alg->ops->decrypt_cbc, elem->key_extended, iv, iv_ver, src, dst, len, elem->pool);

	case EX_ENCRYPT_LRW_MODE_FLAG:
		return decrypt_lrw(alg->ops->decrypt_ecb, elem->key_extended, alg->ops->block_size_bytes, iv, src, dst, len, elem->pool, elem->lrw_context);

	case EX_ENCRYPT_XTS_MODE_FLAG:
		return encrypt_decrypt_xts(alg->ops->encrypt_ecb, alg->ops->decrypt_ecb, *((ULONGLONG*)iv), src, dst, len, elem->key_extended, elem->xts_context);

	default:
		return -EINVAL;
	}
}


char bc_alg_c[]="$Id: bc_alg.c,v 1.10 2006-04-12 05:32:39 nail Exp $";
