#ifndef __BC_DRV_ALG_H__
#define __BC_DRV_ALG_H__

#include <linux/types.h>
#include <linux/module.h>

#include "bc_ioctl.h"
#include "wintypes.h"


#define KEY_TABLE_LENGTH 32

#ifndef __user
#define __user
#endif

/**
 * @brief	Encryption key entry.
 */
struct bc_key_element
{
	unsigned char	*key_extended;
	unsigned char	*lrw_context;
	unsigned char	*xts_context;
	unsigned char	*pool;
	int		lock;
};


/** @brief CBC mode encrypt\decrypt */
typedef VOID(*bc_vencrypt_cbc)(DWORD *IVector, DWORD *KeyAdress, DWORD *SrcBuffer, DWORD *DstBuffer, DWORD Length);

/** @brief ECB mode encrypt\decrypt */
typedef VOID(*bc_vencrypt_ecb)(BYTE* Input, BYTE* Output, BYTE* ExtendedKey); 

/** @brief Key Extention */
typedef BOOL(*bc_vkey_extend)(PUCHAR KeySource, PDWORD KeyDestination);


/**
 * @brief	Virtual algorithm definitions.
 */
struct bc_valg
{
	/* algorithm unique id */
	bc_alg_id		id;
	
	/* algorithm alias */
	char* 			name;

	/* algorithm encryption key length in bits*/
	unsigned int 		key_size_bits;

	/* algorithm extended key storage size in bytes */
	unsigned int 		ext_key_size_bytes;

	/* algorithm operation block size in bytes */
	unsigned int 		block_size_bytes;

	/* owning module to control reference count */
	struct module* 		owner;

	/* encrypt in cbc mode operation */
	bc_vencrypt_cbc		encrypt_cbc;

	/* decrypt in cbc mode operation */
	bc_vencrypt_cbc		decrypt_cbc;

	/* encrypt in ecb mode operation */
	bc_vencrypt_ecb		encrypt_ecb;

	/* decrypt in ecb mode */
	bc_vencrypt_ecb		decrypt_ecb;

	/* extend raw key data into algorithm representation */
	bc_vkey_extend		key_extend;
};


/**
 * @brief	Description of a registered and initialized algorithm.
 * @more	This is a more "inside" data, used internally by algorithm core.
 */
struct bc_algorithm
{
	/* algorithm definition */
	struct bc_valg* 	ops;

	/* 	key table.
		storing key table here is a compromise between security and centralized management.
	 */
	struct bc_key_element*	key_table;
};

	
//
//
//	Key Management
//
//


/**
 * @brief	Test algorithm key by handle.
 */
int bc_alg_test_key (struct bc_algorithm* alg, bc_key_handle kh);


/**
 * @brief	Lock algorithm key by handle.
 */
int bc_alg_lock_key (struct bc_algorithm* alg, bc_key_handle kh, int lock);


/**
 * @brief	Free algorithm key by handle.
 */
int bc_alg_free_key (struct bc_algorithm* alg, bc_key_handle kh);


/**
 * @brief	Allocate memory and store algorithm key, return user a handle.
 */
int bc_alg_make_key(struct bc_algorithm* alg, u_char* __user key, u_int keylen, u_char* __user pool, bc_key_handle* out_handle);


//
//
//	Encryption
//
//


/**
 * @brief	Encrypt block of data.
 */
int bc_alg_encrypt (struct bc_algorithm* alg, bc_key_handle kh, u_char *iv, u_char *src, u_char *dst, u_int buflen, u_int mode);


/**
 * @brief	Decrypt block of data.
 */
int bc_alg_decrypt (struct bc_algorithm* alg, bc_key_handle kh, u_char *iv, u_char *src, u_char *dst, u_int len, u_int mode);


#endif

