/*******************************************************************
 *	Copyright (c) 2003 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_dev26.c
 *
 *	Description:	BestCrypt pseudo-device driver.
 *
 *	Scope:		BestCrypt pseudo-device driver
 *
 *	Platforms:	Linux
 *
 *	Ideas:		Igor M. Arsenin
 *			Serge A. Frolov
 *			Vitaliy G. Zolotarev
 *
 *	Author:		Nail R. Kaipov
 *
 *	Created:	xx-Aug-2003
 *
 *	Revision:	20-Dec-2003.	updated to comply 2.6.0
 *
 *******************************************************************/

//#define	KBUILD_MODNAME	bc
#include "bc_cfg.h"
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>		/* module_param()		*/
#include <linux/vermagic.h>
#include <linux/compiler.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>		/* copy_from/to_user()		*/
#include <linux/hardirq.h>
#include <linux/vmalloc.h>		/* vmalloc()/vmfree()		*/
#include <linux/fs.h>
#include <linux/fs_struct.h>
#include <linux/file.h>			/* fget()			*/
#include <linux/proc_fs.h>		/* proc_fs support		*/
#include <linux/buffer_head.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
#include <linux/devfs_fs_kernel.h>	/* devfs_fs support		*/
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,25))
#include <linux/pid.h>
#include <linux/pid_namespace.h>
#endif

#include <linux/genhd.h>		/* gendisk interface		*/
#include <linux/bio.h>			/* struct bio			*/
#include <linux/blkdev.h>		/* rerquest_queue_t		*/
#include <linux/hdreg.h>		/* HDIO_GETGEO, hd_geometry	*/
//#include <bctypes.h>
#include "bc_ioctl.h"
#include "bc_mgr.h"
#include "bc_dev.h"
#include "bc_alg.h"

#if defined(CONFIG_SMP) || defined(CONFIG_PREEMPT)
#include <linux/smp_lock.h>		/* Shawn Leard added, resolve	*/
					/* Unknown symbol(s) in SuSE linux 9.2 */
#endif

#include <linux/dcache.h>
#include <linux/mount.h>

/* module parameters & friends */
static int		bc_devices = DEFAULT_BC_DEVICES;
static int		bc_partitions = DEFAULT_BC_PARTITIONS;
static int		bc_minor_shift = 0;

static struct bc_device	*bc_dev = NULL;	/* bc devices' array  */
static struct proc_dir_entry*	proc_bcrypt = NULL;

static inline void reset_dev(struct bc_device *bc)
{
	// do not touch bc_control, bc_refcnt, bc_number here
	if (NULL == bc)
		return;
	bc->bc_offset	= 0L;
	bc->bc_start_sector	= 0L;
	bc->bc_num_sectors	= 0L;
	bc->bc_dev	= 0; //to_kdev_t(0);
	bc->bc_gfpmask	= 0;
	bc->bc_biohead	= NULL;
	bc->bc_biotail	= NULL;
	bc->bc_process_bio = NULL;
	bc->bc_dentry	= 0;
	bc->bc_file	= 0;
	memset(&bc->bc_flags, 0, sizeof(bc->bc_flags));
	atomic_set(&bc->bc_pending, 0);
	spin_lock_init(&bc->bc_lock);
	sema_init(&bc->bc_thread, 0);
	sema_init(&bc->bc_run, 0);
	if (NULL != bc->bc_buffer)
		vfree(bc->bc_buffer);
	bc->bc_buffer = NULL;

}

/*- strategy routine stuff -*/
static int process_bio_fops(struct bc_device *bc, struct bio *bio)
{
	int 		i, rw;
	struct bio_vec	*bvec;
	char		*buf, *IV;
	loff_t		pos, sec;
	u16		iv[4];
	u64		iv64, iv64le;
	struct file	*file;
	size_t		len, ret, total; //, size;
	mm_segment_t	fs;

	if (READA == bio_rw(bio)) {
		BIO_ENDIO(bio, bio->bi_size, -EWOULDBLOCK);
		return 0;
	}
	rw    = bio_data_dir(bio);
	file  = bc->bc_file;
	sec   = (loff_t)bio->bi_sector + bc->bc_start_sector;
	pos   = sec * 512 + bc->bc_offset;
	total = 0;

	__bio_for_each_segment(bvec, bio, i, 0) {
		if (bc->bc_flags.iv_64bit) {
			iv64 = sec + (total >> 9) + 1; // iv numbering begins from 1
			iv64le = __cpu_to_le64(iv64);
			IV = (char *)&iv64le;
		} else {
			iv[0] = iv[1] = iv[2] = iv[3] = (sec + (total >> 9)) & 0xFFFF;
			IV = (char *)iv;
		}
		len = bvec->bv_len;

		kmap(bvec->bv_page);
		total += len;
		buf = page_address(bvec->bv_page)+bvec->bv_offset;
		fs = get_fs();

		if (READ == rw) {
			set_fs(KERNEL_DS);
			ret = file->f_op->read(file, buf, len, &pos);
			set_fs(fs);
			bc_alg_decrypt(bc->bc_alg, bc->bc_key, IV, buf, buf, len, bc->bc_alg_mode);
		} else {
			bc_alg_encrypt(bc->bc_alg, bc->bc_key, IV, buf, bc->bc_buffer, len, bc->bc_alg_mode);
			set_fs(KERNEL_DS);
			ret = file->f_op->write(file, bc->bc_buffer, len, &pos);
			set_fs(fs);
		}
		kunmap(bvec->bv_page);

		if (ret != len) {
			printk(KERN_ERR "bc: %s error wanted %ld, got %ld\n",
				READ == rw ? "read" : "write",
				(unsigned long)bvec->bv_len,
				(unsigned long)ret);
			goto error_out;
		}
	}
	BIO_ENDIO(bio, bio->bi_size, 0);
	return 0;
error_out:
	BIO_ENDIO(bio, bio->bi_size, -EIO);
	return 0;
}

static int bc_thread(void *data)
{
	struct bc_device	*bc = (struct bc_device *)data;
	struct bio		*bio;

	daemonize("bcrypt%d", bc->bc_number);
	set_user_nice(current,  -20l);
	set_user_nice(current,  -10l);
#if defined(PF_IOTHREAD)
	current->flags |= PF_IOTHREAD;
#endif
#if defined(PF_NOFREEZE)
	current->flags |= PF_NOFREEZE;
#endif
#if defined(PF_LESS_THROTTLE)
	current->flags |= PF_LESS_THROTTLE;
#endif

	spin_lock_irq(&bc->bc_lock);
	bc->bc_flags.active = 1;
	atomic_inc(&bc->bc_pending);
	spin_unlock_irq(&bc->bc_lock);

	up(&bc->bc_thread);
	while (1) {
		(void)down_interruptible(&bc->bc_run);

		if (0 == atomic_read(&bc->bc_pending))
			break;

		// fetch bio
		spin_lock_irq(&bc->bc_lock);
		bio = bc->bc_biohead;
		if (NULL == bio) {
			spin_unlock_irq(&bc->bc_lock);
			continue;
		}
		if (bc->bc_biotail == bio)
			bc->bc_biotail = NULL;
		bc->bc_biohead = bio->bi_next;
		bio->bi_next = NULL;
		spin_unlock_irq(&bc->bc_lock);

		// process bio
		bc->bc_activity = 1;
		if (NULL == bc->bc_process_bio) {
			BIO_ENDIO(bio, bio->bi_size, -EIO);
		} else {
			bc->bc_process_bio(bc, bio);
		}

		if (atomic_dec_and_test(&bc->bc_pending))
		    break;
	}
	up(&bc->bc_thread);
	return 0;
}

static int bc_make_request(struct request_queue *q, struct bio *bio)
{
	struct bc_device	*bc;
	int			err;
	bc = q->queuedata;

	if (!bc || !bc->bc_dentry || !bc->bc_dentry->d_inode) {
		printk(KERN_ERR "bc: invalid device for request.\n");
		err = -ENODEV;
		goto error_out;
	}
/*
	if (!bd->bd_flags.configured && !bc->bc_flags.busy) {
		printk(KERN_ERR "bc: device not configured yet.\n");
		goto error_out;
	}
*/

	err = -EIO;
	switch (bio_rw(bio)) {
	case WRITE:
		if (bc->bc_flags.readonly) {
			printk(KERN_ERR "bc: attempt to write on readonly device.\n");
			goto error_out;
		}
	case READ:
		break;
	case READA:
		err = -EWOULDBLOCK;
		goto error_out;
	default:
		printk(KERN_ERR "bc: unknown command (%lu).\n", bio_rw(bio));
		goto error_out;
	}

	spin_lock_irq(&bc->bc_lock);
	if (!bc->bc_flags.active) {
		spin_unlock_irq(&bc->bc_lock);
		goto error_out;
	}
	atomic_inc(&bc->bc_pending);
	if (bc->bc_biotail)	bc->bc_biotail->bi_next = bio;
	else			bc->bc_biohead = bio;
	bc->bc_biotail = bio;
	spin_unlock_irq(&bc->bc_lock);
	up(&bc->bc_run);

	return 0;

error_out:
	BIO_ENDIO(bio, bio->bi_size, err);
	return 0;
}
/*----------------------------------------------------------------------*/

static int bc_get_version(struct bc_device *bc, struct bc_version *arg)
{
	struct bc_version	query;

	BC_GET_ARG(arg, query);

	put_user(BC_VERSION_MAJOR, (int *)&arg->ver_major);
	put_user(BC_VERSION_MINOR, (int *)&arg->ver_minor);

	return 0;
}

static int bc_is_free(struct bc_device* bc, int* is_free)
{
	if( !is_free ) return -EINVAL;
	put_user(!bc->bc_flags.configured, is_free);
	return 0;
}

/*
 *	Compatibility to give pre 1.7 software a chance to gracefully report version mismatch
 */
static int bc_get_info(struct bc_device* bc, struct bc_info* arg)
{
	struct bc_info	query;
//	int busy;

	BC_GET_ARG(arg, query);

	put_user(BC_VERSION_MAJOR, (int *)&arg->ver_major);
	put_user(BC_VERSION_MINOR, (int *)&arg->ver_minor);

	return bc_is_free(bc, (int*)&arg->busy);
}


static int install_process_bio(struct bc_device *bc, struct file *file, struct block_device *bdev, int ro)
{
	int		error;
	struct inode	*inode;

	error = -EINVAL;
	inode = file->f_dentry->d_inode;

	if (!inode) {
		printk(KERN_ERR "bc_set_fd: NULL inode.\n");
		goto error_out;
	}
	if (!file->f_op) {
		printk(KERN_ERR "bc_set_fd: NULL file_operations.\n");
		goto error_out;
	}
	if (!file->f_op->read) {
		printk(KERN_ERR "bc_set_fd: Can't perform reads.\n");
		goto error_out;
	}

	error = -EACCES;
	if (S_ISBLK(inode->i_mode)) {
//		error = blkdev_open(inode, file);
//		if (error)
//			goto error_out;
		bc->bc_dev   = inode->i_rdev;
		ro = ro || bdev_read_only(inode->i_bdev);
	} else if (S_ISREG(inode->i_mode)) {
//		bc->bc_dev   = inode->i_dev;
		bc->bc_dev   = inode->i_rdev; // !!! error !!!
	} else
		goto error_out;

	bc->bc_flags.readonly = ro || IS_RDONLY(inode) ||
				(NULL == file->f_op->write);

	if (!bc->bc_flags.readonly) {
		error = get_write_access(inode);
		if (error) {
			printk (KERN_ERR "bc_set_fd: Can't get write access.\n");
		        goto error_out;
		}
//		invalidate_inode_pages (inode);
	}
	set_device_ro(bdev, !!bc->bc_flags.readonly);
	bc->bc_process_bio = process_bio_fops;
	return 0;

error_out:
	return error;
}

#if defined(QUEUE_FLAG_PLUGGED)
// never called actually.
static void bc_unplug_fn(struct request_queue *q)
{
	struct bc_device *bc = q->queuedata;

	clear_bit(QUEUE_FLAG_PLUGGED, &q->queue_flags);
	blk_run_address_space(bc->bc_file->f_mapping);
}
#endif

static int scan_device(struct bc_device *bc, struct block_device *bdev)
{
	loff_t		bytes;
	sector_t	sectors;

	if (NULL == bc->bc_dentry)
		return -ENXIO;
/*  from blkrrpart */
//	if (bc->bc_refcnt > 1)
//		return -EBUSY;
//	bc->bc_flags.busy = 1;

	bytes  = i_size_read(bc->bc_dentry->d_inode->i_mapping->host);
	bytes -= bc->bc_offset;

	sectors = (sector_t)(bytes >> 9);

	if ((loff_t)sectors != (bytes >> 9))
		return -EFBIG;

	bdev->bd_inode->i_size = bytes;		/* major headache if forgotten */
	set_capacity(bc->bc_gendisk, sectors);

	return 0;
}

static int bc_set_fd(struct bc_device *bc, struct block_device *bdev, struct bc_file64 *arg)
{
	struct file		*file = NULL;
	struct bc_algorithm	*bc_alg = NULL;
	struct bc_file64	query;
	int 			error;

	try_module_get(THIS_MODULE);
	BC_GET_ARG(arg, query)

	error = -EBUSY;
	if (bc->bc_dentry) {
		goto error_out;
	}

	bc_alg = bc_mgr_get_alg(query.alg_id);
	if (NULL == bc_alg || bc_alg_test_key(bc_alg, query.key_handle) != 0) {
		error = -EINVAL;
		goto error_out;
	}

	if (!(file = fget(query.fd))) {
		error = -EBADF;
		goto error_out;
	}

	reset_dev(bc);

	bc->bc_buffer   = vmalloc(BC_BUFFER_SIZE);
	if (NULL == bc->bc_buffer)
		goto error_out;

	error = install_process_bio(bc, file, bdev, query.flags & BC_FLAGS_READONLY);
	if (0 != error)
		goto error_out;

	bc->bc_key	= query.key_handle;
	bc->bc_alg	= bc_alg;
	bc->bc_alg_mode = query.alg_mode;
	bc->bc_file	= file;
	bc->bc_dentry	= dget(file->f_dentry);
	bc->bc_offset   = (loff_t)query.offset;
	bc->bc_start_sector  = (loff_t)query.start_sector;
	bc->bc_num_sectors   = (loff_t)query.num_sectors;

	error = scan_device(bc, bdev);
	if (0 != error) {
		dput(file->f_dentry);
		goto error_out;
	}

	bc_alg_lock_key(bc_alg, query.key_handle, 1);
	bc->bc_gfpmask = mapping_gfp_mask(bc->bc_dentry->d_inode->i_mapping);
	mapping_set_gfp_mask(bc->bc_dentry->d_inode->i_mapping, bc->bc_gfpmask & ~(__GFP_IO | __GFP_FS));

	bc->bc_flags.iv_64bit   = query.flags & BC_FLAGS_IV_64BIT || 0;
	bc->bc_flags.multpart   = query.flags & BC_FLAGS_MULTPART || 0;
	bc->bc_flags.configured = 1;

	blk_queue_make_request(bc->bc_queue, bc_make_request);
	bc->bc_queue->queuedata = bc;

#if defined(QUEUE_FLAG_PLUGGED)
	bc->bc_queue->unplug_fn = bc_unplug_fn;
#endif

	kernel_thread(bc_thread, bc, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	down(&bc->bc_thread);

	return 0;

error_out:

	if (file)
		fput(file);
	module_put(THIS_MODULE);
	return error;
}

static int bc_clr_fd(struct bc_device *bc, struct block_device *bdev, /*struct inode *dev_inode,*/ int force)
{
	int		n;//, minor;
	//size_t		ret;
	//loff_t		pos;
	//unsigned char	magic;
	//mm_segment_t	fs;


//	if (!is_parent(bd))
//		return -EPERM;
	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == bc->bc_file)
		return -ENXIO;
	if (bc->bc_refcnt > 1 && !force)
		return -EBUSY;

	fsync_bdev(bdev);

	spin_lock_irq(&bc->bc_lock);
	if (atomic_dec_and_test(&bc->bc_pending))
		up(&bc->bc_run);
	bc->bc_flags.active = 0;
	spin_unlock_irq(&bc->bc_lock);
	down(&bc->bc_thread);
//	waitpid(-1, NULL, __WCLONE|WNOHANG);

	mapping_set_gfp_mask(bc->bc_dentry->d_inode->i_mapping, bc->bc_gfpmask);

	n = 1 << bc_minor_shift;
/*
	for (i = 0; i < n; i++) {
		minor = bc->bc_number*n+i;
		sync_dev(MKDEV(MAJOR_NR, minor));
		invalidate_buffers(MKDEV(MAJOR_NR, minor));
		while (i) {
			hdevfs = devfs_find_handle(bc->bc_hdevfs, NULL,
						MAJOR_NR, minor,
						DEVFS_SPECIAL_BLK, 0);
			if (NULL == hdevfs) {
				bc_hd[minor].de = 0;
				break;
			}
			if (hdevfs != bc->bc_hdevfs)
				devfs_unregister(hdevfs);
		}
		bc_sizes[minor] = 0;
	}
*/
/*
	if (bc->bc_save_fops) {
		dev_inode->i_fop = bc->bc_save_fops;
	}
*/


// since v8 encrypted header containers, we can not unlock containers this way.
/*
	if (!bc->bc_flags.readonly && bc->bc_offset) {
		pos = 0;
		magic = 0xEB;
		fs = get_fs();
		set_fs(KERNEL_DS);
		ret = bc->bc_file->f_op->write(bc->bc_file, &magic, 1, &pos);
		set_fs(fs);
		if (ret != 1)
			printk(KERN_ERR "bc: error %ld writing magic\n",
			(unsigned long)ret);
	}
*/
/*
	if (S_ISBLK(bc->bc_dentry->d_inode->i_mode))
#ifdef _BLKDEV_2_
		blkdev_put(bc->bc_dentry->d_inode->i_bdev, BDEV_FILE);
#else
		blkdev_put(bc->bc_dentry->d_inode->i_bdev);
#endif
*/
	if (!bc->bc_flags.readonly)
		put_write_access(bc->bc_dentry->d_inode);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))
	invalidate_bdev(bdev, 0);
#else
	invalidate_bdev(bdev);
#endif

	set_capacity(bc->bc_gendisk, 0);

	dput(bc->bc_dentry);
	fput(bc->bc_file);

	bc_alg_free_key(bc->bc_alg, bc->bc_key);

	reset_dev(bc);

	module_put(THIS_MODULE);

	return 0;
}


/*
 *	Find bc vfs mounted entry
 *	Added 10.02.09, inso.
 */
static struct vfsmount* bc_find_mount( struct vfsmount** mnt, struct block_device* bdev )
{
	struct vfsmount *root, *p;
	struct list_head* head;

	if( mnt == NULL ) return NULL;
	else *mnt = NULL;

	#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25))
		root = current->fs->rootmnt;
	#else
		root = current->fs->root.mnt;
	#endif

	list_for_each(head, &root->mnt_list) {
		p = list_entry(head, struct vfsmount, mnt_list);
		if (NULL == p ||
		    NULL == p->mnt_sb || (struct super_block*)0x1 == p->mnt_sb ||
		    NULL == p->mnt_root ||
		    NULL == p->mnt_devname)
			continue;

		if( bdev->bd_inode->i_rdev == p->mnt_sb->s_dev ) {
			/*printk( KERN_ERR "found mounted device: %s\n", p->mnt_devname);*/
			*mnt = p;
			break;
		}
	}

	return *mnt;
}

static int bc_lock_dev(struct block_device *bdev, int lock)
{
	struct vfsmount    *mnt;
	struct bc_device* bc = bdev->bd_disk->private_data;

	if (!bc->bc_flags.configured) {
		if (!lock)
			return 0;
		printk(KERN_ERR "bc: attempt to lock free device.\n");
		return -ENXIO;
	}
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
		spin_lock(&dcache_lock);
	#endif
	if (lock && !bc->bc_flags.mounted) {

		if( bc_find_mount( &mnt, bdev ) != NULL ) {
			mntget(mnt);
			if (mnt->mnt_sb->s_flags & MS_SYNCHRONOUS) {
				mnt->mnt_sb->s_flags &= ~MS_SYNCHRONOUS;
			}
		}

		bc->bc_flags.mounted = 1;
		bc->bc_refcnt++;

	} else if (!lock && bc->bc_flags.mounted) {

		if( bc_find_mount( &mnt, bdev ) != NULL ) {
			mntput(mnt);
			if (mnt->mnt_sb->s_flags & MS_SYNCHRONOUS) {
				mnt->mnt_sb->s_flags &= ~MS_SYNCHRONOUS;
			}
		}

		bc->bc_flags.mounted = 0;
		bc->bc_refcnt--;
	}
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
		spin_unlock(&dcache_lock);
	#endif

	return 0;
}

static int bc_force_unlock(struct block_device *bdev)
{
	struct vfsmount    *mnt;
	struct bc_device* bc = bdev->bd_disk->private_data;

	if (!bc->bc_flags.configured)
		return 0;

	fsync_bdev(bdev);

	if (bc->bc_flags.mounted) {
		#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
			spin_lock(&dcache_lock);
		#endif

		if( bc_find_mount( &mnt, bdev ) != NULL ) {
		#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)) || !defined(CONFIG_SMP)
			while (atomic_read(&mnt->mnt_count) > 1) mntput(mnt);
		#else 
			
		#endif		
		}

		#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
			spin_unlock(&dcache_lock);
		#endif
		bc->bc_flags.mounted = 0;
		bc->bc_refcnt--;
	}

	return 0;
}

static int bc_stat_dev(struct block_device* bdev, struct bc_dev_info* arg)
{
	struct bc_device* 	dev = bdev->bd_disk->private_data;
	struct bc_dev_info	query;

	copy_from_user(&query, arg, sizeof(query));

	if(dev->bc_flags.configured) {
		char* buf;
		char* ppath;
		unsigned pathlen = query.bufsiz;

		buf = kmalloc(pathlen, GFP_KERNEL);
		if(!buf) {
			BC_DEBUG_PRINT((KERN_ERR "bc_stat_dev: failed allocating path buffer of size %d\n", pathlen));
			return -ENOMEM;
		}

		memset(buf, '\0', pathlen);

		/* d_path build path from the end of the buffer so unused buffer space will be from the beginning. */
		ppath = d_path(&dev->bc_file->f_path, buf, pathlen);
		if((unsigned long) ppath <= MAX_ERRNO) {
			BC_DEBUG_PRINT((KERN_ERR "bc_stat_dev: d_path failed with 0x%lx\n", PTR_ERR(ppath)));
			kfree(buf);
			return -PTR_ERR(ppath);
		}

		pathlen -= (ppath - buf);
		copy_to_user(query.pathbuf, ppath, pathlen);
		kfree(buf);

		query.is_configured = 1;
		query.bufsiz = pathlen;
	} else {
		query.is_configured = 0;
	}

	copy_to_user(arg, &query, sizeof(query));

	return 0;
}


/*-- algorithm driver ioctls --------------------------------*/

static int bc_vrfy_alg(struct bc_alg *arg)
{
	struct bc_alg		query;
	int exists;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC )
		return -EINVAL;

	exists = (bc_mgr_get_alg(query.alg_id) != NULL);
	put_user(exists, (int *)&arg->verified);

	return 0;
}

static int bc_make_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || !query.key || !query.pool)
		return -EINVAL;

	bc_alg = bc_mgr_get_alg(query.alg_id);//, &query.alg_module[0]);
	if (!bc_alg)
		return -EINVAL;

	/* WARNING!!! using user mode memory areas */
	error = bc_alg_make_key(bc_alg, query.key, query.key_len, query.pool, &query.key_handle);
	if (error)
		return error;

	if (copy_to_user(arg, &query, sizeof(query)))
		return -EFAULT;

	return 0;
}

static int bc_free_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC)
		return -EINVAL;

	bc_alg = bc_mgr_get_alg(query.alg_id);
	if (!bc_alg)
		return -EINVAL;
	error = bc_alg_free_key(bc_alg, query.key_handle);
	return error;
}

static int bc_process(struct bc_block *arg, int operation)
{
	struct bc_block		query;
	struct bc_algorithm	*bc_alg;
	int			error;
	char			*buffer;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC ||
	   !query.buffer ||
	   query.buffer_len > MAX_BLOCK_SIZE)
	{
		return -EINVAL;
	}

	bc_alg = bc_mgr_get_alg(query.alg_id);
	if (!bc_alg)
	{
		BC_DEBUG_PRINT((KERN_ERR "bc_process: request failed, algorithm id %d not found\n", query.alg_id));
		return -EINVAL;
	}

	buffer = vmalloc(query.buffer_len);
	if (!buffer)
		return -ENOMEM;

	if (copy_from_user(buffer, query.buffer, query.buffer_len)) {
		error = -EFAULT;
		goto error_out;
	}

	switch (operation) {
	case BC_ENCRYPT_BLOCK:
		error = bc_alg_encrypt(bc_alg, query.key_handle, &query.iv[0],
				buffer, buffer, query.buffer_len, query.alg_mode);
		break;
	case BC_DECRYPT_BLOCK:
		error = bc_alg_decrypt(bc_alg, query.key_handle, &query.iv[0],
				buffer, buffer, query.buffer_len, query.alg_mode);
		break;
	default:
		error = -EINVAL;
		goto error_out;
	}
	if (0 == error)
		error = copy_to_user(query.buffer, buffer, query.buffer_len);

error_out:
	if (buffer) {
		memset(buffer, 0, query.buffer_len);
		vfree(buffer);
		buffer = NULL;
	}
	return error;
}
/*-------------------------------------------------------------------*/

static int hdio_getgeo(struct block_device *bdev, struct hd_geometry *geo)
{
	unsigned long 		i, j, size;
	unsigned char 		head, sect;
	unsigned short 		cyl;
	struct bc_device*	bc;

	bc = bdev->bd_disk->private_data;

	if (NULL == bc->bc_dentry)
		return -ENXIO;
	if (NULL == geo)
		return -EINVAL;

	//size = bc->bc_gendisk->capacity << 1;
	size = get_capacity( bc->bc_gendisk ) << 1;	// FIX for 2.6.28, works on all 2.6.x versions

	for (i = -1, j = size; j > 0; j >>= 1, i++);

	if (0 > i)
		return -ENXIO;

	if (16 > i) {
		sect = 1;
		head = 1;
		cyl  = size;
	} else if (23 > i) {
		sect = 1;
		head = 1 << (i-15);
		cyl  = size >> (i-15);
	} else {
		sect = 1 << (i-22);
		head = 1 << 7;
		cyl  = size >> (i-15);
	}
	put_user (head, (u8 *)&geo->heads);
	put_user (sect, (u8 *)&geo->sectors);
	put_user (cyl, (u16 *)&geo->cylinders);
	put_user (0,   (u32 *)&geo->start);
	return 0;
}

/*
 *	bc_ioctl helper
 *	operates on new 2.6.28 block API
 */
static int bc_ioctl_generic(struct block_device *bdev, unsigned mode, u_int cmd, u_long arg) {
	struct bc_device	*bc;
	int		error;

	BC_DEBUG_PRINT((KERN_ERR "bc: ioctl code 0x%x\n", cmd));

	if (NULL == bdev)
		return -EINVAL;

	bc = bdev->bd_disk->private_data;
	if (NULL == bc)
		return -ENODEV;

//	bd = get_bcdsk(inode->i_rdev);
//	if (NULL == bd)
//		return -ENODEV;

	error = -EINVAL;
	down(&bc->bc_control);
	switch (cmd) {

	BC_HANDLER("get_info", BC_GET_INFO,   bc_get_info(bc, (struct bc_info*) arg));
	BC_HANDLER("get_version", BC_GET_VERSION,   bc_get_version(bc, (struct bc_version*) arg));
	BC_HANDLER("is_free",  BC_IS_FREE,		bc_is_free( bc, (int*)arg ));
	BC_HANDLER("set_fd  ", BC_SET_FD,        bc_set_fd  (bc, bdev, (struct bc_file64 *) arg));
	BC_HANDLER("clr_fd  ", BC_CLR_FD,        bc_clr_fd  (bc, bdev, /*inode,*/ (int)arg));

	BC_HANDLER("lock_dev", BC_LOCK_DEV,      bc_lock_dev(bdev, 1));
	BC_HANDLER("ulck_dev", BC_UNLOCK_DEV,    bc_lock_dev(bdev, 0));
	BC_HANDLER("stat_dev", BC_STAT_DEV,	 bc_stat_dev(bdev, (struct bc_dev_info*) arg));

	BC_HANDLER("frc_ulck", BC_FORCE_UNLOCK,  bc_force_unlock(bdev));
//	BC_HANDLER("chk_actv", BC_CHECK_ACTIVITY,bc_check_activity(bc, inode->i_rdev, (long *)arg));

	BC_HANDLER("hdio_geo", HDIO_GETGEO,      hdio_getgeo(bdev, (struct hd_geometry *) arg));
//	BC_HANDLER("rrd_part", BLKRRPART,        scan_device(bd, bc));

//	BC_HANDLER("check_fd", BC_CHECK_FD,      bc_check   ((struct bc_check*) arg));

	BC_HANDLER("vrfy_alg", BC_VERIFY_ALG,    bc_vrfy_alg((struct bc_alg*)   arg));
	BC_HANDLER("make_key", BC_MAKE_KEY,      bc_make_key((struct bc_key*)   arg));
	BC_HANDLER("free_key", BC_FREE_KEY,      bc_free_key((struct bc_key*)   arg));
	BC_HANDLER("encr_blk", BC_ENCRYPT_BLOCK, bc_process ((struct bc_block*) arg, BC_ENCRYPT_BLOCK));
	BC_HANDLER("decr_blk", BC_DECRYPT_BLOCK, bc_process ((struct bc_block*) arg, BC_DECRYPT_BLOCK));
	}
	up(&bc->bc_control);
	return error;
}


/*
 *	bc_open helper
 *	operates on new 2.6.28 block API
 */
static int bc_open_generic( struct block_device* bldev, unsigned mode ) {
  	struct bc_device *bc;

	if (NULL == bldev)
	{
		BC_DEBUG_PRINT((KERN_ERR "Debug: NULL == bldev\n"));
		return -EINVAL;
	}


	bc = bldev->bd_disk->private_data;
	if (NULL == bc) {
		BC_DEBUG_PRINT((KERN_ERR "Debug: NULL == bc\n"));
		return -EINVAL;
	}

	down(&bc->bc_control);
	bc->bc_refcnt++;
	up(&bc->bc_control);

	return 0;
}


/*
 *	bc_release helper
 *	operates on new 2.6.28 block API
 */
static int bc_release_generic(struct gendisk * gendisk, unsigned mode) {

	struct bc_device *bc;

	if (NULL == gendisk)
	    return -EINVAL;

	bc = gendisk->private_data;

	down(&bc->bc_control);
	bc->bc_refcnt--;
	up(&bc->bc_control);

	return 0;
}




/*-----------------------------------------------------------*/
int proc_bc_stat(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	char *p = page;

	p += sprintf(p, "# BestCrypt system v" BC_VERSION_STRING ":\n");
	p += sprintf(p, "# Compiled at "__TIME__ " " __DATE__"\n");
	p += sprintf(p, "devices     %d\n", bc_devices);
	p += sprintf(p, "partitions  %d\n", bc_partitions);
	p += sprintf(p, "major       %d\n", MAJOR_NR);
	p += sprintf(p, "minor       0\n");

//	p += sprintf(p, "pending     %d\n", atomic_read(&bc_dev[0].bc_pending));

	len = p - page;
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

void bc_cleanup(void)
{
	int	i;

#ifdef CONFIG_PROC_FS
	if (NULL != proc_bcrypt) {
		remove_proc_entry(DEVICE_CONF, proc_bcrypt);
		remove_proc_entry(DEVICE_STAT, proc_bcrypt);
	}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))
	remove_proc_entry (DEVICE_NAME, &proc_root);
#else
	remove_proc_entry(DEVICE_NAME, NULL);
#endif

#endif

	if (bc_dev) {
		for (i = 0; i < bc_devices;  i++) {
			if (bc_dev[i].bc_gendisk){
				del_gendisk(bc_dev[i].bc_gendisk);
			}
			if (bc_dev[i].bc_queue) {
				blk_cleanup_queue(bc_dev[i].bc_queue);
			}
			if (bc_dev[i].bc_gendisk) {
				put_disk(bc_dev[i].bc_gendisk);
			}
		}
		kfree(bc_dev);
	}
	unregister_blkdev(MAJOR_NR, DEVICE_NAME);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28))

static struct block_device_operations bc_bdops = {
	.owner =	THIS_MODULE,
	.open =		bc_open_generic,
	.release =	bc_release_generic,
	.ioctl =	bc_ioctl_generic,
	.getgeo = 	hdio_getgeo,
};

#else

static int bc_open(struct inode *inode, struct file *file)
{
	if( NULL == inode ) return -EINVAL;
	return (file != NULL) ? bc_open_generic( inode->i_bdev, file->f_mode ) : bc_open_generic(inode->i_bdev, 0);
}

static int bc_release(struct inode *inode, struct file *file)
{
	if (NULL == inode)
	    return -EINVAL;

	return (file != NULL) ? bc_release_generic(inode->i_bdev->bd_disk, file->f_mode) : bc_release_generic(inode->i_bdev->bd_disk, 0);
}

static int bc_ioctl(struct inode *inode, struct file *file, u_int cmd, u_long arg)
{
	if (NULL == inode)
		return -EINVAL;


	return (file != NULL) ? bc_ioctl_generic(inode->i_bdev, file->f_mode, cmd, arg) : bc_ioctl_generic(inode->i_bdev, 0, cmd, arg);
}

static struct block_device_operations bc_bdops = {
	.owner =	THIS_MODULE,
	.open =		bc_open,
	.release =	bc_release,
	.ioctl =	bc_ioctl,

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
		.getgeo =   hdio_getgeo,
	#endif
};

#endif

int __init bc_init(void)
{
//	char	buf[32];
	int	i, max_devices, rc = 0;

	if (bc_devices > MAX_MINORS || bc_devices <= 0) {
		bc_devices = DEFAULT_BC_DEVICES;
	}

	if (bc_partitions > 63)
		bc_partitions = 63;

	if (bc_partitions > 1) {
		for (i = 0; bc_partitions; i++)
			bc_partitions >>= 1;
		bc_partitions = 1 << i;
		bc_minor_shift = i;
		if (MAX_MINORS/bc_partitions < bc_devices)
			bc_devices = MAX_MINORS/bc_partitions;
	} else {
		bc_partitions = 1;
		bc_minor_shift = 0;
	}

	bc_partitions = 1;
	bc_minor_shift = 0;
	max_devices = bc_devices * bc_partitions;

	printk(KERN_INFO "Loading %s (%d[%d] devices)...\n",
			    BC_VERSION_STRING,
			    bc_devices,
			    bc_partitions);
	if (register_blkdev(MAJOR_NR, DEVICE_NAME))	// nothing special, just reserve Major
		return -EIO;
	if (0 > bc_mgr_init())		// initialize algorithm manager
		return -EINVAL;

	bc_dev = kmalloc( bc_devices * sizeof(struct bc_device), GFP_KERNEL);
	if (NULL == bc_dev) {
		rc = -ENOMEM;
		goto error_out;
	}
	memset(bc_dev,       0, bc_devices * sizeof(struct bc_device));
	for (i = 0; i < bc_devices;  i++) {
		bc_dev[i].bc_number = i;
		sema_init(&bc_dev[i].bc_control, 1);
		sema_init(&bc_dev[i].bc_thread, 0);
		sema_init(&bc_dev[i].bc_run, 0);
		spin_lock_init(&bc_dev[i].bc_lock);

		// gendisk allocation and initialization

		bc_dev[i].bc_gendisk = alloc_disk(bc_partitions);
		if (NULL == bc_dev[i].bc_gendisk)
			goto error_out;

		bc_dev[i].bc_queue = blk_alloc_queue(GFP_KERNEL);
		if (NULL == bc_dev[i].bc_queue)
			goto error_out;

		bc_dev[i].bc_gendisk->major = MAJOR_NR;
		bc_dev[i].bc_gendisk->first_minor = i << bc_minor_shift;
		bc_dev[i].bc_gendisk->minors = bc_partitions;
		bc_dev[i].bc_gendisk->fops = &bc_bdops;

		bc_dev[i].bc_gendisk->queue = bc_dev[i].bc_queue;
		set_capacity( bc_dev[i].bc_gendisk, 0L );	// fix after 2.6.28, works for all 2.6 family
		bc_dev[i].bc_gendisk->flags = 0;	// ????

		bc_dev[i].bc_gendisk->private_data = &bc_dev[i];
		sprintf(bc_dev[i].bc_gendisk->disk_name, DEVICE_NAME "%d", i);
		add_disk(bc_dev[i].bc_gendisk);
	}

#ifdef CONFIG_PROC_FS
	proc_bcrypt = create_proc_entry(DEVICE_NAME, S_IFDIR, 0);
	if (NULL != proc_bcrypt) {
		create_proc_read_entry(DEVICE_CONF, S_IFREG | S_IRUGO,
					proc_bcrypt, bc_mgr_proc_conf, NULL);
		create_proc_read_entry(DEVICE_STAT, S_IFREG | S_IRUGO,
					proc_bcrypt, proc_bc_stat, NULL);
	}
#endif

	return 0;
error_out:
	bc_cleanup();
	return rc;
}

void __exit bc_exit(void)
{
	bc_cleanup();
	printk("BestCrypt driver is unloaded...\n");
}

MODULE_AUTHOR("Jetico, Inc.");
MODULE_DESCRIPTION("BestCrypt encrypted block device driver.");
MODULE_LICENSE("Jetico, Inc.");
MODULE_PARM_DESC(bc_devices, "Max. number of devices (1-256).");
MODULE_PARM_DESC(bc_partitions, "Max. partitions per device (0-63).");
MODULE_INFO(vermagic, VERMAGIC_STRING);

#ifdef _NEED_THIS_MODULE
struct module __this_module
__attribute__((section(".gnu.linkonce.this_module"))) = {
#ifdef _KBUILD_QUOTES
 .name = KBUILD_MODNAME,
#else
 .name = __stringify(KBUILD_MODNAME),
#endif
 .init = init_module,
#ifdef CONFIG_MODULE_UNLOAD
 .exit = cleanup_module,
#endif
};

static const char __module_depends[]
#ifdef __attribute_used__
__attribute_used__
#else
__used
#endif
__attribute__((section(".modinfo"))) =
"depends=";
#endif


module_init(bc_init);
module_exit(bc_exit);
module_param(bc_devices, int, 0);
module_param(bc_partitions, int, 0);

EXPORT_SYMBOL(bc_mgr_register_alg);
EXPORT_SYMBOL(bc_mgr_unregister_alg);





