#ifndef BC_LINUX_IOCTL_H
#define BC_LINUX_IOCTL_H

#define MAX_DEV_PATH 32
#define MAX_DEVICES 16

/* MAX key length in bits */
#define MAX_KEY_LENGTH		512
/* pool of random bytes */
#define POOL_SIZE_BYTES		512
/* MAX module name length in bytes. limited by kerneld */
#define MOD_NAME_LEN		64
/* MAX block size for BC_(DE)ENCRYPT_BLOCK in bytes */
#define MAX_BLOCK_SIZE		4096

#define BC_MAGIC		0x0000BCBC

#define EX_ENCRYPT_LRW_MODE_FLAG  0xBC000001
#define EX_ENCRYPT_CBC_MODE_FLAG  0xBC000002
#define EX_ENCRYPT_XTS_MODE_FLAG  0xBC000004
#define EX_ENCRYPT_ECB_MODE_FLAG  0xBC000008

#define ECB_MODE_NAME  "ECB"
#define CBC_MODE_NAME  "CBC"
#define LRW_MODE_NAME  "LRW"
#define XTS_MODE_NAME  "XTS"

typedef unsigned int bc_key_handle;
typedef unsigned int bc_alg_id;

/* IOCTL commands  */
#define BC_GET_INFO		0xBC00


struct bc_info {
	int	magic;			/* magic number protection */
	int	error;			/* return error code       */
	int	ver_major;		/* major version number    */
	int	ver_minor;		/* minor version number    */
	int	busy;			/* is device busy	   */
};


#define BC_SET_FD		0xBC01
#define BC_CLR_FD		0xBC02

#define BC_LOCK_DEV		0xBC03
#define BC_UNLOCK_DEV		0xBC04

#define BC_CHECK_FD		0xBC05
#define BC_NOTIFY_PID		0xBC06
#define BC_GET_PRIV		0xBC07

#define BC_FORCE_UNLOCK		0xBC09
#define BC_CHECK_ACTIVITY	0xBC0A

#define BC_VERIFY_ALG		0xBC10

#define BC_MAKE_KEY		0xBC20
#define BC_TEST_KEY		0xBC21
#define BC_FREE_KEY		0xBC22

#define BC_ENCRYPT_BLOCK	0xBC30
#define BC_DECRYPT_BLOCK	0xBC31

#define BC_IS_FREE		0xBC40
#define BC_GET_VERSION		0xBC41
#define BC_STAT_DEV		0xBC42

/* BC driver user flags definition */
#define BC_FLAGS_READONLY	0x00000001
#define BC_FLAGS_MULTPART	0x00000002
#define	BC_FLAGS_IV_64BIT	0x00000004

struct bc_file64 {
	int	magic;			/* magic number protection         */
	int	error;			/* return error code               */
	int	fd;			/* container's file descriptor     */
	unsigned long long offset;	/* data offset. ALIGNED by 8 bytes */
	unsigned long long start_sector;/* starting sector for actual data */
	unsigned long long num_sectors;	/* actual data size                */
	unsigned int	flags;		/* binding flags                   */
	bc_alg_id	alg_id;		/* encryption algorithm id         */
	int		alg_mode;	/* encryption mode */
	bc_key_handle	key_handle;	/* key handle (invisible to user?) */
};

#define BC_CONTAINER_UNUSED	0
#define BC_CONTAINER_USED	1
#define BC_CONTAINER_MOUNTED	2

struct bc_check {
	int	magic;			/* magic number protection         */
	int	error;			/* return error code               */
	int	fd;			/* container's file descriptor     */
	int	busy;			/* BC_CONTAINER_* flags (return)   */
};

#define BC_PID_EXCLUDE		0
#define BC_PID_INCLUDE		1

struct bc_pid {
	int	magic;			/* magic number protection */
	int	error;			/* return error code       */
	int	pid;			/* pid to notify	   */
	int	sig;			/* signal to send	   */
	int	op;			/* BC_PID_EXCLUDE/INCLUDE  */
};


struct bc_alg {
	int		magic;			/* magic number protection  */
	int		error;			/* return error code        */
	bc_alg_id	alg_id;			/* encryption algorithm id  */
	char		alg_module[MOD_NAME_LEN];	/* algorithm module name */
	int		verified;		/* encryption algorithm is present in driver */
};





struct bc_key {
	int		magic;			/* magic number protection  */
	int		error;			/* return error code        */
	bc_alg_id	alg_id;			/* encryption algorithm id  */
	char		alg_module[MOD_NAME_LEN];	/* algorithm module name */
	unsigned char	*key;		/* pointer to user key	    */
	unsigned int	key_len;	/* key length in bits       */
	unsigned char	*pool;		/* pointer to rnd data pool */	
	bc_key_handle 	key_handle;	/* key handle               */
};


struct bc_block {
	int		magic;			/* magic number protection */
	int		error;			/* return error code       */
	bc_alg_id	alg_id;			/* encryption algorithm id  */
	int		alg_mode;		/* encryption mode */
	bc_key_handle 	key_handle;	/* key handle               */
	unsigned char	iv[8];		/* initial vector          */
	unsigned char	*buffer;	/* block to be processed   */
	unsigned int	buffer_len;	/* block length. ALIGNED?? */
};


struct bc_version
{
	int magic;
	int error;
	int ver_major;
	int ver_minor;
};


/****************
 * 2.0
 */

struct bc_dev_info 
{
	char* 		pathbuf;
	unsigned	bufsiz;
	int		is_configured;
};


/**
 * Set key request.
 */
struct bc_key_ctrl
{
	/* in */
	unsigned char* 	data;	/* key data including any mode context */
	unsigned int	size;	/* total key data size in bytes */

	/* out */
	int 		error;	/* returned bc specific error */
	unsigned int	handle;	/* returned handle to the key */
};


/**
 * Version request.
 */
struct bc_ver_ctrl
{
	/* out */
	int major;	/* major version number */
	int minor;	/* minor version number */
};


/**
 * Set container loop request.
 */
struct bc_setloop_ctrl
{
	/* in */
	int fd;
	int key_handle;
	int alg_id;
	int mode_id;
	unsigned long long data_offset;
	unsigned long long data_length;

	/* out */
	int error;
};	


#endif 
