/*******************************************************************
 *	Copyright (c) 1994-1998 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_mgr.c
 *
 *	Description:	algorithm manager, /proc filesystem support
 *
 *	Scope:		BestCrypt pseudo-device driver
 *
 *	Platforms:	Linux
 *
 *	Author:		Nail R. Kaipov
 *
 *	Created:	10-Nov-1998
 *
 *	Revision:	11-Aug-1999.	Bug fixed in get_bc_algo()
 *			21-Mar-2000.	/proc/bcrypt output change (v0.4b-2)
 *			23-Jun-2000.	/proc/bcrypt output change (v0.4b-5)
 *			07-Feb-2001.	proc get_info->read_proc change
 *
 *******************************************************************/

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/slab.h>
#include <asm/segment.h>

#include "bc_mgr.h"
#include "bc_cfg.h"
#include "gf_mul.h"

static struct bc_alg_mod	*alg_head;



static struct bc_alg_mod *find_mod(bc_alg_id id)
{
	struct bc_alg_mod* mod;

	for (mod = alg_head; mod; mod = mod->next)
		if (mod->alg.ops->id == id) return mod;

	return NULL;
}




//
//
//	Interface
//
//




int bc_mgr_init()
{
	alg_head = NULL;
	return 0;
}


int bc_mgr_proc_conf(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len;
	char *p = page;
	struct bc_alg_mod* tmp;

	p += sprintf(p, "# BestCrypt system v" BC_VERSION_STRING ":\n");
	p += sprintf(p, "# Compiled at "__TIME__ " " __DATE__"\n");
	for (tmp = alg_head; tmp; tmp = tmp->next)
		p += sprintf(p, "module\t%s\t0x%02x\t%d\t%s\n",
				tmp->mod_name[0] ? tmp->mod_name:"unknown",
				tmp->alg.ops->id,
				tmp->alg.ops->key_size_bits,
				tmp->alg.ops->name);
	len = p - page;
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}


static int init_mod(struct bc_alg_mod* mod, struct bc_valg* valg)
{
	memset(mod, 0, sizeof(struct bc_alg_mod));
	memset(mod->mod_name, 0, sizeof(mod->mod_name));

	mod->alg.key_table = kmalloc(sizeof(struct bc_key_element) * KEY_TABLE_LENGTH, GFP_KERNEL);
	if(!mod->alg.key_table)
		return -ENOMEM;

	memset(mod->alg.key_table, 0, sizeof(struct bc_key_element) * KEY_TABLE_LENGTH);

	mod->alg.ops = kmalloc(sizeof(struct bc_valg), GFP_KERNEL);
	if(!mod->alg.ops)
		return -ENOMEM;

	memset(mod->alg.ops, 0, sizeof(struct bc_valg));

	memcpy(mod->alg.ops, valg, sizeof(struct bc_valg));

	mod->mod_refcount = 0;
	mod->mod_flags = 0;

	return 0;
}


int bc_mgr_register_alg (struct bc_valg* alg)
{
	struct bc_alg_mod *new_mod;
	int rc;

	if (!alg || !alg->name) return -EINVAL;
	if (find_mod(alg->id)) {
		printk(KERN_ERR "bc: algorithm %s (%d) is already registered.\n",
			alg->name,
			alg->id);
		return -EBUSY;
	}

	if (	!alg->encrypt_ecb ||
		!alg->decrypt_ecb ||
		!alg->encrypt_cbc ||
		!alg->decrypt_cbc ||
		!alg->key_extend
	)
	{
		printk(KERN_ERR "bc: can't register algorithm functions.\n");
		return -EINVAL;
	}

	new_mod = kmalloc(sizeof(struct bc_alg_mod), GFP_KERNEL);
	if (!new_mod)
		return -ENOMEM;

	rc = init_mod(new_mod, alg);
	if( rc != 0 ) {
		kfree(new_mod);
		return rc;
	}

	new_mod->next = alg_head;
	alg_head = new_mod;

	return 0;
}


struct bc_algorithm* bc_mgr_get_alg (bc_alg_id id)
{
	struct bc_alg_mod* mod;

	mod = find_mod(id);
	if (mod)
		return &mod->alg;

	printk(KERN_ERR "bc: unsupported algorithm (id=%d).\n", id);
	return NULL;
}

static void release_mod(struct bc_alg_mod* mod)
{
	// shred and free key table
	unsigned int i;
	struct bc_key_element* elem;
	struct bc_algorithm* alg = &mod->alg;

	for (i = 0; i < KEY_TABLE_LENGTH; i++)
	{
		elem = &mod->alg.key_table[i];

		if (elem->key_extended) {
			memset(elem->key_extended, 0, alg->ops->ext_key_size_bytes);
			kfree (elem->key_extended);
		}
		if (elem->pool) {
			memset(elem->pool, 0, POOL_SIZE_BYTES);
			kfree (elem->pool);
		}
		if (elem->lrw_context) {
			memset(elem->lrw_context, 0, sizeof(GfmulCtx));
			kfree (elem->lrw_context);
		}
		if (elem->xts_context) {
			memset(elem->xts_context, 0, alg->ops->ext_key_size_bytes);
			kfree (elem->xts_context);
		}
		memset(elem, 0, sizeof(struct bc_key_element));
	}

	kfree(mod->alg.key_table);

	// shred and free mod
	memset(mod, 0, sizeof(struct bc_alg_mod));
	kfree(mod);
}

int bc_mgr_unregister_alg (bc_alg_id id)
{
	struct bc_alg_mod *todel, *prev, *next;

	todel = find_mod(id);
	if (!todel) {
		printk(KERN_ERR "bc: algorithm %d already unregistered.\n", id);
		return -EINVAL;
	}

	next = todel->next;

	if (todel == alg_head) {
		alg_head = next;
	} else {
		prev = alg_head;
		while (prev->next != todel)
			prev = prev->next;

		prev->next = todel->next;
	}

	release_mod(todel);
	return 0;
}


/*
struct bc_algorithm *get_bc_algo(ALG_ID id, char *module_name)
{
	struct bc_alg_mgr *algo;

	algo = find_bc_algo(id);
	if (algo) return &(algo->algo);

	if (!module_name) return NULL;
	request_module("%s", module_name);

	if ((NULL != alg_head) && (alg_head->algo.id == id)) {
		strncpy(alg_head->module_name, module_name, MODULE_NAMELEN);
		return &(alg_head->algo);
	}
	printk(KERN_ERR "bc: unsupported algorithm (id=%d).\n", id);
	return NULL;
}
*/

/*
int register_bc_algo(struct bc_algorithm *algo)
{
	struct bc_alg_mgr *new_algo;

	if (!algo || !algo->name) return -EINVAL;
	if (find_bc_algo(algo->id)) {
		printk(KERN_ERR "bc: algorithm %s (%d) is already registered.\n",
			algo->name,
			algo->id);
		return -EBUSY;
	}

	if (	!algo->encrypt  ||
		!algo->decrypt  ||
		!algo->make_key ||
		!algo->test_key ||
		!algo->free_key ||
		!algo->lock_key ) {
		printk(KERN_ERR "bc: can't register algorithm functions.\n");
		return -EINVAL;
	}
	new_algo = kmalloc(sizeof(struct bc_alg_mgr), GFP_KERNEL);
	if (!new_algo) return -ENOMEM;
	memset(new_algo, 0, sizeof(new_algo));
	memset(new_algo->module_name, 0, sizeof(new_algo->module_name));

	new_algo->algo = *algo;
	new_algo->algo_refcount = 0;
	new_algo->algo_flags = 0;
	new_algo->next = alg_head;
	alg_head = new_algo;

	return 0;
}
*/



/*
int unregister_bc_algo(ALG_ID id)
{
	struct bc_alg_mgr *tmp, *todel;

	todel = find_bc_algo(id);
	if (!todel) {
		printk(KERN_ERR "bc: algorithm %d already unregistered.\n", id);
		return -EINVAL;
	}

	if (todel == alg_head) {
		alg_head = todel->next;
	} else {
		tmp = alg_head;
		while (tmp->next != todel) tmp = tmp->next;
		tmp->next = todel->next;
	}

	memset(todel, 0, sizeof(struct bc_alg_mgr));
	kfree(todel);

	return 0;
}
*/


char bc_mgr_c[]="$Id: bc_mgr.c,v 1.3 2002-10-29 07:11:46 crypt Exp $";
