/*
 *	Copyright Jetico, Oy, 2011.
 *	
 *	Mount helper tool.
 *	A suid tool that helps us mount bestcrypt virtual devices and set mount point ownership.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "common.h"


// Return 1 if filesystem type supports ownership/mode options.
// Return 0 otherwise.
static int fs_supports_ownership(const char* fs)
{
	if(fs == NULL) {
		return 0;
	}

	if(0 == strcmp(fs, "ntfs")) {
		return 1;
	} else if(0 == strcmp(fs, "vfat")) {
		return 1;
	} else if(0 == strcmp(fs, "msdos")) {
		return 1;
	} else {
		return 0;
	}
}


// Process options and launch /bin/mount.
// Change mount point ownership if requested.
static int do_mount(const char* dev, const char* mpoint, const char* type, int ro, int uid, int gid, int mode, const char* opts)
{

	pid_t pid = fork();
	if(pid < 0) {
		fprintf(stderr, "Error: fork failed\n");
		return EXIT_FAILURE;
	}

	if( 0 == pid ) 
	{
		const char* env[] = {"PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL};
	
		char ownership_opts[0xff] = {0};
		if(type && fs_supports_ownership(type)) {
			// add ownership options
			// for ntfs and fat
			snprintf(ownership_opts, sizeof(ownership_opts) - 1, "uid=%d,gid=%d,umask=%o", uid, gid, ~(mode) & 07777 /* convert to umask */);
		}

		// pass requested ownership options.
		// not all filesystems support this, so we can fail here.
	//	bc_get_root();

		setuid(0);
		execle(	"/bin/mount", 

			// mount arguments 
			"/bin/mount",
			ro ? "-r" : "-w",
			"-t",
			(type ? type : "autofs"),
			"-o",
			(opts ? opts : "user,exec,suid,nodev"),
			"-o",
			(fs_supports_ownership(type) ? ownership_opts : ""),
			"-o",
			"sync",
			
			"--",
			dev,
			mpoint,
			NULL,

			env);

		// exec failed if we're here
		return EXIT_FAILURE;

	}

	// parent
	int status;
	waitpid(pid, &status, 0);
	return WEXITSTATUS(status);
}


int main(int argc, char** argv)
{
	if(argc < 3) {
		fprintf(stderr, "Error: please specify a valid BestCrypt virtual block device and a mount point\n");
		return EXIT_FAILURE;
	}

	int ro = 0;
	int uid = getuid();
	int gid = getgid();
	int mode = 0700;
	const char* type = NULL;
	const char* options = NULL;

	char c;
    	while ((c = getopt(argc, argv, "rt:u:g:m:o:")) != -1) {
	switch (c) {
		case 'r': ro = 1; break;
		case 't': type = optarg; break;
		case 'u': uid = atoi(optarg); break;
		case 'g': gid = atoi(optarg); break;
		case 'm': mode = atoi(optarg); break;
		case 'o': options = optarg; break;
		default: fprintf(stderr, "Error: invalid option\n"); return EXIT_FAILURE;
	}
	}

	argv += optind;
	const char* blkdev = argv[0];
	const char* mpoint = argv[1];

	// verify that this is our block device.
	struct stat statbuf;
	if(0 != stat(blkdev, &statbuf)) {
		fprintf(stderr, "Error: stat failed\n");
		return EXIT_FAILURE;
	}

	if(!S_ISBLK(statbuf.st_mode)) {
		fprintf(stderr, "Error: %s is not a block device\n", blkdev);
		return EXIT_FAILURE;
	}

	if(bc_get_dev_major() != major(statbuf.st_rdev)) {
		fprintf(stderr, "Error: %s not a bestcrypt block device\n", blkdev);
	}

	// call mount to do the job
	int rc = do_mount(blkdev, mpoint, type, ro, uid, gid, mode, options);
	if(rc != EXIT_SUCCESS) {
		return rc;
	}

	// change mount point ownership if requested
	if(!ro) {
		bc_get_root();

		if( (0 != chown(mpoint, uid, gid)) || (0 != chmod(mpoint, mode)) ) {
			fprintf(stderr, "Warning: failed changing mount point ownership\n");
		}

		bc_drop_root();
	}

	return EXIT_SUCCESS;		
}


