/*
 *
 *      Copyright (c) 1994-2008 Jetico, Inc., Finland
 *      All rights reserved.
 *
 *      File: bctool.cpp
 *      Revision: $Id: bctool.c,v 1.19 2005-05-14 10:58:51 crypt Exp $
 *      Revision: bctool.cpp v 2.0 2008-12-22 inso
 *      Description: implementation of command line interface
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stddef.h>

#include "bctool.h"
#include "version.h"
#include "error.h"
#include "config.h"
#include "misc.h"

#include <alg/alg.h>
#include <backend/fs.h>
#include <common/log.h>
#include <common/exception.h>
#include <common/core.h>

#include "info.h"
#include "new.h"
#include "format.h"
#include "link.h"
#include "fsck.h"
#include "passwd.h"
#include "hidden.h"
#include "mount.h"
#include "umount.h"
#include "unlock.h"
#include "reencrypt.h"


extern bool bclist(void);

using namespace bdk;


const char* o_type;		/* -c container type */
char *o_description;      	/* -d description of container */
const char *o_fstype;           /* -t type of used file system */
char *o_alg;                    /* -a algorithm name to be used */
int o_algmode;
char *o_options;                /* miscellaneous options */
long long int o_cont_size;      /* -s container size in bytes */
int o_verbose;			/* -v verbose */
int o_ro;			/* -r read only */
int o_mode;                     /* -m mountpoint mode */
int o_uid;                      /* -u user id for mount point (root only)*/
int o_gid;                      /* -g group id for mount point (root only)*/
int o_random_password;          /* -p use random password for raw containers */
int o_force;                    /* -f force operation (like umount) */
int o_umount_all;               /* -A umnount all comtainers for umount command*/

uid_t ruid,euid;		/* real and effective user ID's */

ConsoleLog console_log;

extern bool bc_test_algs();

bool run_command( const char* commandname, int argc, char** argv ) throw(std::exception)
{
    	if ( !strcmp(commandname, "info") ) 		return bcinfo(argc, argv);
    	if ( !strcmp(commandname, "list") ) 		return bclist();
	if ( !strcmp(commandname, "new") ) 		return bcnew(argc, argv, o_alg, o_algmode, o_type, o_cont_size, o_description, o_fstype, (o_verbose > 0));
    	if ( !strcmp(commandname, "link") )  		return bclink(argc, argv, o_ro);
    	if ( !strcmp(commandname, "unlink") )  		return bcunlink(argc, argv);
    	if ( !strcmp(commandname, "raw_link") )  	return bclinkraw(argc, argv, o_alg, o_ro, o_random_password);
	if ( !strcmp(commandname, "format") )		return bcformat(argc, argv, o_fstype);
    	if ( !strcmp(commandname, "fsck") )  		return bcfsck(argc, argv, o_fstype);
    	if ( !strcmp(commandname, "add_passwd") )  	return bcpasswd_add(argc, argv);
    	if ( !strcmp(commandname, "passwd") )  		return bcpasswd_change(argc, argv);
    	if ( !strcmp(commandname, "del_passwd") )  	return bcpasswd_del(argc, argv);
    	if ( !strcmp(commandname, "make_hidden") )	return bchidden(argc, argv, o_alg, o_fstype, o_verbose);
    	if ( !strcmp(commandname, "mount") ) 		return bcmount(argc, argv, o_fstype, o_uid, o_gid, o_mode, o_ro, o_options);
    	if ( !strcmp(commandname, "umount") )  		return bcumount(argc, argv, o_force, o_verbose);
    	if ( !strcmp(commandname, "umount_all") )	return bcumountall(argc, argv, o_force, o_verbose);
	if ( !strcmp(commandname, "unlock") )  		return bcunlock(argc, argv);
    	if ( !strcmp(commandname, "reencrypt") )  	return bcreencrypt(argc, argv, o_alg, o_verbose);

    	bc_get_log(BC_MSG_CRIT) << "Unknown command " << commandname << " cannot continue.";
    	usage(stderr);

    	return false;
}


static bool init_runtime()
{
	if( !read_config() ) {
		fprintf(stderr, "read config failed\n");
		return false;
	}

	return true;
}

static void load_modules()
{
	if( !bdk::bc_core()->module_manager()->load_module( "libmodv7.so" ) )
	{
		bc_get_log(BC_MSG_WARN) << "Warning: Failed loading v7 format module. BestCrypt will run without kgsha\\kgsha256 containers support.";
	}

	if( !bdk::bc_core()->module_manager()->load_module( "libmodv8.so" ) )
	{
		bc_get_log(BC_MSG_WARN) << "Warning: Failed loading v8 format module. BestCrypt will run without kgghost containers support.";
	}
}


bool register_process()
{
	// init logging console output
	bc_get_log().register_listener(&console_log);

	char major, minor;
	backend::bc_crypto_version( &major, &minor );

	if( (major != VERSION_MAJOR) || (minor != VERSION_MINOR) ) {
		bc_get_log(BC_MSG_CRIT) << "Error: One or more components of BestCrypt is outdated.\n"
					<< "Reinstalling correct version of program may help.\n"
					<< " driver version: " << major << "." << minor << "\n"
					<< " bctool version: " << VERSION_MAJOR << "." << VERSION_MINOR << "\n";
		return false;
	}

	return true;
}


void bc_exit(int retcode)
{
	exit(retcode);
}


int main (int argc, char ** argv)
{
    	int  c;
    	char *progname;
    	char *commandname;

    	//Error=0;
    	o_fstype = NULL;
    	o_alg = NULL;
	o_algmode = AM_CBC;
    	o_options = NULL;
    	o_cont_size = 0;
    	o_description = "";
	o_type = "BCv7";
    	o_verbose = 0;
    	o_ro = 0;
    	o_uid = getuid();
    	o_gid = getgid();
   	o_mode = 0700;

    	setvbuf(stderr,(char*)NULL,_IONBF,0);
    	setvbuf(stdout,(char*)NULL,_IONBF,0);

	ruid = getuid();
	euid = geteuid();

	if( !register_process() ) {
		bc_get_log(BC_MSG_CRIT) << "Error: process registration failed";
		exit(EXIT_FAILURE);
	}

    	progname = strdup(argv[0]);

    	while ( (c=getopt(argc,argv,"AVvfhprt:a:o:s:d:c:m:u:g:"))!=-1 )
	{
        	switch ( (char)c )
	        {
        	case 'V':
                	fprintf(stdout, "BestCrypt version %d.%d-%d\nCopyright Jetico, Inc. See LICENSE for details.\n", VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH);
                	exit(EXIT_SUCCESS);
            	case 'v':
                	o_verbose++;
	                break;
        	case 'A':
                	o_umount_all++;
                	break;
            	case 'p':
                	o_random_password=0xff;
                	break;
            	case 'r':
                	o_ro=0xff;
	                break;
            	case 'f':
                	o_force=0xff;
	                break;
            	case 'h':
                	usage(stdout);
	                exit(-1);
            	case 't':
                	o_fstype=optarg;
	                break;
            	case 'a':
                    	o_alg = bc_str_to_lower(optarg);
                	break;
            	case 'o':
                	o_options=optarg;
                	break;
            	case 'd':
                	o_description=optarg;
 	               break;
		case 'c':
			o_type = optarg;
			break;
            	case 's':
                    	o_cont_size = bc_atol(optarg);
                	if ( o_cont_size==0 )
                	{
                    		bc_get_log(BC_MSG_CRIT) << "Invalid container size";
                    		exit(EXIT_FAILURE);
                	}
                	break;
            	case 'm':
                	o_mode=strtol(optarg,NULL,8);
               	 	o_mode=o_mode & 07777;
                	if (0 == o_mode)
                	{
                    		bc_get_log(BC_MSG_WARN) << "Invalid mode. Using 0700";
                    		o_mode=0700;
                	}
                	break;
            	case 'u':
                	if (0 != ruid)
                	{
                    		bc_get_log(BC_MSG_CRIT) << "Only root can specify mountpoint's owner";
                    		exit(EXIT_FAILURE);
                	}
                	o_uid = get_uid_by_name(optarg);
                	if (-1 == o_uid)
                	{
                    		bc_get_log(BC_MSG_CRIT) << "Invalid username " << optarg;
                    		exit(EXIT_FAILURE);
                	}
                	break;
            	case 'g':
                	o_gid = get_gid_by_name(optarg);
                	if (-1 == o_gid)
                	{
                    		bc_get_log(BC_MSG_CRIT) << "Invalid group name " << optarg;
                    		exit(EXIT_FAILURE);
                	}
                	break;
            	case '?': /* fall thru */
            	default:
                	usage(stderr);
                	exit(EXIT_FAILURE);
        	} /* end of switch */

    	} /* end of while */

    	argc -= optind;
    	argv += optind;
        commandname = basename(progname);

    	if ( 0 != strcmp(commandname,"bctool") )
    	{
        	if ( 0 == strncmp(commandname,"bc",2))
        	{
            		commandname +=2;
            		argv--;
            		argc++;
        	}
		else
        	{
            		bc_get_log(BC_MSG_CRIT) << "Unknown command " << commandname;
            		exit(EXIT_FAILURE);
        	}
     	}
	else
    	{
        	commandname = argv[0];
    	}

    	if ( NULL == commandname )
    	{
        	bc_get_log(BC_MSG_CRIT) << "Missed command";
        	usage(stderr);
        	exit(EXIT_FAILURE);
    	}

    	if ( !init_runtime())
    	{
		bc_get_log(BC_MSG_CRIT) << "bctool: failed initializing algorithm core";
		exit(EXIT_FAILURE);
    	}

    	load_modules();

	bool rc;

	try
	{
		rc = run_command(commandname, argc, argv);
	}
	catch(std::exception const & e)
	{
		printf("%s\n", e.what());
	}

    	if (progname) {
        	free(progname);
    	}

	(rc) ? bc_exit(EXIT_SUCCESS) : bc_exit(EXIT_FAILURE);
} /* end of main */


void usage(FILE *fd) {
    const char *usage_string="Usage:\n\
bctool command [options] arguments\n\
\n\
commands:\n\
   list	       - list all linked devices, their container files and mount points\n\
   new         - create container   			(filename -a alg_name -s size [-d description] [-t fstype] [-v])\n\
   format      - format container   			(filename [-t fstype])\n\
   mount       - mount container    			(filename mountpoint [-t fstype] [-m mode] [-u user] [-g group] [-r])\n\
   umount      - unmount container  			(mountpoint [-f])\n\
   umount_all  - unmount all mounted bc containers 	([-v] [-f])\n\
   info        - get info on container 			(filename)\n\
   passwd      - change container's password 		(filename)\n\
   add_passwd  - add additional password for container 	(filename)\n\
   del_passwd  - delete additional password 		(filename)\n\
   fsck        - filesystem check 			(filename [-t fstype])\n\
   reencrypt   - reencrypt container 			(filename -a alg_name [-v])\n\
   make_hidden - make hidden part in container 		(filename size_of_hidden_part [-t fstype] [-v])\n\
                 where: size_of_hidden_part xxx%% in percents of container size\n\
   link        - link container to BestCrypt device 	(filename devicename [-r])\n\
   raw_link    - link raw file to BestCrypt device 	(filename devicename -a alg_name [-p] [-r])\n\
   unlink      - unlink container from device 		(devicename)\n\
   unlock      - unlock container 			(filename)\n\
   keygen_cmd  - container key generator command	(filename command)\n\
\n\
options:\n\
   -t fstype   - to specify filesystem type (default: msdos)\n\
   -a alg_name - to specify algorithm name\n\
   -s size     - to specify container size\n\
                 example: -s xxx - in bytes; -s xxxK - in Kb; -s xxxM - in Mb; -s xxxG - in Gb\n\
   -d desc     - to specify description of container (up to 64 symbols)\n\
   -r          - to mount read only\n\
   -m mode     - to specify mountpoint mode in octal (default 0700)\n\
   -u user     - to specify mountpoint's owner (root only can do it)\n\
   -g group    - to specify mountpoint's group\n\
   -p 	       - use random password\n\
   -v 	       - be verbose\n\
   -V          - print version and exit\n\
\n";

    fprintf(fd, "%s\n", usage_string);
}

