
#include "config.h"

#include <common/build.h>
#include <common/core.h>

#include <alg/service.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>

#define CONFIG_FILE "/etc/bc.conf"

#define IS_WHITESPACE(c) 	(((c) == ' ') || ((c) == '\t') || ((c) == '\n') || ((c) == EOF))
#define IS_COMMENT(c) 		((c) == '#')
#define BUF_SIZE(buf)		(sizeof((buf)) * sizeof(*(buf)))

using namespace bdk;


// config file
static FILE* cfile;


static bool bc_config_open()
{
	cfile = fopen(CONFIG_FILE, "r");
	return (cfile != NULL);
}


static bool bc_config_close()
{
	if( !cfile ) return false;
	return fclose(cfile);
}


static bool read_here(FILE* file, char* buf, size_t bufsiz)
{
	int c;
	size_t nread = 0;

	while( ( c = fgetc(file) ) >= 0 )
	{
		if( IS_WHITESPACE(c) || (nread == bufsiz) ) break;
		buf[nread++] = tolower(c);
	}

	buf[nread] = '\0';

	if( nread == bufsiz ) {
		BC_DEBUG_PRINT(("config: -- warning -- truncated token\n"));
		return false; // truncated token
	}

	return true;
}


static bool read_token(FILE* file, char* buf, size_t bufsiz)
{
	int c;
    
	while ( 1 )
    	{
        	if ( ((c = fgetc(file)) < 0) || feof(file) ) break;

		// skip trailing whitespace
		if( !IS_WHITESPACE(c)) {
			// read from here please
			ungetc(c, file);
			return read_here(file, buf, bufsiz);
		}
	}

    	return false;	// nothing was read
}


static bool read_as_string(FILE *file, char* buf, size_t bufsiz)
{
	memset(buf, '\0', bufsiz);
    	return read_token(file, buf, bufsiz);
}


static bool read_as_number(FILE *file, int* num, char* buf, size_t bufsiz)
{
    	if( !read_token(file, buf, bufsiz))  return false;

	char* err;
	*num = strtol(buf, &err, 0);

    	if ( *err != '\0' ) return false;
    	return true;
}


bool read_config()
{
	if( !bc_config_open() ) {
		return false;
	}

	char buf[0xff] = {0};

	bc_alg_info_t info;

	while( !feof(cfile) ) {

        	if( !read_as_string(cfile, buf, BUF_SIZE(buf) ) ) continue; // we hit an eof likely, confirm on next iteration
        
		if ( !strcmp(buf, "module") )
       		{
       			if ( !read_as_string(cfile, buf, BUF_SIZE(buf)) ) return false;
			info.prvname.assign(buf, BUF_SIZE(buf));

            		if ( !read_as_number(cfile, &info.id, buf, BUF_SIZE(buf)) ) return false;

       			if ( !read_as_number(cfile, (int*)&info.key_size, buf, BUF_SIZE(buf)) ) return false;

       			if ( !read_as_number(cfile, (int*)&info.block_size, buf, BUF_SIZE(buf)) ) return false;

       			if ( !read_as_number(cfile, (int*)&info.ext_key_size, buf, BUF_SIZE(buf)) ) return false;

      			if ( !read_as_string(cfile, buf, BUF_SIZE(buf)) ) return false;
			info.name.assign(buf, BUF_SIZE(buf));

			info.mode_map[AM_CBC].key_size = info.key_size;
			info.mode_map[AM_CBC].is_supported = true;

			info.mode_map[AM_LRW].key_size = info.ext_key_size;
			info.mode_map[AM_LRW].is_supported = true;

			info.mode_map[AM_XTS].key_size = info.ext_key_size;
			info.mode_map[AM_XTS].is_supported = (info.block_size == 16);

			if( !bc_core()->alg_manager()->load_algorithm( info ) ) {
				fprintf(stderr, "load alg failed\n");
				return false;
			}
       		}
	}

	bc_config_close();
	return true;
}	


static bc_dev_conf_t read_dev_config()
{
	bc_dev_conf_t conf;
	memset ( &conf, 0, sizeof ( conf ) );

	FILE* proc = fopen ( "/proc/bcrypt/bc.stat", "r" );
	if ( !proc )
	{
		BC_DEBUG_PRINT ( ( "Debug: failed proc entry reading\n" ) );
		return conf;
	}

	int ret = 0;
	char line[0xff] = {0};
	char field[0xff] = {0};
	int value;
	while ( fgets ( line, sizeof ( line ), proc ) != NULL )
	{
		if ( line[0] == '#' )
			continue;

		ret = sscanf ( line, "%s %d", field, &value );
		if ( ret == 0 )
			continue;

		if ( strcmp ( field, "devices" ) == 0 )
			conf.total_devs = value;
		else if ( strcmp ( field, "major" ) == 0 )
			conf.major = value;
		else if ( strcmp ( field, "minor" ) == 0 )
			conf.minor = value;
	}

	fclose ( proc );
	return conf;
}


bc_dev_conf_t get_dev_config()
{
	static bc_dev_conf_t the_conf = read_dev_config();
	return the_conf;
}

