/*
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 */

#include "core.h"
#include "volume.h"
#include "device.h"
#include "user.h"
#include "process.h"


using namespace bestcrypt;


// Partition table info
struct part_info_t
{
	unsigned int part_count;
	unsigned int ext_part_count;
	bc_space_unit_t table[4];
};

// partition record in MBR
struct partition
{
	unsigned char active;
	unsigned char start_head;
	unsigned char start_sec;
	unsigned char start_cyl;
	unsigned char os_type;
	unsigned char end_head;
	unsigned char end_sec;
	unsigned char end_cyl;
	unsigned int lba_start_sec;
	unsigned int lba_nr_sec;
}__attribute__((packed));


// Parse MBR partition table if any in the supplied sector
static bool parse_partition_table(const void* buf, part_info_t* out_info)
{
	memset(out_info, '\0', sizeof(part_info_t));

	// check msdos partition signature
	if (0x55 != buf[0x1FE] || 0xAA != buf[0x1FF])
		return 0;

	// read partiotion table
	struct partition* prt = (struct partition *) (buf + 0x1BE);
	bool found = false;

	for (unsigned int i = 0; i < 4; ++i, ++prt)
	{
		//x[i] = 0;
		if (0 == prt->lba_nr_sec)
			continue; // no partition in this slot

		// found one
		switch (prt->os_type)
		{
		case 0x05: // DOS extended
		case 0x0F: // Windows 95 extended
		case 0x85: // Linux extended
		{
			// found extended partition
			found = true;
			out_info->ext_part_count++;
			break;
		}
		case 0x07:
		{
			found = true;
			out_info->part_count++;
			out_info->table[i].offset = prt->lba_start_sec;
			out_info->table[i].length = prt->lba_nr_sec;
		}
		};
	}

	return found;
}

struct filesystem 
{
	int 		id;		// kBCFS_XXX id
	const char* 	type;		// string type used for mkfs, fsck, mount, ...
	bool		is_native;	// is this fs a linux native, which means it accepts standart options.
					// if it is not native, it will require some -o stuff specific for it.
};


static const struct filesystem g_filesystems[] = 
{
	{ kBCFS_Unknown, NULL, false }, // unknown fs trap
	{ kBCFS_MSDOS, "msdos", true },
	{ kBCFS_VFAT, "vfat", true },
	{ kBCFS_EXT2, "ext2", true },
	{ kBCFS_EXT3, "ext3", true },
	{ kBCFS_REISERFS, "reiserfs", true },
	{ kBCFS_NTFS, "fuseblk", false }, // try fuseblk ntfs type
};

static bool is_fs_supported_by_kernel(int fsid) 
{
	FILE* f = fopen("/proc/filesystems", O_RDONLY);
	if(!f) {
		BC_TRACE(("Failed opening /proc/filesystems"));
		return false;
	}

	char buf[0xff];
	size_t nread;
	while(!feof(f)) {
		fscanf(f, "%s\n", buf);
		
		
	
}


namespace bestcrypt {


	int bc_link_volume(bc_volume* volume, bool ro)
	{
		bc_device dev;
		// device will be locked on success, don't forget to put it on failure.
		if(!bc_get_free_device(&dev)) {
			BC_TRACE(("Link: no free bestcrypt device to link to\n"));
			return kBCCore_NoFreeDevice;
		}

		// We may have our volume partitioned.
		// In this case we need to link to a device group with whole disk at main bc device and each partition at separate new device.
		// Device group naming is:
		// /dev/bcrypt[0-255] for main disk,
		// /dev/bcrypt[0-255]s[0-255] for each partition.

		unsigned char mbrdata[512] = {0};
		volume->read_sector(mbrdata, 0);

		part_info_t partinfo;
		if(bc_parse_partition_table(volume, &partinfo)) {
			bc_warning("Detected partition table inside volume:\n"
				" %d primary partitions\n"
				" %d extended partitions\n"
				"Will create device group for partitions\n",
				partinfo.prim_count,
				partinfo.ext_count);
		}

		// configure bc disk driver with this volume.
		// bc_bdev_set(volume, dev);
	
		// update volume info.

		volume->set_mounted(dev);

		return kBCCore_Success;
	}


	/**
	 * Unlink previously linked volume from device.
	 *
	 * @p volume	Volume to unlink. 
	 */
	void bc_unlink_volume(bc_volume* volume)
	{
		bc_device* dev = volume->device;
		if(!dev) {
			// XXX: maybe assert this?
			bc_warning("Volume is not linked, nothing to do\n");
			return;
		}

		bc_bdev_reset(dev);	// reset block device attachement to container file.
		bc_put_device(dev);	// put device to mark it availiable.
		volume->device = NULL;
	}


	int bc_format_volume(bc_volume* volume, const char* fstype) 
	{
		
		// our job is to correctly prepare options for mkfs and then just fire it up and let it do its job

		if(fs_is_native_linux(fsid)) {
			
		} else if(fs_is_ntfs(fsid)) {

		}

		bc_process mkfs = bc_run_process("/sbin/mkfs", argc, argv);
		
		// tell volume that we changed its filesystem
		// volume needs to encode its state.

		volume->update();
		return kBCCore_Success;
	}


	int bc_check_volume_fs(bc_volume* volume, const char* userfstype)
	{
		// our job is to correctly prepare options for fsck and then just fire it up and let it do its job
		// volume may or may not know its file system.
		// in case it doesn't know, rely on user supplied type.
	}


	int bc_mount_volume(bc_volume* volume, const char* mpoint, const char* fstype, bool ro, int uid, int gid, const char* options)
	{
		// out job is link volume to block dev and to correctly preprare options for mount.
		
	}
}

