
/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: fs.h
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	Filesystem interface.
 */

#include <sys/mount.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <mntent.h>
#include <linux/fs.h>

#ifndef MNT_DETACH
	#define MNT_DETACH 2
#endif

#include "fs.h"
#include "misc.h"
#include "process.h"

#define MTAB_PATH_STR "/etc/mtab"
#define MOUNT_PATH_STR "/bin/mount"

using namespace bdk;


class NativeFSException : public BCSystemException
{
public:
	NativeFSException(std::string const & action, std::string const & fstype)
	{
		m_msg = fstype + ":" + action;
	}
};
fsmap_t & NativeFS::get_fs_map()
{
	static fsmap_t the_map;

	if(the_map.empty())
	{
		the_map[FS_TAG_MSDOS] = "msdos";
		the_map[FS_TAG_EXT2] = "ext2";
		the_map[FS_TAG_MINIX] = "minix";
		the_map[FS_TAG_EXT3] = "ext3";
		the_map[FS_TAG_VFAT1] = "vfat";
		the_map[FS_TAG_NTFS] = "ntfs";
		the_map[FS_TAG_REISERFS] = "reiserfs";
		the_map[FS_TAG_XFS] = "xfs";
		the_map[FS_TAG_VFAT2] = "vfat";
		the_map[FS_TAG_UNDEFINED] = "undefined";
	}

	return the_map;
}

const char* NativeFS::str_by_id(unsigned int id)
{
	fsmap_t::iterator i = get_fs_map().find(id);
	fsmap_t::iterator end = get_fs_map().end();

	return (i != end) ? i->second : NULL;
}


FSHandle NativeFS::get_by_id(unsigned int id)
{
	fsmap_t::iterator i = get_fs_map().find(id);
	fsmap_t::iterator end = get_fs_map().end();

	return (i != end) ? FSHandle(new NativeFS(i->first)) : FSHandle(NULL);
}


FSHandle NativeFS::get_by_name(const char* name)
{
	fsmap_t::iterator i = get_fs_map().begin();
	fsmap_t::iterator end = get_fs_map().end();

	while(i != end)
	{
		if(strcmp(i->second, name) == 0)
			return FSHandle(new NativeFS(i->first));
		++i;
	}

	return FSHandle(NULL);
}


#if 0
void NativeFS::mount(bdk::BCFSObject const & mpoint, bdk::BCDeviceHandle device, bcu32_t uid, bcu32_t gid, bcu32_t mode, bool ro, std::string const & opts)
{
	BC_DEBUG_PRINT(("mount request mode=0%o", mode));

	Process mount("/bin/mount");
	std::vector < std::string > args;

	if(this->id() != FS_TAG_UNDEFINED)
	{
	        args.push_back("-t");
	        args.push_back(this->type());
	}

	args.push_back(((ro) ? "-r" : "-w"));
	args.push_back("-o");

	if(!opts.empty())
	{
		args.push_back(opts);
	}
	else
	{
		std::string def_options;
		def_options += ((getuid() != 0) ? "user,nosuid,nodev" : "user,suid,dev");
		args.push_back(def_options);
	}

	args.push_back("-o");
	args.push_back("sync");

	if( (!strcmp(this->type(), "vfat")) || (!strcmp(this->type(), "ntfs")) ) {
		args.push_back("-o");
		char vfat_usr[0xff];
		snprintf(vfat_usr, 0xff, "uid=%d,gid=%d,umask=%o", uid, gid, ~(mode) & 07777 /* convert from umask to mask */);
		args.push_back(vfat_usr);
	}

	args.push_back("--");
	args.push_back(device->system_name());
	args.push_back(mpoint.path());

	ElevatedScope sudo;

	int rc = mount.exec(args);
	if( 0 != rc )
		throw NativeFSException(this->type(), "mount failed");
}


bool NativeFS::umount(BCFSObject const & mpoint)
{
	Process mount("/bin/umount");
	std::vector < std::string > args;

	args.push_back("--");
	args.push_back(mpoint.path());

	ElevatedScope suid;
	int rc = mount.exec(args);
	return (rc == 0);
}


void NativeFS::detach(BCFSObject const & mpoint)
{
	Process mount("/bin/umount");
	std::vector < std::string > args;

	args.push_back("-l"); // lazy
	args.push_back("--");
	args.push_back(mpoint.path());

	ElevatedScope suid;
	int rc = mount.exec(args);
	if( 0 != rc )
		throw NativeFSException(this->type(), "detach failed");
}

#else

static char* itoa(bcu32_t i, char* buf, unsigned nbytes) {
	memset(buf, '\0', nbytes);
	snprintf(buf, nbytes - 1, "%d", i);
	return buf;
}

void NativeFS::mount(bdk::BCFSObject const & mpoint, bdk::BCDeviceHandle device, bcu32_t uid, bcu32_t gid, bcu32_t mode, bool ro, std::string const & opts)
{
	BC_DEBUG_PRINT(("mount request mode=0%o", mode));

	Process mount("/sbin/bcmount_helper");
	std::vector < std::string > args;

	if(this->id() != FS_TAG_UNDEFINED) {
	        args.push_back("-t");
	        args.push_back(this->type());
	}

	if(ro) {
		args.push_back("-r");
	}

	if(!opts.empty()) {
		args.push_back("-o");
		args.push_back(opts);
	}

	char value[0xff] = {0};

	args.push_back("-u");
	args.push_back(itoa(uid, value, sizeof(value)));
	
	args.push_back("-g");
	args.push_back(itoa(gid, value, sizeof(value)));

	args.push_back("-m");
	args.push_back(itoa(mode, value, sizeof(value)));

	args.push_back(device->system_name());
	args.push_back(mpoint.path());

	if(0 != mount.exec(args)) {
		throw NativeFSException(this->type(), "mount failed");
	}
}


bool NativeFS::umount(BCFSObject const & mpoint)
{
	Process umount("/sbin/bcumount_helper");
	std::vector < std::string > args;

	args.push_back(mpoint.path());

	return (0 == umount.exec(args));
}


void NativeFS::detach(BCFSObject const & mpoint)
{
	Process umount("/sbin/bcumount_helper");
	std::vector < std::string > args;

	args.push_back("-f");
	args.push_back(mpoint.path());

	if(0 != umount.exec(args)) {
		throw NativeFSException(this->type(), "detach failed");
	}
}

#endif


void NativeFS::format(BCDeviceHandle dev)
{
	Process mkfs("/sbin/mkfs");
	std::vector < std::string > args;
	const char* fstype = this->type();
	const char* device = dev->system_name();

	if (0 == strcasecmp(fstype, "vfat"))
	{
		args.push_back("-t");
		args.push_back("msdos");
		args.push_back("-F");
		args.push_back("32");
		args.push_back(device);
	}
	else if (0 == strcasecmp(fstype, "ext2") || 0 == strcasecmp(fstype, "ext3") )
	{
		args.push_back("-t");
		args.push_back(fstype);
		args.push_back("-m");
		args.push_back("0");
		args.push_back(device);
	}
	else if(0 == strcasecmp(fstype, "ntfs"))
	{
		args.push_back("-t");
		args.push_back(fstype);
		args.push_back("-v");
		args.push_back(device);
	}
	else
	{
	        args.push_back("-t");
	        args.push_back(fstype);
	        args.push_back(device);
	}

	int rc = mkfs.exec(args);
	if( 0 != rc )
		throw NativeFSException(this->type(), "format failed");
}


void NativeFS::check(BCDeviceHandle dev)
{
	Process fsck("/sbin/fsck");
	std::vector < std::string > args;
	const char* fstype = this->type();
	const char* device = dev->system_name();

	args.push_back("-V");
	if(this->id() != FS_TAG_UNDEFINED)
	{
	        args.push_back("-t");
	        args.push_back(fstype);
	}

	args.push_back(device);

	int rc = fsck.exec(args);
	if( 0 != rc )
		throw NativeFSException(this->type(), "check failed");
}

