
/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: fs.h
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	Filesystem interface.
 */

#ifndef BC_FILESYSTEM_H
#define BC_FILESYSTEM_H

#include <string>
#include <memory>

#include <common/types.h>
#include <common/fs.h>
#include <alg/device.h>


/**
 * @brief	Abstract file system, mounted from linked bc container.
 */
class IBCFileSystem
{
public:


	/**
	 * @brief	Mount file system.
	 *
	 * @p mpoint	Mount point.
	 * @p device	Device object to mount on.
	 * @p uid	User id for mounted filesystem ownership.
	 * @p gid	Group id for mounted filesystem ownership.
	 * @p mode	Access mode.
	 * @p ro 	Read only file system.
	 * @p opts	File system specific options.
	 */
	virtual 	void mount(bdk::BCFSObject const & mpoint, bdk::BCDeviceHandle device, bcu32_t uid, bcu32_t gid, bcu32_t mode, bool ro, std::string const & opts) = 0;


	/**
	 * @brief	Unmount the file system
 	 */
	virtual		bool umount(bdk::BCFSObject const & mpoint) = 0;


	virtual		void detach(bdk::BCFSObject const & mpoint) = 0;


	/**
	 * @brief	Format file system.
	 *
	 * @p dev	Device the fs is mounted on.
	 */
	virtual		void format(bdk::BCDeviceHandle dev) = 0;


	/**
	 * @brief	Check file system.
 	 *
	 * @p dev	Device the fs is mounted on.
	 */
	virtual		void check(bdk::BCDeviceHandle dev) = 0;


	/**
	 * @brief	Get file system id tag
	 */
	virtual 	bcu32_t	id() const = 0;


	/**
	 * @brief	Get file system type name
	 */
	virtual		const char* type() const = 0;
};


enum bc_fs_tag_t
{
	FS_TAG_MSDOS 	= 0,
	FS_TAG_EXT2 	= 1,
	FS_TAG_MINIX 	= 2,
	FS_TAG_EXT3 	= 3,
	FS_TAG_VFAT1 	= 4,
	//FS_TAG_RESERVED_5,
	//FS_TAG_RESERVED_6,
	FS_TAG_NTFS 	= 7,
	FS_TAG_REISERFS = 8,
	FS_TAG_XFS 	= 9,
	//FS_TAG_RESERVED_A,
	FS_TAG_VFAT2 	= 11,
	//FS_TAG_RESERVED_C,
	//FS_TAG_RESERVED_D,
	//FS_TAG_RESERVED_E,
	//FS_TAG_RESERVED_F,

	FS_TAG_UNDEFINED,    // always last
};


typedef bdk::BCSharedPtr<IBCFileSystem> FSHandle;
typedef std::map<int, const char*> fsmap_t;


/**
 * @brief	Native linux file system
 */
class NativeFS : public IBCFileSystem
{
public:

	NativeFS() : m_id(0) {};

	NativeFS(bcu32_t id) : m_id(id) {};


	static 	fsmap_t & get_fs_map();


	static 	FSHandle get_by_id(unsigned int id);


	static 	FSHandle get_by_name(const char* name);


	static 	const char* str_by_id(unsigned int id);


	/**
	 * @brief	Mount native file system with mount(2) call.
	 */
	BC_INHERITED 	void mount(bdk::BCFSObject const & mpoint, bdk::BCDeviceHandle device, bcu32_t uid, bcu32_t gid, bcu32_t mode, bool ro, std::string const & opts);


	/**
	 * @brief	Mount native file system with umount(2) call.
	 *
	 * @p device	BC device that the fs is linked on.
	 *
	 * @return	True if umounting was successfull, false if not.
	 */
	BC_INHERITED	bool umount(bdk::BCFSObject const & mpoint);


	/**
	 * @brief	Force file system to be detached.
	 * @more	Will try to detach the file system and force-clear the linked device.
	 */
	BC_INHERITED	void detach(bdk::BCFSObject const & mpoint);


	/**
	 * @brief	Format file system with mkfs call.
	 */
	BC_INHERITED	void format(bdk::BCDeviceHandle dev);


	/**
	 * @brief	Check fs with fsck call.
	 */
	BC_INHERITED	void check(bdk::BCDeviceHandle dev);


	/**
	 * @brief	Native fs tag
	 */
	BC_INHERITED	bcu32_t id() const
	{
		return m_id;
	}


	/**
	 * @brief	Native fs type string
	 */
	BC_INHERITED	const char* type() const
	{
		return get_fs_map().find(this->id())->second;
	}


private:

	/* native fs id */
	bcu32_t m_id;
};



#endif

