
/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: hidden.h
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description: 
 *	linux client 'make_hidden' command.
 */


#include "hidden.h"
#include "misc.h"
#include "iofile.h"
#include "fs.h"

#include <common/log.h>


using namespace bdk;


static bc_size_t get_size_of_hidden_part(const char *string, size_t cont_size /*in 512 units*/ ) 
{

    char *endptr;
    long long result;

    result=strtoul(string,&endptr,0);
    if ( string==endptr ) return 0;
    if ( (string[0]!='\0')&&(endptr[0]=='\0') ) return result;
    if ( (endptr[0]=='K' || endptr[0]=='k' )&&(endptr[1]=='\0') )
    {
        result *=1024;
        return result;
    };
    if ( (endptr[0]=='M' || endptr[0]=='m')&&(endptr[1]=='\0') )
    {
        result *=1024*1024;
        return result;
    };
    if ( (endptr[0]=='%')&&(endptr[1]=='\0') )
    {
        result = (result * cont_size / 100 );
        return result;
    };
    
    return 0L;

}


bool bchidden(int argc, char** argv, const char* algname, const char* fstype, bool verbose)
{
	if( argc < 3 ) {
		bc_get_log(BC_MSG_CRIT) << "Please specify container name and size of hidden part";
		return false;
	} else if ( argc > 3 ) {
		bc_get_log(BC_MSG_CRIT) << "Too many parameters";
		return false;
	}

	const char* fname = argv[1];
	const char* sizestr = argv[2];


	BCFSObject cont_path(fname);
	
	if( !cont_path.is_valid() ) {
		bc_get_log(BC_MSG_CRIT) << "File " << cont_path << " does not exist";
		return false;
	}

	if( !cont_path.access(BC_ACCESS_WRITE) ) {
		bc_get_log(BC_MSG_CRIT) << "You do not have write permissions for file " << cont_path;
		return false;
	}


	ContainerFile cont_file(fname, BC_FILE_LARGE|BC_FILE_READ|BC_FILE_WRITE);
	BCContainerHandle cont = load_container_file(cont_file);

	if( !cont ) {
		bc_get_log(BC_MSG_CRIT) << "Could not parse container from file " << cont_path; 
		return false;
	}

	if( cont->format()->is_locked() ) {
		bc_locked_warning(fname);
		return false;
	}

	
	// calculate size of hidden part
	bc_size_t size = get_size_of_hidden_part( sizestr, cont->format()->size_bytes() );
	if( (size <= 0) || (size > cont->format()->size_bytes() ) ) {
		bc_get_log(BC_MSG_CRIT) << "Error: invalid hidden part size.";
		return false;
	} else if ( (size % 512) != 0 ) {
		size = (((size >> 9 ) + 1) << 9);
		bc_get_log(BC_MSG_WARN) << "Warning: specified container size is not even to block size\n."
					<< "Will increase to " << size << " bytes";
	}



	// add hidden part and get a data context for it
	BCContainerViewHandle view = cont->format()->create_hidden_view(&cont_file, size);
	if(!view) {
		bc_get_log(BC_MSG_CRIT) << "Error: failed adding hidden part to container";
		return false;
	}

	if( verbose ) 
		bc_get_log(BC_MSG_INFO) << "Creating hidden part sized " << (size * 512) / (1024 * 1024) << "MB";


	if( fstype ) 
	{
		FSHandle fs = NativeFS::get_by_name(fstype);
		if(!fs)
		{
			bc_get_log(BC_MSG_CRIT) << "Error: unknown file system " << fstype;
		}
		else
		{
			BCDeviceHandle dev = BCDevice::get_free();
			BCContainerAutoBinding binding = view->create_binding(&cont_file, dev, false);

			fs->format(dev);	
		}
	} 
	else 
	{
		bc_get_log(BC_MSG_INFO) << "Please do not forget to format the container";
	}

	cont->format()->serialize(&cont_file);
	return true;
}

