/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: link.cpp
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	Implementation of linux client 'link', 'raw_link' and 'unlink' commands.
 *
 *	history:
 *
 */

#include <string.h>
#include <unistd.h>

#include "link.h"
#include "misc.h"
#include "iofile.h"

#include <alg/device.h>
#include <common/log.h>
#include <common/core.h>

#include <../CertificateLib/hashAlgs/SHA256/sha256.h>

using namespace bdk;


bool bclink( int argc, char** argv, bool ro )
{
	if( argc < 3 ) {
		bc_get_log(BC_MSG_CRIT) << "Error: Please specify device name and container name";
		return false;
	} else if (argc > 3) {
		bc_get_log(BC_MSG_CRIT) << "Error: Too many parameters";
		return false;
	}

	BCFSObject cname(argv[1]), devname(argv[2]);

	if( !cname.is_valid())
	{
		bc_get_log(BC_MSG_CRIT) << "File " << cname << " does not exist.";
		return false;
	}

	if( !cname.access(BC_ACCESS_READ) )
	{
		bc_get_log(BC_MSG_CRIT) << "You do not have read access for file " << cname;
		return false;
	}

    	if( !ro && !cname.access(BC_ACCESS_WRITE) )
	{
		bc_get_log(BC_MSG_CRIT) << "You do not have write permissions for container file " << cname << ".\n"
					<< "Run with -r option to link in read-only mode.";
		return false;
    	}


	// open device
	if( !devname.is_valid() || !devname.is_special() )
	{
		bc_get_log(BC_MSG_CRIT) << "Device file " << devname << " does not exist.";
		return false;
	}

	BCDeviceHandle device = BCDevice::get(devname.path());
	if( !device->is_free() )
	{
		bc_get_log(BC_MSG_CRIT) << "Device " << devname << " is busy";
		return false;
	}


	// open container file source
	ContainerFile cont_file(cname, BC_FILE_LARGE|BC_FILE_READ|((!ro) ? BC_FILE_WRITE : 0));
	BCContainerHandle cont = load_container_file(cont_file);
	if( !cont )
	{
		bc_get_log(BC_MSG_CRIT) << "Could not parse container from file " << cname;
		return false;
	}
	else if(!cont->is_recognized())
	{
		bc_get_log(BC_MSG_WARN) << "Container was not recognized. Assuming kg-ghost encrypted header";
	}
	else if(cont->format()->is_locked())
	{
		bc_locked_warning(cname.path().c_str());
		return false;
	}


	// ask user secret and load view
	BCContainerViewHandle view = cont->format()->get_view(&cont_file);
	if(!view)
	{
		bc_get_log(BC_MSG_CRIT) << "Error: failed authorizing container";
		return false;
	}


	// bind view to system device.
	BCContainerAutoBinding binding = view->create_binding(&cont_file, device, ro);

	/*
	if( cont->is_recognized() && !ro )
	{
		BC_DEBUG_PRINT(("Debug: locking container\n"));

		// lock container
		cont->format()->lock();

		// write lock to media
		cont->format()->serialize(&cont_file);
	}
        */

	// we're done, lock binded view
	binding.lock();
	bc_get_log(BC_MSG_INFO) << "Successfuly linked container " << cname << " to device " << devname;
	return true;
}


bool bcunlink( int argc, char** argv )
{
	if( argc < 2 ) {
		bc_get_log(BC_MSG_CRIT) << "Error: Please specify device name";
		return false;
	} else if (argc > 2) {
		bc_get_log(BC_MSG_CRIT) << "Error: Too many parameters";
		return false;
	}

	BCFSObject devname(argv[1]);

	// check device file existance
	if( !devname.is_valid() || !devname.is_special() )
	{
		bc_get_log(BC_MSG_CRIT) << "Device file " << devname << " does not exist.";
		return false;
	}

	// open device
	BCDeviceHandle device = BCDevice::get(devname.path());

	if(device->is_free())
	{
		bc_get_log(BC_MSG_INFO) << "Device " << devname << " is free, nothing to unlink.";
		return true;
	}

	// clear device
	if(!device->clear())
	{
		bc_get_log(BC_MSG_CRIT) << "Failed clearing device " << devname;
		return true;
	}

	bc_get_log(BC_MSG_INFO) << "Successfuly unlinked device " << devname;
	return true;
}


static bool GetRandomSeed( unsigned char *seed, int seedLength )
{

    FILE *fd;
    int  i;

    if ( ! (fd=fopen("/dev/urandom","r")) ) return false;

    for ( i=0; i<seedLength; i++ )
        seed[i]=(unsigned char)fgetc(fd);

    fclose (fd);
    return true;
}


bool bclinkraw(int argc, char** argv, const char* algname, bool ro, bool random_password)
{
	if( argc < 3 ) {
		bc_get_log(BC_MSG_CRIT) << "Error: Please specify device name and container name";
		return false;
	} else if (argc > 3) {
		bc_get_log(BC_MSG_CRIT) << "Error: Too many parameters";
		return false;
	}

	if( !algname ) {
		bc_get_log(BC_MSG_CRIT) << "Error: Please specify encryption algorithm";
		return false;
	}

	unsigned char pool[512];
	unsigned char* password;
	const char* fname;
	const char* devname;

	BCFSObject cont_path(fname), device_path(devname);

	// check container file existance
	if( !cont_path.is_valid() || !cont_path.is_file() )
	{
		bc_get_log(BC_MSG_CRIT) << "File " << cont_path << " does not exist.";
		return false;
	}


	// check device file existance
	if( !device_path.is_valid() || !device_path.is_special() )
	{
		bc_get_log(BC_MSG_CRIT) << "Device file " << devname << " does not exist.";
		return false;
	}


	// check container file read access
	if( !cont_path.access(BC_ACCESS_READ) )
	{
		bc_get_log(BC_MSG_CRIT) << "You do not have read access for file " << fname;
		return false;
	}


	// check container file write access in case of rw request
    	if( !ro && !cont_path.access(BC_ACCESS_WRITE) )
	{
		bc_get_log(BC_MSG_CRIT) << "You do not have write permissions for container file " << fname
					<< "Try running with -r option to link in read-only mode.";
		return false;
    	}


	// open device
	BCDeviceHandle device = BCDevice::get(devname);


	// check that device is free
	if( !device->is_free() )
	{
		bc_get_log(BC_MSG_CRIT) << "Device " << devname << " is busy";
		return false;
	}


	// validate alg
	BCAlgService* serv = bc_core()->get_alg_service( algname );
	if( !serv ) {
		bc_get_log(BC_MSG_CRIT) << "Error: Unknown algorithm " << algname;
		return false;
	}


	// open file
	ContainerFile file(fname, BC_FILE_READ|BC_FILE_LARGE|(ro) ? 0 : BC_FILE_WRITE);


	// get random seed or password
	bc_alg_mode_t mode;
	if( !serv->mode_info(mode, AM_CBC) ) {
		bc_get_log(BC_MSG_CRIT) << "Error: mode is not supported.";
		return false;
	}

	unsigned int key_size = (unsigned int)serv->key_size() >> 3;
	unsigned char* key = (unsigned char*) malloc(key_size);

	if( !key ) {
		bc_get_log(BC_MSG_CRIT) << "Error: no memory to allocate key size " << key_size;
		return false;
	}

	memset(key, '\0', key_size);

	if( random_password ) {

		if( !GetRandomSeed(pool, sizeof(pool)) ) {
			bc_get_log(BC_MSG_CRIT) << "Error: could not get random seed.";
			return false;
		}

		if( !SHA256_RandomGenerate(key, key_size, pool, sizeof(pool))) {
			bc_get_log(BC_MSG_CRIT) << "Error: could not generate random key";
			return false;
		}


	} else {

		password = (unsigned char*)getpass("Enter temporary password:");
		if( !password ) {
			bc_get_log(BC_MSG_CRIT) << "Error: Cannot read password";
			return false;
		}

		size_t passlen = strlen((char*)password);

		if( !SHA256_RandomGenerate(key, key_size, password, passlen)) {
			bc_get_log(BC_MSG_CRIT) << "Error: could not generate random key";
			return false;
		}

		memset(password, '\0', passlen);
		free(password);
	}


	memset(pool, '\0', sizeof(pool));

	// init algorithm
	BCAlgHandle alg = BCAlg::init(algname, AM_CBC, key, pool);
	memset(key, '\0', key_size);
	free(key);

	// craft context manually
	bc_space_unit_t space;
	space.offset = 0;
	space.length = file.size();

	BCContainerView full_view(alg, space, BC_CRYPTO_IV_64BIT, false);
	BCContainerAutoBinding binding = full_view.create_binding(&file, device, ro);

	binding.lock();
	bc_get_log(BC_MSG_INFO) << "Successfuly linked to device " << devname;
	return true;
}


