#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <pwd.h>
#include <grp.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#ifndef BLKGETSIZE
	#include <linux/fs.h>
#endif

#include "misc.h"

#include <common/build.h>
#include <common/log.h>

#include <backend/fs.h>

#include "../mod/bc_ioctl.h"

#define BC_ARRAY_SIZE(arr) sizeof((arr))/sizeof(*(arr))


bc_answer_t bc_ask(const char* string, bc_answer_t def_answer)
{
	char c;
	fprintf(stderr, "%s", string);
	c = tolower(fgetc(stdin));

	if( '\n' == c ) return def_answer;
	while( '\n' != fgetc(stdin) );

	if( 'y' == c ) return BC_ANSWER_YES;
	else return BC_ANSWER_NO;
}


bool bc_get_bdev_size(bc_size_t& size, const char* bdevname)
{
	int fd = open(bdevname, O_RDONLY);

	if( fd < 0 ) {
		return false;
	}

	if( 0 != ioctl( fd, BLKGETSIZE, &size ) ) {
		return false;
	}

	size *= S_BLKSIZE;

	return true;
}


bool find_mounted_device( const char* mpoint, char* buf, size_t bufsiz )
{
	struct stat st;
	
	if( lstat(mpoint, &st) != 0 ) return false;
	dev_t dev = st.st_dev;

	for( int i = 0; i < MAX_DEVICES; ++i ) {

		snprintf(buf, bufsiz, "/dev/bcrypt%d", i);
		if( (lstat( buf, &st ) == 0) && (dev == st.st_rdev) ) return true;

	}

	return false;
}


void bc_locked_warning(const char* cont_name)
{
	using namespace bdk;

	printf( "WARNING! Container \"%s\" is probably in use!\n\n"
            	"Two reasons for the report are possible:\n"
            	"    1. The container is already mounted. In this case do not use the container\n"
            	"    until someone dismounts it.\n"
            	"    2. The container was not unmounted properly last time, so a lock mask \n"
            	"    for the container was not reset. If so, unlock the container by running the command:\n"
            	"        \"bctool unlock %s\" \n\n",
		cont_name, cont_name);
}


bool bc_replace_file( const char* orig_name, const char* new_name, const char* backup_name )
{
	using namespace bdk;

	struct stat st;
	if( stat(orig_name, &st ) != 0 ) {
		//bc_get_log(BC_MSG_WARN) << "Error: stat failed, assuming default stat: " << strerror(errno);
		st.st_mode = 0644;
		st.st_uid = 0;
		st.st_gid = 0;
	}

	if( rename(orig_name, backup_name) != 0 ) {
		//bc_get_log(BC_ERR) << "Error: failed renaming: " << strerror(errno);
		return false;
	}


	if( rename(new_name, orig_name) != 0 ) {
		//bc_get_log(BC_ERR) << "Error: failed renaming: " << strerror(errno);
		return false;
	}

		
	if( chown(orig_name, st.st_uid, st.st_gid) != 0 ) {
		//bc_get_log(BC_ERR).msg("Error: failed setting new \"%s\"owner: %s\n", orig_name, strerror(errno));
		return false;
	}

	if( chmod(orig_name, st.st_mode) != 0 ) {
		//bc_get_log(BC_ERR).msg("Error: failed setting new \"%s\" mode: %s\n", orig_name, strerror(errno));
		return false;
	}

	if( unlink(backup_name) != 0 ) {
		//bc_get_log(BC_ERR).msg("Error: failed cleaning backup file \"%s\": %s\n", backup_name, strerror(errno));
		return false;
	}


	return true;
}


void set_sighandler( sighandler_t handler ) 
{
	signal(SIGINT, handler);
	signal(SIGQUIT, handler);
	signal(SIGTERM, handler);
	signal(SIGABRT, handler);
}


int get_uid_by_name(const char* name)
{
	struct passwd *pass;
    	if ( NULL == name ) return -1;

	uid_t uid = atol(name);

    	if( (pass = getpwnam(name)) || (pass = getpwuid(uid)) )
		return pass->pw_uid;
	
	return -1;
}


int get_gid_by_name(const char* name)
{
	
	struct group *grp;
    	if ( NULL == name ) return -1;

	gid_t gid = atol(name);

    	if( (grp = getgrnam(name)) || (grp = getgrgid(gid)) )
		return grp->gr_gid;

	return -1;
}


long long bc_atol ( const char *string ) 
{
	char *endptr;
        long long result;

        result=strtoul ( string,&endptr,0 );
        if ( string==endptr ) return 0;
        if ( ( string[0]!='\0' ) && ( endptr[0]=='\0' ) ) return result;
        if ( ( endptr[0]=='K' || endptr[0]=='k' ) && ( endptr[1]=='\0' ) )
        {
                result *=1024;
                return result;
        };
        if ( ( endptr[0]=='M' || endptr[0]=='m' ) && ( endptr[1]=='\0' ) )
        {
                result *=1024*1024;
                return result;
        };
        if ( ( endptr[0]=='G' || endptr[0]=='g' ) && ( endptr[1]=='\0' ) )
        {
                result *=1024*1024*1024;
                return result;
        };

        return 0L;
}

        
char* bc_str_to_lower ( char *string ) 
{
	size_t i;
	for ( i = 0; i <= strlen ( string ); i++ )
		string[i] = tolower ( string[i] );

	return string;
}


char* bc_str_to_lower ( const char* src, char* dst, size_t bufsiz )
{
	size_t i;
	for ( i = 0; i <= strlen ( src ) && i <= bufsiz; i++ )
	{
		dst[i] = tolower ( src[i] );
	}

	return dst;
}

