/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: mount.cpp
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	linux client 'mount' command.
 */

#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "mount.h"
#include "misc.h"
#include "iofile.h"
#include "fs.h"
#include "process.h"

#include <common/exception.h>
#include <common/log.h>
#include <common/smartptr.h>

#include <alg/device.h>

using namespace bdk;


static int get_fs_type(BCDeviceHandle& device, std::string& out_type)
{
	Process blkid("/sbin/blkid");
	std::vector < std::string > args;
	args.push_back("-o");
	args.push_back("value");
	args.push_back("-s");
	args.push_back("TYPE");
	args.push_back(device->system_name());

	int rc = blkid.exec(args, out_type);
	if(rc == EXIT_SUCCESS) {
		// remove trailing \n
		size_t pos = out_type.rfind("\n");
		if(pos) {
			out_type.erase(pos);
		}
	}

	return rc;
}


bool bcmount(int argc, char** argv, const char* fstype, bcu32_t uid, bcu32_t gid, bcu32_t mode, bool ro, const char* opts )
{
	if( argc < 3 ) {
		bc_get_log(BC_MSG_CRIT) << "Please specify container name and mount point";
		return false;
	} else if( argc > 3 ) {
		bc_get_log(BC_MSG_CRIT) << "Too many parameters";
		return false;
	}

	BCFSObject cont_path(argv[1]), mp_path(argv[2]);

	if(!cont_path.is_valid())
	{
		bc_get_log(BC_MSG_CRIT) << "Error: File " << cont_path << " does not exist";
		return false;
	}

    	if( !ro && !cont_path.access(BC_ACCESS_WRITE) )
	{
		bc_get_log(BC_MSG_CRIT) << "Error: You do not have write permissions for this container.\n"
					<< "Run with -r option to mount read-only";
        	return false;
    	}

	if(!mp_path.is_valid() || !mp_path.is_directory())
	{
		bc_get_log(BC_MSG_CRIT) << "Error: mount point " << mp_path << " does not exist";
		return false;
	}

    	if( (getuid() != 0) && (!mp_path.is_owner()) ) {
        	bc_get_log(BC_MSG_CRIT) << "Error: You are not the owner of " << mp_path;
        	return false;
	}

	// canonize both paths.
	cont_path.canonize();
	mp_path.canonize();


	//
	// load container to mount
	//


	ContainerFile cont_file(cont_path, BC_FILE_LARGE|BC_FILE_READ|((ro) ? 0 : BC_FILE_WRITE));
	BCContainerHandle cont = load_container_file(cont_file);

	if( !cont ) {
		bc_get_log(BC_MSG_CRIT) << "Could not parse container from file " << cont_path;
		return false;
	}


	// check if container is locked
	if( cont->format()->is_locked() )
	{
		bc_locked_warning(cont_path.path().c_str());
		return false;
	}


	// auth
	BCContainerViewHandle view = cont->format()->get_view(&cont_file);
	if( !view ) {
		bc_get_log(BC_MSG_CRIT) << "Error: failed authorizing container";
		return false;
	}


	// link
	BCDeviceHandle dev = BCDevice::get_free();
	if( !dev ) {
		bc_get_log(BC_MSG_CRIT) << "Error: failed allocating free device";
		return false;
	}

	BCContainerAutoBinding binding = view->create_binding(&cont_file, dev, ro);


	// check fstype
	FSHandle fs;
	std::string implicit_fstype;
	get_fs_type(dev, implicit_fstype);

	if (!fstype) {
		if( implicit_fstype.empty() ) {
			// blkid failed, maybe it does not exist on this system
			bc_get_log(BC_MSG_WARN) << "Warning: could not determine file system type and explicit file system type (-t option) was not provided. Mount point ownership problems can occur.";
			fs = NativeFS::get_by_id(FS_TAG_UNDEFINED);
		} else {
			bc_get_log(BC_MSG_INFO) << "Parsed file system type on container as " << implicit_fstype << ". Use -t to override.";
			fs = NativeFS::get_by_name(implicit_fstype.c_str());
		}
	} else {
		if(0 != strcmp(fstype, implicit_fstype.c_str()) ) {
			bc_get_log(BC_MSG_WARN) << "Warning: parsed file system type " << implicit_fstype << " is not as explicitly specified type " << fstype;
		}

		fs = NativeFS::get_by_name(fstype);
		if(!fs)
		{
		        bc_get_log(BC_MSG_CRIT) << "Error: unknown file system type " << fstype;
		        return false;
		}
	}

	// proceed to mount
	fs->mount(mp_path, dev, uid, gid, mode, ro, ((opts)?opts:""));

	binding.lock();
	bc_get_log(BC_MSG_INFO) << "Container successfuly mounted";

	return true;
}


