/*
 *	File: bcnew.h
 *	Module: Best Crypt Linux
 *	Created: 22.12.08 evgeny yakovlev
 *	Description: bcnew command definition
 */

#include <unistd.h>
#include <signal.h>

#include "new.h"
#include "misc.h"
#include "iofile.h"
#include "fs.h"

#include <alg/manager.h>
#include <common/log.h>
#include <common/core.h>

#include <../CertificateLib/hashAlgs/SHA256/sha256.h>

using namespace bdk;

const char* g_cont_name;


// catch termination signals when writing to container file to delete it.
static void sighandler(int sig)
{
        if ( BC_ANSWER_NO == bc_ask("Do you really want to terminate process? (y/[n]):", BC_ANSWER_NO /*default no*/) )
                return;

        bc_get_log(BC_MSG_INFO) << "new: process terminated by signal " << sig;

        if(g_cont_name)
                unlink(g_cont_name);

        exit(EXIT_FAILURE);
}


// randomize container data
static void randomize_container_data(ContainerFile* file, BCContainerViewHandle view, bool verbose)
{
        int pr, opr = 0;
        time_t t = time(NULL);

        bc_sector_t sector;
        bc_ulong_t total_sectors = view->total_sectors();
        bc_ulong_t cur_sector = 0;
        bc_ulong_t written = 0;

        // randomize first sector manually
        SHA256_RandomGenerate((unsigned char*)&sector.sec_data, view->sector_size(), (unsigned char*)&t, sizeof(t));

        while( cur_sector < total_sectors  )
        {
                // write current randomized sector directly to body
                view->direct_write_sector(file, sector, cur_sector);

                if( verbose )
                {
                        // output verbose log
                        pr = (cur_sector * 100 / total_sectors);
                        written = cur_sector * view->sector_size();

                        fprintf(stdout, "generating random data: %llu\\%llu KB (%3d%%)		\r",
                                written / 1024,
                                (total_sectors * view->sector_size()) / 1024,
                                pr
                        );

                        opr = pr;
                }

                // generate another sector
                SHA256_RandomGenerate((unsigned char*)&sector.sec_data, view->sector_size(), (unsigned char*)&sector.sec_data, view->sector_size());

                ++cur_sector;
        }


        if( verbose )
        {
                written = cur_sector * view->sector_size();
                fprintf(stdout, "generating random data: %llu\\%llu KB (%3d%%)		\n", written, written, 100 );
        }
}


bool bcnew(int argc, char** argv, const char* algname, int algmode, const char* type, bc_ulong_t size, const char* desc, const char* fstype, bool verbose)
{
        if( argc < 2 ) {
                bc_get_log(BC_MSG_CRIT) << "Please specify container name";
                return false;
        } else if ( argc > 2 ) {
                bc_get_log(BC_MSG_CRIT) << "Too many parameters";
                return false;
        } else if(!algname) {
                bc_get_log(BC_MSG_CRIT) << "Please specify algorithm name";
                return false;
        }

        BCAlgService* alg = bc_core()->get_alg_service(algname);

        if( !alg )
        {
                bc_get_log(BC_MSG_CRIT) << "Error: Unknown encryption algorithm " << algname;
                return false;
        }


        BCFSObject fspath(argv[1]);
        ContainerFile cont_file;

        if( fspath.is_valid() )
        {
                if( fspath.is_special() )
                {
                        if( BC_ANSWER_NO == bc_ask("Warning: All data on specified block device will be lost! Continue? (y/[n]):", BC_ANSWER_NO) ) {
                                return false;
                        }

                        if( size != 0 ) {
                                bc_get_log(BC_MSG_WARN) << "File " << fspath << " is a block device, ignoring \"-s\" option";
                        }

                        if( !bc_get_bdev_size(size, fspath.path().c_str()) ) {
                                bc_get_log(BC_MSG_CRIT) << "Error: failed to get block device size";
                                return false;
                        }
                }
                else
                {
                        bc_get_log(BC_MSG_CRIT) << "File " << fspath << " already exists and is not a block device";
                        return false;
                }

                cont_file.set_file(BCFile::open(fspath.path(), BC_FILE_READ|BC_FILE_WRITE|BC_FILE_LARGE));
        }
        else
        {
                if( size == 0 ) {
                        bc_get_log(BC_MSG_CRIT) << "Please specify container size";
                        return false;
                } else if ( (size % 512) != 0 ) {
                        size = (((size >> 9 ) + 1) << 9);
                        bc_get_log(BC_MSG_WARN) << "Warning: specified container size is not even to block size.\n"
                        << "Will increase to " << size << " bytes";
                }

                cont_file.set_file(BCFile::create(fspath.path(), BC_FILE_READ|BC_FILE_WRITE|BC_FILE_LARGE|BC_FILE_CREATE|BC_FILE_EXCL, BC_PERM_DEFAULT));
        }


        // create container
        BCContainerHandle cont = BCContainer::make(type, alg->id(), algmode, size, desc );
        if( !cont )
        {
                bc_get_log(BC_MSG_CRIT) << "Error: could not initialize new container of type " << type;
                return false;
        }


        // the stuff below might take a while
        g_cont_name = fspath.path().c_str();
        signal(SIGINT, sighandler);
        signal(SIGQUIT, sighandler);
        signal(SIGTERM, sighandler);
        signal(SIGABRT, sighandler);


        BCContainerViewHandle view = cont->format()->create_view();
        if(!view)
        {
                bc_get_log(BC_MSG_CRIT) << "Error: could not create container auth";
                return false;
        }


        if( !fspath.is_special() )
        {
                randomize_container_data(&cont_file, view, verbose);
        }

        cont->format()->serialize(&cont_file);
        bc_get_log(BC_MSG_INFO) << "Container successfuly created";

        // check if user requested formatting
        if(fstype)
        {
                BCDeviceHandle dev = BCDevice::get_free();

                FSHandle fs = NativeFS::get_by_name(fstype);
                if(!fs)
                {
                        bc_get_log(BC_MSG_WARN) << "Warning: unknown file system " << fstype
                        << "\nContainer will not be formatted.";
                }
                else
                {
                        BCContainerAutoBinding binding = view->create_binding(&cont_file, dev, false);
                        fs->format(dev);
                }
        }
        else
        {
                fprintf(stdout, "Please do not forget to format the container\n");
        }

        return true;
}

