/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: reencrypt.cpp
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	linux client 'reencrypt' command.
 */

#include <string.h>
#include <unistd.h>
#include <signal.h>

#include "reencrypt.h"
#include "misc.h"
#include "iofile.h"
#include "bctool.h"

#include <common/fs.h>
#include <common/log.h>
#include <common/core.h>
#include <common/thread.h>
#include <alg/alg.h>

using namespace bdk;


RRProcess::RRProcess(BCFSObject const & contpath) :
        m_cur_sector(0),
        m_total_sectors(0),
        m_aborted(false)
{
        BCFSObject rrstate_path(contpath.path() + ".rrstate");
        BCFSObject rrbackup_path(contpath.path() + ".rrbackup");

        if (rrstate_path.is_valid())
        {
                // state file already present, which means interrupted process.
                m_status = RR_INTERRUPTED;

                // restore state
                m_state_file = BCFile::open(rrstate_path, BC_FILE_READ | BC_FILE_WRITE);
                m_state_file->pread((bc_byte_t*) &m_cur_sector, sizeof(m_cur_sector), 1, 0);
        }
        else
        {
                m_status = RR_NEW;
                m_cur_sector = 0;
                m_state_file = BCFile::create(rrstate_path, BC_FILE_READ | BC_FILE_WRITE
                                | BC_FILE_CREATE | BC_FILE_EXCL);
        }

        m_backup_file = BCFile::open(rrbackup_path, BC_FILE_READ|BC_FILE_WRITE|BC_FILE_CREATE);
}

RRProcess::~RRProcess()
{
        if (m_status != RR_INTERRUPTED)
        {
                m_state_file->fs_object().remove();
                m_backup_file->fs_object().remove();
        }
}

void RRProcess::worker_thread()
{
        // start per-sector reencryption
        bc_sector_t sector;
        while (m_cur_sector < m_total_sectors)
        {
                m_old_view->read_sector(m_io, sector, m_cur_sector);
                m_new_view->write_sector(m_io, sector, m_cur_sector);

                // update state
                ++m_cur_sector;
                this->save_state();

                // check interrupt state
                if (this->is_aborted())
                {
                        sync();
                        m_status = RR_INTERRUPTED;
                        m_callback->on_abort(this);
                        break;
                }
                else
                {
                        m_callback->on_sector(this);
                }
        }

        m_status = RR_FINISHED;
        m_callback->on_end(this);
}

void RRProcess::start(BCContainerViewHandle old_view, BCContainerViewHandle new_view,
                      IBCContainerIO* io, IRRProgressCallback* callback)
{
        m_total_sectors = old_view->total_sectors();
        m_callback = callback;
        m_old_view = old_view;
        m_new_view = new_view;
        m_io = io;

        if(this->status() == RR_NEW)
                m_callback->on_begin(this);
        else if(this->status() == RR_INTERRUPTED)
                m_callback->on_resume(this);
        else
                BC_ASSERT(false);


        BC_DEBUG_PRINT(("Debug: initialized reencryption: start sector %llu, total sectors %llu, old alg name %s, new alg name %s\n",
                                        m_cur_sector,
                                        m_total_sectors,
                                        m_old_view->algorithm()->name(),
                                        m_new_view->algorithm()->name()
                        ));

        m_status = RR_IN_PROGRESS;

        try
        {
        //        BCThread worker;
        //        worker.start(new RRFunction(this));
                this->worker_thread();
        }
        catch(std::exception const & e)
        {
                this->abort();
        }
}

void RRProcess::store_backup(BCContainerHandle cont)
{
        ContainerFile cfile(m_backup_file);
        cont->format()->serialize(&cfile);
}

BCContainerHandle RRProcess::load_backup()
{
        ContainerFile cfile(m_backup_file);
        return load_container_file(cfile);
}

void RRProcess::save_state()
{
        m_state_file->pwrite((unsigned char*) &m_cur_sector, sizeof(m_cur_sector), 1, 0);
}

void RRProcess::load_state()
{
        m_state_file->pread((unsigned char*) &m_cur_sector, sizeof(m_cur_sector), 1, 0);
}
//------------------------------------------------------------------
// global process ptr for interrupt signals
RRProcess* g_rr_proc;


void sighandler_nointr(int sig)
{
        if (BC_ANSWER_YES == bc_ask("Do you really want to terminate process?\n"
                                    "Terminating reencryption will result in invalid container state that can only be fixed by resuming reencryption later\n"
                                    "(y/[n]):", BC_ANSWER_NO /*default no*/)
        )
        {
                g_rr_proc->abort();
        }

        return;
}

void sighandler_intr(int sig)
{
        if (BC_ANSWER_YES == bc_ask("Do you really want to terminate process? (y/[n]):",
                                    BC_ANSWER_NO /*default no*/))
        {
                bc_get_log(BC_MSG_INFO) << "Process terminated by signal";
                bc_exit(EXIT_FAILURE);
        }

        return;
}

/**
 * @brief       Our default progress reporter.
 */
class RRProgressCallback : public IRRProgressCallback
{
public:

        explicit RRProgressCallback(bool verbose) : m_verbose(verbose) {}

        BC_INHERITED    void on_begin(RRProcess* proc)
        {
                fprintf(stdout, "Starting reencryption\n");
                BC_DEBUG_PRINT(("Debug: starting at sector %llu\n", proc->cur_sector()));
        }

        BC_INHERITED    void on_end(RRProcess* proc)
        {
                fprintf(stdout, "\nContainer successfully reencrypted\n");
        }

        BC_INHERITED    void on_abort(RRProcess* proc)
        {
                fprintf(stdout, "\nReeincryption process interrupted\n");
                BC_DEBUG_PRINT(("Debug: aborted at sector %llu\n", proc->cur_sector()));
                bc_exit(EXIT_SUCCESS);
        }

        BC_INHERITED    void on_resume(RRProcess* proc)
        {
                fprintf(stdout, "Resuming reencryption\n");
                BC_DEBUG_PRINT(("Debug: resuming at sector %llu\n", proc->cur_sector()));
        }

        BC_INHERITED    void on_sector(RRProcess* proc)
        {
                if (m_verbose)
                {
                        // output verbose log
                        fprintf(stdout, "reencryption process: %llu\\%llu KB (%3d%%)         \r",
                                        (proc->cur_sector() * 512) / 1024,
                                        (proc->total_sectors() * 512) / 1024,
                                        (int)(proc->cur_sector() * 100 / proc->total_sectors())
                        );
                }

        }

private:

        /* perform verbose progress output */
        bool m_verbose;
};


bool bcreencrypt(int argc, char** argv, const char* algname, bool verbose)
{
        if (argc < 2)
        {
                bc_get_log(BC_MSG_CRIT) << "Error: Please specify container name";
                return false;
        }
        else if (argc > 2)
        {
                bc_get_log(BC_MSG_CRIT) << "Error: too many parameters";
                return false;
        }

        if (!algname)
        {
                bc_get_log(BC_MSG_CRIT) << "Error: Please specify algorithm name";
                return false;
        }

        // check new algorithm validity
        BCAlgService* new_alg = bc_core()->alg_manager()->find_algorithm(algname);
        if (!new_alg)
        {
                bc_get_log(BC_MSG_CRIT) << "Error: Unknown new algorithm " << algname;
                return false;
        }

        // check container file and load it
        BCFSObject cont_path(argv[1]);

        if (!cont_path.is_valid())
        {
                bc_get_log(BC_MSG_CRIT) << "Error: File " << cont_path << " does not exist";
                return false;
        }

        if (!cont_path.access(BC_ACCESS_READ | BC_ACCESS_WRITE))
        {
                bc_get_log(BC_MSG_CRIT) << "Error: You do not have rw permissions for file "
                                << cont_path;
                return false;
        }

        ContainerFile cfile(cont_path, BC_FILE_READ | BC_FILE_WRITE | BC_FILE_LARGE);
        BCContainerHandle cont = load_container_file(cfile);
        if (!cont)
        {
                bc_get_log(BC_MSG_CRIT) << "Error: failed parsing container file " << cont_path;
                return false;
        }

        BCContainerViewHandle old_view;
        BCContainerViewHandle new_view;

        // get current view
        bc_get_log(BC_MSG_INFO) << "Enter old container password";
        old_view = cont->format()->get_view(&cfile);
        if (old_view == BCContainerView::invalid_view())
        {
                bc_get_log(BC_MSG_CRIT) << "Error: auth failed";
                return false;
        }

        // check that loaded context is not hidden
        if (old_view->is_hidden())
        {
                bc_get_log(BC_MSG_CRIT) << "Error: Sorry, hidden part reencryption is not supported";
                return false;
        }

        // -------------------------------------------------------------------------------------------
        //	Load container reencryption state and initialize new context depending on load results
        //


        // load initial rr state
        RRProcess rr(cont_path);

        // we have a fresh start
        if (rr.status() == RRProcess::RR_NEW)
        {
                // init context with new algorithm
                cont->format()->set_algorithm_id(new_alg->id());
                new_view = cont->format()->create_view();
                if (new_view == BCContainerView::invalid_view())
                {
                        bc_get_log(BC_MSG_CRIT) << "Error: failed initializing new container auth";
                        return false;
                }

                // print ominous warning
                if (BC_ANSWER_NO == bc_ask("Warning: You are about to reencrypt your container.\n"
                                           "Please note that all additional passwords and hidden parts (if any) will be lost except for the new password you entered.\n"
                                           "Are you sure you want to continue? (y/[n])",
                                           BC_ANSWER_NO)
                )
                {
                        return false;
                }

                // store new container header backup
                rr.store_backup(cont);
        }
        // we were reencrypting this container before, but were interrupted
        // load backup and continue
        else if (rr.status() == RRProcess::RR_INTERRUPTED)
        {
                // load container reencrypted backup
                BCContainerHandle backup = rr.load_backup();

                bc_get_log(BC_MSG_INFO) << "Re-enter new container password";
                new_view = backup->format()->get_view(&cfile);

                cont = backup;
        }
        // debug error
        else
        {
                BC_ASSERT(false);
                bc_get_log(BC_MSG_CRIT) << "Error: Invalid reencryption state";
                return false;
        }

        //
        // 	Start initialized rr process
        //

        g_rr_proc = &rr;
        set_sighandler(sighandler_nointr);

        rr.start(old_view, new_view, &cfile, new RRProgressCallback(verbose));
        while(rr.status() == RRProcess::RR_IN_PROGRESS)
                ;

        cont->format()->serialize(&cfile);
        return true;
}

