
/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: reencrypt.h
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	linux client 'reencrypt' command.
 */


#ifndef BC_REENCRYPT_H
#define BC_REENCRYPT_H

#include <common/types.h>
#include <common/class_defs.h>
#include <common/file.h>
#include <common/thread.h>
#include <container/view.h>
#include <container/io.h>
#include <container/type.h>


class RRProcess;

/**
 * @brief       Reencryption progress report callback.
 */
class IRRProgressCallback
{
        BC_INTERFACE(IRRProgressCallback)
public:

        /**
         * @brief       Called on process start.
         */
        virtual void on_begin(RRProcess* proc) {};

        /**
         * @brief       Called on process end.
         */
        virtual void on_end(RRProcess* proc) {};

        /**
         * @brief       Called on process abort.
         */
        virtual void on_abort(RRProcess* proc) {};

        /**
         * @brief       Called on process resume.
         */
        virtual void on_resume(RRProcess* proc) {};

        /**
         * @brief       Called after each sector reencryption.
         */
        virtual void on_sector(RRProcess* proc) {};
};


/**
 * @brief       Default progress callback.
 */
typedef IRRProgressCallback RRNullCallback;


/**
 * @brief       Reencryption process.
 */
class RRProcess
{
BC_FORBID_COPY(RRProcess);
public:

        /**
         * @brief       Reencryption current status.
         */
        enum rr_status_t
        {
                RR_NEW = 0,             /**< fresh start */
                RR_IN_PROGRESS = 1,     /**< in progress */
                RR_INTERRUPTED = 2,     /**< was interrupted */
                RR_FINISHED = 9,        /**< finished */
        };

        /**
         * @brief       Initialize reencryption process
         * @more        If process was interrupted, will load last saved state.
         *
         * @p cont      Container file path.
         */
        RRProcess(bdk::BCFSObject const & cont);

        /**
         * @brief       Release reencryption process.
         * @more        Cleanup backup files depending on current status.
         */
        ~RRProcess();

        /**
         * @brief       Start reencryption process.
         *
         * @p old_view  Current container configured view.
         * @p new_view  New container configured view (new algorithm config).
         * @p io        Container io object.
         * @p callback  Progress callback object.
         */
        void start(bdk::BCContainerViewHandle old_view, bdk::BCContainerViewHandle new_view,
                   bdk::IBCContainerIO* io, IRRProgressCallback* callback);

        /**
         * @brief      Abort process.
         */
        void abort()
        {
                m_aborted = true;
                m_status = RR_INTERRUPTED;
        }

        /**
         * @brief       Was process aborted?
         */
        bool is_aborted() const
        {
                return m_aborted;
        }

        /**
         * @brief       Last reencrypted sector.
         */
        bc_ulong_t cur_sector() const
        {
                return m_cur_sector;
        }

        /**
         * @brief       Total sectors to reencrypt.
         */
        bc_ulong_t total_sectors() const
        {
                return m_total_sectors;
        }

       /**
         * @brief       Current reencryption process status
         */
        rr_status_t status() const
        {
                return m_status;
        }

        void save_state();

        void load_state();

        /**
         * @brief       Store reencrypted container backup
         */
        void store_backup(bdk::BCContainerHandle);

        /**
         * @brief       Load reencrypted container backup
         */
        bdk::BCContainerHandle load_backup();

protected:

        /**
         * @brief       Reencryption runs in the separate thread.
         */
        void worker_thread();

        /**
         * @brief       Reencryption thread functor.
         */
        class RRFunction : public bdk::IBCThreadFunction
        {
        public:

                RRFunction(RRProcess* proc) : m_process(proc) {}

                /**
                 * @brief       Call reencryption thread worker
                 */
                BC_INHERITED    void operator() ()
                {
                        m_process->worker_thread();
                }

        private:

                RRProcess* m_process;
        };

private:

        /* status flag */
        rr_status_t     m_status;

        /* last fully processed sector */
        volatile bc_ulong_t      m_cur_sector;

        /* total sectors to reencrypt */
        bc_ulong_t      m_total_sectors;

        /* registered progress calback */
        IRRProgressCallback* m_callback;

        /* old container data config */
        bdk::BCContainerViewHandle m_old_view;

        /* new container data config */
        bdk::BCContainerViewHandle m_new_view;

        /* container data io interface */
        bdk::IBCContainerIO* m_io;

        /* state next_sector backup */
        bdk::BCFileHandle    m_state_file;

        /* state next_sector backup */
        bdk::BCFileHandle    m_backup_file;

        /* interrupt process */
        volatile bool   m_aborted;
};


/**
 * @brief       Command entry.
 */
bool bcreencrypt( int argc, char** argv, const char* new_alg, bool verbose );


#endif

