/**
 *
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	file: umount.cpp
 *	scope: linux client
 *	author: evgeny yakovlev
 *
 *	description:
 *	linux client 'umount' command.
 */

#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <sys/user.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <unistd.h>
#include <mntent.h>

#include "misc.h"
#include "umount.h"
#include "fs.h"
#include "config.h"

#include <alg/device.h>
#include <common/log.h>

using namespace bdk;


static bool is_bc_mount_entry(struct mntent* pmnt)
{
	struct stat st;

	if( lstat(pmnt->mnt_dir, &st) != 0 )
		return false;

	dev_t dev = st.st_dev;

	bc_dev_conf_t conf = get_dev_config();
	return (major(dev) == conf.major);
}


static BCDeviceHandle get_mounted_device(BCFSObject const & path)
{
	struct stat st;

	if( lstat(path.path().c_str(), &st) != 0 )
		throw BCSystemException("lstat failed");

	dev_t dev = st.st_dev;
	bc_dev_conf_t conf = get_dev_config();
	if(major(dev) != conf.major)
	{
		return BCDevice::invalid_device();
	}
	else
	{
		char devname[0xff] = {0};
		snprintf(devname, 0xff, "/dev/bcrypt%d", minor(dev));
		return BCDevice::get(devname);
	}
}


/*
 *	General umount entry
 */
static bool do_umount( BCFSObject const & mpoint, bool forced )
{
	// find bc device that hosts mounted f
	BCDeviceHandle device = get_mounted_device(mpoint);
	if(!device)
	{
		bc_get_log(BC_MSG_CRIT) << "Error: failed getting mounted device for moint point " << mpoint;
		return false;
	}


	// remove persistent device lock that prevents it from being unlinked by user.
	// see bcmount implementation for paired lock.
	device->unlock();

	// get mounted fs handle
	FSHandle fs = NativeFS::get_by_id(0);

	// umount fs
	// strategy it to try simple stuff first and if it fails, then give up or try forced stuff
	// if anything exceptional occurs, just detach the device if forced

	try
	{
		if(!fs->umount(mpoint))
		{
			BC_DEBUG_PRINT(("Debug: umount failed. moving to detach\n"));
			// if asked for forced umount and simple umount failed, detach fs instead.
			if(forced) {
				bc_get_log(BC_MSG_INFO) << "Forcing unmount...";
				fs->detach(mpoint);
			} else {
				return false;
			}
		}

		if(!device->clear())
		{
			BC_DEBUG_PRINT(("Debug: clear failed. moving to detach\n"));
			if(forced)
				device->detach();
			else
				return false;
		}

		return true;
	}
	catch ( std::exception const & e )
	{
		BC_DEBUG_PRINT(("Debug: caught umount exception\n"));
		if(forced)
			device->detach();

		throw;
	}
}


bool bcumount(int argc, char** argv, bool force, bool verbose)
{
	if( argc < 2 ) {
		bc_get_log(BC_MSG_CRIT) << "Please specify mount point";
		return false;
	} else if( argc > 2 ) {
		bc_get_log(BC_MSG_CRIT) << "Too many parameters";
		return false;
	}

	BCFSObject mp(argv[1]);

	if(!mp.is_valid())
	{
		bc_get_log(BC_MSG_CRIT) << "Error: mount point " << mp << " does not exist";
		return false;
	}

	mp.canonize();

	if( (getuid() != 0) && !mp.is_owner() )
	{
		bc_get_log(BC_MSG_CRIT) << "Error: sorry you are not the owner of mount point " << mp;
		return false;
	}

	// dump disk caches
	sync();


	// umount
	if( !do_umount(mp, force) )
	{
		bc_get_log(BC_MSG_CRIT) << "Error: unmounting failed";
		return false;
	}

	bc_get_log(BC_MSG_INFO) << "Container successfuly unmounted";
	return true;
}


bool bcumountall(int argc, char** argv, bool force, bool verbose)
{
	if( argc > 1 ) {
		bc_get_log(BC_MSG_CRIT) << "Too many parameters";
		return false;
	}

	// dump disk caches
	sync();


	FILE* mtab = setmntent("/etc/mtab", "r");
	if( NULL == mtab )
		throw BCSystemException("Failed opening mtab");

	// scan mtab for our entries
	std::vector < BCFSObject > entries;
	uid_t ruid = getuid();
	struct mntent* pmnt;

	while((pmnt = getmntent(mtab)) && !feof(mtab))
	{
		BCFSObject mp(pmnt->mnt_dir);
		if(is_bc_mount_entry(pmnt) && ((ruid == 0) || mp.is_owner()))
		{
			entries.push_back(mp);
		}
	}

	endmntent(mtab);


	// umount all scanned entries
	std::vector < BCFSObject >::iterator i = entries.begin();
	std::vector < BCFSObject >::iterator end = entries.end();
	size_t umounted_count = 0;

	while(i != end)
	{
		if(!do_umount(*i, force))
			bc_get_log(BC_MSG_WARN) << "Warning: failed to umount " << *i << ", proceeding";
		else
			++umounted_count;

		++i;
	}

	if( verbose )
		bc_get_log(BC_MSG_INFO) << "Unmounted entries " << umounted_count << " of " << entries.size();


	return true;
}

