/*
 *	Copyright (c) 2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Encrypted volume interface.
 */

#ifndef BC_VOLUME_H
#define BC_VOLUME_H

namespace bestcrypt {

	struct bc_container;
	struct bc_device;
	struct bc_cipher;
	struct bc_key_info;


	/**
	 * Volume is an encrypted disk image encoded in bc container.
	 */
	struct bc_volume
	{
		bc_container*	cont; 	// container that encodes this volume.
		bc_cipher	cipher;	// cipher that this volume is encrypted with.
		bc_key_info	key;	// volume encryption key info.

		// if volume is linked, those will be initialized
		bc_device*	device;		// whole volume device
		bc_device**	ext_devices;	// if partitions were detected inside, this will be initialized for each partition.
	};


	/**
	 * Read volume data.
	 * Will read data from volume media and decrypt it.
	 *
	 * @p vol	Volume to read sector data from.
	 * @p data	Data buffer.
	 * @p secnum	Sector number to start reading from.
	 * @p nsectors	Number of sectors to read.
	 *
	 * @return	Total number of sectors read.
	 */
	size_t bc_read_sectors(bc_volume* vol, void* data, size_t secnum, size_t nsectors);

	
	/**
	 * Same as bc_read_sectors but does not decrypt data.
	 */
	size_t bc_read_raw_sectors(bc_volume* vol, void* data, size_t secnum, size_t nsectors);


	/**
	 * Write volume data.
	 * Will encrypt data and write it to volume.
	 *
	 * @p vol	Volume to write sector data to.
	 * @p data	Data buffer.
	 * @p secnum	Sector number to start writing from.
	 * @p nsectors	Number of sectors to write.
	 *
	 * @return	Total number of sectors written.
	 */
	size_t bc_write_sectors(bc_volume* vol, const void* data, size_t secnum, size_t nsectors);

	
	/**
	 * Same as bc_write_sectors but does not encrypt data.
	 */
	size_t bc_write_raw_sectors(bc_volume* vol, const void* data, size_t secnum, size_t nsectors);


	/**
	 * Parse partition table on volume.
	 *
	 * @p vol	Volume.
	 * @p out_part	Initialized partition table info.
	 *
	 * @return 	was any partition info found or not in volume.
	 */
	bool bc_parse_partition_table(bc_volume* vol, bc_part_info* out_part);
}

#endif

